/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef __MODEL__
#define __MODEL__

#include "modelgen.h"
#include "spritegn.h"


/*
==============================================================================

COMMON

==============================================================================
*/

// version used for the render
typedef struct polyverts_s
{
	float	tv[3];		// pre-transformed vertex coords
	float	st[2];		// standard tex-coords
	float	lm[2];		// lightmap tex-coords
	float	rgba[4];	// colour
} polyverts_t;


/*

d*_t structures are on-disk representations
m*_t structures are in-memory

*/

// entity effects

#define	EF_BRIGHTFIELD			1
#define	EF_MUZZLEFLASH 			2
#define	EF_BRIGHTLIGHT 			4
#define	EF_DIMLIGHT 			8


// extra effects
// these are set per model rather than per entity
#define EF_FIREFLARE			(1 << 1)
#define EF_TORCHFLARE			(1 << 2)
#define EF_ROCKETFLARE			(1 << 3)
#define EF_SHAMBLERMAGIC		(1 << 4)
#define EF_FLICKER				(1 << 5)
#define EF_S_LIGHT				(1 << 6)
#define EF_BUBBLESPRITE			(1 << 7)


/*
==============================================================================

TEXTURES

==============================================================================
*/

typedef struct gltexture_s
{
	// texture numbers
	int		texnum;
	int		lumatexnum;

	// info
	char	identifier[64];
	int		width, height;

	// where the texture comes from
	char	sourcefile[64];
	int		offset;

	// checksum
	byte	md5digest[16];

	// texture flags
	int flags;

	// linked list management
	struct gltexture_s *next;
	qboolean registered;
} gltexture_t;

typedef struct
{
	struct gltexture_s *glt;
	float	sl, tl, sh, th;
} glpic_t;

// texture upload flags
#define TEX_NOFLAGS		0
#define TEX_MIPMAP		1
#define TEX_LUMASPEC	2
#define TEX_ALIAS		8
#define TEX_BSP			16
#define TEX_SPRITE		32
#define TEX_2D			64
#define TEX_EXTERNAL	128
#define TEX_LIQUID		256
#define TEX_SCRAP		512
#define TEX_VIEWMODEL	1024
#define TEX_RESOURCE	2048
#define TEX_ALPHAMASK	4096
#define TEX_SBARWIDGET	8192

void GL_UploadTexture (gltexture_t *glt, unsigned *data, int width, int height, int bpp, int flags);
gltexture_t *GL_LoadTexture (char *identifier, int width, int height, byte *data, int flags);
gltexture_t *GL_FindGLT (int texnum);
void GL_LoadBitmapResourceTexture (gltexture_t *glt, int ResourceID, unsigned int flags);


/*
==============================================================================

BRUSH MODELS

==============================================================================
*/


#define LIT_IDENT 0x54494c51

// litfile header - darkplaces standard
typedef struct litheader_s
{
	char ident[4];
	int version;
} litheader_t;


//
// in memory representation
//
// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	vec3_t		position;
} mvertex_t;

#define	SIDE_FRONT	0
#define	SIDE_BACK	1
#define	SIDE_ON		2


// plane_t structure
// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct mplane_s
{
	vec3_t	normal;
	float	dist;
	byte	type;			// for texture axis selection and fast side tests
	byte	signbits;		// signx + signy<<1 + signz<<1
	byte	pad[2];
} mplane_t;

typedef struct texture_s
{
	char		name[16];
	unsigned	width, height;
	struct gltexture_s	*glt;
	struct msurface_s	*texturechain;
	int			anim_total;				// total tenths in sequence ( 0 = no)
	int			anim_min, anim_max;		// time for this frame min <=time< max
	struct texture_s *anim_next;		// in the animation sequence
	struct texture_s *alternate_anims;	// bmodels in frmae 1 use these
	unsigned	offsets[MIPLEVELS];		// four mip maps stored
} texture_t;


#define	SURF_PLANEBACK		2
#define	SURF_DRAWSKY		4
#define SURF_DRAWTURB		16
#define SURF_UNDERWATER		128
#define SURF_DRAWLAVA		256
#define SURF_DRAWSLIME		512
#define SURF_DRAWTELE		1024


#define	LM_BLOCK_WIDTH		128
#define	LM_BLOCK_HEIGHT		128


// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct
{
	unsigned short	v[2];
	unsigned int	cachededgeoffset;
} medge_t;

typedef struct
{
	float		vecs[2][4];
	float		mipadjust;
	texture_t	*texture;
	int			flags;
} mtexinfo_t;


typedef struct glpoly_s
{
	struct	glpoly_s	*next;
	struct	glpoly_s	*chain;

	// number of verts
	int		numverts;

	// most efficient drawing primitive
	// can't use a GLenum cos we don't have gl.h here
	unsigned int PRIMITIVE_TYPE;

	int		flags;			// for SURF_UNDERWATER

	// base "v" verts
	vec3_t		*basev;

	// offset into array for verts
	int			vertoffset;

	// general verts
	polyverts_t	*verts;
} glpoly_t;

typedef struct msurface_s
{
	int			visframe;		// should be drawn when node is crossed

	mplane_t	*plane;
	int			flags;

	int			firstedge;	// look up in model->surfedges[], negative numbers
	int			numedges;	// are backwards edges
	
	short		texturemins[2];
	short		extents[2];

	int			light_s, light_t;	// gl lightmap coordinates

	// smax and tmax precalced
	int			smax;
	int			tmax;

	// pointer to lightmap data for the surf
	byte		*lightmapdata;

	glpoly_t	*polys;				// multiple if warped
	struct	msurface_s	*texturechain;

	mtexinfo_t	*texinfo;

	// fixed waterwarp
	byte		warpindex[2];

	// lighting info
	int			dlightframe;

	// make this tolerant of increasing MAX_DLIGHTS
	byte		dlightbits[(MAX_DLIGHTS + 7) / 8];

	// lightmap struct number
	int			lightmaptexturenum;

	// texture object number
	int			gllmtexnum;

	byte		styles[MAXLIGHTMAPS];
	int			cached_light[MAXLIGHTMAPS];	// values currently used in lightmap
	qboolean	cached_dlight;				// true if dynamic light in cache
	byte		*samples;		// [numstyles*surfsize]
} msurface_t;

typedef struct mnode_s
{
// common with leaf
	int			contents;		// 0, to differentiate from leafs
	int			visframe;		// node needs to be traversed if current
	
	float		minmaxs[6];		// for bounding box culling

	struct mnode_s	*parent;

// node specific
	mplane_t	*plane;
	struct mnode_s	*children[2];	

	unsigned short		firstsurface;
	unsigned short		numsurfaces;
} mnode_t;



typedef struct mleaf_s
{
// common with node
	int			contents;		// wil be a negative contents number
	int			visframe;		// node needs to be traversed if current

	float		minmaxs[6];		// for bounding box culling

	struct mnode_s	*parent;

// leaf specific
	byte		*visdata;
	efrag_t		*efrags;

	msurface_t	**firstmarksurface;
	int			nummarksurfaces;
	int			key;			// BSP sequence number for leaf's contents
	byte		ambient_sound_level[NUM_AMBIENTS];
} mleaf_t;

// !!! if this is changed, it must be changed in asm_i386.h too !!!
typedef struct
{
	dclipnode_t	*clipnodes;
	mplane_t	*planes;
	int			firstclipnode;
	int			lastclipnode;
	vec3_t		clip_mins;
	vec3_t		clip_maxs;
} hull_t;

/*
==============================================================================

SPRITE MODELS

==============================================================================
*/


// FIXME: shorten these?
typedef struct mspriteframe_s
{
	int		width;
	int		height;
	float	up, down, left, right;
	struct gltexture_s	*glt;
} mspriteframe_t;

typedef struct
{
	int				numframes;
	float			*intervals;
	mspriteframe_t	*frames[1];
} mspritegroup_t;

typedef struct
{
	spriteframetype_t	type;
	mspriteframe_t		*frameptr;
} mspriteframedesc_t;

typedef struct
{
	int					type;
	int					maxwidth;
	int					maxheight;
	int					numframes;
	float				beamlength;		// remove?
	void				*cachespot;		// remove?
	mspriteframedesc_t	frames[1];
} msprite_t;


/*
==============================================================================

ALIAS MODELS

Alias models are position independent, so the cache manager can move them.
==============================================================================
*/

typedef struct
{
	int					firstpose;
	int					numposes;
	float				interval;
	trivertx_t			bboxmin;
	trivertx_t			bboxmax;
	int					frame;
	char				name[16];
} maliasframedesc_t;

typedef struct
{
	trivertx_t			bboxmin;
	trivertx_t			bboxmax;
	int					frame;
} maliasgroupframedesc_t;

typedef struct
{
	int						numframes;
	int						intervals;
	maliasgroupframedesc_t	frames[1];
} maliasgroup_t;

// !!! if this is changed, it must be changed in asm_draw.h too !!!
typedef struct mtriangle_s {
	int					facesfront;
	int					vertindex[3];
} mtriangle_t;


#define	MAX_SKINS	32

typedef struct aliashdr_s
{
	int			ident;
	int			version;
	vec3_t		scale;
	vec3_t		scale_origin;
	float		boundingradius;
	vec3_t		eyeposition;
	int			numskins;
	int			skinwidth;
	int			skinheight;
	int			numverts;
	int			numtris;
	int			numframes;
	synctype_t	synctype;
	int			flags;
	float		size;

	int					numposes;
	int					poseverts;
	int					posedata;	// numposes*poseverts trivert_t
	int					commands;	// gl command list with embedded s/t
	struct gltexture_s	*glt[MAX_SKINS][4];
	int					texels[MAX_SKINS];	// only for player skins

	// actual numbers used for rendering
	int			numrenderpolys;
	int			numrenderverts;

	// this must be the *last* entry in the struct!!!
	maliasframedesc_t	frames[1];	// variable sized
} aliashdr_t;


// prescaled in software
typedef struct
{
	float	v[3];
	byte	lightnormalindex;
	qboolean nodraw;
} drawvertx_t;


#define	MAXALIASVERTS	2048
#define	MAXALIASFRAMES	256
#define	MAXALIASTRIS	2048
extern	aliashdr_t	*pheader;
extern	stvert_t	stverts[MAXALIASVERTS];
extern	mtriangle_t	triangles[MAXALIASTRIS];
extern	trivertx_t	*poseverts[MAXALIASFRAMES];

// imports
void GL_CountAliasPolys (aliashdr_t *paliashdr);
void AllocAliasPolyVerts (void);


//===================================================================

//
// Whole model
//

typedef enum {mod_brush, mod_sprite, mod_alias} modtype_t;

#define	EF_ROCKET	1			// leave a trail
#define	EF_GRENADE	2			// leave a trail
#define	EF_GIB		4			// leave a trail
#define	EF_ROTATE	8			// rotate (bonus items)
#define	EF_WIZARDGOO	16			// green split trail
#define	EF_ZOMGIB	32			// small blood trail
#define	EF_KNIGHTFIRE	64			// orange split trail + rotate
#define	EF_VORE	128			// purple trail

typedef struct model_s
{
	char		name[MAX_QPATH];
	qboolean	needload;		// bmodels and sprites don't cache normally

	modtype_t	type;
	int			numframes;
	synctype_t	synctype;
	
	int			flags;

	// volume occupied by the model graphics
	vec3_t		mins, maxs;
	float		radius;

	// solid volume for clipping 
	qboolean	clipbox;
	vec3_t		clipmins, clipmaxs;

//
// brush model
//
	int			firstmodelsurface, nummodelsurfaces;

	int			numsubmodels;
	dmodel_t	*submodels;

	int			numplanes;
	mplane_t	*planes;

	int			numleafs;		// number of visible leafs, not counting 0
	mleaf_t		*leafs;

	int			numvertexes;
	mvertex_t	*vertexes;

	int			numedges;
	medge_t		*edges;

	int			numnodes;
	mnode_t		*nodes;

	int			numtexinfo;
	mtexinfo_t	*texinfo;

	int			numsurfaces;
	msurface_t	*surfaces;

	int			numsurfedges;
	int			*surfedges;

	int			numclipnodes;
	dclipnode_t	*clipnodes;

	int			nummarksurfaces;
	msurface_t	**marksurfaces;

	hull_t		hulls[MAX_MAP_HULLS];

	int			numtextures;
	texture_t	**textures;

	byte		*lightdata;
	char		*entities;

	// visibility
	byte		*visdata;

	// effects
	int			effects;

	qboolean	transwater;

//
// additional model data
//
	cache_user_t	cache;		// only access through Mod_Extradata

} model_t;

//============================================================================

void	Mod_Init (void);
void	Mod_ClearAll (void);
model_t *Mod_ForName (char *name, qboolean crash);
void	*Mod_Extradata (model_t *mod);	// handles caching
void	Mod_TouchModel (char *name);

mleaf_t *Mod_PointInLeaf (float *p, model_t *model);
byte	*Mod_LeafPVS (mleaf_t *leaf, model_t *model);

#endif	// __MODEL__
