/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"


typedef enum
{
	pt_static,
	pt_grav,
	pt_slowgrav,
	pt_fire,
	pt_explode,
	pt_explode2,
	pt_blob,
	pt_blob2,
	pt_gunsmoke,
	pt_singleframe,
	pt_expfade,
	pt_explosmoke,
	pt_bloodtrail,
	pt_bloodspray
} ptype_t;

typedef struct particle_s
{
	// driver-usable fields
	vec3_t		org;
	float		color;

	int			drawtype;
	glpoly_t	*poly;
	float		radius;
	float		alpha;
	float		scale;

	// exponential fade
	float		update;
	float		think;

	// drivers never touch the following fields
	struct particle_s	*next;
	vec3_t		vel;
	float		ramp;
	float		die;
	ptype_t		type;
} particle_t;

// default max # of particles at one time
#define MAX_PARTICLES			8192

// particle ramps.  these are indexes into the quake palette
int		ramp1[8] = {0x6f, 0x6d, 0x6b, 0x69, 0x67, 0x65, 0x63, 0x61};
int		ramp2[8] = {0x6f, 0x6e, 0x6d, 0x6c, 0x6b, 0x6a, 0x68, 0x66};
int		ramp3[8] = {0x6d, 0x6b, 6, 5, 4, 3};

particle_t	*active_particles, *free_particles;

particle_t	*particles;


// particles are drawn as quads so we need 4 verts per poly
// beware of setting MAX_PARTICLES too high!!!  this requires 364 bytes of RAM per particle
// so the sums are:
// 1024		: 372K
// 2048		: 744K
// 4096		: about 1.5MB
// 8192		: about 3MB
glpoly_t ParticlePolys[MAX_PARTICLES];
polyverts_t ParticleVerts[4 * MAX_PARTICLES];

// particle draw types
#define P_POINTS		666

// base types
#define P_SMOKE			0
#define P_FLARE1		1
#define P_FLARE2		2
#define P_BLAST			3

#define P_BLOOD1		4
#define P_BLOOD2		5
#define P_BLOOD3		6
#define P_BLOOD4		7

#define P_SPARK			8
#define P_FLASH			9
#define P_BUBBLE		10
#define P_11			11

#define P_HIT			12
#define P_13			13
#define P_14			14
#define P_15			15

#define P_GUNSMOKE		16

// additional types re-using the base type texture coords

typedef struct ptexcoord_t
{
	float bases;
	float baset;
	float sadd;
	float tadd;
} ptexcoord_s;

ptexcoord_s particle_texcoords[] =
{
	// base types
	{0, 0, 0.25, 0.5},			// smoke
	{0.25, 0.5, 0.25, 0.5},		// flare 1
	{0.5, 0.5, 0.25, 0.5},		// flare 2
	{0, 0.5, 0.25, 0.5},		// blast

	{0.375, 0, 0.125, 0.25},	// blood 1
	{0.5, 0, 0.125, 0.25},		// blood 2
	{0.625, 0, 0.125, 0.25},	// blood 3
	{0.75, 0, 0.125, 0.25},		// blood 4

	{0.875, 0, 0.125, 0.25},	// spark
	{0.25, 0.25, 0.125, 0.25},	// flash
	{0.375, 0.25, 0.125, 0.25},	// bubble
	{0.75, 0.5, 0.25, 0.25},	

	{0.625, 0.25, 0.125, 0.25},		// wall hit
	{0.25, 0.75, 0.25, 0.25},	
	{0.5, 0.75, 0.25, 0.25},	
	{0.75, 0.75, 0.25, 0.25},	

	// additional types re-using the base type texture coords
	{0, 0, 0.25, 0.5},			// gun smoke
};


// torch smoke
// keep this high enough to kill framerate dependency as much as possible
#define NUM_TORCH_SMOKE_PARTICLES 1024

particle_t R_SmokeParticles[NUM_TORCH_SMOKE_PARTICLES];
glpoly_t SmokePolys[NUM_TORCH_SMOKE_PARTICLES];
polyverts_t SmokeVerts[4 * NUM_TORCH_SMOKE_PARTICLES];


void R_InitTorchSmokeParticle (particle_t *p)
{
	// set the base particle state
	if (rand () & 1)
		p->color = 0 + (rand () % 7);
	else p->color = 96 + (rand () % 7);

	p->org[0] = -2 + (rand () % 5);
	p->org[1] = -2 + (rand () % 5);
	p->org[2] = -2 + (rand () % 5);

	p->vel[0] = p->vel[1] = p->vel[2] = 0;

	p->drawtype = P_SMOKE;
	p->radius = 3;
	p->alpha = 0.666;
}


void R_UpdateTorchSmokeParticle (particle_t *p, float updatetime)
{
	p->org[2] += p->vel[2] * updatetime;
	p->radius += (updatetime * 60);
	p->vel[2] += (updatetime * 45);
	p->alpha -= (updatetime / 2);
}


void R_UpdateTorchSmokeParticles (void)
{
	int i;
	float lowestalpha = 666;
	int lowestalphanum = 0;

	for (i = 0; i < NUM_TORCH_SMOKE_PARTICLES; i++)
	{
		R_UpdateTorchSmokeParticle (&R_SmokeParticles[i], cl_frametime);

		// find the particle with the lowest alpha
		if (R_SmokeParticles[i].alpha < lowestalpha)
		{
			lowestalpha = R_SmokeParticles[i].alpha;
			lowestalphanum = i;
		}
	}

	// add a new particle to the list to replace the particle with the lowest alpha
	// done here to ensure that only one new particle is added each frame, otherwise the
	// smoke animation experiences some strange blips every now and then
	R_InitTorchSmokeParticle (&R_SmokeParticles[lowestalphanum]);
}


void R_InitTorchSmokeParticles (void)
{
	int i;
	int j;
	particle_t *p;
	float updatetime;

	for (i = 0; i < NUM_TORCH_SMOKE_PARTICLES; i++)
	{
		R_InitTorchSmokeParticle (&R_SmokeParticles[i]);

		// link in poly and verts
		R_SmokeParticles[i].poly = &SmokePolys[i];
		SmokePolys[i].numverts = 4;
		SmokePolys[i].PRIMITIVE_TYPE = GL_QUADS;
		SmokePolys[i].vertoffset = i * 4;
		SmokePolys[i].verts = &SmokeVerts[i * 4];
		SmokePolys[i].chain = NULL;
		SmokePolys[i].flags = 0;
		SmokePolys[i].next = NULL;

		// run a fake update to simulate adding an extra particle each frame
		for (j = 0; j < i; j++)
		{
			updatetime = 6 + (rand () % 9);
			updatetime /= 1000.0;

			R_UpdateTorchSmokeParticle (&R_SmokeParticles[j], updatetime);
		}
	}
}


/*
===============
R_InitParticles
===============
*/
void R_InitParticles (void)
{
	int i;
	int j;
	polyverts_t *pv = ParticleVerts;
	glpoly_t *p = ParticlePolys;
	float temp_cl_frametime;

	particles = (particle_t *) Hunk_AllocName (MAX_PARTICLES * sizeof (particle_t), "particles");

	// link up the polys and verts, and set initial poly properties
	for (i = 0; i < MAX_PARTICLES; i++, pv += 4, p++)
	{
		p->chain = NULL;
		p->flags = 0;
		p->next = NULL;
		p->numverts = 4;
		p->PRIMITIVE_TYPE = GL_QUADS;
		p->verts = pv;
		p->vertoffset = i * 4;
	}

	// hack the frametime for the fake updates
	temp_cl_frametime = cl_frametime;

	// initialize torch smokes
	R_InitTorchSmokeParticles ();

	// run some fake updates to get a viable baseline
	for (j = 0; j < 512; j++)
	{
		// generate a fake frametime, keep it random for validity
		cl_frametime = 13.333f + (rand () % 9);
		cl_frametime /= 1000.0;

		// run a fake update of the particles
		R_UpdateTorchSmokeParticles ();
	}

	// restore the original frametime
	cl_frametime = temp_cl_frametime;
}


/*
===============
R_EntityParticles
===============
*/

#define NUMVERTEXNORMALS	162
extern	float	r_avertexnormals[NUMVERTEXNORMALS][3];
vec3_t	avelocities[NUMVERTEXNORMALS];
float	beamlength = 16;
vec3_t	avelocity = {23, 7, 3};
float	partstep = 0.01;
float	timescale = 0.01;

void R_EntityParticles (entity_t *ent)
{
	int			msgcount;
	int			i;
	particle_t	*p;
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	vec3_t		forward;
	float		dist;
	
	dist = 64;
	msgcount = 50;

	if (!avelocities[0][0])
	{
		for (i = 0; i < NUMVERTEXNORMALS * 3; i++)
		{
			avelocities[0][i] = (rand () & 255) * 0.01;
		}
	}

	for (i = 0; i < NUMVERTEXNORMALS; i++)
	{
		// moved this up to here so that we can avoid unnecessary sin/cos calcs if we run out of particles
		if (!free_particles)
			return;

		angle = cl.time * avelocities[i][0];
		sy = sin (angle);
		cy = cos (angle);

		angle = cl.time * avelocities[i][1];
		sp = sin (angle);
		cp = cos (angle);

		angle = cl.time * avelocities[i][2];
		sr = sin (angle);
		cr = cos (angle);

		forward[0] = cp * cy;
		forward[1] = cp * sy;
		forward[2] = -sp;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.01;
		p->color = 0x6f;
		p->type = pt_explode;

		p->org[0] = ent->origin[0] + r_avertexnormals[i][0] * dist + forward[0] * beamlength;			
		p->org[1] = ent->origin[1] + r_avertexnormals[i][1] * dist + forward[1] * beamlength;			
		p->org[2] = ent->origin[2] + r_avertexnormals[i][2] * dist + forward[2] * beamlength;			
	}
}


/*
===============
R_ClearParticles
===============
*/
void R_ClearParticles (void)
{
	int		i;
	polyverts_t *pv = ParticleVerts;
	glpoly_t *p = ParticlePolys;

	free_particles = &particles[0];
	active_particles = NULL;

	for (i = 0; i < MAX_PARTICLES; i++)
	{
		particles[i].next = &particles[i + 1];

		particles[i].alpha = 1.0f;
		particles[i].drawtype = P_POINTS;
		particles[i].radius = 1.0f;
		particles[i].scale = 0.001f;

		// link particles to polys
		particles[i].poly = &ParticlePolys[i];
	}

	particles[MAX_PARTICLES - 1].next = NULL;

	// link up the polys and verts, and set initial poly properties
	// this is done every map load to get valid properties
	for (i = 0; i < MAX_PARTICLES; i++, pv += 4, p++)
	{
		p->chain = NULL;
		p->flags = NULL;
		p->next = NULL;
		p->numverts = 4;
		p->PRIMITIVE_TYPE = GL_QUADS;
		p->verts = pv;
	}
}


void R_ReadPointFile_f (void)
{
	FILE	*f;
	vec3_t	org;
	int		r;
	int		c;
	particle_t	*p;
	char	name[MAX_OSPATH];
	
	sprintf (name,"maps/%s.pts", sv.name);

	COM_FOpenFile (name, &f);

	if (!f)
	{
		Con_Printf ("couldn't open %s\n", name);
		return;
	}
	
	Con_Printf ("Reading %s...\n", name);
	c = 0;

	for (;;)
	{
		r = fscanf (f, "%f %f %f\n", &org[0], &org[1], &org[2]);

		// didn't read 3 components
		if (r != 3) break;

		c++;
		
		if (!free_particles)
		{
			Con_Printf ("Not enough free particles\n");
			break;
		}

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;
		
		p->die = 99999;
		p->color = (-c) & 15;
		p->type = pt_static;
		VectorCopy (vec3_origin, p->vel);
		VectorCopy (org, p->org);
	}

	fclose (f);
	Con_Printf ("%i points read\n", c);
}

void R_BloodParticleSplash (vec3_t org, vec3_t dir, int color, int msgcount)
{
	int i;
	int j;
	particle_t *p;

	// 10 times the pain, baby!!!
	for (i = 0; i < msgcount * 10; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 666;

		p->color = (int) ((color & ~7) + (rand () & 7));

		p->type = pt_bloodspray;

		p->org[0] = org[0] + ((rand () & 15) - 8);
		p->org[1] = org[1] + ((rand () & 15) - 8);
		p->org[2] = org[2] + ((rand () & 15) - 8);

		// spray that gore!!!
		p->vel[0] = ((rand () & 127) - 64);
		p->vel[1] = ((rand () & 127) - 64);
		p->vel[2] = ((rand () & 127) - 64);

		p->drawtype = P_BLOOD1 + (rand () % 4);
		p->radius = 1 + (rand () % 5);
		p->alpha = 0.6 + (float) (rand () % 5) / 10;
		p->scale = 0.001f;
	}
}


void R_HipnoticParticleField (vec3_t org, vec3_t dir, int color, int msgcount)
{
	int			i, j;
	particle_t	*p;

	for (i = 0; i < msgcount * 2; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.3 * (rand () % 5);

		p->color = (int) ((color & ~7) + (rand() & 7));

		p->type = pt_slowgrav;

		p->org[0] = org[0] + ((rand () & 15) - 8);
		p->org[1] = org[1] + ((rand () & 15) - 8);
		p->org[2] = org[2] + ((rand () & 15) - 8);

		p->vel[0] = dir[0] * 15;
		p->vel[1] = dir[1] * 15;
		p->vel[2] = dir[2] * 15;

		p->drawtype = P_SPARK;
		p->radius = 1.5;
		p->alpha = 1;
		p->scale = 0.00002f;
	}
}


/*
===============
R_ParseParticleEffect

Parse an effect out of the server message
===============
*/
void R_ParseParticleEffect (void)
{
	vec3_t		org, dir;
	int			i, msgcount, color;

	for (i = 0; i < 3; i++) org[i] = MSG_ReadCoord ();
	for (i = 0; i < 3; i++) dir[i] = MSG_ReadChar () * (1.0 / 16);

	msgcount = MSG_ReadByte ();
	color = MSG_ReadByte ();

	// hack to increase the number of particles for rocket explosions and the like
	if (hipnotic && msgcount <= 4)
	{
		// particle field
		R_HipnoticParticleField (org, dir, color, msgcount);
	}
	else if (msgcount == 255)
	{
		// explosion
		R_ParticleExplosion (org);
	}
	else if (color == 73 || color == 225)
	{
		// blood splashes
		R_BloodParticleSplash (org, dir, color, msgcount);
	}
	else
	{
		// generic particle effect type thing (should be very few of these...)
		R_RunParticleEffect (org, dir, color, msgcount);
	}
}
	
/*
===============
R_ParticleExplosion

===============
*/
void R_ParticleExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;

	// add a blast
	for (i = 0; i < 32; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		// set the die high so that they go through the full ramps
		p->die = cl.time + 666;
		p->color = ramp1[0];
		p->ramp = (rand () % 3);

		if (i & 1)
			p->type = pt_explode;
		else p->type = pt_explode2;

		p->drawtype = P_BLAST;
		p->radius = 48 + (rand () % 32);
		p->alpha = 0.1f + (float) (rand () % 10) / 10.0f;
		p->scale = 0.002f;

		// unwound these loops cos there's quite a lot of them
		p->org[0] = org[0] + ((rand () % 32) - 16);
		p->org[1] = org[1] + ((rand () % 32) - 16);
		p->org[2] = org[2] + ((rand () % 32) - 16);

		p->vel[0] = (rand () % 16) - 8;
		p->vel[1] = (rand () % 16) - 8;
		p->vel[2] = (rand () % 16) - 8;
	}

	// and some smoke
	for (i = 0; i < 128; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		// set the die high so that they go through the full ramps
		p->die = cl.time + 666;

		if (i & 1)
			p->color = 0 + (rand () % 7);
		else p->color = 96 + (rand () % 7);

		p->drawtype = P_SMOKE;
		p->radius = 10 + (rand () % 16);
		p->alpha = 0.6f + (float) (rand () % 5) / 10.0f;
		p->scale = 0.002f;

		// unwound these loops cos there's quite a lot of them
		p->org[0] = org[0] + ((rand () & 63) - 32);
		p->org[1] = org[1] + ((rand () & 63) - 32);
		p->org[2] = org[2] + ((rand () & 63) - 32);

		p->vel[0] = (rand () & 63) - 32;
		p->vel[1] = (rand () & 63) - 32;
		p->vel[2] = (rand () & 7) - 4;
		p->type = pt_explosmoke;
	}
}


/*
===============
R_ParticleExplosion2

Colour mapped explosion
===============
*/
void R_ParticleExplosion2 (vec3_t org, int colorStart, int colorLength)
{
	int i, j;
	particle_t *p;
	int colorMod = 0;

	for (i = 0; i < 512; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.3;
		p->color = colorStart + (colorMod % colorLength);
		colorMod++;

		p->type = pt_blob;

		// unwound these loops cos there's quite a lot of them
		p->org[0] = org[0] + ((rand () % 32) - 16);
		p->org[1] = org[1] + ((rand () % 32) - 16);
		p->org[2] = org[2] + ((rand () % 32) - 16);

		p->vel[0] = (rand () % 512) - 256;
		p->vel[1] = (rand () % 512) - 256;
		p->vel[2] = (rand () % 512) - 256;
	}
}

/*
===============
R_BlobExplosion

===============
*/
void R_BlobExplosion (vec3_t org)
{
	int i, j;
	particle_t *p;

	for (i = 0; i < 1024; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 1 + (rand () & 8) * 0.05;

		if (i & 1)
		{
			p->type = pt_blob;
			p->color = 66 + rand () % 6;
		}
		else
		{
			p->type = pt_blob2;
			p->color = 150 + rand () % 6;
		}

		// unwound these loops cos there's quite a lot of them
		p->org[0] = org[0] + ((rand () % 32) - 16);
		p->org[1] = org[1] + ((rand () % 32) - 16);
		p->org[2] = org[2] + ((rand () % 32) - 16);

		p->vel[0] = (rand () % 512) - 256;
		p->vel[1] = (rand () % 512) - 256;
		p->vel[2] = (rand () % 512) - 256;
	}
}

/*
===============
R_RunParticleEffect

===============
*/
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int msgcount)
{
	int i, j;
	particle_t *p;

	for (i = 0; i < msgcount; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.1 * (rand () % 5);
		p->color = (color & ~7) + (rand () & 7);
		p->type = pt_slowgrav;

		p->org[0] = org[0] + ((rand () & 15) - 8);
		p->org[1] = org[1] + ((rand () & 15) - 8);
		p->org[2] = org[2] + ((rand () & 15) - 8);

		p->vel[0] = dir[0] * 15;
		p->vel[1] = dir[1] * 15;
		p->vel[2] = dir[2] * 15;
	}
}


/*
===============
R_LavaSplash

===============
*/
void R_LavaSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;
	int			c;

	for (c = 0, i = -16; i < 17; i++)
	{
		for (j = -16; j < 17; j++)
		{
			for (k = 0; k < 1; k++, c++)
			{
				if (!free_particles) return;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;
		
				p->die = cl.time + 2 + (rand () & 31) * 0.02;

				if (i == j)
					p->color = 96 + (rand () % 7);
				else p->color = 0 + (rand () % 7);

				p->type = pt_slowgrav;

				dir[0] = j * 8 + (rand () & 7);
				dir[1] = i * 8 + (rand () & 7);
				dir[2] = 256;
	
				p->org[0] = org[0] + dir[0];
				p->org[1] = org[1] + dir[1];
				p->org[2] = org[2] + (rand () & 63);

				VectorNormalize (dir);						
				vel = 50 + (rand () & 63);
				VectorScale (dir, vel, p->vel);

				p->drawtype = P_SMOKE;
				p->radius = 20 + (rand () % 20);
				p->alpha = 0.666;
				p->scale = 0.001;
			}
		}
	}
}

/*
===============
R_TeleportSplash

===============
*/
void R_TeleportSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i = -16; i < 17; i += 8)
	{
		for (j = -16; j < 17; j += 8)
		{
			for (k = -24; k < 33; k += 8)
			{
				if (!free_particles) return;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;
		
				p->die = cl.time + 0.2 + (rand () & 7) * 0.02;
				p->color = 7 + (rand () & 7);
				p->type = pt_slowgrav;

				dir[0] = j * 8;
				dir[1] = i * 8;
				dir[2] = k * 8;

				p->org[0] = org[0] + i + (rand () & 3);
				p->org[1] = org[1] + j + (rand () & 3);
				p->org[2] = org[2] + k + (rand () & 3);

				VectorNormalize (dir);						
				vel = 50 + (rand () & 63);
				VectorScale (dir, vel, p->vel);

				p->drawtype = P_SPARK;
				p->radius = 1.0f + (float) (rand () % 4);
				p->alpha = 1.0f;
			}
		}
	}
}


void R_SmokeTrail (vec3_t start, vec3_t end, int rampscale)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	while (len > 0)
	{
		// decrement faster
		len -= 3;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		// BEGIN SPECIFIC
		// this is the only difference between the 2 types of smoke trail
		p->ramp = (rand () & 3) + rampscale;

		p->color = ramp3[(int) p->ramp];
		p->type = pt_fire;

		p->org[0] = start[0] + ((rand () % 6) - 3);
		p->org[1] = start[1] + ((rand () % 6) - 3);
		p->org[2] = start[2] + ((rand () % 6) - 3);

		p->drawtype = P_SMOKE;
		p->radius = 1.0f;
		p->alpha = 0.0f;
		p->scale = 0.001;
		// END SPECIFIC

		VectorAdd (start, vec, start);
	}
}


void R_BloodTrail (vec3_t start, vec3_t end, int extradec)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	while (len > 0)
	{
		// decrement faster
		len -= (3 + extradec);

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		// BEGIN SPECIFIC
		p->type = pt_bloodtrail;
		p->color = 67 + (rand () & 3);

		p->org[0] = start[0] + ((rand () % 6) - 3);
		p->org[1] = start[1] + ((rand () % 6) - 3);
		p->org[2] = start[2] + ((rand () % 6) - 3);

		p->drawtype = P_BLOOD1 + (rand () % 4);
		p->radius = 2 + (rand () % 5);
		p->alpha = 1.0f;
		p->scale = 0.001f;
		// END SPECIFIC

		VectorAdd (start, vec, start);
	}
}


void R_VoreTrail (vec3_t start, vec3_t end)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	while (len > 0)
	{
		// decrement faster
		len -= 3;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		// BEGIN SPECIFIC
		p->type = pt_static;
		p->die = cl.time + 0.3;
		p->color = 152 + (rand () & 3);

		p->org[0] = start[0] + ((rand () % 15) - 8);
		p->org[1] = start[1] + ((rand () % 15) - 8);
		p->org[2] = start[2] + ((rand () % 15) - 8);
		// END SPECIFIC

		VectorAdd (start, vec, start);
	}
}


void R_GreenGoo (vec3_t start, vec3_t end)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	static int	tracercount = 0;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	while (len > 0)
	{
		// decrement faster
		len -= 3;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		// BEGIN SPECIFIC
		p->die = cl.time + 0.5;
		p->type = pt_static;
		p->color = 52 + ((tracercount & 4) << 1);

		tracercount++;

		VectorCopy (start, p->org);

		if (tracercount & 1)
		{
			p->vel[0] = 30 * vec[1];
			p->vel[1] = 30 * -vec[0];
		}
		else
		{
			p->vel[0] = 30 * -vec[1];
			p->vel[1] = 30 * vec[0];
		}

		p->vel[2] = 0;
		// END SPECIFIC

		VectorAdd (start, vec, start);
	}
}


void R_FireTrail (vec3_t start, vec3_t end)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	static int	tracercount = 0;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);

	while (len > 0)
	{
		// decrement faster
		len -= 3;

		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		// BEGIN SPECIFIC
		p->die = cl.time + 0.5;
		p->type = pt_static;
		p->color = 230 + ((tracercount & 4) << 1);

		tracercount++;

		VectorCopy (start, p->org);

		if (tracercount & 1)
		{
			p->vel[0] = 30 * vec[1];
			p->vel[1] = 30 * -vec[0];
		}
		else
		{
			p->vel[0] = 30 * -vec[1];
			p->vel[1] = 30 * vec[0];
		}

		p->vel[2] = 0;
		// END SPECIFIC

		VectorAdd (start, vec, start);
	}
}


void R_SmokeAndFlashEffect (vec3_t org, int flashcount, int flashcolour, int smokecount, int smokecolour)
{
	int i;
	particle_t *p;

	// add a flash
	for (i = 0; i < flashcount; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 666;
		p->type = pt_expfade;
		p->update = 0;
		p->think = 0.05;

		p->color = flashcolour + (rand () % 7);

		p->org[0] = org[0] - 2 + (rand () % 4);
		p->org[1] = org[1] - 2 + (rand () % 4);
		p->org[2] = org[2] - 2 + (rand () % 4);

		p->vel[0] = 0;
		p->vel[1] = 0;
		p->vel[2] = 0;

		p->drawtype = P_FLASH;
		p->radius = 8 + (rand () % 6);
		p->alpha = 0.3 + (0.01 * (float) (rand () % 20));
		p->scale = 0.0f;
	}

	// and some rising smoke
	for (i = 0; i < smokecount; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		// set the die ludicrously high so that the smoke will always go through the full alpha ramp
		p->die = cl.time + 666;

		// although the type is "gunsmoke" it can be used for any generic rising smoke type
		p->type = pt_gunsmoke;

		p->color = smokecolour + (rand () % 7);

		p->org[0] = org[0] - 2 + (rand () % 5);
		p->org[1] = org[1] - 2 + (rand () % 5);
		p->org[2] = org[2] - 2 + (rand () % 5);

		p->vel[0] = 0;
		p->vel[1] = 0;
		p->vel[2] = 0;

		p->drawtype = P_GUNSMOKE;
		p->radius = 3;
		p->alpha = 1.0f;
		p->scale = 0.0f;
	}
}


void R_FlareSprite (vec3_t org, int colour, float radius, float scale, float alpha)
{
	particle_t *p;

	// no available particles
	if (!free_particles) return;

	// link in
	p = free_particles;
	free_particles = p->next;
	p->next = active_particles;
	active_particles = p;

	// set the die high enough to ensure that it doesn't get acidentally killed
	p->die = cl.time + 666;

	// particle only lasts 1 frame
	p->type = pt_singleframe;

	p->color = colour;

	p->org[0] = org[0];
	p->org[1] = org[1];
	p->org[2] = org[2];

	p->vel[0] = 0;
	p->vel[1] = 0;
	p->vel[2] = 0;

	// flicker the radius
	p->radius = radius;

	p->drawtype = P_FLARE1;
	p->alpha = alpha;
	p->scale = scale;
}


void R_SLightSprite (vec3_t org)
{
	particle_t *p;

	// no available particles
	if (!free_particles) return;

	// link in
	p = free_particles;
	free_particles = p->next;
	p->next = active_particles;
	active_particles = p;

	// set the die high enough to ensure that it doesn't get acidentally killed
	p->die = cl.time + 666;

	// particle only lasts 1 frame
	p->type = pt_singleframe;

	p->color = 108;

	p->org[0] = org[0];
	p->org[1] = org[1];
	p->org[2] = org[2];

	p->vel[0] = 0;
	p->vel[1] = 0;
	p->vel[2] = 0;

	// flicker the radius
	p->radius = 64;

	p->drawtype = P_FLARE2;
	p->alpha = 0.9;
	p->scale = 0.001;
}


void R_BubbleSprite (vec3_t org)
{
	particle_t *p;

	// no available particles
	if (!free_particles) return;

	// link in
	p = free_particles;
	free_particles = p->next;
	p->next = active_particles;
	active_particles = p;

	// set the die high enough to ensure that it doesn't get acidentally killed
	p->die = cl.time + 666;

	// particle only lasts 1 frame
	p->type = pt_singleframe;

	p->color = 46;

	p->org[0] = org[0];
	p->org[1] = org[1];
	p->org[2] = org[2];

	p->vel[0] = 0;
	p->vel[1] = 0;
	p->vel[2] = 0;

	// flicker the radius
	p->radius = 4;

	p->drawtype = P_BUBBLE;
	p->alpha = 0.9;
	p->scale = 0.001;
}


/*
===============
R_DrawParticles
===============
*/
extern	cvar_t	sv_gravity;

#define DRAW_NONE				0
#define DRAW_POINTS				1
#define DRAW_TEXES				2
#define DRAW_SMOKE				4

// we want to draw with 2 seperate blend types so keep 2 chains
glpoly_t *ChainA1 = NULL;
glpoly_t *ChainA1MinusA = NULL;

// remove all particles that have died
// returns 0 if there are point particles only, 
int R_SetupParticles (void)
{
	int i;
	int DrawType = 0;
	particle_t *p, *kill;

	// keep things clean
	polyverts_t *v;

	// scaling
	float scale;
	vec3_t pUp;
	vec3_t pRight;
	vec3_t pOrg;

	ChainA1 = NULL;
	ChainA1MinusA = NULL;

	// set up smokes
	// all we do at this stage is determine if we need to draw it
	for (i = 0; i < cl_numvisedicts; i++)
	{
		// occluded
		if (!cl_visedicts[i]->OcclusionBits) continue;

		if ((cl_visedicts[i]->model->effects & EF_FIREFLARE) || (cl_visedicts[i]->model->effects & EF_TORCHFLARE))
		{
			// draw textured particles
			DrawType |= (DRAW_TEXES | DRAW_SMOKE);

			// one hit is good enough
			break;
		}
	}

	for (;;) 
	{
		kill = active_particles;

		if (kill && kill->die < cl.time)
		{
			// reset stuff on kill
			kill->alpha = 1.0f;
			kill->drawtype = P_POINTS;
			kill->radius = 1.0f;
			kill->scale = 0.001f;

			active_particles = kill->next;
			kill->next = free_particles;
			free_particles = kill;

			continue;
		}

		break;
	}

	for (p = active_particles; p; p = p->next)
	{
		for (;;)
		{
			kill = p->next;

			if (kill && kill->die < cl.time)
			{
				// reset stuff on kill
				kill->alpha = 1.0f;
				kill->drawtype = P_POINTS;
				kill->radius = 1.0f;
				kill->scale = 0.001f;

				p->next = kill->next;
				kill->next = free_particles;
				free_particles = kill;
				continue;
			}

			break;
		}

		// chain them together here
		if (p->drawtype == P_POINTS)
		{
			// notify the caller that we have point particles to draw
			DrawType |= DRAW_POINTS;

			// next particle
			continue;
		}

		// calculate the scale
		scale = (p->org[0] - r_origin[0]) * vpn[0] + (p->org[1] - r_origin[1]) * vpn[1] + (p->org[2] - r_origin[2]) * vpn[2];

		// adjust for scale
		if (scale < 10)
		{
			// don;t draw if too close
			continue;
		}
		else if (scale < 20)
		{
			// normal size
			scale = 1;
		}
		else
		{
			// hack a scale up to keep particles from disapearing
			scale = 1 + scale * p->scale;
		}

		// don;t draw if invisible!!!
		if (p->alpha <= 0) continue;
		if (p->alpha > 1) p->alpha = 1;

		// set the proper chain
		switch (p->drawtype)
		{
		case P_SPARK:
		//case P_SMOKE:
		case P_FLASH:
		case P_BLAST:
		//case P_GUNSMOKE:
		case P_FLARE1:
		case P_FLARE2:
		case P_HIT:
			// GL_SRC_ALPHA, GL_ONE blendfunc
			p->poly->next = ChainA1;
			ChainA1 = p->poly;
			break;

		default:
			// GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA blendfunc
			p->poly->next = ChainA1MinusA;
			ChainA1MinusA = p->poly;
			break;
		}

		// take a pointer to the verts to keep the code clean and avoid multiple indirections
		v = p->poly->verts;

		// calculate the scaled vertex offset factors
		VectorScale (vup, p->radius * scale, pUp);
		VectorScale (vright, p->radius * scale, pRight);

		// these particles will be off-center, so correct that by calcing a new origin
		// to offset the verts from, which is half the offset away in each direction
		pOrg[0] = p->org[0] - (pRight[0] + pUp[0]) / 2;
		pOrg[1] = p->org[1] - (pRight[1] + pUp[1]) / 2;
		pOrg[2] = p->org[2] - (pRight[2] + pUp[2]) / 2;

		// fill in the drawing vertexes
		v[0].tv[0] = pOrg[0];
		v[0].tv[1] = pOrg[1];
		v[0].tv[2] = pOrg[2];

		v[1].tv[0] = pOrg[0] + pUp[0];
		v[1].tv[1] = pOrg[1] + pUp[1];
		v[1].tv[2] = pOrg[2] + pUp[2];

		v[2].tv[0] = v[1].tv[0] + pRight[0];
		v[2].tv[1] = v[1].tv[1] + pRight[1];
		v[2].tv[2] = v[1].tv[2] + pRight[2];

		v[3].tv[0] = pOrg[0] + pRight[0];
		v[3].tv[1] = pOrg[1] + pRight[1];
		v[3].tv[2] = pOrg[2] + pRight[2];

		// fill in the texcoords
		v[0].st[0] = particle_texcoords[p->drawtype].bases;
		v[0].st[1] = particle_texcoords[p->drawtype].baset;

		v[1].st[0] = v[0].st[0] + particle_texcoords[p->drawtype].sadd;
		v[1].st[1] = v[0].st[1];

		v[2].st[0] = v[1].st[0];
		v[2].st[1] = v[0].st[1] + particle_texcoords[p->drawtype].tadd;

		v[3].st[0] = particle_texcoords[p->drawtype].bases;
		v[3].st[1] = v[2].st[1];

		// fill in the colours
		v[0].rgba[0] = v[1].rgba[0] = v[2].rgba[0] = v[3].rgba[0] = d_8to24tableFloat[(int) p->color][0];
		v[0].rgba[1] = v[1].rgba[1] = v[2].rgba[1] = v[3].rgba[1] = d_8to24tableFloat[(int) p->color][1];
		v[0].rgba[2] = v[1].rgba[2] = v[2].rgba[2] = v[3].rgba[2] = d_8to24tableFloat[(int) p->color][2];
		v[0].rgba[3] = v[1].rgba[3] = v[2].rgba[3] = v[3].rgba[3] = p->alpha;

		// notify the caller that we have textured particles to draw
		DrawType |= DRAW_TEXES;
	}

	return DrawType;
}


void R_UpdateParticles (void)
{
	particle_t		*p;
	float			grav;
	int				i;
	float			time4;
	float			time2, time3;
	float			time1;
	float			dvel;
	float			frametime;
	float			halftime;
	float			doubletime;

	frametime = cl.time - cl.oldtime;

	time4 = frametime * 20;
	time3 = frametime * 15;
	time2 = frametime * 10;
	time1 = frametime * 5;
	halftime = frametime / 2;
	doubletime = frametime * 2;

	grav = frametime * sv_gravity.value * 0.05;
	dvel = 4 * frametime;

	for (p = active_particles; p; p = p->next)
	{
		p->org[0] += p->vel[0] * frametime;
		p->org[1] += p->vel[1] * frametime;
		p->org[2] += p->vel[2] * frametime;
		
		switch (p->type)
		{
		case pt_static:
			break;

		case pt_fire:
			p->ramp += time1;

			if (p->ramp >= 6)
				p->die = -1;
			else
				p->color = ramp3[(int) p->ramp];

			p->vel[2] += grav;

			p->alpha += doubletime;
			p->radius += time4;

			if (p->alpha > 0.5f) p->alpha = 0.5f;

			break;

		case pt_explode:
			p->ramp += time2;

			if (p->ramp >= 8)
				p->die = -1;
			else
			{
				p->color = ramp1[(int) p->ramp];

				p->alpha -= halftime;
				if (p->alpha <= 0.0f) p->die = -1;
			}
			break;

		case pt_explode2:
			p->ramp += time3;

			if (p->ramp >= 8)
				p->die = -1;
			else
			{
				p->color = ramp2[(int)p->ramp];

				p->alpha -= halftime;
				if (p->alpha <= 0.0f) p->die = -1;
			}
			break;

		case pt_blob:
			p->vel[0] += p->vel[0] * dvel;
			p->vel[1] += p->vel[1] * dvel;
			p->vel[2] += (p->vel[2] * dvel) - grav;
			break;

		case pt_blob2:
			p->vel[0] -= p->vel[0] * dvel;
			p->vel[1] -= p->vel[1] * dvel;
			p->vel[2] -= (p->vel[2] * dvel) - grav;
			break;

		case pt_gunsmoke:
			// gunsmoke grows, rises and fades
			p->vel[2] += grav;
			p->radius += time2 * 3;
			p->alpha -= frametime;
			if (p->alpha < 0) p->die = -1;
			break;

		case pt_explosmoke:
			// explo smoke grows and fades
			p->vel[2] += (grav / 2);
			p->radius += time2 * 2;
			p->alpha -= (halftime / 2);
			if (p->alpha < 0) p->die = -1;
			break;

		case pt_expfade:
			p->update += frametime;

			if (p->update > p->think)
			{
				p->alpha /= 2.0f;
				p->update = 0.0f;
			}

			if (p->alpha < 0.1) p->die = -1;
			break;

		case pt_singleframe:
			// kill immediately
			p->die = -1;
			break;

		case pt_bloodspray:
			p->alpha -= doubletime;
			if (p->alpha <= 0.0f) p->die = -1;
			p->vel[2] -= grav;
			break;

		case pt_bloodtrail:
			// fade blood and fall through to grav
			p->alpha -= halftime;
			if (p->alpha <= 0.0f) p->die = -1;

		case pt_grav:
		case pt_slowgrav:
			p->vel[2] -= grav;
			break;
		}
	}
}


void R_DrawPointParticles (void)
{
	particle_t *p;
	float attenuations[3] = {0.01, 0, 0.01};

	glPointParameterf (GL_POINT_SIZE_MIN, 2);
	glPointParameterf (GL_POINT_SIZE_MAX, 40);
	glPointParameterfv (GL_POINT_DISTANCE_ATTENUATION, attenuations);

	glDisable (GL_TEXTURE_2D);
	glEnable (GL_POINT_SMOOTH);

	glPointSize (40);

	glBegin (GL_POINTS);
	
	for (p = active_particles; p; p = p->next)
	{
		if (p->drawtype != P_POINTS) continue;

		glColor3ubv ((byte *) &d_8to24tableRGBA[(int) p->color]);
		glVertex3fv (p->org);
	}

	glEnd ();
	glEnable (GL_TEXTURE_2D);
	glDisable (GL_POINT_SMOOTH);
}


void R_SmokeModel (vec3_t org, float ofs)
{
	int i;

	vec3_t pUp;
	vec3_t pRight;
	vec3_t pOrg;

	// these are just containers to keep things neat for the particle currently being drawn
	particle_t p;
	polyverts_t *v = SmokeVerts;

	int numquads = 0;

	for (i = 0; i < NUM_TORCH_SMOKE_PARTICLES; i++)
	{
		// lower-bound alpha
		if (R_SmokeParticles[i].alpha < 0.001) R_SmokeParticles[i].alpha = 0.001;

		p.org[0] = org[0] + R_SmokeParticles[i].org[0];
		p.org[1] = org[1] + R_SmokeParticles[i].org[1];
		p.org[2] = org[2] + (R_SmokeParticles[i].org[2] * 1.25) + ofs;

		// calculate the scaled vertex offset factors
		VectorScale (vup, R_SmokeParticles[i].radius, pUp);
		VectorScale (vright, R_SmokeParticles[i].radius, pRight);

		// these particles will be off-center, so correct that by calcing a new origin
		// to offset the verts from, which is half the offset away in each direction
		pOrg[0] = p.org[0] - (pRight[0] + pUp[0]) / 2;
		pOrg[1] = p.org[1] - (pRight[1] + pUp[1]) / 2;
		pOrg[2] = p.org[2] - (pRight[2] + pUp[2]) / 2;

		// fill in the drawing vertexes
		v[0].tv[0] = pOrg[0];
		v[0].tv[1] = pOrg[1];
		v[0].tv[2] = pOrg[2];

		v[1].tv[0] = pOrg[0] + pUp[0];
		v[1].tv[1] = pOrg[1] + pUp[1];
		v[1].tv[2] = pOrg[2] + pUp[2];

		v[2].tv[0] = v[1].tv[0] + pRight[0];
		v[2].tv[1] = v[1].tv[1] + pRight[1];
		v[2].tv[2] = v[1].tv[2] + pRight[2];

		v[3].tv[0] = pOrg[0] + pRight[0];
		v[3].tv[1] = pOrg[1] + pRight[1];
		v[3].tv[2] = pOrg[2] + pRight[2];

		// fill in the texcoords
		v[0].st[0] = particle_texcoords[P_SMOKE].bases;
		v[0].st[1] = particle_texcoords[P_SMOKE].baset;

		v[1].st[0] = v[0].st[0] + particle_texcoords[P_SMOKE].sadd;
		v[1].st[1] = v[0].st[1];

		v[2].st[0] = v[1].st[0];
		v[2].st[1] = v[0].st[1] + particle_texcoords[P_SMOKE].tadd;

		v[3].st[0] = particle_texcoords[P_SMOKE].bases;
		v[3].st[1] = v[2].st[1];

		// fill in the colours
		v[0].rgba[0] = v[1].rgba[0] = v[2].rgba[0] = v[3].rgba[0] = d_8to24tableFloat[(int) R_SmokeParticles[i].color][0];
		v[0].rgba[1] = v[1].rgba[1] = v[2].rgba[1] = v[3].rgba[1] = d_8to24tableFloat[(int) R_SmokeParticles[i].color][1];
		v[0].rgba[2] = v[1].rgba[2] = v[2].rgba[2] = v[3].rgba[2] = d_8to24tableFloat[(int) R_SmokeParticles[i].color][2];
		v[0].rgba[3] = v[1].rgba[3] = v[2].rgba[3] = v[3].rgba[3] = R_SmokeParticles[i].alpha;

		v += 4;
		numquads += 4;
	}

	// draw particle quads
	// this should be * 4 to be absolutely correct, but it's not.
	// go figure.
	glDrawArrays (GL_QUADS, 0, NUM_TORCH_SMOKE_PARTICLES);
}


void R_DrawSmokeParticles (void)
{
	int i;

	// use a different vertex array for these
	GL_EnableVertexArrays (SmokeVerts, VA_ST | VA_RGBA);

	for (i = 0; i < cl_numvisedicts; i++)
	{
		// occluded
		if (!cl_visedicts[i]->OcclusionBits) continue;

		// different types of smoke...
		if (cl_visedicts[i]->model->effects & EF_FIREFLARE) R_SmokeModel (cl_visedicts[i]->origin, -4);
		if (cl_visedicts[i]->model->effects & EF_TORCHFLARE) R_SmokeModel (cl_visedicts[i]->origin, 4);
	}

	GL_DisableVertexArrays ();
}


void R_DrawParticles (void)
{
	int DrawType;

	// always update these even if none are to be drawn
	R_UpdateTorchSmokeParticles ();

	// setup to draw and remove dead particles
	DrawType = R_SetupParticles ();

	// no particles to draw
	if (!DrawType) return;

	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);

	if (DrawType & DRAW_POINTS) R_DrawPointParticles ();

	if (DrawType & DRAW_TEXES)
	{
		// change the depth range a little to prevent the particles clipping against nearby world geometry
		// by bringing them slightly forward in the z order while retaining their position in the world.
		// this still gives a few artefacts, but it's nowhere near as ghastly as without it
		glDepthRange (0.1, 1.0);

		glBindTexture (GL_TEXTURE_2D, r_particletexture.texnum);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		GL_EnableVertexArrays (ParticleVerts, VA_ST | VA_RGBA);

		if (ChainA1)
		{
			glBlendFunc (GL_SRC_ALPHA, GL_ONE);
			GL_DrawVertexArrays (ChainA1);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		}

		if (ChainA1MinusA)
			GL_DrawVertexArrays (ChainA1MinusA);

		GL_DisableVertexArrays ();

		// restore depth range
		glDepthRange (0.3, 1.0);

		if (DrawType & DRAW_SMOKE)
			R_DrawSmokeParticles ();

		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	}

	glDisable (GL_BLEND);
	glColor4f (1, 1, 1, 1);
	glDepthMask (GL_TRUE);

	// update particles
	R_UpdateParticles ();
}

