/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


/*
INTERFACE DESCRIPTION
---------------------
The following are the PUBLIC functions contained in this file.  They should have prototypes declared in the
appropriate globally accessible header file.

void GL_AllocTexture (int *texnum);
	Allocates a texture number from the texture pool.
*/


#include "quakedef.h"

// includes for image libs
#include "jpeglib.h"
#include "png.h"

// needed for exe resource bitmap loading
#include "resource.h"

// the bsps only have 600-odd textures.  factor in a few more for models and 2D stuff and 1024 should be plenty
#define	MAX_GLTEXTURES	1024

// texture linked list
gltexture_t	*gltextures = NULL;

void MD5_Checksum (unsigned char *data, int dataLen, unsigned char *checksum);
char MD5_Compare (unsigned char *digest1, unsigned char *digest2);
void MD5_Copy (unsigned char *srcdigest, unsigned char *dstdigest);

#define MAX_QUAKE_TEXTURE_SIZE	1024

byte LumaSpecTable[256];

//====================================================================


void GL_TextureInit (void)
{
	int i;
	int j;
	float f;

	// build a table to allow us to pull a luma out of any texture
	for (i = 0; i < 256; i++)
	{
		// take a copy
		f = i;

		// scale 0 to 1
		f /= 255.0f;

		// eliminate darker pixels
		f -= 0.5;
		if (f < 0) f = 0;

		// rescale 0 to 1
		f *= 2.0f;

		// square a few times to get the effect
		f *= f;
		f *= f;

		// rescale 0 to 255 and clamp
		f *= 255.0f;
		f *= 2.0f;
		if (f > 255.0f) f = 255.0f;

		// copy back
		LumaSpecTable[i] = (byte) f;
	}
}


/*
================
GL_AllocTexture

Allocates a texture_extension_number for a texture.

This is currently just a wrapper for glGenTextures and serves no purpose other than to keep all
the texture management code in the one place.
================
*/
void GL_AllocTexture (int *texnum)
{
	glGenTextures (1, texnum);
}


/*
================
GL_FindTexture
================
*/
int GL_FindTexture (char *identifier)
{
	gltexture_t	*glt;

	for (glt = gltextures; glt; glt = glt->next)
	{
		if (!strcmp (identifier, glt->identifier))
			return glt->texnum;
	}

	return -1;
}


gltexture_t *GL_FindGLT (int texnum)
{
	gltexture_t	*glt;

	for (glt = gltextures; glt; glt = glt->next)
	{
		if (glt->texnum == texnum)
			return glt;
	}

	return NULL;
}

/*
================
GL_ResampleTexture
================
*/
typedef struct resamplepoint_s
{
	int p1;
	int p2;
	int p1percent;
	int p2percent;
} resamplepoint_t;


resamplepoint_t resamplew[MAX_QUAKE_TEXTURE_SIZE];
resamplepoint_t resampleh[MAX_QUAKE_TEXTURE_SIZE];

void MakeResamplePoint (resamplepoint_t *resamp, int inscale, int outscale, int outmult)
{
	int i;

	// set up the resample points across
	for (i = 0; i < outscale; i++)
	{
		// calculate the equalivalent point in the "in" buffer
		float fp = (float) (inscale - 1) * ((float) i / (float) (outscale - 1));

		resamp[i].p1 = (int) floor (fp);
		resamp[i].p2 = (int) ceil (fp);

		// the decimal part is the percentage of p2 used
		resamp[i].p2percent = (int) ((fp - (float) resamp[i].p1) * 100.0f);

		// the rest is the percentage of p2 used
		resamp[i].p1percent = 100 - resamp[i].p2percent;

		resamp[i].p1 *= outmult;
		resamp[i].p2 *= outmult;
	}
}


void GL_ResampleTexture (byte *in, int inwidth, int inheight, unsigned *out, int outwidth, int outheight, int bpp)
{
	int h;
	int w;
	int outpoint;
	byte *indata[4];
	byte *outdata;
	int i;
	int inrgba[4];
	int inpercents[4];
	int resampmult;
	int pix;

	if (bpp == 8)
		resampmult = 1;
	else resampmult = 4;

	// set up our resample points in 2 lookup tables for added speed
	MakeResamplePoint (resamplew, inwidth, outwidth, 1 * resampmult);
	MakeResamplePoint (resampleh, inheight, outheight, inwidth * resampmult);

	for (h = 0, outpoint = 0; h < outheight; h++)
	{
		for (w = 0; w < outwidth; w++, outpoint++)
		{
			// get a valid pointer to the outdata
			outdata = (byte *) &out[outpoint];

			// set up the percents
			inpercents[0] = (resamplew[w].p1percent + resampleh[h].p1percent);
			inpercents[1] = (resamplew[w].p2percent + resampleh[h].p1percent);
			inpercents[2] = (resamplew[w].p1percent + resampleh[h].p2percent);
			inpercents[3] = (resamplew[w].p2percent + resampleh[h].p2percent);

			// initially 0
			inrgba[0] = inrgba[1] = inrgba[2] = inrgba[3] = 0;

			// set up the data points
			if (bpp == 8)
			{
				// in-place upsampling and resampling
				// convert to bgra while we're at it
				indata[0] = (byte *) &d_8to24tableBGRA[in[resamplew[w].p1 + resampleh[h].p1]];
				indata[1] = (byte *) &d_8to24tableBGRA[in[resamplew[w].p2 + resampleh[h].p1]];
				indata[2] = (byte *) &d_8to24tableBGRA[in[resamplew[w].p1 + resampleh[h].p2]];
				indata[3] = (byte *) &d_8to24tableBGRA[in[resamplew[w].p2 + resampleh[h].p2]];
			}
			else
			{
				// just resample
				indata[0] = &in[resamplew[w].p1 + resampleh[h].p1];
				indata[1] = &in[resamplew[w].p2 + resampleh[h].p1];
				indata[2] = &in[resamplew[w].p1 + resampleh[h].p2];
				indata[3] = &in[resamplew[w].p2 + resampleh[h].p2];
			}

			// accumulate the data for each point
			for (pix = 0; pix < 4; pix++)
			{
				// accumulate as ints to avoid byte data type wrapping
				inrgba[pix] += (int) indata[0][pix] * inpercents[0];
				inrgba[pix] += (int) indata[1][pix] * inpercents[1];
				inrgba[pix] += (int) indata[2][pix] * inpercents[2];
				inrgba[pix] += (int) indata[3][pix] * inpercents[3];

				// copy across
				outdata[pix] = (byte) (inrgba[pix] / 400);
			}
		}
	}
}


/*
===============
GL_UploadTexture
===============
*/
unsigned *scaled = NULL;
byte *externaldata = NULL;
int externalsize = 0;


void GL_UploadTextureObject (int texnum, unsigned *data, GLenum format, int width, int height, qboolean mipmap)
{
	// bind the texture object
	glBindTexture (GL_TEXTURE_2D, texnum);

	// hardware mipmapping (OpenGL 1.4 or higher an absolute must) 
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, mipmap ? GL_TRUE : GL_FALSE);

	// don't bother compressing cos Quake textures are fairly light on the old video ram
	// always take it up in RGBA format cos that's what OpenGL uses (or at least is supposed
	// to use, according to the specs) internally anyway.  take 8 bit texes in bgra
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, format, GL_UNSIGNED_BYTE, data);

	// apply proper filtering
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mipmap ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY, mipmap ? GL_Caps.GL_MaxAnisotropy / 2.0f : 0.0f);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
}


void GL_AlphaMaskTexture (byte *data, int width, int height)
{
	int i;

	for (i = 0; i < width * height; i++, data += 4)
	{
		// set alpha channel to the value of the red channel
		// (it's assumed these are greyscale)
		data[3] = data[0];

		// set rgb to white
		data[0] = 255;
		data[1] = 255;
		data[2] = 255;
	}
}


void SetImagePointer (int currsize)
{
	// do we need to allocate a new buffer?
	if (currsize > externalsize && externaldata)
	{
		// yes - release the current buffer and NULL it
		free (externaldata);
		externaldata = NULL;
		externalsize = 0;
	}

	// if we don't have a buffer pointer, set one up to hold a 32 bit tex
	// set the size too...
	if (!externaldata)
	{
		externaldata = malloc (currsize * sizeof (int));
		externalsize = currsize;
	}

	// satan's cock
	if (!externaldata) Sys_Error ("Failed to allocate %i KB RAM", currsize * sizeof (int) / 1024);
}


float sqrttable[256] = {-1};

float GSPixels (byte *rgba)
{
	float gs;
	int i, j;
	int ntsc[] = {30, 59, 11};
	int total = 0;

	// init the sqrttable the first time we come in here
	if (sqrttable[0] < 0)
		for (i = 0; i < 256; i++)
			sqrttable[i] = sqrt ((float) i / 1020.0f);

	/*
	for (i = 0; i < 3; i++)
	{
		// square for height intensity
		j = rgba[i];
		j *= (int) rgba[i];

		// accumulate to total
		total += (j >> 8) * ntsc[i];
	}

	return sqrttable[total / 100];
	*/

	for (i = 0; i < 3; i++)
	{
		// square for height intensity
		j = rgba[i];

		// accumulate to total
		total += j * ntsc[i];
	}

	return sqrttable[total / 100];
}


/*
=============================
MakeDOT3

Convert a height field image into a normal map. This involves
differencing each texel with its right and upper neighbor, then
normalizing the cross product of the two difference vectors.

Original code from "Practical and robust bumpmapping" hacked to death by me...
=============================
*/
unsigned short ExpandW[MAX_QUAKE_TEXTURE_SIZE + 1];
unsigned int ExpandH[MAX_QUAKE_TEXTURE_SIZE + 1];

void MakeDOT3 (gltexture_t *glt, unsigned int *src, unsigned int *dst, int w, int h)
{
	int i, j;
	float sqlen, reciplen, nx, ny, nz, c, cx, cy, dcx, dcy;
	float scale = 3;
	byte *dstrgba = (byte *) dst;
	unsigned int pix;

	// set up the expansion tables - height
	// always multiplies by W
	for (i = 0; i < h; i++)
		ExpandH[i] = i * w;

	// last element wraps to the first
	ExpandH[h] = 0;
	ExpandH[h + 1] = w;
	ExpandH[h + 2] = w * 2;

	// set up the expansion tables - width
	for (j = 0; j < w; j++)
		ExpandW[j] = j;

	// last element wraps to the first
	ExpandW[w] = 0;
	ExpandW[w + 1] = 1;
	ExpandW[w + 2] = 2;

	for (i = 0; i < h; i += 2)
	{
		for (j = 0; j < w; j += 2, dstrgba += 4)
		{
			// Expand [0,255] texel values to the [0,1] range.
			c = GSPixels ((byte *) &src[ExpandH[i] + ExpandW[j]]);
			c += GSPixels ((byte *) &src[ExpandH[i + 1] + ExpandW[j]]);
			c += GSPixels ((byte *) &src[ExpandH[i + 1] + ExpandW[j + 1]]);
			c += GSPixels ((byte *) &src[ExpandH[i] + ExpandW[j + 1]]);

			// Expand the texel to its right.
			cx = GSPixels ((byte *) &src[ExpandH[i] + ExpandW[j + 2]]);
			cx += GSPixels ((byte *) &src[ExpandH[i + 1] + ExpandW[j + 2]]);
			cx += GSPixels ((byte *) &src[ExpandH[i + 1] + ExpandW[j + 3]]);
			cx += GSPixels ((byte *) &src[ExpandH[i] + ExpandW[j + 3]]);

			// Expand the texel one down.
			cy = GSPixels ((byte *) &src[ExpandH[i + 2] + ExpandW[j]]);
			cy += GSPixels ((byte *) &src[ExpandH[i + 3] + ExpandW[j]]);
			cy += GSPixels ((byte *) &src[ExpandH[i + 3] + ExpandW[j + 1]]);
			cy += GSPixels ((byte *) &src[ExpandH[i + 2] + ExpandW[j + 1]]);

			dcx = scale * (c - cx);
			dcy = scale * (c - cy);

			// Normalize the vector.  (nz is 1 so add 1 * 1 here...)
			sqlen = dcx * dcx + dcy * dcy + 1;
			reciplen = 1.0f / sqrt (sqlen);

			nx = dcy * reciplen;
			ny = -dcx * reciplen;
			nz = reciplen;

			// Repack the normalized vector into an RGB unsigned byte
			// vector in the normal map image.
			dstrgba[2] = 128 + 127 * nx;
			dstrgba[1] = 128 + 127 * ny;
			dstrgba[0] = 128 + 127 * nz;
		}
	}
}


void GL_UploadTexture (gltexture_t *glt, unsigned *data, int width, int height, int bpp, int flags)
{
	int i;
	int scaled_width, scaled_height;
	byte *pix;

	// allocate memory for scaled buffer as needed
	// note: we deliberately cap the max usable texture size below the max size supported by the card
	// because otherwise we could have very silly texture sizes here. 
	if (!scaled) scaled = malloc (sizeof (int) * MAX_QUAKE_TEXTURE_SIZE * MAX_QUAKE_TEXTURE_SIZE);

	if (!scaled) Sys_Error ("Failed to allocate %i RAM for scaled texture buffer", sizeof (int) * MAX_QUAKE_TEXTURE_SIZE * MAX_QUAKE_TEXTURE_SIZE);

	// get power of 2 dimensions
	for (scaled_width = 1; scaled_width < width; scaled_width *= 2);
	for (scaled_height = 1; scaled_height < height; scaled_height *= 2);

	// cap at the max size supported by the implementation
	if (scaled_width > GL_Caps.GL_MaxTextureSize) scaled_width = GL_Caps.GL_MaxTextureSize;
	if (scaled_height > GL_Caps.GL_MaxTextureSize) scaled_height = GL_Caps.GL_MaxTextureSize;

	// cap at the max size we can be bothered supporting
	if (scaled_width > MAX_QUAKE_TEXTURE_SIZE) scaled_width = MAX_QUAKE_TEXTURE_SIZE;
	if (scaled_height > MAX_QUAKE_TEXTURE_SIZE) scaled_height = MAX_QUAKE_TEXTURE_SIZE;

	// expand to 32 bpp and/or resample, as required
	// if both are required, they are both done in a single pass
	if (bpp == 32 && scaled_width == width && scaled_height == height)
	{
		// copy it straight in
		for (i = 0; i < width * height; i++) 
		{
			scaled[i] = data[i];
		}
	}
	else if (bpp == 8 && scaled_width == width && scaled_height == height)
	{
		// convert to bytes
		byte *data8bpp = (byte *) data;

		// expand 8bpp to 32 bpp
		for (i = 0; i < width * height; i++)
		{
			scaled[i] = d_8to24tableBGRA[data8bpp[i]];
		}
	}
	else GL_ResampleTexture ((byte *) data, width, height, scaled, scaled_width, scaled_height, bpp);

	// certain flags require specific actions before we can proceed with the upload
	if (flags & TEX_ALPHAMASK) GL_AlphaMaskTexture ((byte *) scaled, scaled_width, scaled_height);

	GL_UploadTextureObject (glt->texnum, scaled, bpp == 8 ? GL_BGRA : GL_RGBA, scaled_width, scaled_height, (flags & TEX_MIPMAP));

	if (flags & TEX_LUMASPEC)
	{
		qboolean lumaOK = false;

		if (bpp == 8 && scaled_width == width && scaled_height == height)
		{
			// rebuild the texture into the scaled buffer
			byte *data8bpp = (byte *) data;
			byte *pix2;

			// expand 8bpp to 32 bpp
			for (i = 0; i < width * height; i++)
			{
				// copy in the regular fullbright
				if (data8bpp[i] > 223)
					scaled[i] = d_8to24tableBGRA[data8bpp[i]];
				else scaled[i] = 0;

				pix2 = (byte *) &d_8to24tableBGRA[data8bpp[i]];
				pix = (byte *) &scaled[i];

				// take the brighter of the 2
				if (LumaSpecTable[pix2[0]] > pix[0]) pix[0] = LumaSpecTable[pix2[0]];
				if (LumaSpecTable[pix2[1]] > pix[1]) pix[1] = LumaSpecTable[pix2[1]];
				if (LumaSpecTable[pix2[2]] > pix[2]) pix[2] = LumaSpecTable[pix2[2]];

				if (pix[0] > 0 || pix[1] > 0 || pix[2] > 0) lumaOK = true;
			}
		}
		else if (bpp == 8)
		{
			byte *data8bpp = (byte *) data;
			byte *pix2;
			unsigned int *data32bpp;

			// take the external image memory here, and increase it by 2 so that we have some
			// headroom for future use
			SetImagePointer (width * height * 2);

			// take a copy as unsigned ints so we can array-index it properly
			data32bpp = (unsigned int *) externaldata;

			// expand to 32bpp into the external memory, converting to luma pixels as we go
			for (i = 0; i < width * height; i++)
			{
				// copy in the regular fullbright
				if (data8bpp[i] > 223)
					data32bpp[i] = d_8to24tableBGRA[data8bpp[i]];
				else data32bpp[i] = 0;

				pix2 = (byte *) &d_8to24tableBGRA[data8bpp[i]];
				pix = (byte *) &data32bpp[i];

				// take the brighter of the 2
				if (LumaSpecTable[pix2[0]] > pix[0]) pix[0] = LumaSpecTable[pix2[0]];
				if (LumaSpecTable[pix2[1]] > pix[1]) pix[1] = LumaSpecTable[pix2[1]];
				if (LumaSpecTable[pix2[2]] > pix[2]) pix[2] = LumaSpecTable[pix2[2]];

				if (pix[0] > 0 || pix[1] > 0 || pix[2] > 0) lumaOK = true;
			}

			// now resample the converted image
			GL_ResampleTexture (externaldata, width, height, scaled, scaled_width, scaled_height, 32);
		}
		else
		{
			// change the colours
			for (i = 0; i < scaled_width * scaled_height; i++)
			{
				pix = (byte *) &scaled[i];

				pix[0] = LumaSpecTable[pix[0]];
				pix[1] = LumaSpecTable[pix[1]];
				pix[2] = LumaSpecTable[pix[2]];

				if (pix[0] > 0 || pix[1] > 0 || pix[2] > 0) lumaOK = true;
			}
		}

		lumaOK = false;

		// don't make a luma if there are no luma pixels
		if (!lumaOK)
			glt->lumatexnum = 0;
		else GL_UploadTextureObject (glt->lumatexnum, scaled, bpp == 8 ? GL_BGRA : GL_RGBA, scaled_width, scaled_height, (flags & TEX_MIPMAP));
	}
}


/*
===============
GL_LoadPNG

PNG Loading
===============
*/
qboolean GL_LoadPNG (gltexture_t *glt, FILE *f)
{
	unsigned char header[8], **rowpointers;
	png_structp	png;
	png_infop	pnginfo;
	int		y, width, height, bitdepth, colortype;
	int		interlace, compression, filter, bytesperpixel;
	unsigned long	rowbytes;
	int w, h;

	fread (header, 1, 8, f);

	if (png_sig_cmp (header, 0, 8)) return false;

	if (!(png = png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL))) return false;

	if (!(pnginfo = png_create_info_struct (png)))
	{
		png_destroy_read_struct (&png, &pnginfo, NULL);
		return false;
	}

	if (setjmp (png->jmpbuf))
	{
		png_destroy_read_struct (&png, &pnginfo, NULL);
		return false;
	}

	png_init_io (png, f);
	png_set_sig_bytes (png, 8);
	png_read_info (png, pnginfo);
	png_get_IHDR (png, pnginfo, (png_uint_32 *) &width, (png_uint_32 *) &height, &bitdepth, &colortype, &interlace, &compression, &filter);

	if (colortype == PNG_COLOR_TYPE_PALETTE)
	{
		png_set_palette_to_rgb (png);
		png_set_filler (png, 255, PNG_FILLER_AFTER);			
	}

	if (colortype == PNG_COLOR_TYPE_GRAY && bitdepth < 8)
		png_set_gray_1_2_4_to_8 (png);
	
	if (png_get_valid(png, pnginfo, PNG_INFO_tRNS))
		png_set_tRNS_to_alpha (png);

	if (colortype == PNG_COLOR_TYPE_GRAY || colortype == PNG_COLOR_TYPE_GRAY_ALPHA)
	{
		png_set_gray_to_rgb (png);
		png_set_filler (png, 255, PNG_FILLER_AFTER);			
	}

	if (colortype != PNG_COLOR_TYPE_RGBA)
		png_set_filler (png, 255, PNG_FILLER_AFTER);
	
	if (bitdepth < 8)
		png_set_expand (png);
	else if (bitdepth == 16)
		png_set_strip_16 (png);

	png_read_update_info (png, pnginfo);
	rowbytes = png_get_rowbytes (png, pnginfo);
	bytesperpixel = png_get_channels (png, pnginfo);
	bitdepth = png_get_bit_depth (png, pnginfo);

	// fill in properties
	// just using width and height are *not* good here...!
	// rowbytes / 4 should also be good for use here, but we'll do it the official way
	w = png_get_image_width (png, pnginfo);
	h = png_get_image_height (png, pnginfo);

	if (bitdepth != 8 || bytesperpixel != 4)
	{	
		png_destroy_read_struct (&png, &pnginfo, NULL);
		return false;
	}

	// allocate an extra row for the rowpointers
	SetImagePointer ((w + 1) * h);

	// set up the row pointers
	rowpointers = (unsigned char **) &externaldata[w * h * 4];

	// and for each row
	for (y = 0; y < height; y++)
		rowpointers[y] = externaldata + y * rowbytes;

	// read it all in
	png_read_image (png, rowpointers);
	png_read_end (png, NULL);

	png_destroy_read_struct (&png, &pnginfo, NULL);

	// upload
	glt->flags |= TEX_EXTERNAL;
	GL_UploadTexture (glt, externaldata, w, h, 32, glt->flags);

	return true;
}


/*
=============
GL_LoadTGA

TGA Loading stuff
=============
*/
typedef struct TGAHeader_s
{
	unsigned char id_length;
	unsigned char colormap_type;
	unsigned char image_type;
	unsigned short colormap_index;
	unsigned short colormap_length;
	unsigned char colormap_size;
	unsigned short x_origin;
	unsigned short y_origin;
	unsigned short width;
	unsigned short height;
	unsigned char pixel_size;
	unsigned char attributes;
} TGAHeader_t;


int TGAfgetLittleShort (FILE *f)
{
	unsigned char b1 = fgetc (f);
	unsigned char b2 = fgetc (f);

	return (short) (b1 + b2 * 256);
}


qboolean GL_LoadTGA (gltexture_t *glt, FILE *f)
{
	TGAHeader_t targa_header;
	int columns;
	int rows;
	int numPixels;
	unsigned char *pixbuf;
	int row, column;
	int flipper;

	// read the header
	targa_header.id_length = fgetc (f);
	targa_header.colormap_type = fgetc (f);
	targa_header.image_type = fgetc (f);

	targa_header.colormap_index = TGAfgetLittleShort (f);
	targa_header.colormap_length = TGAfgetLittleShort (f);
	targa_header.colormap_size = fgetc (f);
	targa_header.x_origin = TGAfgetLittleShort (f);
	targa_header.y_origin = TGAfgetLittleShort (f);
	targa_header.width = TGAfgetLittleShort (f);
	targa_header.height = TGAfgetLittleShort (f);
	targa_header.pixel_size = fgetc (f);
	targa_header.attributes = fgetc (f);

	// they call him flipper, flipper, faster than lightning, no-one you see, is smarter than he.
	// thanks to shadow for the heads up on this.
	flipper = !((targa_header.attributes & 0x20) >> 5);

	// validate
	if (targa_header.image_type != 2 && targa_header.image_type != 10) return false;
	if (targa_header.colormap_type != 0 || (targa_header.pixel_size != 32 && targa_header.pixel_size != 24)) return false;

	columns = targa_header.width;
	rows = targa_header.height;
	numPixels = columns * rows;

	// the image will presumably be OK now...
	SetImagePointer (numPixels);

	// skip TARGA image comment
	if (targa_header.id_length != 0) fseek (f, targa_header.id_length, SEEK_CUR);

	if (targa_header.image_type == 2)
	{
		// Uncompressed, RGB images
		for (row = rows - 1; row >= 0; row--)
		{
			if (flipper)
				pixbuf = externaldata + row * columns * 4;
			else pixbuf = externaldata + ((row - 1) - row) * columns * 4;

			for (column = 0; column < columns; column++)
			{
				unsigned char blue = fgetc (f);
				unsigned char green = fgetc (f);
				unsigned char red = fgetc (f);
				unsigned char alphabyte;

				*pixbuf++ = red;
				*pixbuf++ = green;
				*pixbuf++ = blue;

				switch (targa_header.pixel_size)
				{
				case 24:
					*pixbuf++ = 255;
					break;

				case 32:
					alphabyte = fgetc (f);
					*pixbuf++ = alphabyte;
					break;
				}
			}
		}
	}
	else if (targa_header.image_type == 10)
	{
		// Runlength encoded RGB images
		unsigned char red;
		unsigned char green;
		unsigned char blue;
		unsigned char alphabyte;
		unsigned char packetHeader;
		unsigned char packetSize;
		unsigned char j;

		for (row = rows - 1; row >= 0; row--)
		{
			if (flipper)
				pixbuf = externaldata + row * columns * 4;
			else pixbuf = externaldata + ((row - 1) - row) * columns * 4;

			for (column = 0; column < columns;)
			{
				packetHeader = fgetc (f);
				packetSize = 1 + (packetHeader & 0x7f);

				if (packetHeader & 0x80)
				{
					// run-length packet
					blue = fgetc (f);
					green = fgetc (f);
					red = fgetc (f);

					switch (targa_header.pixel_size)
					{
					case 24:
						alphabyte = 255;
						break;

					case 32:
						alphabyte = fgetc (f);
						break;
					}
	
					for (j = 0; j < packetSize; j++)
					{
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = alphabyte;

						column++;

						if (column == columns)
						{
							// run spans across rows
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							if (flipper)
								pixbuf = externaldata + row * columns * 4;
							else pixbuf = externaldata + ((row - 1) - row) * columns * 4;
						}
					}
				}
				else
				{
					// non run-length packet
					for (j = 0; j < packetSize; j++)
					{
						blue = fgetc (f);
						green = fgetc (f);
						red = fgetc (f);

						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;

						switch (targa_header.pixel_size)
						{
						case 24:
							*pixbuf++ = 255;
							break;

						case 32:
							alphabyte = fgetc (f);
							*pixbuf++ = alphabyte;
							break;
						}

						column++;

						if (column == columns)
						{
							// pixel packet run spans across rows
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							if (flipper)
								pixbuf = externaldata + row * columns * 4;
							else pixbuf = externaldata + ((row - 1) - row) * columns * 4;
						}
					}
				}
			}

			breakOut:;
		}
	}

	// upload
	glt->flags |= TEX_EXTERNAL;
	GL_UploadTexture (glt, externaldata, targa_header.width, targa_header.height, 32, glt->flags);

	// success
	return true;
}


/*
=================
GL_LoadBMP

BMP Loading Malarkey.  God, this format is horrible...!
=================
*/
#ifndef BI_RGB
#define BI_RGB 0L
#endif

#ifndef BI_RLE8
#define BI_RLE8 1L
#endif

// MH - preferred to ID's version in the q3a source as mine supports compression and 4bpp
qboolean GL_LoadBMP (gltexture_t *glt, FILE *f)
{
	int i, j;
	int run;
	int runlength;
	unsigned char bmIn;
	BITMAPINFO bmInfo;
	BITMAPFILEHEADER bmHeader;
	RGBQUAD bmPalette[256];
	unsigned char *bmData;
	int pad;

	// read the file header
	fread (&bmHeader, 1, sizeof (bmHeader), f);

	// validate the type
	if (bmHeader.bfType != 0x4D42) return false;

	// read the BITMAPINFO
	fread (&bmInfo, 1, sizeof (bmInfo), f);

	if (bmInfo.bmiHeader.biCompression != BI_RGB)
	{
		if (!(bmInfo.bmiHeader.biCompression == BI_RLE8 && bmInfo.bmiHeader.biBitCount == 8))
			return false;
	}

	// support 4, 8 and 24 BPP modes
	if (bmInfo.bmiHeader.biBitCount != 4 && 
		bmInfo.bmiHeader.biBitCount != 8 && 
		bmInfo.bmiHeader.biBitCount != 24) return false;

	SetImagePointer (bmInfo.bmiHeader.biWidth * abs (bmInfo.bmiHeader.biHeight));
	bmData = externaldata;

	if (bmInfo.bmiHeader.biBitCount != 24)
	{
		// read the palette entries
		// the first one is in the BITMAPINFO struct
		bmPalette[0].rgbBlue = bmInfo.bmiColors[0].rgbBlue;
		bmPalette[0].rgbGreen = bmInfo.bmiColors[0].rgbGreen;
		bmPalette[0].rgbRed = bmInfo.bmiColors[0].rgbRed;
		bmPalette[0].rgbReserved = bmInfo.bmiColors[0].rgbReserved;

		// remaining entries follow the struct
		for (i = 1; i < bmInfo.bmiHeader.biClrUsed; i++)
		{
			fread (&bmPalette[i], 1, sizeof (RGBQUAD), f);
		}
	}

	// each row in a bitmap is padded to align it on a 4 unsigned char boundary, so find
	// out how much padding we need (this may not work with RLE images)
	for (i = 0; ; i++)
	{
		pad = i * 4;
		if (pad >= (bmInfo.bmiHeader.biWidth * 3)) break;
	}

	pad -= bmInfo.bmiHeader.biWidth * 3;

	// now seek to the proper offset of the data
	fseek (f, bmHeader.bfOffBits, SEEK_SET);

	// load the data
	for (i = 0; i < bmInfo.bmiHeader.biWidth * abs (bmInfo.bmiHeader.biHeight); i++)
	{
		if (bmInfo.bmiHeader.biBitCount == 4)
		{
			bmIn = fgetc (f);

			bmData[0] = bmPalette[bmIn >> 4].rgbRed;
			bmData[1] = bmPalette[bmIn >> 4].rgbGreen;
			bmData[2] = bmPalette[bmIn >> 4].rgbBlue;
			bmData[3] = 255;

			bmData[4] = bmPalette[bmIn % 16].rgbRed;
			bmData[5] = bmPalette[bmIn % 16].rgbGreen;
			bmData[6] = bmPalette[bmIn % 16].rgbBlue;
			bmData[7] = 255;

			bmData += 8;

			// increment i again cos we're reading two entires at a time
			i++;

			continue;
		}
		else if (bmInfo.bmiHeader.biBitCount == 8)
		{
			if (bmInfo.bmiHeader.biCompression == BI_RLE8)
			{
				bmIn = fgetc (f);

				if (bmIn == 0)
				{
					bmIn = fgetc (f);
					i++;

					switch (bmIn)
					{
					case 0:
						// End of Line - just do nothing - trust me, it works.
						break;
					case 1:
						// End of Bitmap
						goto bmDone;
						break;
					case 2:
						// Delta
						// (tests indicate this isn't actually used... most likely cos the bitmap format
						// doesn't specify what the skipped pixels are to be filled with!!!)
						bmIn = fgetc (f);
						i++;

						bmData += bmIn * 4;

						bmIn = fgetc (f);
						i++;

						bmData += bmIn * bmInfo.bmiHeader.biWidth * 4;

						break;
					default:
						// absolute mode
						runlength = bmIn;

						for (run = 0; run < runlength; run++)
						{
							bmIn = fgetc (f);
							i++;

							bmData[0] = bmPalette[bmIn].rgbRed;
							bmData[1] = bmPalette[bmIn].rgbGreen;
							bmData[2] = bmPalette[bmIn].rgbBlue;
							bmData[3] = 255;

							bmData += 4;
						}

						if (runlength % 2)
						{
							// align on word boundaries
							bmIn = fgetc (f);
							i++;
						}

						break;
					}
				}
				else
				{
					// encoded mode
					runlength = bmIn;

					bmIn = fgetc (f);
					i++;

					for (run = 0; run < runlength; run++)
					{
						bmData[0] = bmPalette[bmIn].rgbRed;
						bmData[1] = bmPalette[bmIn].rgbGreen;
						bmData[2] = bmPalette[bmIn].rgbBlue;
						bmData[3] = 255;

						bmData += 4;
					}
				}

				continue;
			}

			bmIn = fgetc (f);

			bmData[0] = bmPalette[bmIn].rgbRed;
			bmData[1] = bmPalette[bmIn].rgbGreen;
			bmData[2] = bmPalette[bmIn].rgbBlue;
			bmData[3] = 255;

			bmData += 4;

			continue;
		}
		else
		{
			// 24 BPP
			bmData[2] = fgetc (f);
			bmData[1] = fgetc (f);
			bmData[0] = fgetc (f);
			bmData[3] = 255;

			bmData += 4;
		}

		// read the padding
		for (j = 0; j < pad; j++) fgetc (f);
	}

bmDone:;

	// check orientation
	if (bmInfo.bmiHeader.biHeight > 0)
	{
		// bottom-up bitmap so we must invert it (sigh)
		for (i = 0; i < abs (bmInfo.bmiHeader.biHeight) / 2; i++)
		{
			for (j = 0; j < bmInfo.bmiHeader.biWidth; j++)
			{
				unsigned int srcpos = (i * bmInfo.bmiHeader.biWidth) + j;
				unsigned int dstpos = (((abs (bmInfo.bmiHeader.biHeight) - 1) - i) * bmInfo.bmiHeader.biWidth) + j;
				unsigned int tmp;

				tmp = ((unsigned int *) externaldata)[dstpos];
				((unsigned int *) externaldata)[dstpos] = ((unsigned int *) externaldata)[srcpos];
				((unsigned int *) externaldata)[srcpos] = tmp;
			}
		}
	}

	// upload
	glt->flags |= TEX_EXTERNAL;
	GL_UploadTexture (glt, externaldata, bmInfo.bmiHeader.biWidth, abs (bmInfo.bmiHeader.biHeight), 32, glt->flags);

	return true;
}


/*
===============
GL_LoadJPG

JPG Loading
===============
*/
qboolean GL_LoadJPG (gltexture_t *glt, FILE *f)
{
	struct jpeg_decompress_struct cinfo;
	JDIMENSION num_scanlines;
	JSAMPARRAY in;
	struct jpeg_error_mgr jerr;
	int numPixels;
	int row_stride;
	unsigned char *out;
	int count;
	int i;
	int r, g, b;

	// set up the decompression.
	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress (&cinfo);

	// inititalize the source
	jpeg_stdio_src (&cinfo, f);

	// initialize decompression
	(void) jpeg_read_header (&cinfo, TRUE);
	(void) jpeg_start_decompress (&cinfo);

	numPixels = cinfo.image_width * cinfo.image_height;

	// initialize the input buffer - we'll use the in-built memory management routines in the
	// JPEG library because it will automatically free the used memory for us when we destroy
	// the decompression structure.  cool.
	row_stride = cinfo.output_width * cinfo.output_components;
	in = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

	// bit of error checking
	if (cinfo.output_components != 3) goto error;
	if ((numPixels * 4) != ((row_stride * cinfo.output_height) + numPixels)) goto error;

	// read the jpeg
	count = 0;

	// initialize the return data
	externaldata = malloc ((numPixels * 4));

	while (cinfo.output_scanline < cinfo.output_height) 
	{
		num_scanlines = jpeg_read_scanlines(&cinfo, in, 1);
		out = in[0];

		for (i = 0; i < row_stride;)
		{
			r = externaldata[count++] = out[i++];
			g = externaldata[count++] = out[i++];
			b = externaldata[count++] = out[i++];

			// alpha is always 255 with JPEGs
			externaldata[count++] = 255;
		}
	}

	// finish decompression and destroy the jpeg
	(void) jpeg_finish_decompress (&cinfo);

	// upload
	glt->flags |= TEX_EXTERNAL;
	GL_UploadTexture (glt, externaldata, cinfo.image_width, cinfo.image_height, 32, glt->flags);

	return true;

error:;
	// have to do this rather than a simple "return false" cos we need to destroy the reader too!!!
	jpeg_destroy_decompress (&cinfo);
	return false;
}


/*
===============
GL_LoadPCX

PCX Loading
===============
*/
typedef struct pcx_s
{
	char manufacturer;
	char version;
	char encoding;
	char bits_per_pixel;

	unsigned short xmin;
	unsigned short ymin;
	unsigned short xmax;
	unsigned short ymax;

	unsigned short hres;
	unsigned short vres;

	unsigned char palette[48];

	char reserved;
	char color_planes;

	unsigned short bytes_per_line;
	unsigned short palette_type;

	char filler[58];

	unsigned int data;			// unbounded
} pcx_t;


qboolean GL_LoadPCX (gltexture_t *glt, FILE *f)
{
	pcx_t *pcx, pcxbuf;
	unsigned char palette[768];
	unsigned char *pix;
	int x, y;
	int dataByte, runLength;
	int count;

	// parse the PCX file
	fread (&pcxbuf, 1, sizeof (pcxbuf), f);

	pcx = &pcxbuf;

	if (pcx->manufacturer != 0x0a || pcx->version != 5 || pcx->encoding != 1 || pcx->bits_per_pixel != 8 || pcx->xmax >= 1024 || pcx->ymax >= 1024)
		return false;

	// seek to palette
	fseek (f, -768, SEEK_END);
	fread (palette, 1, 768, f);

	fseek (f, sizeof (pcxbuf) - 4, SEEK_SET);

	count = (pcx->xmax + 1) * (pcx->ymax + 1);
	SetImagePointer (count);

	for (y = 0; y <= pcx->ymax; y++)
	{
		pix = externaldata + 4 * y * (pcx->xmax + 1);

		for (x = 0; x <= pcx->xmax;)
		{
			dataByte = fgetc (f);

			if ((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				dataByte = fgetc (f);
			}
			else runLength = 1;

			while (runLength-- > 0)
			{
				pix[0] = palette[dataByte * 3];
				pix[1] = palette[dataByte * 3 + 1];
				pix[2] = palette[dataByte * 3 + 2];

				// PCX files never have alpha
				pix[3] = 255;
				pix += 4;
				x++;
			}
		}
	}

	// upload
	glt->flags |= TEX_EXTERNAL;
	GL_UploadTexture (glt, externaldata, pcx->xmax + 1, pcx->ymax + 1, 32, glt->flags);

	return true;
}


typedef struct imageformats_s
{
	char *extension;
	qboolean (*Texloader) (gltexture_t *, FILE *);
} imageformats_t;

imageformats_t imageformats[] =
{
	// image loaders. jpg is twice cos the jpeg extension is also valid
	{"png", GL_LoadPNG},
	{"tga", GL_LoadTGA},
	{"bmp", GL_LoadBMP},
	{"jpg", GL_LoadJPG},
	{"jpeg", GL_LoadJPG},
	{"pcx", GL_LoadPCX},

	// always end the list with this
	{NULL, NULL},
};

qboolean GL_TryExternalTexture (gltexture_t *glt, char *path)
{
	char tempident[MAX_OSPATH];
	char textureloadname[MAX_OSPATH];
	int i;
	int j;
	FILE *f;

	// copy the name cos we're going to modify it
	strcpy (tempident, glt->identifier);

	// replace the water identification character
	if (tempident[0] == '*') tempident[0] = '#';

	// remove the extension
	for (i = strlen (tempident) - 1; i; i--)
	{
		if (tempident[i] == '.')
		{
			tempident[i] = 0;
			break;
		}
	}

	// try each format
	for (i = 0; ; i++)
	{
		// end of the list
		if (!imageformats[i].extension) return false;

		sprintf (textureloadname, "%s%s.%s", path, tempident, imageformats[i].extension);

		// attempt to open
		COM_FOpenFile (textureloadname, &f);

		if (f)
		{
			if (imageformats[i].Texloader (glt, f))
			{
				// it worked
				fclose (f);
				return true;
			}

			// didn't work
			fclose (f);
		}
	}

	// never gets here, but the compiler must be satisfied
	return false;
}


/*
================
GL_LoadTexture

Somewhat rewritten to use md5 checksums rather than identifiers and dimensions for cache matching.
Both the checksum and the identifier must match for the surface refresh to be valid, otherwise the
render will break if 2 textures with the same data but different names are in the one map! (Who 
built dm2 anyway???)
================
*/
gltexture_t *GL_LoadTexture (char *identifier, int width, int height, byte *data, int flags)
{
	int			i;
	gltexture_t	*glt;
	byte md5digest[16];
	byte ShotgunShells[] = {202, 6, 69, 163, 17, 112, 190, 234, 102, 56, 225, 242, 212, 175, 27, 187};

	// make an md5 of the texture data
	MD5_Checksum (data, (width * height), md5digest);

	// fix white line at base of shotgun shells box
	// one of id's artists needs a good slap here...
	// done before the cache check to prevent reloads of this texture...
	if (MD5_Compare (md5digest, ShotgunShells))
		Q_memcpy (data, data + 32 * 31, 32);

	// check the cache
	for (glt = gltextures; glt; glt = glt->next)
	{
		// texture matches
		// both the names and the data must match for it to be valid
		if (MD5_Compare (md5digest, glt->md5digest) && !strcmp (identifier, glt->identifier))
		{
			Con_DPrintf ("Matched texture %s\n", identifier);
			glt->registered = true;
			return glt;
		}
	}

	// allocate a texture slot
	if (gltextures)
	{
		// walk the list to the end
		for (glt = gltextures; glt; glt = glt->next)
		{
			// found the end
			if (!glt->next)
			{
				// add to the end
				glt->next = (gltexture_t *) malloc (sizeof (gltexture_t));

				// point at the last one
				glt = glt->next;

				// done
				break;
			}
		}
	}
	else
	{
		// begin the list and point glt at it
		gltextures = (gltexture_t *) malloc (sizeof (gltexture_t));

		// point glt at it
		glt = gltextures;
	}

	// point next at nothing
	glt->next = NULL;

	// mark it in use
	glt->registered = true;

	// take a texture object
	GL_AllocTexture (&glt->texnum);

	// store the md5
	MD5_Copy (md5digest, glt->md5digest);

	// check for fullbrights
	if (flags & TEX_LUMASPEC)
	{
		// take a texture object for the luma
		// we'll check it for real later on
		GL_AllocTexture (&glt->lumatexnum);
	}
	else glt->lumatexnum = 0;

	// ID's artists strike again!!!  model skins don't have a consistent background colour, so
	// hack in alpha for them.  these 2 are the main variants
	if (flags & TEX_ALIAS)
	{
		// take top-left colour as the background
		byte bgcolor = data[0];

		// now replace them all
		for (i = 0; i < (width * height); i++)
		{
			if (data[i] == bgcolor) data[i] = 255;
		}
	}

	// copy in the name and flags
	strcpy (glt->identifier, identifier);
	glt->flags = flags;

	// try for external
	//if (GL_TryExternalTexture (glt, "textures/"))
	//	return glt;

	// set info
	glt->width = width;
	glt->height = height;

	// upload
	GL_UploadTexture (glt, data, width, height, 8, flags);

	return glt;
}

/*
================
GL_LoadPicTexture
================
*/
gltexture_t *GL_LoadPicTexture (char *identifier, qpic_t *pic)
{
	return GL_LoadTexture (identifier, pic->width, pic->height, pic->data, TEX_2D);
}


/*
================
GL_RegisterTextures

Registers any textures that were used
================
*/
void GL_RegisterTextures (void)
{
	gltexture_t *glt = gltextures;
	gltexture_t *prev = NULL;
	gltexture_t *next;
	int NumTextures = 0;
	int UnRegistered = 0;

	for (;;)
	{
		// register for the next load - increment by 1 because registration sequence has not advanced yet
		if (!(glt->flags & TEX_MIPMAP)) glt->registered = true;
		if (glt->flags & TEX_ALIAS) glt->registered = true;
		if (glt->flags & TEX_SPRITE) glt->registered = true;
		if (glt->flags & TEX_2D) glt->registered = true;
		if (glt->flags & TEX_SCRAP) glt->registered = true;
		if (glt->flags & TEX_VIEWMODEL) glt->registered = true;
		if (glt->flags & TEX_RESOURCE) glt->registered = true;
		if (glt->flags & TEX_ALPHAMASK) glt->registered = true;
		if (glt->flags & TEX_SBARWIDGET) glt->registered = true;

		if (!glt->registered)
		{
			// update pointers
			if (prev)
			{
				// point previous->next to this->next so we skip over this
				prev->next = glt->next;
			}
			else
			{
				// start of list
				gltextures = glt->next;
			}
		}

		// store previous texture
		prev = glt;

		// store next
		next = glt->next;

		if (!glt->registered)
		{
			// do the dirt on glt
			if (glt->texnum) glDeleteTextures (1, &glt->texnum);
			if (glt->lumatexnum) glDeleteTextures (1, &glt->lumatexnum);
			free (glt);
			UnRegistered++;
		}

		// get out?
		if (!next) break;

		// restore glt at the next
		glt = next;
	}

	// unregister them all
	for (glt = gltextures; glt; glt = glt->next)
	{
		glt->registered = false;

		NumTextures++;
	}

	Con_Printf ("GL_RegisterTextures: %i textures in use, released %i\n", NumTextures, UnRegistered);
}


/*
================
R_TextureFrame

Run every frame to release memory from any temp buffers in use
================
*/
void R_TextureFrame (void)
{
	// free the memory
	if (scaled) free (scaled);
	if (externaldata) free (externaldata);

	// the pointers are now "undefined", so NULL them
	scaled = NULL;
	externaldata = NULL;
}


/*
================
GL_LoadBitmapResourceTexture

Loads a BMP resource texture

only bitmaps can be loaded from resources, cos versions of Windows prior to XP don't support other
image formats in resources.
================
*/
void GL_LoadBitmapResourceTexture (gltexture_t *glt, int ResourceID, unsigned int flags)
{
	int i, j;
	int run;
	int runlength;
	unsigned char *bmIn;
	BITMAPINFO *bmInfo;
	BITMAPFILEHEADER bmHeader;
	RGBQUAD bmPalette[256];
	RGBQUAD *PaletteIn;
	unsigned char *bmData;
	unsigned char *bmRaw;
	int pad;
	HINSTANCE hInstance;
	HANDLE hBitmap;

	// get a handle to the MHQuake executable
	hInstance = GetModuleHandle (NULL);

	// load the resource
	hBitmap = LoadResource (hInstance, FindResource (hInstance, MAKEINTRESOURCE (ResourceID), RT_BITMAP));

	// lock it
	bmRaw = (byte *) LockResource (hBitmap);

	// get the bm info header
	// read the header
	bmInfo = (BITMAPINFO *) bmRaw;

	// skip past the header (for 24 BPP RGB images the pixel data immediately follows, otherwise a palette is next)
	bmRaw += sizeof (BITMAPINFO);

	if (bmInfo->bmiHeader.biCompression != BI_RGB)
	{
		if (!(bmInfo->bmiHeader.biCompression == BI_RLE8 && bmInfo->bmiHeader.biBitCount == 8))
			goto BadBMP;
	}

	// support 4, 8 and 24 BPP modes
	if (bmInfo->bmiHeader.biBitCount != 4 && 
		bmInfo->bmiHeader.biBitCount != 8 && 
		bmInfo->bmiHeader.biBitCount != 24) goto BadBMP;

	SetImagePointer (bmInfo->bmiHeader.biWidth * abs (bmInfo->bmiHeader.biHeight));
	bmData = externaldata;

	if (bmInfo->bmiHeader.biBitCount != 24)
	{
		// read the palette entries
		// the first one is in the BITMAPINFO struct
		bmPalette[0].rgbBlue = bmInfo->bmiColors[0].rgbBlue;
		bmPalette[0].rgbGreen = bmInfo->bmiColors[0].rgbGreen;
		bmPalette[0].rgbRed = bmInfo->bmiColors[0].rgbRed;
		bmPalette[0].rgbReserved = bmInfo->bmiColors[0].rgbReserved;

		// remaining entries follow the struct
		for (i = 1; i < bmInfo->bmiHeader.biClrUsed; i++)
		{
			PaletteIn = (RGBQUAD *) bmRaw;

			bmPalette[i].rgbBlue = PaletteIn->rgbBlue;
			bmPalette[i].rgbGreen = PaletteIn->rgbGreen;
			bmPalette[i].rgbRed = PaletteIn->rgbRed;
			bmPalette[i].rgbReserved = PaletteIn->rgbReserved;

			bmRaw += sizeof (RGBQUAD);
		}
	}

	// each row in a bitmap is padded to align it on a 4 unsigned char boundary, so find
	// out how much padding we need (this may not work with RLE images)
	for (i = 0; ; i++)
	{
		pad = i * 4;
		if (pad >= (bmInfo->bmiHeader.biWidth * 3)) break;
	}

	pad -= bmInfo->bmiHeader.biWidth * 3;

	// load the data
	for (i = 0; i < bmInfo->bmiHeader.biWidth * abs (bmInfo->bmiHeader.biHeight); i++)
	{
		if (bmInfo->bmiHeader.biBitCount == 4)
		{
			bmIn = bmRaw++;

			bmData[0] = bmPalette[*bmIn >> 4].rgbRed;
			bmData[1] = bmPalette[*bmIn >> 4].rgbGreen;
			bmData[2] = bmPalette[*bmIn >> 4].rgbBlue;
			bmData[3] = 255;

			bmData[4] = bmPalette[*bmIn % 16].rgbRed;
			bmData[5] = bmPalette[*bmIn % 16].rgbGreen;
			bmData[6] = bmPalette[*bmIn % 16].rgbBlue;
			bmData[7] = 255;

			bmData += 8;

			// increment i again cos we're reading two entires at a time
			i++;

			continue;
		}
		else if (bmInfo->bmiHeader.biBitCount == 8)
		{
			if (bmInfo->bmiHeader.biCompression == BI_RLE8)
			{
				bmIn = bmRaw++;

				if (*bmIn == 0)
				{
					bmIn = bmRaw++;
					i++;

					switch (*bmIn)
					{
					case 0:
						// End of Line - just do nothing - trust me, it works.
						break;
					case 1:
						// End of Bitmap
						goto bmDone;
						break;
					case 2:
						// Delta
						// (tests indicate this isn't actually used... most likely cos the bitmap format
						// doesn't specify what the skipped pixels are to be filled with!!!)
						bmIn = bmRaw++;
						i++;

						bmData += *bmIn * 4;

						bmIn = bmRaw++;
						i++;

						bmData += *bmIn * bmInfo->bmiHeader.biWidth * 4;

						break;
					default:
						// absolute mode
						runlength = *bmIn;

						for (run = 0; run < runlength; run++)
						{
							bmIn = bmRaw++;
							i++;

							bmData[0] = bmPalette[*bmIn].rgbRed;
							bmData[1] = bmPalette[*bmIn].rgbGreen;
							bmData[2] = bmPalette[*bmIn].rgbBlue;
							bmData[3] = 255;

							bmData += 4;
						}

						if (runlength % 2)
						{
							// align on word boundaries
							bmIn = bmRaw++;
							i++;
						}

						break;
					}
				}
				else
				{
					// encoded mode
					runlength = *bmIn;

					bmIn = bmRaw++;
					i++;

					for (run = 0; run < runlength; run++)
					{
						bmData[0] = bmPalette[*bmIn].rgbRed;
						bmData[1] = bmPalette[*bmIn].rgbGreen;
						bmData[2] = bmPalette[*bmIn].rgbBlue;
						bmData[3] = 255;

						bmData += 4;
					}
				}

				continue;
			}

			bmIn = bmRaw++;

			bmData[0] = bmPalette[*bmIn].rgbRed;
			bmData[1] = bmPalette[*bmIn].rgbGreen;
			bmData[2] = bmPalette[*bmIn].rgbBlue;
			bmData[3] = 255;

			bmData += 4;

			continue;
		}
		else
		{
			// 24 BPP
			bmData[2] = *bmRaw++;
			bmData[1] = *bmRaw++;
			bmData[0] = *bmRaw++;
			bmData[3] = 255;

			bmData += 4;
		}

		// read the padding
		for (j = 0; j < pad; j++) bmRaw++;
	}

bmDone:;

	// check orientation
	if (bmInfo->bmiHeader.biHeight > 0)
	{
		// bottom-up bitmap so we must invert it (sigh)
		for (i = 0; i < abs (bmInfo->bmiHeader.biHeight) / 2; i++)
		{
			for (j = 0; j < bmInfo->bmiHeader.biWidth; j++)
			{
				unsigned int srcpos = (i * bmInfo->bmiHeader.biWidth) + j;
				unsigned int dstpos = (((abs (bmInfo->bmiHeader.biHeight) - 1) - i) * bmInfo->bmiHeader.biWidth) + j;
				unsigned int tmp;

				tmp = ((unsigned int *) externaldata)[dstpos];
				((unsigned int *) externaldata)[dstpos] = ((unsigned int *) externaldata)[srcpos];
				((unsigned int *) externaldata)[srcpos] = tmp;
			}
		}
	}

	// copy the flags - make sure it's properly flagged as 32 bit
	glt->flags = (flags | TEX_EXTERNAL | TEX_RESOURCE);

	// take a texture object for it
	GL_AllocTexture (&glt->texnum);

	// no luma
	glt->lumatexnum = 0;

	// upload
	GL_UploadTexture (glt, externaldata, bmInfo->bmiHeader.biWidth, abs (bmInfo->bmiHeader.biHeight), 32, glt->flags);

	return;

BadBMP:;
	Sys_Error ("Failed to Load Bitmap Resource %i", ResourceID);
}


