/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_warp.c -- sky and water polygons

#include "quakedef.h"

extern	model_t	*loadmodel;

int		skytexturenum;

int		solidskytexture;
int		alphaskytexture;
float	speedscale;		// for top sky and bottom sky

msurface_t	*warpface;

#define GL_SUBDIVIDE_SIZE 64

void BoundPoly (int numverts, float *verts, vec3_t mins, vec3_t maxs)
{
	int		i, j;
	float	*v;

	mins[0] = mins[1] = mins[2] = 999999;
	maxs[0] = maxs[1] = maxs[2] = -999999;
	v = verts;

	for (i = 0; i < numverts; i++)
	{
		if (*v < mins[0]) mins[0] = *v;
		if (*v > maxs[0]) maxs[0] = *v;

		if (*v < mins[1]) mins[1] = *v;
		if (*v > maxs[1]) maxs[1] = *v;

		if (*v < mins[2]) mins[2] = *v;
		if (*v > maxs[2]) maxs[2] = *v;
	}
}


int numwaterverts;

void SubdividePolygon (int numverts, float *verts)
{
	int		i, j, k;
	vec3_t	mins, maxs;
	float	m;
	float	*v;
	vec3_t	front[64], back[64];
	int		f, b;
	float	dist[64];
	float	frac;
	glpoly_t	*poly;
	float	s, t;
	vec3_t  total;

	if (numverts > 60)
		Sys_Error ("numverts = %i", numverts);

	BoundPoly (numverts, verts, mins, maxs);

	for (i=0 ; i<3 ; i++)
	{
		m = (mins[i] + maxs[i]) * 0.5;
		m = GL_SUBDIVIDE_SIZE * floor (m/GL_SUBDIVIDE_SIZE + 0.5);
		if (maxs[i] - m < 8)
			continue;
		if (m - mins[i] < 8)
			continue;

		// cut it
		v = verts + i;
		for (j=0 ; j<numverts ; j++, v+= 3)
			dist[j] = *v - m;

		// wrap cases
		dist[j] = dist[0];
		v-=i;
		VectorCopy (verts, v);

		f = b = 0;
		v = verts;
		for (j=0 ; j<numverts ; j++, v+= 3)
		{
			if (dist[j] >= 0)
			{
				VectorCopy (v, front[f]);
				f++;
			}
			if (dist[j] <= 0)
			{
				VectorCopy (v, back[b]);
				b++;
			}
			if (dist[j] == 0 || dist[j+1] == 0)
				continue;
			if ( (dist[j] > 0) != (dist[j+1] > 0) )
			{
				// clip point
				frac = dist[j] / (dist[j] - dist[j+1]);
				for (k=0 ; k<3 ; k++)
					front[f][k] = back[b][k] = v[k] + frac*(v[3+k] - v[k]);
				f++;
				b++;
			}
		}

		SubdividePolygon (f, front[0]);
		SubdividePolygon (b, back[0]);
		return;
	}

	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->basev = Hunk_Alloc (sizeof (vec3_t) * (numverts + 2));
	poly->next = warpface->polys;
	warpface->polys = poly;
	poly->numverts = numverts + 2;

	poly->vertoffset = numwaterverts;
	numwaterverts += poly->numverts;

	total[0] = total[1] = total[2] = 0;

	for (i = 0; i < numverts; i++, verts += 3)
	{
		// world verts only, don't need to bother with owt else
		poly->basev[i + 1][0] = verts[0];
		poly->basev[i + 1][1] = verts[1];
		poly->basev[i + 1][2] = verts[2];

		// accumulate verts into total
		total[0] += verts[0] / (float) numverts;
		total[1] += verts[1] / (float) numverts;
		total[2] += verts[2] / (float) numverts;
	}

	// copy the average of the totals to first...
	poly->basev[0][0] = total[0];
	poly->basev[0][1] = total[1];
	poly->basev[0][2] = total[2];

	// ...and copy vert 1 to last...
	poly->basev[numverts + 1][0] = poly->basev[1][0];
	poly->basev[numverts + 1][1] = poly->basev[1][1];
	poly->basev[numverts + 1][2] = poly->basev[1][2];

	// always a trifan
	poly->PRIMITIVE_TYPE = GL_TRIANGLE_FAN;

	// set up texcoords
	// rotate the plane so that v[2] is 0 or as close to 0 as poss for all verts, then v[0] and v[1] will be the texcoords
}

/*
================
GL_SubdivideSurface

Breaks a polygon up along axial 64 unit
boundaries so that turbulent and sky warps
can be done reasonably.
================
*/
int GL_SubdivideSurface (model_t *mod, msurface_t *fa, int waterverts)
{
	vec3_t		verts[64];
	int			numverts;
	int			i;
	int			lindex;
	float		*vec;
	texture_t	*t;

	warpface = fa;
	warpface->polys = NULL;

	// convert edges back to a normal polygon
	numverts = 0;
	numwaterverts = waterverts;

	for (i=0 ; i<fa->numedges ; i++)
	{
		lindex = mod->surfedges[fa->firstedge + i];

		if (lindex > 0)
			vec = mod->vertexes[mod->edges[lindex].v[0]].position;
		else
			vec = mod->vertexes[mod->edges[-lindex].v[1]].position;

		VectorCopy (vec, verts[numverts]);
		numverts++;
	}

	SubdividePolygon (numverts, verts[0]);

	return numwaterverts;
}

//=========================================================



/*
=============
EmitWaterPolys

Does a water warp on the pre-fragmented glpoly_t chain
=============
*/
void EmitWaterPolys (msurface_t *fa)
{
	glpoly_t	*p;
	int			i;

	for (p = fa->polys; p; p = p->next)
	{
		glBegin (p->PRIMITIVE_TYPE);

		for (i = 0; i < p->numverts; i++)
		{
			glTexCoord2fv (p->verts[i].st);
			glVertex3fv (p->verts[i].tv);
		}

		glEnd ();
	}
}


/*
=============
EmitSkyPolys
=============
*/
void EmitSkyPolys (msurface_t *fa)
{
	glpoly_t	*p;
	int			i;
	float	st[2];
	vec3_t	dir;
	float	length;

	for (p = fa->polys; p; p = p->next)
	{
		glBegin (p->PRIMITIVE_TYPE);

		for (i = 0; i < p->numverts; i++)
		{
			VectorSubtract (p->verts[i].tv, r_origin, dir);
			dir[2] *= 3;	// flatten the sphere

			length = dir[0]*dir[0] + dir[1]*dir[1] + dir[2]*dir[2];
			length = sqrt (length);
			length = 6*63/length;

			dir[0] *= length;
			dir[1] *= length;

			st[0] = (speedscale + dir[0]) * (1.0/128);
			st[1] = (speedscale + dir[1]) * (1.0/128);

			glTexCoord2fv (st);
			glVertex3fv (p->verts[i].tv);
		}

		glEnd ();
	}
}

/*
===============
EmitBothSkyLayers

Does a sky warp on the pre-fragmented glpoly_t chain
This will be called for brushmodels, the world
will have them chained together.
===============
*/
void EmitBothSkyLayers (msurface_t *fa)
{
	int			i;
	int			lindex;
	float		*vec;

	glBindTexture (GL_TEXTURE_2D, solidskytexture);
	speedscale = realtime*8;
	speedscale -= (int)speedscale & ~127 ;

	EmitSkyPolys (fa);

	glEnable (GL_BLEND);
	glBindTexture (GL_TEXTURE_2D, alphaskytexture);
	speedscale = realtime*16;
	speedscale -= (int)speedscale & ~127 ;

	EmitSkyPolys (fa);

	glDisable (GL_BLEND);
}

/*
=================
R_DrawSkyChain
=================
*/
void R_DrawSkyChain (msurface_t *s)
{
	msurface_t	*fa;

	glBindTexture (GL_TEXTURE_2D, solidskytexture);
	speedscale = realtime*8;
	speedscale -= (int)speedscale & ~127 ;

	for (fa=s ; fa ; fa=fa->texturechain)
		EmitSkyPolys (fa);

	glEnable (GL_BLEND);
	glBindTexture (GL_TEXTURE_2D, alphaskytexture);
	speedscale = realtime*16;
	speedscale -= (int)speedscale & ~127 ;

	for (fa=s ; fa ; fa=fa->texturechain)
		EmitSkyPolys (fa);

	glDisable (GL_BLEND);
}


//===============================================================

/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
==============
*/
void R_InitSky (texture_t *tx)
{
	int			i, j, p;
	byte		*src;
	unsigned	trans[128 * 128];
	unsigned	transpix;
	int			r, g, b;
	unsigned	*rgba;
	extern	int skytexturenum;

	src = (byte *) (tx + 1);

	// make an average value for the back to avoid
	// a fringe on the top level
	r = g = b = 0;

	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			// load the back texture data
			p = src[i * 256 + j + 128];
			rgba = &d_8to24tableBGRA[p];
			trans[(i * 128) + j] = *rgba;

			// accumulate the average
			r += ((byte *) rgba)[0];
			g += ((byte *) rgba)[1];
			b += ((byte *) rgba)[2];
		}
	}

	// get the average
	((byte *) &transpix)[0] = r / (128 * 128);
	((byte *) &transpix)[1] = g / (128 * 128);
	((byte *) &transpix)[2] = b / (128 * 128);
	((byte *) &transpix)[3] = 0;

	if (!solidskytexture) GL_AllocTexture (&solidskytexture);

	glBindTexture (GL_TEXTURE_2D, solidskytexture);
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, 128, 128, 0, GL_BGRA, GL_UNSIGNED_BYTE, trans);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// load up the top data
	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			p = src[i * 256 + j];

			// set to the average if it's alpha
			if (p == 0)
				trans[(i * 128) + j] = transpix;
			else
				trans[(i * 128) + j] = d_8to24tableBGRA[p];
		}
	}

	if (!alphaskytexture) GL_AllocTexture (&alphaskytexture);

	glBindTexture (GL_TEXTURE_2D, alphaskytexture);
	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, 128, 128, 0, GL_BGRA, GL_UNSIGNED_BYTE, trans);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
}


gltexture_t r_skyboxtextures[6];

void R_InitSkyTextures (void)
{
	int i;

	for (i = 0; i < 6; i++)
	{
		GL_AllocTexture (&r_skyboxtextures[i].texnum);

		r_skyboxtextures[i].flags = 0;
	}
}


char *skypath[] = 
{
	"",
	"gfx/",
	"env/",
	"gfx/env/",
	NULL
};

char *skysuf[] =
{
	"ft",
	"rt",
	"bk",
	"lf",
	"up",
	"dn"
};

qboolean GL_TryExternalTexture (gltexture_t *glt, char *path);

qboolean R_SetSkyBox (char *name)
{
	int i;
	int j;
	char basename[256];
	int ok = 0;
	FILE *f;

	for (i = 0; i < 6; i++)
	{
		// clear the flags so we can test if it's OK
		r_skyboxtextures[i].flags = 0;

		for (j = 0; ; j++)
		{
			// nowhere else to test
			if (!skypath[j])
				break;

			// build the base name
			sprintf (basename, "%s%s%s", skypath[j], name, skysuf[i]);

			Con_DPrintf ("Trying %s... ", basename);

			if (GL_TryExternalTexture (&r_skyboxtextures[i], basename))
			{
				ok++;
				Con_DPrintf ("OK\n");
				break;
			}
			else Con_DPrintf ("Failed\n");
		}
	}

	if (ok == 6)
		return true;

	return false;
}


// fast vertex calculation
float QuadVecs[4][6][3] =
{
	{
		{4096.0f, 4096.0f, -4096.0f},
		{-4096.0f, 4096.0f, -4096.0f},
		{-4096.0f, -4096.0f, -4096.0f},
		{4096.0f, -4096.0f, -4096.0f},
		{-4096.0f, 4096.0f, 4096.0f},
		{-4096.0f, -4096.0f, -4096.0f}
	},
	{
		{4096.0f, 4096.0f, 4096.0f},
		{-4096.0f, 4096.0f, 4096.0f},
		{-4096.0f, -4096.0f, 4096.0f},
		{4096.0f, -4096.0f, 4096.0f},
		{-4096.0f, -4096.0f, 4096.0f},
		{-4096.0f, 4096.0f, -4096.0f}
	},
	{
		{4096.0f, -4096.0f, 4096.0f},
		{4096.0f, 4096.0f, 4096.0f},
		{-4096.0f, 4096.0f, 4096.0f},
		{-4096.0f, -4096.0f, 4096.0f},
		{4096.0f, -4096.0f, 4096.0f},
		{4096.0f, 4096.0f, -4096.0f}
	},
	{
		{4096.0f, -4096.0f, -4096.0f},
		{4096.0f, 4096.0f, -4096.0f},
		{-4096.0f, 4096.0f, -4096.0f},
		{-4096.0f, -4096.0f, -4096.0f},
		{4096.0f, 4096.0f, 4096.0f},
		{4096.0f, -4096.0f, -4096.0f}
	}
};


void MakeSkyVecs (int quad, float s, float t, int axis)
{
	vec3_t		v;

	v[0] = (QuadVecs[quad][axis][0]) + r_origin[0];
	v[1] = (QuadVecs[quad][axis][1]) + r_origin[1];
	v[2] = (QuadVecs[quad][axis][2]) + r_origin[2];

	glTexCoord2f (s, t);
	glVertex3fv (v);
}


void R_RenderSky (void)
{
	int i;

	// test the textures to make sure we have them all
	for (i = 0; i < 6; i++)
	{
		if (r_skyboxtextures[i].flags == 0)
		{
			return;
		}
	}

	// switch the projection matrix for a greater z
	// (don;t bother switching state back cos this is the very last thing rendered)
	glMatrixMode (GL_PROJECTION);
	glPushMatrix ();
    glLoadIdentity ();
	gluPerspective (r_refdef.fov_y, (float) glwidth / (float) glheight, 4, 8192);

	glDisable (GL_CULL_FACE);

	// we're good for a skybox now
	for (i = 0; i < 6; i++)
	{
		glBindTexture (GL_TEXTURE_2D, r_skyboxtextures[i].texnum);

		glBegin (GL_QUADS);

		MakeSkyVecs (0, 0.001953, 0.998047, i);
		MakeSkyVecs (1, 0.001953, 0.001953, i);
		MakeSkyVecs (2, 0.998047, 0.001953, i);
		MakeSkyVecs (3, 0.998047, 0.998047, i);

		glEnd ();
	}

	glEnable (GL_CULL_FACE);
	glPopMatrix ();
	glMatrixMode (GL_MODELVIEW);
}
