/*
===============================================================================================================
	Copyright (C) 2006 MH

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

	See the GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
===============================================================================================================
*/

/*
===============================================================================================================
	MEMORY.C

	Memory management routines

	Memory is allocated as a single contiguous block either at startup, or the first time a request for memory
	is made.  All subsequent requests for memory are taken from that block.

	Memory is given in chunks of 1K.  Each such chunk has a header containing descriptive info.

	Memory can be allocated on a permanent (entire run-time) basis, or on a temporary (current map only) basis.

	PUBLIC INTERFACE

	void Mem_Init (void)
	Called on startup or at first request for memory, may also be called any time to re-initialize the memory
	system.

	void Mem_Shutdown (void)
	Called on shutdown (although not necessarily as the OS will free process memory.  May also be called any
	time to flush the memory system.

	void Mem_Touch (void)
	Should be called each frame (or group of frames) to ensure that the OS doesn't page out our memory.

	void *Mem_Alloc (unsigned int AllocSize, char isTemp)
	Allocates a block of memory.  Takes the requested size and a boolean for if it's temp or not, returns a
	pointer to the allocated memory.  Called each time memory is required.

	void Mem_FreeTemp (void)
	Frees temporary blocks, and consolidates contiguous temporary blocks.  Called at the start of a map load.

===============================================================================================================
*/

#include <windows.h>
#include <stdlib.h>


// our full memory allocation
typedef struct Global_Memory_s
{
	unsigned char *MemBlocks;
	unsigned int MemSize;
	unsigned int MemBegin;
} Global_Memory_t;

static Global_Memory_t GlobalMemory = {NULL, 0, 0};

// memory block header struct
typedef struct Memory_Block_s
{
	// 1 if free, 0 if used
	char isFree;

	// size of block in bytes (including header)
	unsigned int MemSize;

	// 0 if permanent, 1 if temporary
	char isTemp;
} Memory_Block_t;

/*
=====================
Mem_Init

Initialize memory allocation
=====================
*/
void Mem_Init (void)
{
	MEMORYSTATUSEX lpBuffer;
	int MemoryStatusResult;
	int i;
	Memory_Block_t *Block;

	// retrieve memory status from the OS
	// use the 2000/XP version of the function to support PCs with > 4GB RAM
	lpBuffer.dwLength = sizeof (MEMORYSTATUSEX);

	MemoryStatusResult = GlobalMemoryStatusEx (&lpBuffer);

	// to do - GlobalMemoryStatusEx failed
	if (!MemoryStatusResult)
	{
	}

	// take half available physical memory, but never less than 32MB and never more than 256MB
	if (lpBuffer.ullAvailPhys > 0x20000000)
		GlobalMemory.MemSize = 0x10000000;
	else if (lpBuffer.ullAvailPhys < 0x4000000)
		GlobalMemory.MemSize = 0x2000000;
	else GlobalMemory.MemSize = (unsigned int) (lpBuffer.ullAvailPhys / 2);

	// release memory if it was previously in use
	if (GlobalMemory.MemBlocks) free (GlobalMemory.MemBlocks);

	// allocate memory block
	GlobalMemory.MemBlocks = (unsigned char *) malloc (GlobalMemory.MemSize);

	// to do - allocation failed
	if (!GlobalMemory.MemBlocks)
	{
	}

	// mark as free blocks in groups of 1024
	for (i = 0; i < GlobalMemory.MemSize; i += 1024)
	{
		Block = (Memory_Block_t *) (&GlobalMemory.MemBlocks[i]);
		Block->isFree = 1;
	}

	// clear to all free and as one contiguous block
	Block = (Memory_Block_t *) GlobalMemory.MemBlocks;
	Block->isFree = 1;
	Block->isTemp = 1;
	Block->MemSize = GlobalMemory.MemSize - 1024;

	// hack - mark the last block as used
	Block = (Memory_Block_t *) (&GlobalMemory.MemBlocks[GlobalMemory.MemSize - 1024]);
	Block->isFree = 0;
	Block->isTemp = 0;
	Block->MemSize = 1024;

	// reset begin point for touching
	GlobalMemory.MemBegin = 0;
}


/*
======================
Mem_Shutdown

Shut down the memory system
======================
*/
void Mem_Shutdown (void)
{
	// release the memory if it was allocated
	if (GlobalMemory.MemBlocks) free (GlobalMemory.MemBlocks);
}


/*
=======================
Mem_Touch

Touch memory to ensure that Windows doesn't page it out
=======================
*/
void Mem_Touch (void)
{
	int i;
	unsigned char c;

	// go in jumps of 1 MB so that we don't have to touch all of the memory
	for (i = GlobalMemory.MemBegin; i < GlobalMemory.MemSize; i += 0x100000)
		c = GlobalMemory.MemBlocks[i];

	// add 64 k to the begin point
	GlobalMemory.MemBegin += 0x10000;

	// wrap
	if (GlobalMemory.MemBegin > GlobalMemory.MemSize) GlobalMemory.MemBegin = 0;
}


/*
====================
Mem_AllocBlock

Allocates an actual block and fixes up block pointers
====================
*/
void *Mem_AllocBlock (Memory_Block_t *Block, unsigned int AllocSize, char isTemp)
{
	int BlockSize;
	Memory_Block_t *Next;

	// get block size as the next multiple of 1024 above the allocation size
	for (BlockSize = 0; BlockSize < AllocSize; BlockSize += 1024);

	// ensure that there's room for the header
	if (AllocSize + sizeof (Memory_Block_t) > BlockSize) BlockSize += 1024;

	// point to the next block
	Next = (Memory_Block_t *) (((unsigned char *) Block) + BlockSize);

	// fix up the memory size in the next block if it's free
	if (Next->isFree) Next->MemSize = Block->MemSize - BlockSize;

	// fill in the block
	Block->isFree = 0;
	Block->isTemp = isTemp;
	Block->MemSize = BlockSize;

	// the data follows immediately after the header
	return (void *) (Block + 1);
}


/*
====================
Mem_Alloc

Allocate a block of memory at the requested size
====================
*/
void *Mem_Alloc (unsigned int AllocSize, char isTemp)
{
	int i;
	Memory_Block_t *Block = (Memory_Block_t *) GlobalMemory.MemBlocks;

	for (i = 0; i < GlobalMemory.MemSize; )
	{
		// get a pointer to the block
		Block = (Memory_Block_t *) (&GlobalMemory.MemBlocks[i]);

		// look for the first unallocated block that's large enough
		if (Block->isFree && Block->MemSize >= AllocSize)
			return Mem_AllocBlock (Block, AllocSize, isTemp);

		// incrememt by size of block
		i += Block->MemSize;
	}

	// to do - if it gets to here it failed to allocate and must therefore crash!

	// shut up compiler
	return NULL;
}


/*
====================
Mem_FreeTemp

Free temporary blocks
====================
*/
void Mem_FreeTemp (void)
{
	int i;
	Memory_Block_t *Block = (Memory_Block_t *) GlobalMemory.MemBlocks;

	// free all temp blocks
	for (i = 0; i < GlobalMemory.MemSize; )
	{
		// get a pointer to the block
		Block = (Memory_Block_t *) (&GlobalMemory.MemBlocks[i]);

		// look for the first unallocated block that's large enough
		if (Block->isTemp) Block->isFree = 1;

		// incrememt by size of block
		i += Block->MemSize;
	}

	// consolidate temporary blocks
	for (i = 0; i < GlobalMemory.MemSize; )
	{
		// get a pointer to the block
		Block = (Memory_Block_t *) (&GlobalMemory.MemBlocks[i]);

		// look for the first unallocated block
		if (Block->isFree)
		{
			// init total amount free
			unsigned int TotalFree = Block->MemSize;
			Memory_Block_t *Next;

			while (1)
			{
				// jump to the next block
				i += Block->MemSize;
				Next = (Memory_Block_t *) (&GlobalMemory.MemBlocks[i]);

				// not free
				if (!Next->isFree) break;

				// accumulate free size
				TotalFree += Next->MemSize;
			}

			// write in new free size
			Block->MemSize = TotalFree;
		}
		else
		{
			// jump to new block
			// this will already be done for a free block...
			i += Block->MemSize;
		}
	}
}
