/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c -- this is the only file outside the refresh that touches the
// vid buffer

#include "quakedef.h"

int			gl_max_size = 1024;

byte		*draw_chars;				// 8*8 graphic characters
qpic_t		*draw_disc;
qpic_t		*draw_backtile;

int			translate_texture;
gltexture_t *char_texture;


byte		conback_buffer[sizeof (qpic_t) + sizeof (glpic_t)];
qpic_t		*conback = (qpic_t *) &conback_buffer;


extern byte menuplyr_pixels[];

// quad drawing
typedef struct drawverts_s
{
	float v[3];
	float st[2];
} drawverts_t;

drawverts_t drawverts[4];


// resource textures
gltexture_t CrossHairTexture;
gltexture_t ParticleTexture;
gltexture_t Font8x8Texture;

extern glpic_t Font8x8[];
int Font8x8Widths[64];

int R_LoadMapShot (char *name);

//=============================================================================
/* Support Routines */


/*
===============
Draw_Init
===============
*/
void Draw_Init (void)
{
	int i;
	int j;
	int c;
	byte *newCharset;
	float	screenaspect;

	// calculate 2d screen size
	screenaspect = (float) vid.width / (float) vid.height;

	// calculate depending on screenaspect, we add to the calculation to compensate
	// for rounding errors, then make them multiples of 8
	if (screenaspect > 1.333333f)
	{
		// widescreen, lock height at 480 and rescale width to compensate
		vid.width = 4 + (int) (480.0f * screenaspect);
		vid.height = 480;
	}
	else
	{
		// tall, lock width at 640 and rescale height to compensate
		vid.width = 640;
		vid.height = 4 + (640.0f / screenaspect);
	}

	// make them multiples of 8
	vid.width &= 0xfff8;
	vid.height &= 0xfff8;

	BuildTextureSearchPaths ("textures/", NULL, NULL, NULL);

	// load the charset manually because we need to hack the charset alpha colour and character
	// spacing because one of id's artists badly needs a good slap around the head...
	draw_chars = W_GetLumpName ("conchars");

	// set the proper translucent colour
	for (i = 0; i < 16384; i++)
	{
		if (draw_chars[i] == 0)
		{
			draw_chars[i] = 255;
		}
	}

	// allocate memory for a new charset
	newCharset = (byte *) malloc (256 * 256);

	// clear to all alpha
	memset (newCharset, 255, 256 * 256);

	// copy them in with better spacing
	for (i = 0, j = 0; i < 128 * 128; i += 8, j += 16)
	{
		// jump down a half row
		if (i && !(i % 1024)) j += 2048;

		for (c = 0; c < 8; c++)
		{
			newCharset[j + c] = draw_chars[i + c];
		}
	}

	// now this is evil.
	// some of the chars are meant as consecutive dividers and the like, so we don;t want them spaced.
	// because we're spacing by the width of a char, we just copy the relevant pixels in
	for (i = 16; i < 24; i++)
	{
		for (j = 224; j < 232; j++)
		{
			newCharset[i * 256 + j - 8] = newCharset[i * 256 + j];
			newCharset[i * 256 + j + 8] = newCharset[i * 256 + j];
		}
	}

	for (i = 144; i < 152; i++)
	{
		for (j = 224; j < 232; j++)
		{
			newCharset[i * 256 + j - 8] = newCharset[i * 256 + j];
			newCharset[i * 256 + j + 8] = newCharset[i * 256 + j];
		}
	}

	for (i = 128; i < 136; i++)
	{
		for (j = 16; j < 24; j++)
		{
			newCharset[i * 256 + j - 8] = newCharset[i * 256 + j];
			newCharset[i * 256 + j + 8] = newCharset[i * 256 + j];
		}
	}

	for (i = 128; i < 136; i++)
	{
		newCharset[i * 256 + 47] = newCharset[i * 256 + 23];
		newCharset[i * 256 + 56] = newCharset[i * 256 + 16];
	}

	// now turn them into textures
	char_texture = GL_LoadTexture (/*"wad/charset"*/ "charset", 256, 256, newCharset, TEX_INCACHE);

	// free memory
	free (newCharset);

	// load console
	conback = Draw_CachePic ("gfx/conback.lmp");

	// hack the properties so it draws correctly
	conback->height = vid.height;
	conback->width = vid.width;

	// save a texture slot for translated picture
	GL_AllocTexture (&translate_texture);

	// get the other pics we need
	draw_disc = Draw_PicFromWad ("disc");
	draw_backtile = Draw_PicFromWad ("backtile");

	// load the resources we need also
	GL_LoadBitmapResourceTexture (&CrossHairTexture, IDB_BITMAP1, TEX_ALPHAMASK);
	GL_LoadBitmapResourceTexture (&ParticleTexture, IDB_BITMAP3, TEX_ALPHAMASK);
	GL_LoadBitmapResourceTexture (&Font8x8Texture, IDB_BITMAP4, TEX_FONT8X8);

	// now set up the rest of the stuff we need for the 8x8 font texture
	for (i = 0; i < 64; i++)
	{
		// get the width of this character
		Font8x8Widths[i] = 224.0f * (Font8x8[i].sh - Font8x8[i].sl);

		// not necessary, but just to keep the struct filled in and valid
		Font8x8[i].gltexture = &Font8x8Texture;
	}
}


int FontMap[128] =
{
//	 0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 63,
	52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, -1, -1, -1, -1, -1,
	-1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
	-1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
	41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1,
};

void Draw_Font8x8String (int x, int y, char *str, float scale)
{
	int MappedFont;
	int yscale = (int) (scale * 28.0f);

	glBindTexture (GL_TEXTURE_2D, Font8x8Texture.texnum);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	while (*str)
	{
		// map it to an array position
		if (*str > 127)
		{
			// red text
			glColor3ub (255, 0, 0);
			MappedFont = FontMap[(*str) - 128];
		}
		else
		{
			// brown text
			glColor3ub (144, 77, 29);
			MappedFont = FontMap[*str];
		}

		// if it's a character we can't handle just emit a space
		// (space is also mapped to a character we can't handle)
		if (MappedFont < 0 || MappedFont > 63)
		{
			str++;
			x += (int) (scale * (float) 16);
			continue;
		}

		if (*str >= 'a' && *str <= 'z') scale *= 0.8;

		drawverts[0].st[0] = Font8x8[MappedFont].sl;
		drawverts[0].st[1] = Font8x8[MappedFont].tl;
		drawverts[0].v[0] = x;
		drawverts[0].v[1] = y;

		drawverts[1].st[0] = Font8x8[MappedFont].sh;
		drawverts[1].st[1] = Font8x8[MappedFont].tl;
		drawverts[1].v[0] = x + (int) (scale * (float) Font8x8Widths[MappedFont]);
		drawverts[1].v[1] = y;

		drawverts[2].st[0] = Font8x8[MappedFont].sh;
		drawverts[2].st[1] = Font8x8[MappedFont].th;
		drawverts[2].v[0] = x + (int) (scale * (float) Font8x8Widths[MappedFont]);
		drawverts[2].v[1] = y + yscale;

		drawverts[3].st[0] = Font8x8[MappedFont].sl;
		drawverts[3].st[1] = Font8x8[MappedFont].th;
		drawverts[3].v[0] = x;
		drawverts[3].v[1] = y + yscale;

		glDrawArrays (GL_QUADS, 0, 4);

		x += (int) (scale * ((float) Font8x8Widths[MappedFont] + 2));

		if (*str >= 'a' && *str <= 'z') scale /= 0.8;

		str++;
	}

	// revert to white
	glColor3f (1, 1, 1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}


float XHairS[8] = {0, 0.25, 0.5, 0.75, 0, 0.25, 0.5, 0.75};
float XHairT[8] = {0, 0, 0, 0, 0.5, 0.5, 0.5, 0.5};
float XHairSAdd = 0.25;
float XHairTAdd = 0.5;

extern cvar_t crosshair;
extern cvar_t cl_crossx;
extern cvar_t cl_crossy;

void Draw_CrossHair (void)
{
	int activecrosshair;

	// make it valid for array indexing
	activecrosshair = crosshair.value - 1;

	// reduce to a valid scale
	while (activecrosshair > 8) activecrosshair -= 8;

	// no crosshair
	if (activecrosshair < 0) return;

	// switch blend func for alphamask texes
	glBlendFunc (GL_SRC_ALPHA, GL_ONE);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// set colour
	glColor3f (1, 1, 1);

	glBindTexture (GL_TEXTURE_2D, CrossHairTexture.texnum);

	// set up verts
	// the crosshair is offset slightly downwards so that it matches up better with where the bullets actually end up.
	// it's still not perfect for all conditions, but is more reasonable for most than the default center position.
	drawverts[0].st[0] = XHairS[activecrosshair];
	drawverts[0].st[1] = XHairT[activecrosshair];
	drawverts[0].v[0] = (vid.width / 2 - 12 + cl_crossx.value);
	drawverts[0].v[1] = (vid.height / 2 - 8 + cl_crossy.value);

	drawverts[1].st[0] = XHairS[activecrosshair] + XHairSAdd;
	drawverts[1].st[1] = XHairT[activecrosshair];
	drawverts[1].v[0] = (vid.width / 2 - 12 + cl_crossx.value) + 24;
	drawverts[1].v[1] = (vid.height / 2 - 8 + cl_crossy.value);

	drawverts[2].st[0] = XHairS[activecrosshair] + XHairSAdd;
	drawverts[2].st[1] = XHairT[activecrosshair] + XHairTAdd;
	drawverts[2].v[0] = (vid.width / 2 - 12 + cl_crossx.value) + 24;
	drawverts[2].v[1] = (vid.height / 2 - 8 + cl_crossy.value) + 24;

	drawverts[3].st[0] = XHairS[activecrosshair];
	drawverts[3].st[1] = XHairT[activecrosshair] + XHairTAdd;
	drawverts[3].v[0] = (vid.width / 2 - 12 + cl_crossx.value);
	drawverts[3].v[1] = (vid.height / 2 - 8 + cl_crossy.value) + 24;

	glDrawArrays (GL_QUADS, 0, 4);

	// restore state
	glColor3f (1, 1, 1);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// test
	/*
	Draw_Font8x8String (64, 60, "Single Player", 0.75f);
	Draw_Font8x8String (64, 80, "Multiplayer", 0.75f);
	Draw_Font8x8String (64, 100, "Options", 0.75f);
	Draw_Font8x8String (64, 120, "Help/Ordering", 0.75f);
	Draw_Font8x8String (64, 140, "Quit", 0.75f);
	*/
}


/*
================
Draw_Character

Draws one 8*8 graphics character with 0 being transparent.
It can be clipped to the top of the screen to allow the console to be
smoothly scrolled off.
================
*/
void Draw_Character (int x, int y, int num)
{
	int				row, col;
	float			frow, fcol, size;

	// space
	if (num == 32) return;

	num &= 255;

	// totally off screen
	if (y <= -8) return;

	row = num >> 4;
	col = num & 15;

	frow = row * 0.0625;
	fcol = col * 0.0625;
	size = 0.03125;

	glBindTexture (GL_TEXTURE_2D, char_texture->texnum);

	drawverts[0].st[0] = fcol;
	drawverts[0].st[1] = frow;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = fcol + 0.03125;
	drawverts[1].st[1] = frow;
	drawverts[1].v[0] = x + 8;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = fcol + 0.03125;
	drawverts[2].st[1] = frow + 0.03125;
	drawverts[2].v[0] = x + 8;
	drawverts[2].v[1] = y + 8;

	drawverts[3].st[0] = fcol;
	drawverts[3].st[1] = frow + 0.03125;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + 8;

	glDrawArrays (GL_QUADS, 0, 4);
}

/*
================
Draw_String
================
*/
void Draw_String (int x, int y, char *str)
{
	while (*str)
	{
		Draw_Character (x, y, *str);
		str++;
		x += 8;
	}
}

/*
================
Draw_DebugChar

Draws a single character directly to the upper right corner of the screen.
This is for debugging lockups by drawing different chars in different parts
of the code.
================
*/
void Draw_DebugChar (char num)
{
}

/*
=============
Draw_AlphaPic
=============
*/
void Draw_AlphaPic (int x, int y, qpic_t *pic, float alpha)
{
	glpic_t			*gl;

	gl = (glpic_t *) pic->data;

	// nothing to draw
	if (!gl->gltexture) return;

	glColor4f (1, 1, 1, alpha);

	glBindTexture (GL_TEXTURE_2D, gl->gltexture->texnum);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	drawverts[0].st[0] = gl->sl;
	drawverts[0].st[1] = gl->tl;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = gl->sh;
	drawverts[1].st[1] = gl->tl;
	drawverts[1].v[0] = x + pic->width;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = gl->sh;
	drawverts[2].st[1] = gl->th;
	drawverts[2].v[0] = x + pic->width;
	drawverts[2].v[1] = y + pic->height;

	drawverts[3].st[0] = gl->sl;
	drawverts[3].st[1] = gl->th;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + pic->height;

	glDrawArrays (GL_QUADS, 0, 4);

	glColor4f (1, 1, 1, 1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}


/*
=============
Draw_Pic
=============
*/
void Draw_Pic (int x, int y, qpic_t *pic)
{
	glpic_t			*gl;

	gl = (glpic_t *) pic->data;

	// nothing to draw
	if (!gl->gltexture) return;

	glColor4f (1, 1, 1, 1);

	glBindTexture (GL_TEXTURE_2D, gl->gltexture->texnum);

	drawverts[0].st[0] = gl->sl;
	drawverts[0].st[1] = gl->tl;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = gl->sh;
	drawverts[1].st[1] = gl->tl;
	drawverts[1].v[0] = x + pic->width;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = gl->sh;
	drawverts[2].st[1] = gl->th;
	drawverts[2].v[0] = x + pic->width;
	drawverts[2].v[1] = y + pic->height;

	drawverts[3].st[0] = gl->sl;
	drawverts[3].st[1] = gl->th;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + pic->height;

	glDrawArrays (GL_QUADS, 0, 4);
}


void Draw_TextBox (int x, int y, int width, int lines)
{
	qpic_t	*p;
	int		cx, cy;
	int		n;

	// draw left side
	cx = x;
	cy = y;

	p = Draw_CachePic ("gfx/box_tl.lmp");
	Draw_Pic (cx, cy, p);

	p = Draw_CachePic ("gfx/box_ml.lmp");

	for (n = 0; n < lines; n++)
	{
		cy += 8;
		Draw_Pic (cx, cy, p);
	}

	p = Draw_CachePic ("gfx/box_bl.lmp");
	Draw_Pic (cx, cy + 8, p);

	// draw middle
	cx += 8;

	while (width > 0)
	{
		cy = y;
		p = Draw_CachePic ("gfx/box_tm.lmp");
		Draw_Pic (cx, cy, p);

		p = Draw_CachePic ("gfx/box_mm.lmp");

		for (n = 0; n < lines; n++)
		{
			cy += 8;

			if (n == 1) p = Draw_CachePic ("gfx/box_mm2.lmp");

			Draw_Pic (cx, cy, p);
		}

		p = Draw_CachePic ("gfx/box_bm.lmp");
		Draw_Pic (cx, cy + 8, p);

		width -= 2;
		cx += 16;
	}

	// draw right side
	cy = y;

	p = Draw_CachePic ("gfx/box_tr.lmp");
	Draw_Pic (cx, cy, p);

	p = Draw_CachePic ("gfx/box_mr.lmp");

	for (n = 0; n < lines; n++)
	{
		cy += 8;
		Draw_Pic (cx, cy, p);
	}

	p = Draw_CachePic ("gfx/box_br.lmp");
	Draw_Pic (cx, cy + 8, p);
}


void Draw_GenericMenuPic (int x, int y, int size)
{
	int xadd = (size * 32);
	int yadd = (size * 24);
	glpic_t *gl = (glpic_t *) conback->data;

	glBindTexture (GL_TEXTURE_2D, gl->gltexture->texnum);

	glColor4f (1, 1, 1, 1);

	drawverts[0].st[0] = 0;
	drawverts[0].st[1] = 0;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = 1;
	drawverts[1].st[1] = 0;
	drawverts[1].v[0] = x + xadd;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = 1;
	drawverts[2].st[1] = 1;
	drawverts[2].v[0] = x + xadd;
	drawverts[2].v[1] = y + yadd;

	drawverts[3].st[0] = 0;
	drawverts[3].st[1] = 1;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + yadd;

	glDrawArrays (GL_QUADS, 0, 4);
}


void Draw_MapshotPic (char *name, int x, int y, int size)
{
	extern int mapshottexturenum;

	int xadd = (size * 32);
	int yadd = (size * 24);

	glBindTexture (GL_TEXTURE_2D, mapshottexturenum);

	glColor4f (1, 1, 1, 1);

	drawverts[0].st[0] = 0;
	drawverts[0].st[1] = 0;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = 1;
	drawverts[1].st[1] = 0;
	drawverts[1].v[0] = x + xadd;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = 1;
	drawverts[2].st[1] = 1;
	drawverts[2].v[0] = x + xadd;
	drawverts[2].v[1] = y + yadd;

	drawverts[3].st[0] = 0;
	drawverts[3].st[1] = 1;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + yadd;

	glDrawArrays (GL_QUADS, 0, 4);
}


/*
=============
Draw_TransPic
=============
*/
void Draw_TransPic (int x, int y, qpic_t *pic)
{
	Draw_Pic (x, y, pic);
}


/*
=============
Draw_TransPicTranslate

Only used for the player color selection menu
=============
*/
void Draw_TransPicTranslate (int x, int y, qpic_t *pic, byte *translation)
{
	int				v, u, c;
	unsigned		trans[64*64], *dest;
	byte			*src;
	int				p;

	glBindTexture (GL_TEXTURE_2D, translate_texture);

	c = pic->width * pic->height;

	dest = trans;

	for (v = 0; v < 64; v++, dest += 64)
	{
		src = &menuplyr_pixels[((v * pic->height) >> 6) * pic->width];

		for (u = 0; u < 64; u++)
		{
			p = src[(u * pic->width) >> 6];

			if (p == 255)
				dest[u] = p;
			else
				dest[u] =  d_8to24table[translation[p]];
		}
	}

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, 64, 64, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);

	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glColor3f (1, 1, 1);

	drawverts[0].st[0] = 0;
	drawverts[0].st[1] = 0;
	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].st[0] = 1;
	drawverts[1].st[1] = 0;
	drawverts[1].v[0] = x + pic->width;
	drawverts[1].v[1] = y;

	drawverts[2].st[0] = 1;
	drawverts[2].st[1] = 1;
	drawverts[2].v[0] = x + pic->width;
	drawverts[2].v[1] = y + pic->height;

	drawverts[3].st[0] = 0;
	drawverts[3].st[1] = 1;
	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + pic->height;

	glDrawArrays (GL_QUADS, 0, 4);
}


/*
================
Draw_ConsoleBackground

================
*/
void Draw_ConsoleBackground (int lines)
{
	int y = (vid.height * 3) >> 2;

	if (lines > y)
		Draw_Pic (0, lines - vid.height, conback);
	else
		Draw_AlphaPic (0, lines - vid.height, conback, 0.75);
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c)
{
	glDisable (GL_TEXTURE_2D);

	glColor3f (host_basepal[c * 3] / 255.0, host_basepal[c * 3 + 1] / 255.0, host_basepal[c * 3 + 2] / 255.0);

	drawverts[0].v[0] = x;
	drawverts[0].v[1] = y;

	drawverts[1].v[0] = x + w;
	drawverts[1].v[1] = y;

	drawverts[2].v[0] = x + w;
	drawverts[2].v[1] = y + h;

	drawverts[3].v[0] = x;
	drawverts[3].v[1] = y + h;

	glDrawArrays (GL_QUADS, 0, 4);

	glColor3f (1, 1, 1);
	glEnable (GL_TEXTURE_2D);
}
//=============================================================================

/*
================
Draw_FadeScreen

================
*/
void Draw_FadeScreen (void)
{
	glEnable (GL_BLEND);
	glDisable (GL_TEXTURE_2D);
	glColor4f (0, 0, 0, 0.5);

	drawverts[0].v[0] = 0;
	drawverts[0].v[1] = 0;

	drawverts[1].v[0] = vid.width;
	drawverts[1].v[1] = 0;

	drawverts[2].v[0] = vid.width;
	drawverts[2].v[1] = vid.height;

	drawverts[3].v[0] = 0;
	drawverts[3].v[1] = vid.height;

	glDrawArrays (GL_QUADS, 0, 4);

	glColor4f (1, 1, 1, 1);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

//=============================================================================


void R_InitAliasRender (void);
void R_EndAliasRender (void);
void R_DrawAliasModel (entity_t *e);
void GL_Set2D (void);

void Draw_QuadModel (void)
{
	entity_t ent;
	model_t *QuadModel = Mod_ForName ("progs/quaddama.mdl", true);

	glViewport (24, 24, glwidth - 24, glheight - 24);
	glMatrixMode (GL_PROJECTION);
	gluPerspective (75, (float) glwidth / (float) glheight, 4, 4000);

	glMatrixMode (GL_MODELVIEW);

    glRotatef (-90, 1, 0, 0);	    // put Z going up
    glRotatef (90, 0, 0, 1);	    // put Z going up

	glCullFace (GL_FRONT);

	// set drawing parms
	if (gl_cull.value)
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);

	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);

	memset (&ent, 0, sizeof (entity_t));
	ent.model = QuadModel;

	ent.origin[0] = 0;
	ent.origin[1] = 300;
	ent.origin[2] = 0;

	R_InitAliasRender ();
	R_DrawAliasModel (&ent);
	R_EndAliasRender ();

	GL_Set2D ();
}


/*
================
Draw_BeginDisc

Draws the little blue disc in the corner of the screen.
Call before beginning any disc IO.

Removed this because disk IO takes no time at all these days.
================
*/
void Draw_BeginDisc (void)
{
}


/*
================
Draw_EndDisc

Erases the disc icon.
Call after completing any disc IO
================
*/
void Draw_EndDisc (void)
{
}

/*
============
R_PolyBlend
============
*/
void R_PolyBlend (void)
{
	if (!gl_polyblend.value) return;
	if (!v_blend[3]) return;

	// console is full screen
	if (scr_con_current == vid.height) return;

	// no polyblend in intermission
	if (cl.intermission) return;

	// not connected
	if (cls.state != ca_connected) return;

	glEnable (GL_BLEND);
	glDisable (GL_TEXTURE_2D);

	glColor4fv (v_blend);

	drawverts[0].v[0] = 0;
	drawverts[0].v[1] = 0;

	drawverts[1].v[0] = vid.width;
	drawverts[1].v[1] = 0;

	drawverts[2].v[0] = vid.width;
	drawverts[2].v[1] = vid.height;

	drawverts[3].v[0] = 0;
	drawverts[3].v[1] = vid.height;

	glDrawArrays (GL_QUADS, 0, 4);

	glEnable (GL_TEXTURE_2D);

	glColor4f (1, 1, 1, 1);
}


void Draw_Brightness (void)
{
	float brightval;

	// no need
	if (gl_brightness.value > 0.98 && gl_brightness.value < 1.02) return;

	// clamp
	if (gl_brightness.value < 0) gl_brightness.value = 0;
	if (gl_brightness.value > 2) gl_brightness.value = 2;

	brightval = gl_brightness.value / 2.0f;

	glEnable (GL_BLEND);
	glBlendFunc (GL_DST_COLOR, GL_SRC_COLOR);
	glDisable (GL_TEXTURE_2D);

	glColor3f (brightval, brightval, brightval);

	drawverts[0].v[0] = 0;
	drawverts[0].v[1] = 0;

	drawverts[1].v[0] = vid.width;
	drawverts[1].v[1] = 0;

	drawverts[2].v[0] = vid.width;
	drawverts[2].v[1] = vid.height;

	drawverts[3].v[0] = 0;
	drawverts[3].v[1] = vid.height;

	glDrawArrays (GL_QUADS, 0, 4);

	glEnable (GL_TEXTURE_2D);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f (1, 1, 1, 1);

	// this is the last thing drawn, so...
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glDisableClientState (GL_VERTEX_ARRAY);
}


/*
================
GL_Set2D

Setup as if the screen was 320*200
================
*/
void Draw_DeathScreen (void);

void GL_Set2D (void)
{
	glViewport (0, 0, glwidth, glheight);

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();

	glOrtho  (0, vid.width, vid.height, 0, -99999, 99999);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);
	glEnable (GL_BLEND);

	glColor4f (1,1,1,1);

	// bring up vertex pointer
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (2, GL_FLOAT, sizeof (drawverts_t), drawverts[0].v);

	// bring up image texture array
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (drawverts_t), drawverts[0].st);

	// do polyblend here rather than in 3d
	R_PolyBlend ();

	// do blood and death before anything else
	// R_DrawBlood ();
	Draw_DeathScreen ();
}

//====================================================================


/*
========================================================================================

	BBB   L     OO    OO   DDD      AA   N  N  DDD     DDD   EEEE   AA  TTTTT  H  H
	B  B  L    O  O  O  O  D  D    A  A  NN N  D  D    D  D  E     A  A   T    H  H
	BBB   L    O  O  O  O  D  D    A  A  N NN  D  D    D  D  EEE   A  A   T    HHHH
	B  B  L    O  O  O  O  D  D    AAAA  N  N  D  D    D  D  E     AAAA   T    H  H
	BBB   LLLL  OO    OO   DDD     A  A  N  N  DDD     DDD   EEEE  A  A   T    H  H

========================================================================================
*/


float DeathAlpha = 0;
float DeathRed = 0.666;

void Draw_DeathScreen (void)
{
	// console is full screen
	if (scr_con_current == vid.height) return;

	// no death in intermission
	if (cl.intermission) return;

	// not connected
	if (cls.state != ca_connected) return;

	if (cl.stats[STAT_HEALTH] > 0)
	{
		DeathAlpha = 0;
		DeathRed = 0.666;
		return;
	}

	glEnable (GL_BLEND);
	glDisable (GL_TEXTURE_2D);
	glColor4f (DeathRed, 0, 0, DeathAlpha);

	drawverts[0].v[0] = 0;
	drawverts[0].v[1] = 0;

	drawverts[1].v[0] = vid.width;
	drawverts[1].v[1] = 0;

	drawverts[2].v[0] = vid.width;
	drawverts[2].v[1] = vid.height;

	drawverts[3].v[0] = 0;
	drawverts[3].v[1] = vid.height;

	glDrawArrays (GL_QUADS, 0, 4);

	glDisableClientState (GL_VERTEX_ARRAY);

	glColor4f (1,1,1,1);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	DeathAlpha += r_frametime / 5.0;
	DeathRed -= r_frametime / 8.0;
}


