/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "quakedef.h"

void BuildSurfaceDisplayList (msurface_t *surf);

/*
==================
R_AnimateLight

Added lightstyle interpolation.  Yummy!!!
==================
*/
void R_AnimateLight (void)
{
	int			j, k;
	float		l;
	int			flight;
	int			clight;
	float		lerpfrac;
	float		backlerp;

	// light animations
	// 'm' is normal light, 'a' is no light, 'z' is double bright
	flight = (int) floor (cl.time * 10);
	clight = (int) ceil (cl.time * 10);
	lerpfrac = (cl.time * 10) - flight;
	backlerp = 1.0f - lerpfrac;

	for (j = 0; j < MAX_LIGHTSTYLES; j++)
	{
		if (!cl_lightstyle[j].length)
		{
			// was 256, changed to 264 for consistency
			d_lightstylevalue[j] = 264;
			continue;
		}
		else if (cl_lightstyle[j].length == 1)
		{
			// single length style so don't bother interpolating
			d_lightstylevalue[j] = 22 * (cl_lightstyle[j].map[0] - 'a');
			continue;
		}

		// interpolate animating light
		// frame just gone
		k = flight % cl_lightstyle[j].length;
		k = cl_lightstyle[j].map[k] - 'a';
		l = (float) (k * 22) * backlerp;

		// upcoming frame
		k = clight % cl_lightstyle[j].length;
		k = cl_lightstyle[j].map[k] - 'a';
		l += (float) (k * 22) * lerpfrac;

		d_lightstylevalue[j] = (int) l;
	}
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/

// this comes from chase.c - major "!!!!" if we ever remove the chasecam...!
void TraceLine (vec3_t start, vec3_t end, vec3_t impact);

// ray casting for dynamic lights.  the dlight must not be blocked on it's way to the surf
// midpoint in a straight line for it to be valid for the surf
qboolean CastRay (msurface_t *surf, vec3_t lightorg)
{
	vec3_t impact;
	float dist;
	int i;

	for (i = 0; i < surf->polys->numverts; i++)
	{
		// run a trace
		TraceLine (surf->polys->verts[i].tv, lightorg, impact);

		dist = DotProduct (lightorg, impact);

		// allow some fuzziness in dist
		if (dist < 10) return true;
	}

	return false;
}


/*
=============
R_MarkLights
=============
*/
void R_MarkLights (dlight_t *light, int lnum, mnode_t *node)
{
	int			i;
	int			j;
	mplane_t	*splitplane;
	float		dist;
	msurface_t	*surf;

	if (node->contents < 0) return;
	if (node->visframe != r_visframecount) return;

	splitplane = node->plane;
	dist = DotProduct (light->origin, splitplane->normal) - splitplane->dist;

	if (dist > light->radius)
	{
		R_MarkLights (light, lnum, node->children[0]);
		return;
	}

	if (dist < -light->radius)
	{
		R_MarkLights (light, lnum, node->children[1]);
		return;
	}

	// add dlights to surfs in this node
	surf = cl.worldmodel->surfaces + node->firstsurface;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		// surf not visible this frame
		if (surf->visframe != r_framecount) continue;

		// !!!!!
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		if (surf->dlightframe != r_framecount)
		{
			// init a new dlight on the surf
			for (j = 0; j < MAX_DLIGHTS; j += 8)
				surf->dlbits[(j >> 3)] = 0;

			surf->dlightframe = r_framecount;
		}

		// mark it
		surf->dlbits[(lnum >> 3)] |= (1 << (lnum & 7));

		// the light has hit a surf now
		light->active = true;
	}

	// recurse down both sides
	R_MarkLights (light, lnum, node->children[0]);
	R_MarkLights (light, lnum, node->children[1]);
}


/*
=============
R_PushDlights
=============
*/
void R_PushDlights (mnode_t *headnode)
{
	int		i;
	dlight_t	*l;

	l = cl_dlights;

	for (i = 0; i < MAX_DLIGHTS; i++, l++)
	{
		// initially inactive
		l->active = false;

		if (l->die < cl.time || l->radius < 0.0f)
			continue;

		R_MarkLights (l, i, headnode);
	}
}


/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

mplane_t		*lightplane;
vec3_t			lightspot;
int				lightcolour[3];

int RecursiveLightPoint (mnode_t *node, vec3_t start, vec3_t end)
{
	int			r;
	float		front, back, frac;
	int			side;
	mplane_t	*plane;
	vec3_t		mid;
	msurface_t	*surf;
	int			s, t, ds, dt;
	int			i;
	mtexinfo_t	*tex;
	byte		*lightmap;
	unsigned	scale;
	int			maps;

	// didn't hit anything
	if (node->contents < 0) return -1;

	// calculate mid point
	plane = node->plane;
	front = DotProduct (start, plane->normal) - plane->dist;
	back = DotProduct (end, plane->normal) - plane->dist;
	side = front < 0;

	if ((back < 0) == side) return RecursiveLightPoint (node->children[side], start, end);

	frac = front / (front-back);
	mid[0] = start[0] + (end[0] - start[0]) * frac;
	mid[1] = start[1] + (end[1] - start[1]) * frac;
	mid[2] = start[2] + (end[2] - start[2]) * frac;

	// go down front side	
	r = RecursiveLightPoint (node->children[side], start, mid);

	// hit something
	if (r >= 0) return r;

	// didn't hit anything
	if ((back < 0) == side) return -1;

	// check for impact on this node
	VectorCopy (mid, lightspot);
	lightplane = plane;

	surf = cl.worldmodel->surfaces + node->firstsurface;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		// no lightmaps
		if (surf->flags & SURF_DRAWTILED) continue;

		tex = surf->texinfo;

		s = DotProduct (mid, tex->vecs[0]) + tex->vecs[0][3];
		t = DotProduct (mid, tex->vecs[1]) + tex->vecs[1][3];;

		if (s < surf->texturemins[0] || t < surf->texturemins[1]) continue;

		ds = s - surf->texturemins[0];
		dt = t - surf->texturemins[1];

		if (ds > surf->extents[0] || dt > surf->extents[1]) continue;

		if (!surf->samples) return 0;

		ds >>= 4;
		dt >>= 4;

		lightmap = surf->samples;
		r = lightcolour[0] = lightcolour[1] = lightcolour[2] = 0;

		if (lightmap)
		{
			lightmap += 3 * (dt * ((surf->extents[0] >> 4) + 1) + ds);

			for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
			{
				scale = d_lightstylevalue[surf->styles[maps]];

				r += *lightmap * scale;
				lightcolour[0] += lightmap[0] * scale;
				lightcolour[1] += lightmap[1] * scale;
				lightcolour[2] += lightmap[2] * scale;

				lightmap += surf->smax * surf->tmax * 3;
			}

			r >>= 8;
			lightcolour[0] >>= 8;
			lightcolour[1] >>= 8;
			lightcolour[2] >>= 8;
		}

		return r;
	}

	// go down back side
	return RecursiveLightPoint (node->children[!side], mid, end);
}

void R_LightPoint (vec3_t p, vec3_t colour)
{
	vec3_t		end;
	int			r;
	int			lnum;
	vec3_t		dist;
	float		add;

	if (!cl.worldmodel->lightdata)
	{
		colour[0] = colour[1] = colour[2] = 255;
		return;
	}

	end[0] = p[0];
	end[1] = p[1];
	end[2] = p[2] - 4096;
	
	r = RecursiveLightPoint (cl.worldmodel->nodes, p, end);
	
	if (r < 0)
		r = 0;

	colour[0] = lightcolour[0];
	colour[1] = lightcolour[1];
	colour[2] = lightcolour[2];

	for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
	{
		if (cl_dlights[lnum].die >= cl.time)
		{
			VectorSubtract (p, cl_dlights[lnum].origin, dist);

			add = cl_dlights[lnum].radius - Length (dist);

			if (add > 0)
			{
				colour[0] += add;
				colour[1] += add;
				colour[2] += add;
			}
		}
	}

	// never allow to go totally dark
	colour[0] = (colour[0] + 16.0f) / 512.0f;
	colour[1] = (colour[1] + 16.0f) / 512.0f;
	colour[2] = (colour[2] + 16.0f) / 512.0f;

	return;
}


/*
================================================================================================================

				LIGHTMAPS

================================================================================================================
*/


unsigned		blocklights[3][18 * 18];

lightmap_t *lightmaps = NULL;


/*
===============
R_AddDynamicLights
===============
*/
void R_AddDynamicLights (msurface_t *surf)
{
	int			lnum;
	int			sd, td;
	float		dist, rad, minlight;
	vec3_t		impact, local;
	int			s, t;
	int			i;
	mtexinfo_t	*tex;

	tex = surf->texinfo;

	for (lnum = 0; lnum < MAX_DLIGHTS; lnum++)
	{
		// dlight is on no surfs
		if (!cl_dlights[lnum].active) continue;

		// dead light
		// id forgot this...!!!
		if (cl_dlights[lnum].die < cl.time) continue;

		// not marked for checking by this light
		if (!(surf->dlbits[(lnum >> 3)] & (1 << (lnum & 7)))) continue;

		rad = cl_dlights[lnum].radius;
		dist = DotProduct (cl_dlights[lnum].origin, surf->plane->normal) - surf->plane->dist;

		rad -= fabs(dist);
		minlight = cl_dlights[lnum].minlight;

		if (rad < minlight)
			continue;

		minlight = rad - minlight;

		for (i=0 ; i<3 ; i++)
			impact[i] = cl_dlights[lnum].origin[i] - surf->plane->normal[i]*dist;

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];
		
		for (t = 0 ; t<surf->tmax ; t++)
		{
			td = local[1] - t*16;

			if (td < 0)
				td = -td;

			for (s=0 ; s<surf->smax ; s++)
			{
				sd = local[0] - s*16;

				if (sd < 0)
					sd = -sd;

				if (sd > td)
					dist = sd + (td>>1);
				else
					dist = td + (sd>>1);

				if (dist < minlight)
				{
					blocklights[0][t * surf->smax + s] += (rad - dist) * cl_dlights[lnum].rgb[0];
					blocklights[1][t * surf->smax + s] += (rad - dist) * cl_dlights[lnum].rgb[1];
					blocklights[2][t * surf->smax + s] += (rad - dist) * cl_dlights[lnum].rgb[2];

					// mark the surf as modified by dynamic
					surf->lightmodified |= LIGHT_MODIFIED_DYNAMIC;
				}
			}
		}
	}
}


extern byte DigitalVibrance[256];

__inline void R_AddBlockLight (int blin, byte *dest)
{
	*dest = (blin > 65280) ? 255 : DigitalVibrance[(byte) (blin >> 8)];
}


/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the 8.8 format in blocklights
===============
*/
void R_BuildLightMap (msurface_t *surf, int stride)
{
	int			i, j, size;
	byte		*lightmap;
	unsigned	scale;
	int			maps;
	unsigned	*bl[3];
	byte		*dest = surf->base;

	surf->cached_dlight = (surf->dlightframe == r_framecount);

	size = surf->smax * surf->tmax;
	lightmap = surf->samples;

	// set to full bright if no light data
	if (!cl.worldmodel->lightdata)
	{
		for (i = 0; i < size; i++)
		{
			blocklights[0][i] = 65280;
			blocklights[1][i] = 65280;
			blocklights[2][i] = 65280;
		}

		goto store;
	}

	// clear to no light
	for (i = 0; i < size; i++)
	{
		blocklights[0][i] = 0;
		blocklights[1][i] = 0;
		blocklights[2][i] = 0;
	}

	// add all the dynamic lights first
	if (surf->dlightframe == r_framecount)
		R_AddDynamicLights (surf);

	// lightmap wasn't modified at all!
	if (!surf->lightmodified) return;

	// add all the lightmaps
	if (lightmap)
	{
		for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
		{
			scale = d_lightstylevalue[surf->styles[maps]];
			surf->cached_light[maps] = scale;

			for (i = 0; i < size; i++)
			{
				blocklights[0][i] += lightmap[i * 3 + 0] * scale;
				blocklights[1][i] += lightmap[i * 3 + 1] * scale;
				blocklights[2][i] += lightmap[i * 3 + 2] * scale;
			}

			// skip to next lightmap
			lightmap += (size * 3);
		}
	}

	// bound, invert, and shift
store:
	bl[0] = blocklights[0];
	bl[1] = blocklights[1];
	bl[2] = blocklights[2];

	stride -= (surf->smax << 2);

	for (i = 0; i < surf->tmax; i++, dest += stride)
	{
		for (j = 0; j < surf->smax; j++)
		{
			R_AddBlockLight (*bl[0]++, &dest[0]);
			R_AddBlockLight (*bl[1]++, &dest[1]);
			R_AddBlockLight (*bl[2]++, &dest[2]);
			dest[3] = 255;

			dest += LIGHTMAP_BYTES;
		}
	}
}


/*
================
R_BlendLightmaps

upload all lightmaps that were modified
================
*/
void R_BlendLightmaps (void)
{
	glRect_t *theRect;
	lightmap_t	*lm;

	for (lm = lightmaps; lm; lm = lm->next)
	{
		if (lm->modified)
		{
			lm->modified = false;
			theRect = &lm->rectchange;

			glBindTexture (GL_TEXTURE_2D, lm->texnum);

			glTexSubImage2D
			(
				GL_TEXTURE_2D,
				0,
				0,
				theRect->t, 
				LM_BLOCK_WIDTH,
				theRect->h,
				GL_RGBA,
				GL_UNSIGNED_BYTE,
				lm->data + (theRect->t * LM_BLOCK_WIDTH * LIGHTMAP_BYTES)
			);

			theRect->l = LM_BLOCK_WIDTH;
			theRect->t = LM_BLOCK_HEIGHT;
			theRect->h = 0;
			theRect->w = 0;
		}
	}
}


/*
================
R_RenderDynamicLightmaps

================
*/
void R_RenderDynamicLightmaps (void)
{
	int			maps;
	glRect_t    *theRect;
	extern msurface_t  *genericchain;
	msurface_t *surf;
	qboolean dynamic;

	for (surf = genericchain; surf; surf = surf->genericchain)
	{
		// initially unmodified
		surf->lightmodified = LIGHT_MODIFIED_UNMODIFIED;

		if (surf->flags & (SURF_DRAWSKY | SURF_DRAWTURB))
			continue;

		dynamic = false;

		// check for lightmap modification
		for (maps = 0; maps < MAXLIGHTMAPS && surf->styles[maps] != 255; maps++)
		{
			if (d_lightstylevalue[surf->styles[maps]] != surf->cached_light[maps])
			{
				// modified by static
				surf->lightmodified |= LIGHT_MODIFIED_STATIC;
				dynamic = true;
			}
		}

		if (surf->cached_dlight)
		{
			// always force to dynamic modification if it previously had a dynamic light as it 
			// will always need to be rebuilt
			surf->lightmodified |= LIGHT_MODIFIED_DYNAMIC;
			dynamic = true;
		}

		if (surf->dlightframe == r_framecount || dynamic)
		{
			// build the lightmap for the surf first
			R_BuildLightMap (surf, LM_BLOCK_WIDTH * LIGHTMAP_BYTES);

			// light wasn't truly modified
			if (!surf->lightmodified) continue;

			// now do the necessary stuff to reflect the modification
			surf->lightmap->modified = true;
			theRect = &surf->lightmap->rectchange;

			if (surf->light_t < theRect->t)
			{
				if (theRect->h)
					theRect->h += theRect->t - surf->light_t;
				theRect->t = surf->light_t;
			}

			if (surf->light_s < theRect->l)
			{
				if (theRect->w)
					theRect->w += theRect->l - surf->light_s;
				theRect->l = surf->light_s;
			}

			if ((theRect->w + theRect->l) < (surf->light_s + surf->smax)) theRect->w = (surf->light_s - theRect->l) + surf->smax;
			if ((theRect->h + theRect->t) < (surf->light_t + surf->tmax)) theRect->h = (surf->light_t - theRect->t) + surf->tmax;
		}
	}
}


/*
=============================================================================

  LIGHTMAP ALLOCATION

=============================================================================
*/

// returns a texture number and the position inside it
lightmap_t *AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	lightmap_t *lm;

	for (lm = lightmaps; lm; lm = lm->next)
	{
		best = LM_BLOCK_HEIGHT;

		for (i = 0; i < LM_BLOCK_WIDTH - w; i++)
		{
			best2 = 0;

			for (j = 0; j < w; j++)
			{
				if (lm->allocated[i + j] >= best) break;
				if (lm->allocated[i + j] > best2) best2 = lm->allocated[i + j];
			}

			if (j == w)
			{
				// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > LM_BLOCK_HEIGHT)
			continue;

		for (i = 0; i < w; i++)
			lm->allocated[*x + i] = best + h;

		return lm;
	}

	// allocate and init a new lightmap
	lm = (lightmap_t *) malloc (sizeof (lightmap_t));
	memset (lm, 0, sizeof (lightmap_t));

	// make a texture
	GL_AllocTexture (&lm->texnum);

	// link it in
	lm->next = lightmaps;
	lightmaps = lm;

	// call recursively to get the new lightmap
	return AllocBlock (w, h, x, y);
}


void MakeLightmapST (msurface_t *surf, vec3_t vec, float *st)
{
	float s;
	float t;

	// lightmap texture coordinates
	s = DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3];
	s -= surf->texturemins[0];
	s += surf->light_s * 16;
	s += 8;
	s /= LM_BLOCK_WIDTH * 16;

	t = DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3];
	t -= surf->texturemins[1];
	t += surf->light_t * 16;
	t += 8;
	t /= LM_BLOCK_HEIGHT * 16;

	while (s > 1.0f) s -= 1.0f;
	while (s < 0.0f) s += 1.0f;

	while (t > 1.0f) t -= 1.0f;
	while (t < 0.0f) t += 1.0f;

	st[0] = s;
	st[1] = t;
}


/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (msurface_t *surf)
{
	if (surf->flags & (SURF_DRAWSKY | SURF_DRAWTURB))
		return;

	// set stored surf data (rather than having to constantly recalc as it's required)
	surf->smax = (surf->extents[0] >> 4) + 1;
	surf->tmax = (surf->extents[1] >> 4) + 1;

	surf->lightmap = AllocBlock (surf->smax, surf->tmax, &surf->light_s, &surf->light_t);

	surf->base = surf->lightmap->data;
	surf->base += (surf->light_t * LM_BLOCK_WIDTH + surf->light_s) * LIGHTMAP_BYTES;

	// tell R_BuildLightMap that we want this to be modified regardless
	surf->lightmodified = LIGHT_MODIFIED_STATIC;

	// and buid the lightmap
	R_BuildLightMap (surf, LM_BLOCK_WIDTH * LIGHTMAP_BYTES);
}


/*
==================
GL_BuildLightmaps

Builds the lightmap texture
with all the surfaces from all brush models
==================
*/
extern mvertex_t	*r_pcurrentvertbase;
extern model_t		*currentmodel;
void GL_DoubleSizeTexture (int w, int h, unsigned *data);

void GL_BuildLightmaps (void)
{
	int		i, j;
	model_t	*m;
	lightmap_t *lm;
	lightmap_t *Free;

	for (i = 0, lm = lightmaps; ;)
	{
		// no more lightmaps
		if (!lm) break;

		// swap pointers
		Free = lm;
		lm = Free->next;

		// free Free
		glDeleteTextures (1, &Free->texnum);

		free (Free);
		i++;
	}

	Con_DPrintf ("Released %i Lightmaps\n", i);

	// explicitly null
	lightmaps = NULL;

	r_framecount = 1;		// no dlightcache

	for (j = 1; j < MAX_MODELS; j++)
	{
		m = cl.model_precache[j];
		if (!m) break;
		if (m->name[0] == '*') continue;

		r_pcurrentvertbase = m->vertexes;
		currentmodel = m;

		for (i = 0; i < m->numsurfaces; i++)
		{
			GL_CreateSurfaceLightmap (m->surfaces + i);

			if (m->surfaces[i].flags & SURF_DRAWTURB)
				continue;

			BuildSurfaceDisplayList (m->surfaces + i);
		}
	}

	// upload all lightmaps that were filled
	for (lm = lightmaps, i = 0; lm; lm = lm->next)
	{
		lm->modified = false;

		lm->rectchange.l = LM_BLOCK_WIDTH;
		lm->rectchange.t = LM_BLOCK_HEIGHT;
		lm->rectchange.w = 0;
		lm->rectchange.h = 0;

		glBindTexture (GL_TEXTURE_2D, lm->texnum);

		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glTexImage2D
		(
			GL_TEXTURE_2D,
			0,
			LIGHTMAP_BYTES,
			LM_BLOCK_WIDTH,
			LM_BLOCK_HEIGHT,
			0,
			GL_RGBA,
			GL_UNSIGNED_BYTE,
			lm->data
		);

		// WriteTGA (va ("lightmap_%i", i), LM_BLOCK_WIDTH, LM_BLOCK_HEIGHT, lm->data, 32);

		i++;
	}

	Con_DPrintf ("Allocated %i Lightmaps\n", i);
}

