/*
   Copyright (C) 1996-1997 Id Software, Inc. This program is free software;
   you can redistribute it and/or modify it under the terms of the GNU General
   Public License as published by the Free Software Foundation;
   either version 2 of the License, or (at your option) any later version. This
   program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
   without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
   PARTICULAR PURPOSE. See the GNU General Public License for more details. You
   should have received a copy of the GNU General Public License along with this program;
   if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite
   330, Boston, MA 02111-1307, USA.
 */
#include "quakedef.h"
#include "r_local.h"

#define MAX_PARTICLES			32768	// default max # of particles at one time
#define ABSOLUTE_MIN_PARTICLES	4096	// no fewer than this no matter what's on the command line

typedef enum
{
	pt_kill,
	pt_static,
	pt_grav,
	pt_slowgrav,
	pt_lavasplash,
	pt_fire,
	pt_explode,
	pt_explode2,
	pt_explode3,
	pt_blob,
	pt_blob2,
	pt_antigrav
} ptype_t;

// st coords. all particles are 64 x 64 font based images, s and t increase by
// 0.25 to make the full image
float particlest[20][4] = 
{
//	 smin    tmin    smax    tmax
	{0.000f, 0.000f, 0.125f, 0.250f},		// pd_blood1
	{0.125f, 0.000f, 0.250f, 0.250f},		// pd_blood2
	{0.250f, 0.000f, 0.375f, 0.250f},		// pd_blood3
	{0.375f, 0.000f, 0.500f, 0.250f},		// pd_blood4

	{0.500f, 0.000f, 0.625f, 0.250f},		// pd_tracer
	{0.625f, 0.000f, 0.750f, 0.250f},		// pd_bubble
	{0.750f, 0.000f, 0.875f, 0.250f},		// pd_bulletmark
	{0.875f, 0.000f, 1.000f, 0.250f},		// pd_burnmark

	{0.000f, 0.250f, 0.250f, 0.750f},		// pd_smoke
	{0.250f, 0.250f, 0.500f, 0.750f},		// pd_flare1
	{0.500f, 0.250f, 0.750f, 0.750f},		// pd_flare2
	{0.750f, 0.250f, 1.000f, 0.750f},		// pd_flash

	{0.000f, 0.750f, 0.125f, 1.000f},		// pd_spark
	{0.125f, 0.750f, 0.250f, 1.000f},		// pd_dot
	{0.250f, 0.000f, 0.375f, 0.250f},		// pd_unknown1
	{0.375f, 0.000f, 0.500f, 0.250f},		// pd_unknown2

	{0.500f, 0.000f, 0.625f, 0.250f},		// pd_unknown3
	{0.625f, 0.000f, 0.750f, 0.250f},		// pd_unknown4
	{0.750f, 0.000f, 0.875f, 0.250f},		// pd_unknown5
	{0.875f, 0.000f, 1.000f, 0.250f},		// pd_unknown6
};

// !!! if this is changed, it must be changed in d_ifacea.h too !!!
typedef struct particle_s
{
	// driver-usable fields
	vec3_t				org;
	float				color;

	// drivers never touch the following fields
	struct particle_s	*next;
	vec3_t				vel;
	float				ramp;
	float				die;
	ptype_t				type;

	float				radius;
	float				alpha;

	int					drawtype;

	float				scale;

	// 2 different blends
	struct particle_s	*chain;
} particle_t;

// 2 different blends
particle_t	*particlechain[3];

int			ramp1[8] = { 0x6f, 0x6d, 0x6b, 0x69, 0x67, 0x65, 0x63, 0x61 };
int			ramp2[8] = { 0x6f, 0x6e, 0x6d, 0x6c, 0x6b, 0x6a, 0x68, 0x66 };
int			ramp3[8] = { 0x6d, 0x6b, 6, 5, 4, 3 };

particle_t	*active_particles, *free_particles;

particle_t	*particles;
int			r_numparticles;

vec3_t		r_pright, r_pup, r_ppn;


/*
 =======================================================================================================================
    R_InitParticles
 =======================================================================================================================
 */
void R_InitParticles (void)
{
	int i;

	i = COM_CheckParm ("-particles");

	if (i)
	{
		r_numparticles = (int) (Q_atoi (com_argv[i + 1]));

		if (r_numparticles < ABSOLUTE_MIN_PARTICLES)
			r_numparticles = ABSOLUTE_MIN_PARTICLES;
	}
	else
		r_numparticles = MAX_PARTICLES;

	Con_Printf ("\nParticle Memory: %i KB\n", (r_numparticles * sizeof (particle_t) + 1023) / 1024);

	particles = (particle_t *) Hunk_AllocName (r_numparticles * sizeof (particle_t), "particles");
}


/* R_EntityParticles */
#define NUMVERTEXNORMALS	162
extern float	r_avertexnormals[NUMVERTEXNORMALS][3];
vec3_t			avelocities[NUMVERTEXNORMALS];
float			beamlength = 16;
vec3_t			avelocity = { 23, 7, 3 };
float			partstep = 0.01;
float			timescale = 0.01;


void R_EntityParticles (entity_t *ent)
{
	int			count;
	int			i;
	particle_t	*p;
	float		angle;
	float		sr, sp, sy, cr, cp, cy;
	vec3_t		forward;
	float		dist;

	dist = 64;
	count = 50;

	if (!avelocities[0][0])
	{
		for (i = 0; i < NUMVERTEXNORMALS * 3; i++)
			avelocities[0][i] = (rand () & 255) * 0.01;
	}

	for (i = 0; i < NUMVERTEXNORMALS; i++)
	{
		angle = cl.time * avelocities[i][0];
		sy = sin (angle);
		cy = cos (angle);
		angle = cl.time * avelocities[i][1];
		sp = sin (angle);
		cp = cos (angle);
		angle = cl.time * avelocities[i][2];
		sr = sin (angle);
		cr = cos (angle);

		forward[0] = cp * cy;
		forward[1] = cp * sy;
		forward[2] = -sp;

		if (!free_particles)
			return ;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.01;
		p->color = 0x6f;
		p->type = pt_explode;

		p->org[0] = ent->origin[0] + r_avertexnormals[i][0] * dist + forward[0] * beamlength;
		p->org[1] = ent->origin[1] + r_avertexnormals[i][1] * dist + forward[1] * beamlength;
		p->org[2] = ent->origin[2] + r_avertexnormals[i][2] * dist + forward[2] * beamlength;
	}
}


/*
 =======================================================================================================================
    R_ClearParticles
 =======================================================================================================================
 */
void R_ClearParticles (void)
{
	int i;

	free_particles = &particles[0];
	active_particles = NULL;

	for (i = 0; i < r_numparticles; i++)
	{
		particles[i].next = &particles[i + 1];

		// initial rendering params
		particles[i].drawtype = pd_dot;
		particles[i].radius = 1.5f;
		particles[i].alpha = 1.0f;
		particles[i].scale = 0.004;
	}

	particles[r_numparticles - 1].next = NULL;
}


void R_ReadPointFile_f (void)
{
	FILE		*f;
	vec3_t		org;
	int			r;
	int			c;
	particle_t	*p;
	char		name[MAX_OSPATH];

	sprintf (name, "maps/%s.pts", sv.name);

	COM_FOpenFile (name, &f);
	if (!f)
	{
		Con_Printf ("couldn't open %s\n", name);
		return ;
	}

	Con_Printf ("Reading %s...\n", name);
	c = 0;
	for (;;)
	{
		r = fscanf (f, "%f %f %f\n", &org[0], &org[1], &org[2]);
		if (r != 3)
			break;
		c++;

		if (!free_particles)
		{
			Con_Printf ("Not enough free particles\n");
			break;
		}

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = 99999;
		p->color = (-c) & 15;
		p->type = pt_static;
		VectorCopy (vec3_origin, p->vel);
		VectorCopy (org, p->org);
	}

	fclose (f);
	Con_Printf ("%i points read\n", c);
}


/*
 =======================================================================================================================
    R_ParticleExplosion
 =======================================================================================================================
 */
void R_ParticleExplosion (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;

	for (i = 0; i < 8; i++)
	{
		if (!free_particles) return;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 5;
		p->color = ramp1[0];
		p->ramp = 0;

		p->drawtype = pd_flash;
		p->radius = 4 + (rand () % 5);
		p->alpha = 4.0f + (float) (rand () % 11) / 10.0f;

		p->type = pt_explode;

		for (j = 0; j < 3; j++)
		{
			p->org[j] = org[j] + ((rand () % 16) - 8);
			p->vel[j] = (rand () % 16) - 8;
		}

		for (k = 0; k < 8; k++)
		{
			p = free_particles;
			free_particles = p->next;
			p->next = active_particles;
			active_particles = p;

			p->die = cl.time + 666;
			p->color = 3 + (rand () % 7);
			p->ramp = 0;

			p->drawtype = pd_smoke;
			p->radius = 20 + (rand () % 20);
			p->alpha = 0.666;

			p->type = pt_antigrav;

			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand () % 64) - 32);
				p->vel[j] = (rand () % 64) - 32;
			}
		}
	}
}


/*
 =======================================================================================================================
    R_ParticleExplosion2
 =======================================================================================================================
 */
void R_ParticleExplosion2 (vec3_t org, int colorStart, int colorLength)
{
	R_ParticleExplosion (org);
}


/*
 =======================================================================================================================
    R_BlobExplosion
 =======================================================================================================================
 */
void R_BlobExplosion (vec3_t org)
{
	int			i, j;
	particle_t	*p;

	for (i = 0; i < 1024; i++)
	{
		if (!free_particles)
			return ;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 1 + (rand () & 8) * 0.05;

		if (i & 1)
		{
			p->type = pt_blob;
			p->color = 66 + rand () % 6;
			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand () % 32) - 16);
				p->vel[j] = (rand () % 512) - 256;
			}
		}
		else
		{
			p->type = pt_blob2;
			p->color = 150 + rand () % 6;
			for (j = 0; j < 3; j++)
			{
				p->org[j] = org[j] + ((rand () % 32) - 16);
				p->vel[j] = (rand () % 512) - 256;
			}
		}
	}
}


/*
 =======================================================================================================================
    R_LavaSplash
 =======================================================================================================================
 */
void R_LavaSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i = -16; i <= 16; i += 2)
	{
		for (j = -16; j <= 16; j += 2)
		{
			for (k = 0; k < 1; k++)
			{
				if (!free_particles)
					return ;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				// full alpha ramp
				p->die = cl.time + 666;

				if (i > -15 && i < 15 && j > -15 && j < 15)
					p->color = 99 + (rand () % 7);
				else
					p->color = 3 + (rand () % 7);

				p->type = pt_lavasplash;

				dir[0] = j * 8 + (rand () & 7);
				dir[1] = i * 8 + (rand () & 7);
				dir[2] = 256;

				p->org[0] = org[0] + dir[0] + (rand () & 127) - 64;
				p->org[1] = org[1] + dir[1] + (rand () & 127) - 64;
				p->org[2] = org[2] + (rand () & 63);

				VectorNormalize (dir);
				vel = 50 + (rand () & 63);
				VectorScale (dir, vel, p->vel);

				p->drawtype = pd_smoke;
				p->radius = 20 + (rand () % 20);
				p->alpha = 0.666;
			}
		}
	}
}


/*
 =======================================================================================================================
    R_TeleportSplash
 =======================================================================================================================
 */
void R_TeleportSplash (vec3_t org)
{
	int			i, j, k;
	particle_t	*p;
	float		vel;
	vec3_t		dir;

	for (i = -16; i < 16; i += 4)
	{
		for (j = -16; j < 16; j += 4)
		{
			for (k = -24; k < 32; k += 4)
			{
				if (!free_particles)
					return ;

				p = free_particles;
				free_particles = p->next;
				p->next = active_particles;
				active_particles = p;

				p->die = cl.time + 0.2 + (rand () & 7) * 0.02;
				p->color = 7 + (rand () & 7);
				p->type = pt_slowgrav;

				dir[0] = j * 8;
				dir[1] = i * 8;
				dir[2] = k * 8;

				p->org[0] = org[0] + i + (rand () & 3);
				p->org[1] = org[1] + j + (rand () & 3);
				p->org[2] = org[2] + k + (rand () & 3);

				VectorNormalize (dir);
				vel = 50 + (rand () & 63);
				VectorScale (dir, vel, p->vel);

				p->drawtype = pd_spark;
				p->radius = 1.5f + (float) (rand () % 8) / 7;
				p->alpha = 1;
			}
		}
	}
}


void R_RocketTrail (vec3_t start, vec3_t end, int type)
{
	vec3_t		vec;
	float		len;
	int			j;
	particle_t	*p;
	int			dec;
	static int	tracercount;

	VectorSubtract (end, start, vec);
	len = VectorNormalize (vec);
	if (type < 128)
		dec = 3;
	else
	{
		dec = 1;
		type -= 128;
	}

	while (len > 0)
	{
		len -= dec;

		if (!free_particles)
			return ;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		VectorCopy (vec3_origin, p->vel);
		p->die = cl.time + 2;

		switch (type)
		{
		case 0: // rocket trail
			p->ramp = (rand () & 3);
			p->color = ramp3[(int) p->ramp];

			p->type = pt_fire;

			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () % 6) - 3);

			p->drawtype = pd_smoke;
			p->radius = 2.0f;
			p->alpha = 1.0f;

			len -= 3;

			break;

		case 1: // smoke smoke
			p->ramp = (rand () & 3) + 2;
			p->color = ramp3[(int) p->ramp];

			p->type = pt_fire;

			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () % 6) - 3);

			p->drawtype = pd_smoke;
			p->radius = 2.0f;
			p->alpha = 1.0f;

			len -= 3;

			break;

		case 2: // blood
			p->type = pt_grav;

			p->color = 71 + (rand () & 3);

			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () % 6) - 3);

			p->drawtype = pd_blood1 + (rand () % 4);
			p->alpha = 0.666f;
			p->radius = 4 + (rand () % 8);
			break;

		case 3: // wizard
		case 5: // knight
			p->die = cl.time + 0.5;
			p->type = pt_static;
			if (type == 3)
				p->color = 52 + ((tracercount & 4) << 1);
			else
				p->color = 230 + ((tracercount & 4) << 1);

			tracercount++;

			VectorCopy (start, p->org);
			if (tracercount & 1)
			{
				p->vel[0] = 30 * vec[1];
				p->vel[1] = 30 * -vec[0];
			}
			else
			{
				p->vel[0] = 30 * -vec[1];
				p->vel[1] = 30 * vec[0];
			}

			if (type == 3)
			{
				// wizard
				p->drawtype = pd_tracer;
				p->alpha = 1.0f;
				p->radius = 4 + (rand () % 8);
			}
			else
			{
				// knight
				p->drawtype = pd_tracer;
				p->radius = 6.0f + (rand () % 9);
				p->alpha = 1.0f;
			}

			break;

		case 4: // slight blood
			p->type = pt_grav;

			p->color = 71 + (rand () & 3);

			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () % 6) - 3);

			p->drawtype = pd_blood1 + (rand () % 4);
			p->alpha = 0.666f;
			p->radius = 4 + (rand () % 8);

			len -= 3;
			break;

		case 6: // voor trail
			p->color = 9 * 16 + 8 + (rand () & 3);
			p->type = pt_static;
			p->die = cl.time + 0.3;
			for (j = 0; j < 3; j++)
				p->org[j] = start[j] + ((rand () & 15) - 8);

			// voor
			p->drawtype = pd_tracer;
			p->alpha = 1.0f;
			p->radius = 4 + (rand () % 8);

			break;
		}

		VectorAdd (start, vec, start);
	}
}


void R_SmokeEffect (vec3_t org, int count, float InitialAlpha, int flashbase)
{
	int			i;
	particle_t	*p;

	for (i = 0; i < count; i++)
	{
		if (!free_particles)
			return ;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		// set the die ludicrously high so that the smoke will always go through the full
		// alpha ramp
		p->die = cl.time + 666;
		p->type = pt_antigrav;

		p->color = 3 + (rand () % 7);

		p->org[0] = org[0] - 2 + (rand () % 5);
		p->org[1] = org[1] - 2 + (rand () % 5);
		p->org[2] = org[2] - 2 + (rand () % 5);

		p->vel[0] = 0;
		p->vel[1] = 0;
		p->vel[2] = 4 + (rand () % 5);

		p->drawtype = pd_smoke;
		p->radius = 3;
		p->alpha = InitialAlpha;
	}

	count *= 2;

	// now add a muzzleflash
	for (i = 0; i < count; i++)
	{
		if (!free_particles)
			return ;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		// these should go through the full alpha scale
		p->die = cl.time + 666;
		p->type = pt_kill;

		p->color = flashbase + (rand () % 8);

		p->org[0] = org[0] - 2 + (rand () % 5);
		p->org[1] = org[1] - 2 + (rand () % 5);
		p->org[2] = org[2] - 2 + (rand () % 5);

		p->vel[0] = 0;
		p->vel[1] = 0;
		p->vel[2] = 0;

		p->drawtype = pd_flash;
		p->radius = 8 + (rand () % 9);
		p->alpha = 1.0f;
	}
}


void R_GunshotParticles (vec3_t org, vec3_t dir, int count)
{
	int			i;
	particle_t	*p;

	// smoke
	for (i = 0; i < count; i++)
	{
		if (!free_particles)
			return ;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 666;

		p->color = 3 + (rand () % 7);

		p->type = pt_antigrav;

		p->org[0] = org[0] + ((rand () & 15) - 8);
		p->org[1] = org[1] + ((rand () & 15) - 8);
		p->org[2] = org[2] + ((rand () & 15) - 8);

		p->vel[0] = dir[0] * 5;
		p->vel[1] = dir[1] * 5;
		p->vel[2] = dir[2] * 5;

		p->drawtype = pd_smoke;
		p->radius = 0.6 + (float) (rand () % 5) / 10;
		p->alpha = 0.333 + (float) (rand () % 5) / 10;
	}

	count *= 2;

	// sparks
	for (i = 0; i < count; i++)
	{
		if (!free_particles)
			return ;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.1 * (rand () % 5);

		p->color = 103 + (rand () % 8);

		p->type = pt_slowgrav;

		p->org[0] = org[0] + ((rand () & 15) - 8);
		p->org[1] = org[1] + ((rand () & 15) - 8);
		p->org[2] = org[2] + ((rand () & 15) - 8);

		p->vel[0] = dir[0] + ((rand () & 63) - 32);
		p->vel[1] = dir[1] + ((rand () & 63) - 32);
		p->vel[2] = dir[2] + ((rand () & 63) - 32);

		p->drawtype = pd_spark;
		p->radius = 1.25f + (float) (rand () % 16) / 30;
		p->alpha = 1.0f;
	}
}


void R_DrawParticleSprite (vec3_t org, int style, int color, int radius, float scale)
{
	particle_t *p;

	if (!free_particles)
		return;

	p = free_particles;
	free_particles = p->next;
	p->next = active_particles;
	active_particles = p;

	p->die = cl.time + 666;

	p->color = color;

	p->type = pt_kill;

	p->org[0] = org[0];
	p->org[1] = org[1];
	p->org[2] = org[2];

	p->vel[0] = 0;
	p->vel[1] = 0;
	p->vel[2] = 0;

	p->drawtype = (style > 15) ? 15 : (style < 0) ? 15 : style;
	p->radius = radius;
	p->alpha = 1;
	p->scale = scale;
}

#define DIV180MULTPI	0.01745329252

void TraceLine (vec3_t start, vec3_t end, vec3_t impact);

qboolean TraceOcclude (vec3_t org)
{
	vec3_t stop;
	float dist;

	TraceLine (org, r_refdef.vieworg, stop);
	dist = DotProduct (r_refdef.vieworg, stop);

	if (dist < 1) return false;

	return true;
}

qboolean R_ModelOcclusion (entity_t *ent)
{
	vec3_t mins, maxs;

	// if any of the occlusion tests pass we return false and draw the flare
	// 1 - the origin
	// most flares will pass at this stage and return false
	if (!TraceOcclude (ent->origin)) return false;

	// get mins and maxs points
	VectorAdd (ent->origin, ent->model->mins, mins);
	VectorAdd (ent->origin, ent->model->maxs, maxs);

	// 2 - mins and maxs points
	if (!TraceOcclude (mins)) return false;
	if (!TraceOcclude (maxs)) return false;

	return true;
}

void R_AddEntModelEffects (entity_t *ent)
{
	float baseorigin[] = {ent->origin[0], ent->origin[1], ent->origin[2]};
	float baseradius = 25.0f;
	int flarecolour = 0;
	float scale = 0.0f;
	float alpha = 0.5f;

	// no effects
	if (ent->model->flags < 256) return;

	// handle sprite replacements first
	if (ent->model->flags & EF_FLARESPRITE)
	{
		R_DrawParticleSprite (ent->origin, pd_flare2, 108, 128, 0.001);
		return;
	}

	if (ent->model->flags & EF_BUBBLE)
	{
		R_DrawParticleSprite (ent->origin, pd_bubble, 46, 6, 0.0);
		return;
	}

	// because we can stack these effects we check them individually rather than making a mess
	if (ent->model->flags & EF_FLARETORCH)
	{
		baseradius = 64;
		flarecolour = 107;

		// hack for torches...
		if (ent->model->name[11] == '.')
			baseorigin[2] += 8;
		else if (ent->model->name[11] == '2')
		{
			if (ent->frame == 0)
				baseorigin[2] += 0;
			else baseorigin[2] += 8;
		}
	}

	if (ent->model->flags & EF_ROCKETFLARE)
	{
		// move the glow behind the rocket - fixme - do a fast sin/cos lookup table for this...
		baseorigin[0] += cos (ent->angles[1] * DIV180MULTPI) * (-20.0f);
		baseorigin[1] += sin (ent->angles[1] * DIV180MULTPI) * (-20.0f);
		baseorigin[2] += sin (ent->angles[0] * DIV180MULTPI) * (-20.0f);

		flarecolour = 107;
		alpha = 0.75f;
		baseradius = 64;
	}

	if (ent->model->flags & EF_FLICKER) baseradius += (rand () % (int) (baseradius / 6));

	// catch anything that hasn't had a colour set
	if (!flarecolour) return;

	// test for occlusion on the bbox
	if (R_ModelOcclusion (ent)) return;

	// add the flare
	R_DrawParticleSprite (baseorigin, pd_flare1, flarecolour, baseradius, scale);
}


void R_BloodParticleSplash (vec3_t org, vec3_t dir, int color, int range, int count)
{
	int			i;
	particle_t	*p;

	for (i = 0; i < count; i++)
	{
		if (!free_particles)
			return ;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.1 * (rand () % 5);

		p->color = color + (rand () % range);

		p->type = pt_slowgrav;

		p->org[0] = org[0] + ((rand () & 15) - 8);
		p->org[1] = org[1] + ((rand () & 15) - 8);
		p->org[2] = org[2] + ((rand () & 15) - 8);

		p->vel[0] = dir[0] * 15;
		p->vel[1] = dir[1] * 15;
		p->vel[2] = dir[2] * 15;

		p->drawtype = pd_blood1 + (rand () % 4);
		p->radius = 8 + (rand () % 8);
		p->alpha = 0.666f;
	}
}


void R_HipnoticParticleField (vec3_t org, vec3_t dir, int color, int count)
{
	int			i;
	particle_t	*p;

	for (i = 0; i < count * 5; i++)
	{
		if (!free_particles)
			return ;

		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.3 * (rand () % 5);

		p->color = (int) ((color &~7) + (rand () & 7));

		p->type = pt_slowgrav;

		p->org[0] = org[0] + ((rand () & 15) - 8);
		p->org[1] = org[1] + ((rand () & 15) - 8);
		p->org[2] = org[2] + ((rand () & 15) - 8);

		p->vel[0] = dir[0] * 15;
		p->vel[1] = dir[1] * 15;
		p->vel[2] = dir[2] * 15;

		p->drawtype = pd_spark;
		p->radius = 2.5f + (float) (rand () % 16) / 15;
		p->alpha = 1;
	}
}


/*
 =======================================================================================================================
    R_RunParticleEffect
 =======================================================================================================================
 */
void R_RunParticleEffect (vec3_t org, vec3_t dir, int color, int count)
{
	int			i, j;
	particle_t	*p;

	// handle some special types
	if (count == 1024)
	{
		R_ParticleExplosion (org);
		return;
	}
	else if (hipnotic && count <= 4)
	{
		// particle field
		R_HipnoticParticleField (org, dir, color, count);
		return;
	}
	else if (count == 255)
	{
		// always an explosion
		R_ParticleExplosion (org);
		return;
	}
	else if (color == 73)
	{
		// blood splashes
		R_BloodParticleSplash (org, dir, 66, 13, count);
		return;
	}
	else if (color == 225)
	{
		// blood splashes
		R_BloodParticleSplash (org, dir, 224, 6, count);
		return;
	}

	// generic
	for (i = 0; i < count; i++)
	{
		if (!free_particles)
			return ;
		p = free_particles;
		free_particles = p->next;
		p->next = active_particles;
		active_particles = p;

		p->die = cl.time + 0.1 * (rand () % 5);
		p->color = (color &~7) + (rand () & 7);
		p->type = pt_slowgrav;

		for (j = 0; j < 3; j++)
		{
			p->org[j] = org[j] + ((rand () & 15) - 8);
			p->vel[j] = dir[j] * 15;	// (rand()%300)-150;
		}
	}
}


/*
 =======================================================================================================================
    R_ParseParticleEffect Parse an effect out of the server message
 =======================================================================================================================
 */
void R_ParseParticleEffect (void)
{
	vec3_t	org, dir;
	int		i, count, color;

	for (i = 0; i < 3; i++)
		org[i] = MSG_ReadCoord ();
	for (i = 0; i < 3; i++)
		dir[i] = MSG_ReadChar () * (1.0 / 16);

	count = MSG_ReadByte ();
	color = MSG_ReadByte ();

	R_RunParticleEffect (org, dir, color, count);
}


/* R_DrawParticles */
extern cvar_t	sv_gravity;

void R_KillParticles (void)
{
	particle_t	*p, *kill;

	particlechain[0] = particlechain[1] = particlechain[2] = NULL;

	// jump to the first active particle in the list
	for (;;)
	{
		kill = active_particles;

		// this particle is dead so remove it from the active list, add it to the free
		// list, and carry on
		if (kill && kill->die < cl.time)
		{
			active_particles = kill->next;
			kill->next = free_particles;
			free_particles = kill;

			kill->drawtype = pd_dot;
			kill->radius = 1.5f;
			kill->alpha = 1.0f;
			kill->scale = 0.004;

			continue;
		}

		// if we get to here we have found a particle that isn't dead, so stop right now
		break;
	}

	// go through the list killing as we go if active_particles == NULL the for loop
	// will handle it admirably
	for (p = active_particles; p; p = p->next)
	{
		// jump to the next active particle in the list
		for (;;)
		{
			kill = p->next;

			// very same method as above, except we don't remove it from active_particles as
			// we want to retain that pointer to the beginning of the list.
			if (kill && kill->die < cl.time)
			{
				p->next = kill->next;
				kill->next = free_particles;
				free_particles = kill;

				kill->drawtype = pd_dot;
				kill->radius = 1.5f;
				kill->alpha = 1.0f;
				kill->scale = 0.004;

				continue;
			}

			// very same method as above
			break;
		}

		// different types of particles require different blendfuncs
		switch (p->drawtype)
		{
		case pd_flash:
		case pd_spark:
		case pd_flare1:
		case pd_flare2:
		case pd_tracer:
			p->chain = particlechain[1];
			particlechain[1] = p;
			break;

		default:
			p->chain = particlechain[0];
			particlechain[0] = p;
			break;
		}
	}
}


typedef struct partverts_s
{
	float	v[3];

	float	st[2];

	float	rgba[4];
} partverts_t;


// draw particles in batches of 1024 verts, = 256 quads note - this define MUST
// be a multiple of 4!!!
#define PARTICLE_BATCH_SIZE 1024

partverts_t			partverts[PARTICLE_BATCH_SIZE];

extern gltexture_t	ParticleTexture;

void R_RenderParticles (void)
{
	int				i;
	particle_t		*p;
	vec3_t			pOrg, pUp, pRight;
	float			scale;
	int				numparts;
	extern float	PRGB[3][256];

	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glBindTexture (GL_TEXTURE_2D, ParticleTexture.texnum);

	// bring up vertex pointer
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (partverts_t), partverts[0].v);

	// bring up image texture array
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (partverts_t), partverts[0].st);

	// bring up colour array
	glEnableClientState (GL_COLOR_ARRAY);
	glColorPointer (4, GL_FLOAT, sizeof (partverts_t), partverts[0].rgba);

	for (i = 0; i < 3; i++)
	{
		if (!particlechain[i])
			continue;

		if (i == 0)
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		else
			glBlendFunc (GL_SRC_ALPHA, GL_ONE);

		for (numparts = 0, p = particlechain[i]; p; p = p->chain)
		{
			// hack a scale up to keep particles from disapearing
			scale = (p->org[0] - r_origin[0]) * vpn[0] + (p->org[1] - r_origin[1]) * vpn[1] + (p->org[2] - r_origin[2]) * vpn[2];

			if (scale < 20)
				scale = 1;
			else
				scale = 1 + scale * p->scale;

			// calculate the scaled vertex offset factors
			VectorScale (vup, p->radius, pUp);
			VectorScale (vright, p->radius, pRight);

			// calculate the offset origin to center the quad on the particle origin
			pOrg[0] = p->org[0] - (pRight[0] + pUp[0]) / 2;
			pOrg[1] = p->org[1] - (pRight[1] + pUp[1]) / 2;
			pOrg[2] = p->org[2] - (pRight[2] + pUp[2]) / 2;

			// build 4 verts
			partverts[numparts].rgba[0] = PRGB[0][(int) p->color];
			partverts[numparts].rgba[1] = PRGB[1][(int) p->color];
			partverts[numparts].rgba[2] = PRGB[2][(int) p->color];
			partverts[numparts].rgba[3] = p->alpha;

			// the weird texcoords are so that the particle will draw right way up!
			partverts[numparts].st[0] = particlest[p->drawtype][0];
			partverts[numparts].st[1] = particlest[p->drawtype][3];

			partverts[numparts].v[0] = pOrg[0];
			partverts[numparts].v[1] = pOrg[1];
			partverts[numparts].v[2] = pOrg[2];

			numparts++;

			partverts[numparts].rgba[0] = PRGB[0][(int) p->color];
			partverts[numparts].rgba[1] = PRGB[1][(int) p->color];
			partverts[numparts].rgba[2] = PRGB[2][(int) p->color];
			partverts[numparts].rgba[3] = p->alpha;

			partverts[numparts].st[0] = particlest[p->drawtype][0];
			partverts[numparts].st[1] = particlest[p->drawtype][1];

			partverts[numparts].v[0] = pOrg[0] + pUp[0];
			partverts[numparts].v[1] = pOrg[1] + pUp[1];
			partverts[numparts].v[2] = pOrg[2] + pUp[2];

			numparts++;

			partverts[numparts].rgba[0] = PRGB[0][(int) p->color];
			partverts[numparts].rgba[1] = PRGB[1][(int) p->color];
			partverts[numparts].rgba[2] = PRGB[2][(int) p->color];
			partverts[numparts].rgba[3] = p->alpha;

			partverts[numparts].st[0] = particlest[p->drawtype][2];
			partverts[numparts].st[1] = particlest[p->drawtype][1];

			partverts[numparts].v[0] = pOrg[0] + pUp[0] + pRight[0];
			partverts[numparts].v[1] = pOrg[1] + pUp[1] + pRight[1];
			partverts[numparts].v[2] = pOrg[2] + pUp[2] + pRight[2];

			numparts++;

			partverts[numparts].rgba[0] = PRGB[0][(int) p->color];
			partverts[numparts].rgba[1] = PRGB[1][(int) p->color];
			partverts[numparts].rgba[2] = PRGB[2][(int) p->color];
			partverts[numparts].rgba[3] = p->alpha;

			partverts[numparts].st[0] = particlest[p->drawtype][2];
			partverts[numparts].st[1] = particlest[p->drawtype][3];

			partverts[numparts].v[0] = pOrg[0] + pRight[0];
			partverts[numparts].v[1] = pOrg[1] + pRight[1];
			partverts[numparts].v[2] = pOrg[2] + pRight[2];

			numparts++;

			// draw when the buffer fills
			if (numparts == PARTICLE_BATCH_SIZE)
			{
				glDrawArrays (GL_QUADS, 0, PARTICLE_BATCH_SIZE);
				numparts = 0;
			}
		}

		// draw anything left over
		if (numparts) glDrawArrays (GL_QUADS, 0, numparts);
	}

	glDisableClientState (GL_COLOR_ARRAY);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glDisableClientState (GL_VERTEX_ARRAY);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glColor4f (1, 1, 1, 1);

	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	glDepthRange (0.1, 0.8);
}


void R_UpdateParticles (void)
{
	particle_t	*p;
	float		grav;
	int			i;
	float		time2, time3, time4;
	float		time1;
	float		dvel;

	time4 = r_frametime * 40;
	time3 = r_frametime * 15;
	time2 = r_frametime * 10;
	time1 = r_frametime * 5;

	grav = r_frametime * sv_gravity.value * 0.05;

	dvel = 4 * r_frametime;

	for (p = active_particles; p; p = p->next)
	{
		p->org[0] += p->vel[0] * r_frametime;
		p->org[1] += p->vel[1] * r_frametime;
		p->org[2] += p->vel[2] * r_frametime;

		switch (p->type)
		{
		case pt_kill:
			p->die = -1;
			break;

		case pt_static:
			break;

		case pt_fire:
			p->ramp += time1;

			if (p->ramp >= 6)
				p->ramp = p->die = -1;
			else
				p->color = ramp3[(int) p->ramp];

			p->vel[2] += grav;
			break;

		case pt_explode:
			p->ramp += time2;

			if (p->ramp >= 8)
				p->ramp = 7;
			else
				p->color = ramp1[(int) p->ramp];

			p->alpha -= time2;

			if (p->alpha < 0) p->die = -1;

			p->radius += time4 * 10;

			break;

		case pt_explode2:
			p->ramp += time3;

			if (p->ramp >= 8)
				p->die = -1;
			else
				p->color = ramp2[(int) p->ramp];

			p->vel[0] -= p->vel[0] * r_frametime;
			p->vel[1] -= p->vel[1] * r_frametime;
			p->vel[2] -= p->vel[2] * r_frametime;
			break;

		case pt_explode3:
			p->ramp += time3;

			if (p->ramp >= 8)
				p->die = -1;
			else
				p->color = ramp2[(int) p->ramp];

			p->vel[0] += p->vel[0] * r_frametime;
			p->vel[1] += p->vel[1] * r_frametime;
			p->vel[2] -= p->vel[2] * r_frametime;

			p->vel[2] -= grav * 4.0f;
			break;

		case pt_blob:
			for (i = 0; i < 3; i++)
				p->vel[i] += p->vel[i] * dvel;
			p->vel[2] -= grav;
			break;

		case pt_blob2:
			for (i = 0; i < 2; i++)
				p->vel[i] -= p->vel[i] * dvel;
			p->vel[2] -= grav;
			break;

		case pt_grav:
		case pt_slowgrav:
		case pt_lavasplash:
			p->vel[2] -= grav;
			break;

		case pt_antigrav:
			p->vel[2] += grav;
			break;
		}

		// handle drawtype special cases
		switch (p->drawtype)
		{
		case pd_smoke:	// lavasplash fades slower
			if (p->type == pt_lavasplash)
				p->alpha -= r_frametime / 10.0f;
			else
				p->alpha -= r_frametime;
			
			if (p->alpha < 0)
				p->die = -1;
			
			p->radius += time4;
			
			break;
		}
	}
}


void R_DrawParticles (void)
{
	// kill them first
	R_KillParticles ();

	// nothing to either render or update
	if (!active_particles)
		return ;

	// now render and update
	R_RenderParticles ();
	R_UpdateParticles ();
}


