/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"

typedef struct spriteverts_s
{
	float v[3];
	float st[2];
} spriteverts_t;

spriteverts_t spriteverts[4];

/*
==============================================================================================================================

			SPRITES

	Most sprites are just "chrome" effects, and will be ultimately replaced by the particle system.  A few, however, are
	evil hacks done by the Hipnotic development team utilising unintended quirks of the engine to achieve an effect it was
	never meant to do in the first place.  Because of this, sprites must die.

==============================================================================================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;

	psprite = currententity->model->cache.data;
	frame = currententity->frame;

	// silent fallback
	if ((frame >= psprite->numframes) || (frame < 0)) frame = 0;

	if (psprite->frames[frame].type == SPR_SINGLE)
		pspriteframe = psprite->frames[frame].frameptr;
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + currententity->syncbase;

		// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
		// are positive, so we don't have to worry about division by 0
		targettime = time - ((int) (time / fullinterval)) * fullinterval;

		for (i = 0; i < (numframes - 1); i++)
			if (pintervals[i] > targettime)
				break;

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel

=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	vec3_t			point;
	mspriteframe_t	*frame;
	float			*up, *right;
	vec3_t			v_forward, v_right, v_up;
	msprite_t		*psprite;

	// don't even bother culling, because it's just a single
	// polygon without a surface cache
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->cache.data;

	if (psprite->type == SPR_ORIENTED)
	{
		// bullet marks on walls
		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		up = v_up;
		right = v_right;
	}
	else
	{
		// normal sprite
		up = vup;
		right = vright;
	}

	glColor3f (1, 1, 1);

	glBindTexture (GL_TEXTURE_2D, frame->gltexture->texnum);

	spriteverts[0].st[0] = 0;
	spriteverts[0].st[1] = 0;
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->left, right, point);
	spriteverts[0].v[0] = point[0];
	spriteverts[0].v[1] = point[1];
	spriteverts[0].v[2] = point[2];

	spriteverts[1].st[0] = 1;
	spriteverts[1].st[1] = 0;
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->left, right, point);
	spriteverts[1].v[0] = point[0];
	spriteverts[1].v[1] = point[1];
	spriteverts[1].v[2] = point[2];

	spriteverts[2].st[0] = 1;
	spriteverts[2].st[1] = 1;
	VectorMA (e->origin, frame->up, up, point);
	VectorMA (point, frame->right, right, point);
	spriteverts[2].v[0] = point[0];
	spriteverts[2].v[1] = point[1];
	spriteverts[2].v[2] = point[2];

	spriteverts[3].st[0] = 0;
	spriteverts[3].st[1] = 1;
	VectorMA (e->origin, frame->down, up, point);
	VectorMA (point, frame->right, right, point);
	spriteverts[3].v[0] = point[0];
	spriteverts[3].v[1] = point[1];
	spriteverts[3].v[2] = point[2];

	glDrawArrays (GL_QUADS, 0, 4);
}


void R_AddEntModelEffects (entity_t *ent);

void R_DrawSpriteEntities (void)
{
	int i;
	qboolean ClearState = false;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		currententity = cl_visedicts[i];

		R_AddEntModelEffects (currententity);

		// sprites flagged for no drawing
		if (currententity->model->flags & EF_NODRAW) continue;

		if (currententity->model->type == mod_sprite)
		{
			if (!ClearState)
			{
				// bring up vertex pointer
				glEnableClientState (GL_VERTEX_ARRAY);
				glVertexPointer (3, GL_FLOAT, sizeof (spriteverts_t), spriteverts[0].v);

				// bring up image texture array
				glEnableClientState (GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer (2, GL_FLOAT, sizeof (spriteverts_t), spriteverts[0].st);

				glEnable (GL_BLEND);
				ClearState = true;
			}

			R_DrawSpriteModel (currententity);
		}
	}

	if (ClearState)
	{
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
		glDisableClientState (GL_VERTEX_ARRAY);
		glDisable (GL_BLEND);
	}
}

//==================================================================================
