/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"
#include "cvar.h"
#include "render.h"
#include "client.h"
#include "gl_model.h"
#include "glquake.h"
#include "glext.h"


qboolean skythisframe;
qboolean waterthisframe;


void R_BlendLightmaps (void);


extern int lightmap_textures;

int		addSurfs;

// we need a "fake" volume for brush models
volume_t bmvol[2];

qboolean fbChain = false;
msurface_t *dtChain;

#define CAUSTIC_SCALE 0.0234375

int gl_currenttmus;

void R_PrebuildLightmaps (model_t *mod);
extern glpoly_t *lightmap_polys[MAX_LIGHTMAPS];

void R_PushDlights (mnode_t *node, qboolean bmnode);
void R_PrebuildSurfaceLightmap (msurface_t *surf);


/*
===============
Sin approximations.  These work in a similar fashion to the old gl_warpsin.h but are
calculated at engine startup rather than stored in a fixed array.  This lets me keep
a higher precision copy of the table in memory without having to use a great big
include file containing loadsa data, plus I can also tweak the precision.  These are
faster than the library sin function and are significantly faster than JoeQuake's
SINTABLE_APPROX function (which was actually *slower* than the library sin function
when I had it in here...)

SINTABLE_SIZE is currently 1024, which gives a good degree of accuracy without the need
for a massive table lookup.  You can increase it by changing the define in glquake.h,
but be aware that beyond a certain level rounding errors start to come in and the
visual result gets incredibly choppy.  Decreasing it to 256 gives comparable results to
the old gl_warpsin.h method.
===============
*/
float sintable[SINTABLE_SIZE];
float sintablex2[SINTABLE_SIZE];

int STBitMask;

float STMult;
float STMultx2;

float warptime;


/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (texture_t *base)
{
	int		relative;
	int		count;

	// save out the base texture in case we have a bad cycle
	texture_t *bad = base;

	if (currententity->frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}
	
	// no animations
	if (!base->anim_total) return base;

	relative = (int) (realtime * 10) % base->anim_total;

	count = 0;

	while (base->anim_min > relative || base->anim_max <= relative)
	{
		base = base->anim_next;

		// bad animation cycle so just return the base texture
		if (!base) return bad;
		if (++count > 100) return bad;
	}

	return base;
}


/*
===================
R_DrawFullbrightChain

Draws the chain of fullbrights sorted by texture
This should be the very last thing drawn as we don't want fullbrights to be
overdrawn by anything else
===================
*/
void R_DrawFullbrightChain (model_t *mod)
{
	int i;
	int j;
	int numverts;
	float *v;
	msurface_t *surf;
	texture_t *t;

	glEnable (GL_ALPHA_TEST);
	glAlphaFunc (GL_GREATER, 0.666);
	glDepthMask (GL_FALSE);

	for (i = 0; i < mod->numtextures; i++)
	{
		if (!mod->textures[i]) continue;
		if (!mod->textures[i]->fbChain) continue;

		t = R_TextureAnimation (mod->textures[i]);

		surf = mod->textures[i]->fbChain;

		glBindTexture (GL_TEXTURE_2D, t->fullbright);

		for (; surf; surf = surf->fbChain)
		{
			v = surf->polys->verts[0];
			numverts = surf->polys->numverts;

			glBegin (GL_POLYGON);

			for (j = 0; j < numverts; j++, v += VERTEXSIZE)
			{
				// fullbright has the same texcoords as the world
				glTexCoord2fv (&v[3]);

				// and draw it
				glVertex3fv (v);
			}

			glEnd ();
		}

		mod->textures[i]->fbChain = NULL;
	}

	glDisable (GL_ALPHA_TEST);
	glDepthMask (GL_TRUE);

	fbChain = false;
}


// drawn as a seperate pass because of the blendfunc, this kills the poor framerate (drops 20)
void R_DrawDetailChain (void)
{
	int i;
	int numverts;
	float *v;
	msurface_t *surf;

	glEnable (GL_BLEND);
	glBlendFunc (GL_DST_COLOR, GL_SRC_COLOR);
	//GL_BlendTest ();
	glDepthMask (GL_FALSE);
	glActiveTexture (GL_TEXTURE0);
	glDisable (GL_TEXTURE_2D);
	glActiveTexture (GL_TEXTURE1);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
	glBindTexture (GL_TEXTURE_2D, detailtexture);

	for (surf = dtChain; surf; surf = surf->dtChain)
	{
		v = surf->polys->verts[0];
		numverts = surf->polys->numverts;

		glBegin (GL_POLYGON);

		for (i = 0; i < numverts; i++, v += VERTEXSIZE)
		{
			// detail has the same texcoords as the world
			glMultiTexCoord2fv (GL_TEXTURE1, &v[7]);

			// and draw it
			glVertex3fv (v);
		}

		glEnd ();
	}

	dtChain = NULL;

	glActiveTexture (GL_TEXTURE0);
	glEnable (GL_TEXTURE_2D);
	glActiveTexture (GL_TEXTURE1);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);
}


/*
=====================
R_DrawTextureChains

Sometimes the old ways are still the best...

Runs correctly and looks consistent in all configurations.

3 render paths are supported:
* standard above-water with world and lightmaps
* 3 TMU below water
* 2 TMU below water (requires a second pass)

I had MHQuake code that supported detail texes (still do, in fact) but it ran v slow,
even on 4TMUs, so it had to go.  Anyway, they messed up my carefully set up colour
and lighting adjustments, and I don't think they're such a big deal to begin with.
=====================
*/
void R_DrawTextureChains3TMU (model_t *mod, int volume, texture_t *caustic)
{
	int i;
	msurface_t *surf;
	texture_t *t;

	// surface drawing vars
	int	j;
	int numverts;
	float *v;
	float nv[2];

	qboolean fog = false;

	// lightmap optimisation
	int oldLMTex;

	// set up any fog here (underwater fog always)
	/*
	MH doesn't like fog!!!
	glFogi (GL_FOG_MODE, GL_LINEAR);
	glFogfv (GL_FOG_COLOR, caustic->uwater_fog);
	glFogf (GL_FOG_START, 10);
	glFogf (GL_FOG_END, 1000);
	glFogf (GL_FOG_DENSITY, 0.2);
	glEnable (GL_FOG);

	fog = true;
	*/

	// caustic TMU
	glActiveTexture (GL_TEXTURE2);
	glEnable (GL_TEXTURE_2D);
	glBindTexture (GL_TEXTURE_2D, caustic->gl_texturenum);

	for (i = 0; i < mod->numtextures; i++)
	{
		// this only happens in e2m3
		if (!mod->textures[i]) continue;

		// not on the draw list
		if (!mod->textures[i]->texturechain[volume]) continue;

		// batch this up
		t = R_TextureAnimation (mod->textures[i]);

		surf = mod->textures[i]->texturechain[volume];

		glActiveTexture (GL_TEXTURE0);
		glBindTexture (GL_TEXTURE_2D, t->gl_texturenum);
		glActiveTexture (GL_TEXTURE1);

		oldLMTex = -1;

		for (; surf; surf = surf->drawChain)
		{
			// detail chain
			//surf->dtChain = dtChain;
			//dtChain = surf;

			// only add to the fullbright chain if the current animation has fullbrights
			if (t->fullbright)
			{
				surf->fbChain = surf->texinfo->texture->fbChain;
				surf->texinfo->texture->fbChain = surf;

				fbChain = true;
			}

			// optimise lightmap texture binding
			if (oldLMTex != surf->lightmaptexturenum)
			{
				glBindTexture (GL_TEXTURE_2D, lightmap_textures + surf->lightmaptexturenum);
				oldLMTex = surf->lightmaptexturenum;
			}

			v = surf->polys->verts[0];
			numverts = surf->polys->numverts;

			glBegin (GL_POLYGON);

			for (j = 0; j < numverts; j++, v+= VERTEXSIZE)
			{
				// world
				glMultiTexCoord2fv (GL_TEXTURE0, &v[3]);

				// caustics
				nv[0] = v[7] + sintable [((int) (warptime + v[10])) & STBitMask];
				nv[1] = v[8] + sintable [((int) (warptime + v[9])) & STBitMask];

				// yeh, you *can* do this...
				glMultiTexCoord2fv (GL_TEXTURE2, nv);

				// lightmap
				glMultiTexCoord2fv (GL_TEXTURE1, &v[5]);

				// and draw it
				glVertex3fv (v);
			}

			glEnd ();

			// reset the visframe so that subsequent maps don't pick it up by accident
			surf->visframe = -1;
		}

		mod->textures[i]->texturechain[volume] = NULL;
	}

	// bring down caustic TMU
	glActiveTexture (GL_TEXTURE2);
	glDisable (GL_TEXTURE_2D);

	// make sure GL_TEXTURE1 is active coming out!!!
	glActiveTexture (GL_TEXTURE1);

	//R_DrawDetailChain ();

	// shut down any fog
	if (fog) glDisable (GL_FOG);
}


void R_DrawTextureChains2Pass (model_t *mod, int volume, texture_t *caustic)
{
	int i;
	msurface_t *surf;
	texture_t *t;

	// surface drawing vars
	int	j;
	int numverts;
	float *v;
	float nv[2];

	qboolean fog = false;

	// set up any fog here (underwater fog always)
	/*
	MH doesn't like fog!!!
	glFogi (GL_FOG_MODE, GL_LINEAR);
	glFogfv (GL_FOG_COLOR, caustic->uwater_fog);
	glFogf (GL_FOG_START, 10);
	glFogf (GL_FOG_END, 1000);
	glFogf (GL_FOG_DENSITY, 0.2);
	glEnable (GL_FOG);

	fog = true;
	*/

	// clear down the lightmap polys chain
	memset (lightmap_polys, 0, sizeof (lightmap_polys));

	// set up for caustics
	glActiveTexture (GL_TEXTURE1);
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glScalef (CAUSTIC_SCALE, CAUSTIC_SCALE, 1);
	glMatrixMode (GL_MODELVIEW);
	glBindTexture (GL_TEXTURE_2D, caustic->gl_texturenum);
	glActiveTexture (GL_TEXTURE0);

	for (i = 0; i < mod->numtextures; i++)
	{
		// this only happens in e2m3
		if (!mod->textures[i]) continue;

		// not on the draw list
		if (!mod->textures[i]->texturechain[volume]) continue;

		// batch this up
		t = R_TextureAnimation (mod->textures[i]);

		surf = mod->textures[i]->texturechain[volume];

		glBindTexture (GL_TEXTURE_2D, t->gl_texturenum);

		for (; surf; surf = surf->drawChain)
		{
			// detail chain
			//surf->dtChain = dtChain;
			//dtChain = surf;

			// only add to the fullbright chain if the current animation has fullbrights
			if (t->fullbright)
			{
				surf->fbChain = surf->texinfo->texture->fbChain;
				surf->texinfo->texture->fbChain = surf;

				fbChain = true;
			}

			// add the polys to the lightmap chain
			surf->polys->lmChain = lightmap_polys[surf->lightmaptexturenum];
			lightmap_polys[surf->lightmaptexturenum] = surf->polys;

			v = surf->polys->verts[0];
			numverts = surf->polys->numverts;

			glBegin (GL_POLYGON);

			for (j = 0; j < numverts; j++, v+= VERTEXSIZE)
			{
				// world
				glMultiTexCoord2fv (GL_TEXTURE0, &v[3]);

				// caustics
				nv[0] = v[7] + sintable [((int) (warptime + v[10])) & STBitMask];
				nv[1] = v[8] + sintable [((int) (warptime + v[9])) & STBitMask];

				glMultiTexCoord2fv (GL_TEXTURE1, nv);

				// and draw it
				glVertex3fv (v);
			}

			glEnd ();

			// reset the visframe so that subsequent maps don't pick it up by accident
			surf->visframe = -1;
		}

		mod->textures[i]->texturechain[volume] = NULL;
	}

	// put us back the way we were
	glActiveTexture (GL_TEXTURE1);
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glDisable (GL_TEXTURE_2D);
	glActiveTexture (GL_TEXTURE0);

	// second pass for lightmap blend - if we're going to have to do a second
	// pass we may as well do it by the most efficient route...
	R_BlendLightmaps ();

	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);

	//R_DrawDetailChain ();

	// shut down any fog
	if (fog) glDisable (GL_FOG);
}


void R_DrawTextureChains2TMU (model_t *mod, int volume)
{
	int i;
	msurface_t *surf;
	texture_t *t;

	// surface drawing vars
	int	j;
	int numverts;
	float *v;

	qboolean fog = false;

	// lightmap optimisation
	int oldLMTex;

	// set up any fog here (world fog always)
	/*
	MH doesn't like fog!!!
	glFogi (GL_FOG_MODE, GL_LINEAR);
	glFogfv (GL_FOG_COLOR, whateverthefuck);
	glFogf (GL_FOG_START, 10);
	glFogf (GL_FOG_END, 1000);
	glFogf (GL_FOG_DENSITY, 0.2);
	glEnable (GL_FOG);

	fog = true;
	*/

	for (i = 0; i < mod->numtextures; i++)
	{
		// this only happens in e2m3
		if (!mod->textures[i]) continue;

		// not on the draw list
		if (!mod->textures[i]->texturechain[volume]) continue;

		// batch this up
		t = R_TextureAnimation (mod->textures[i]);

		surf = mod->textures[i]->texturechain[volume];

		glActiveTexture (GL_TEXTURE0);
		glBindTexture (GL_TEXTURE_2D, t->gl_texturenum);
		glActiveTexture (GL_TEXTURE1);

		oldLMTex = -1;

		for (; surf; surf = surf->drawChain)
		{
			// detail chain
			//surf->dtChain = dtChain;
			//dtChain = surf;

			// only add to the fullbright chain if the current animation has fullbrights
			if (t->fullbright)
			{
				surf->fbChain = surf->texinfo->texture->fbChain;
				surf->texinfo->texture->fbChain = surf;

				fbChain = true;
			}

			// optimise lightmap texture binding
			if (oldLMTex != surf->lightmaptexturenum)
			{
				glBindTexture (GL_TEXTURE_2D, lightmap_textures + surf->lightmaptexturenum);
				oldLMTex = surf->lightmaptexturenum;
			}

			v = surf->polys->verts[0];
			numverts = surf->polys->numverts;

			glBegin (GL_POLYGON);

			for (j = 0; j < numverts; j++, v+= VERTEXSIZE)
			{
				// world
				glMultiTexCoord2fv (GL_TEXTURE0, &v[3]);

				// lightmap
				glMultiTexCoord2fv (GL_TEXTURE1, &v[5]);

				// and draw it
				glVertex3fv (v);
			}

			glEnd ();

			// reset the visframe so that subsequent maps don't pick it up by accident
			surf->visframe = -1;
		}

		mod->textures[i]->texturechain[volume] = NULL;
	}

	//R_DrawDetailChain ();

	// shut down any fog
	if (fog) glDisable (GL_FOG);
}


/*
==================
R_DrawVolumeChains

volume merging and some astutely placed Con_Printfs indicates that
no more than 4 drawing volumes (i.e. volumes after merging) are
used in ID1.  I've upped it to 16 to be utterly safe.

Each "volume" for the purposes of drawing is all the surfs in the map with the same
underwater or abovewater type.  All abovewater surfs are in volume 0 always.  The
underwater type depends on the water texture they are under: if there are 2 different
slime textures in a map, this would give 2 slime volumes.  Volumes are used for setting
the correct fog colour and caustic texture *only* at this stage.
==================
*/
void R_DrawVolumeChains (model_t *mod)
{
	int i;

	// apply all current stains before doing the lightmaps
	// we don't stain BSP brush models (aside from the world model)
	if (mod == cl.worldmodel)
	{
		R_ApplyStains (mod->nodes);
	}
	else if (mod != cl.worldmodel && mod->name[0] == '*')
	{
		R_ApplyStains (mod->nodes + mod->hulls[0].firstclipnode);
	}

	// prebuild all lightmaps for this model
	R_PrebuildLightmaps (mod);

	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// optimise caustic warps
	warptime = STMult * realtime;

	// detail chain initial state
	dtChain = NULL;

	for (i = 0; i < 16; i++)
	{
		// brush models only have one volume set!!!
		if (i >= mod->numVolumes) break;

		// cos all the volumes are moved to the front of the list, once we hit one with
		// a visframe of -2 we can be sure that the rest will have no surfs
		if (mod->volumes[i].visframe == -2) break;

		// the volume has no drawlist for this frame
		if (mod->volumes[i].visframe != r_framecount) continue;

		// caustic is NULL if it's above water
		if (gl_currenttmus >= 3 && mod->volumes[i].caustic)
		{
			// 3 TMU world/caustic/lightmap render
			R_DrawTextureChains3TMU (mod, i, mod->volumes[i].caustic);
		}
		else if (mod->volumes[i].caustic)
		{
			// 2 TMU world/caustic render, then a second pass (R_BlendLightmaps) for lightmaps
			R_DrawTextureChains2Pass (mod, i, mod->volumes[i].caustic);
		}
		else
		{
			// "classic" 2 TMU world/lightmap render
			R_DrawTextureChains2TMU (mod, i);
		}
	}

	// take TMU 1 down (make sure GL_TEXTURE1 is always active coming out...)
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable (GL_TEXTURE_2D);

	glActiveTexture (GL_TEXTURE0);

	// fullbrights are drawn for all volumes at the end of each model.  I *could*
	// multitexture them again but I'd be bringing TMUs up and down like a young
	// lady of the night's undergarments.  Plus I want them to be drawn *after* fog
	if (fbChain) R_DrawFullbrightChain (mod);
}


/*
================
R_DrawWaterSurfaces

Each water texture is given a different scaling factor and blend mode
================
*/
void R_DrawWaterChain (msurface_t *surf, float scale);

void R_DrawWaterSurfaces (void)
{
	// water surfs are chained by texture
	int i;
	msurface_t *surf;
	texture_t *t;
	float scale;

	// only go through here if we have to
	if (!waterthisframe) return;

	// help optimise water warps
	warptime = realtime * STMultx2;

	// go back to the world matrix
    glLoadMatrixf (r_world_matrix);

	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);

	glActiveTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	for (i = 0; i < cl.worldmodel->numtextures; i++)
	{
		t = cl.worldmodel->textures[i];
		if (!t) continue;
		if (t->name[0] != '*') continue;
		if (!t->lqChain) continue;

		surf = t->lqChain;

		// each surf in the chain will have the same flags.  teleports aren't drawn
		// translucent or with any effects on them.  the blend funcs below have been
		// chosen based on what looks best (to me) with the default id textures.
		if (surf->flags & SURF_DRAWSLIME)
		{
			// slime - this makes it look a bit brighter, but that's OK:
			// slime *should* look radioactive!!!
			glBlendFunc (GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
			//glBlendFunc (GL_SRC_COLOR, GL_DST_COLOR);
			glColor4f (1, 1, 1, 0.8);
			scale = 128;
		}
		else if (surf->flags & SURF_DRAWLAVA)
		{
			// lava
			// ooooooh this blendfunc makes it BURN!!!
			// looks really bad with reflection mapping though
			// consider moving back to a single blendfunc for all???
			glBlendFunc (GL_ONE_MINUS_DST_COLOR, GL_ONE);
			//glBlendFunc (GL_SRC_COLOR, GL_DST_COLOR);
			glColor4f (1, 1, 1, 0.7);
			scale = 96;
		}
		else if (surf->flags & SURF_DRAWWATER)
		{
			// the wet stuff
			glBlendFunc (GL_SRC_COLOR, GL_DST_COLOR);
			glColor4f (1, 1, 1, 0.5);
			scale = 256;
		}
		else
		{
			// teleports
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glColor4f (1, 1, 1, 0.99);
			scale = 64;
		}

		R_DrawWaterChain (surf, scale);

		glColor4f (1, 1, 1, 1);

		t->lqChain = NULL;
	}

	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);
}



/*
=================
R_DrawBrushModel
=================
*/
void R_DrawInlineBModel (entity_t *ent, model_t *clmodel)
{
	int i, k;
	mplane_t *pplane;
	float dotp;
	msurface_t *surf;
	mleaf_t *bLeaf;
	vec3_t neworg;

	// calculate dynamic lighting for bmodel
	if (clmodel->firstmodelsurface != 0 && clmodel->name[0] == '*')
	{
		R_PushDlights (clmodel->nodes + clmodel->hulls[0].firstclipnode, true);
	}

	// here we get or set the leaf the bmodel is in
	if (clmodel->name[0] == '*' && waterthisframe && clmodel->modelleaf)
	{
		// see has the bmodel moved
		if (ent->origin[0] != 0 && ent->origin[1] != 0 && ent->origin[2] != 0)
		{
			// move it up a bit
			neworg[0] = clmodel->origin[0] + ent->origin[0];
			neworg[1] = clmodel->origin[1] + ent->origin[1];
			neworg[2] = clmodel->origin[2] + ent->origin[2] + 8;

			bLeaf = Mod_PointInLeaf (neworg, cl.worldmodel);
		}
		else bLeaf = clmodel->modelleaf;
	}
	else
	{
		// just set to the viewleaf
		bLeaf = r_viewleaf;
	}

	// now get it again cos we need it again
	surf = &clmodel->surfaces[clmodel->firstmodelsurface];

	for (i = 0; i < clmodel->nummodelsurfaces; i++, surf++)
	{
		if (surf->flags & (SURF_DRAWSKY | SURF_DRAWTURB)) continue;

		switch (surf->plane->type)
		{
		case PLANE_X:
			dotp = modelorg[0] - surf->plane->dist;
			break;
		case PLANE_Y:
			dotp = modelorg[1] - surf->plane->dist;
			break;
		case PLANE_Z:
			dotp = modelorg[2] - surf->plane->dist;
			break;
		default:
			dotp = DotProduct (modelorg, surf->plane->normal) - surf->plane->dist;
			break;
		}

		if (((surf->flags & SURF_PLANEBACK) && (dotp < -BACKFACE_EPSILON)) || (!(surf->flags & SURF_PLANEBACK) && (dotp > BACKFACE_EPSILON)))
		{
			surf->drawChain = surf->texinfo->texture->texturechain[0];
			surf->texinfo->texture->texturechain[0] = surf;

			addSurfs++;

			// if it's a BSP bmodel don't bother updating light
			if (clmodel->name[0] == '*') R_PrebuildSurfaceLightmap (surf);
		}
	}

	// set up the "fake" volume
	bmvol[0].visframe = r_framecount;
	bmvol[1].visframe = -2;
	bmvol[0].volumenum = 0;

	// ammo and health should never be marked underwater
	if (clmodel->name[0] != '*')
		bmvol[0].caustic = NULL;
	else bmvol[0].caustic = bLeaf->caustic;

	// allocate it to the model
	clmodel->volumes = bmvol;
	clmodel->numVolumes = 2;

	// and draw, blondie you son of a thousand fathers, all bastards like you!!!
	R_DrawVolumeChains (clmodel);
}


void R_RotateForEntity (entity_t *e, qboolean blendLerp);


void R_DrawBrushModel (entity_t *e)
{
	vec3_t mins, maxs;
	int i, j;
	qboolean rotated;
	model_t	*clmodel;

	clmodel = e->model;

	if (clmodel->nummodelsurfaces == 0) return;

	currententity = e;

	if (e->angles[0] || e->angles[1] || e->angles[2])
	{
		rotated = true;

		for (i = 0; i < 3; i++)
		{
			mins[i] = e->origin[i] - clmodel->radius;
			maxs[i] = e->origin[i] + clmodel->radius;
		}
	}
	else
	{
		rotated = false;

		VectorAdd (e->origin, clmodel->mins, mins);
		VectorAdd (e->origin, clmodel->maxs, maxs);
	}

	if (R_CullBox (mins, maxs)) return;

	glColor3f (1, 1, 1);

	VectorSubtract (r_refdef.vieworg, e->origin, modelorg);

	if (rotated)
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (modelorg, temp);
		AngleVectors (e->angles, forward, right, up);
		modelorg[0] = DotProduct (temp, forward);
		modelorg[1] = -DotProduct (temp, right);
		modelorg[2] = DotProduct (temp, up);
	}

	glPushMatrix ();

	// seems i fixed this somewhere else...  do do dooo do do do doooo (looks away)
	//e->angles[0] = -e->angles[0];	// stupid quake bug
	//e->angles[2] = -e->angles[2];	// stupid quake bug

	R_RotateForEntity (e, true);

	// seems i fixed this somewhere else...  do do dooo do do do doooo (looks away)
	//e->angles[0] = -e->angles[0];	// stupid quake bug
	//e->angles[2] = -e->angles[2];	// stupid quake bug

	R_DrawInlineBModel (e, clmodel);

	glPopMatrix ();
}


/*
=============================================================

	WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
int	addLeafs;

void R_RecursiveWorldNode (mnode_t *node)
{
	int i;
	int c;
	int side;
	mplane_t *plane;
	msurface_t *surf;
	msurface_t **mark;
	mleaf_t *pleaf;
	double dist;
	int sidebit;

	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;

	// if a leaf node, mark the surfaces as visible
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *) node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				// mark the surfaces in this leaf for drawing
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		// (only used for static entities - torches and suchlike)
		if (pleaf->efrags) R_StoreEfrags (&pleaf->efrags);

		return;
	}

	// node is just a decision point, so go down the apropriate sides
	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dist = modelorg[0] - plane->dist;
		break;
	case PLANE_Y:
		dist = modelorg[1] - plane->dist;
		break;
	case PLANE_Z:
		dist = modelorg[2] - plane->dist;
		break;
	default:
		dist = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dist >= 0)
	{
		// front side
		side = 0;
		sidebit = 0;
	}
	else
	{
		// back side
		side = 1;
		sidebit = SURF_PLANEBACK;
	}

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	if (node->numsurfaces)
	{
		for (c = node->numsurfaces, surf = cl.worldmodel->surfaces + node->firstsurface; c; c--, surf++)
		{
			// not marked as visible
			if (surf->visframe != r_framecount) continue;

			// node backfaces
			if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;

			// check for surf backfacing
			switch (surf->plane->type)
			{
			case PLANE_X:
				dist = modelorg[0] - surf->plane->dist;
				break;
			case PLANE_Y:
				dist = modelorg[1] - surf->plane->dist;
				break;
			case PLANE_Z:
				dist = modelorg[2] - surf->plane->dist;
				break;
			default:
				dist = DotProduct (modelorg, surf->plane->normal) - surf->plane->dist;
				break;
			}

			// surf backfaces
			if (!(((surf->flags & SURF_PLANEBACK) && (dist < -BACKFACE_EPSILON)) || 
				(!(surf->flags & SURF_PLANEBACK) && (dist > BACKFACE_EPSILON)))) continue;

			if (surf->flags & SURF_DRAWSKY)
			{
				skythisframe = true;
			}
			else if (surf->flags & SURF_DRAWTURB)
			{
				// store whether water is visible this frame to reduce Mod_PointInLeaf checking
				// for brush models, as well as to determine whether or not we need to draw it!!!
				waterthisframe = true;

				surf->lqChain = surf->texinfo->texture->lqChain;
				surf->texinfo->texture->lqChain = surf;

				addSurfs++;
			}
			else
			{
				// build a seperate texture chain for each volume
				surf->drawChain = surf->texinfo->texture->texturechain[surf->volume];
				surf->texinfo->texture->texturechain[surf->volume] = surf;

				// mark the volume as visible
				cl.worldmodel->volumes[surf->volume].visframe = r_framecount;

				// now that we have the lighting info we can update it here.
				R_PrebuildSurfaceLightmap (surf);

				addSurfs++;
			}
		}
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}


/*
=============
R_DrawWorld
=============
*/
float lighttime = 0;

void R_DrawWorld (void)
{
	entity_t	ent;

	// set up our current tmus for the render - only of value for when i want to
	// test how MHQuake looks on cards with less TMUs than i have
	if (gl_activetmus.value > gl_textureunits) Cvar_SetValue ("gl_activetmus", gl_textureunits);
	if (gl_activetmus.value < 2) Cvar_SetValue ("gl_activetmus", 2);

	gl_currenttmus = (int) gl_activetmus.value;

	memset (&ent, 0, sizeof (ent));
	ent.model = cl.worldmodel;

	VectorCopy (r_refdef.vieworg, modelorg);

	currententity = &ent;

	glColor3f (1, 1, 1);

	// default values
	skythisframe = false;
	waterthisframe = false;

	R_PushDlights (cl.worldmodel->nodes, false);

	R_RecursiveWorldNode (cl.worldmodel->nodes);

	// only update lighting 10 times per second.  next to no visible difference.
	if (lighttime >= LIGHT_UPLOAD) lighttime = 0;
	lighttime += frametime;

	// and draw, Texan!!!
	// (The best jokes work on more than one level.  Here I'm not only punning on the word
	// "draw" by putting it in a completely irrelevant Western context, but also bringing in
	// the fact that ID are based in Texas.  I'll get me coat...)
	R_DrawVolumeChains (cl.worldmodel);

	// we make sure here that we only draw sky when required
	if (skythisframe) R_DrawSkyChain ();
}


/*
===============
AddVisibleLeaf

Adds the specified leaf to the PVS
This is tolerant of adding the same leaf twice...
===============
*/
void AddVisibleLeaf (mleaf_t *VLeaf)
{
	byte	*vis;
	mnode_t	*node;
	int		i;
	byte	solid[4096];

	// already been added this frame
	if (VLeaf->waterVisFrame == r_visframecount) return;

	if (r_novis.value)
	{
		vis = solid;

		// set it to 255 so the bits always match
		memset (solid, 0xff, (cl.worldmodel->numleafs + 7) >> 3);
	}
	else vis = Mod_LeafPVS (VLeaf, cl.worldmodel);

	addLeafs = 0;

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			addLeafs++;

			// mark the visible nodes
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				// already added
				if (node->visframe == r_visframecount) break;

				node->visframe = r_visframecount;
				node = node->parent;
			}
			while (node);
		}
	}

	// mark it as having been checked
	VLeaf->waterVisFrame = r_visframecount;
}


/*
===============
R_MarkLeaves
===============
*/
extern cvar_t gl_lockpvs;

void R_MarkLeaves (void)
{
	//addSurfs = 0;

	// still in the same leaf
	if (r_oldviewleaf == r_viewleaf) return;

	if (gl_lockpvs.value) return;

	// this only increments when the viewleaf changes
	r_visframecount++;

	// add the viewleaf (we can add PVS for any leaf we want with this function)
	AddVisibleLeaf (r_viewleaf);

	// is this necessary any more???
	if (r_oldviewleaf)
	{
		if (r_oldviewleaf->contents != r_viewleaf->contents)
		{
			// fix water transitions.  If we have translucent water it's probably not
			// much of a bother as the vis for both leafs is likely to be very similar.
			// The only time this could cause bother is if we teleport from above water
			// to underwater or vice versa, although the hit is probably not that bad.
			AddVisibleLeaf (r_oldviewleaf);
		}
	}

	r_oldviewleaf = r_viewleaf;
}



int bound (int lower, int num, int upper)
{
	if (num < lower) return lower;
	if (num > upper) return upper;
	return num;
}


void GL_BuildPolygonFromSurface (model_t *mod, msurface_t *surf)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	float		*vec;
	float		s, t;
	glpoly_t	*poly;

	// reconstruct the polygon
	pedges = mod->edges;
	lnumverts = surf->numedges;

	// draw texture
	poly = Hunk_Alloc (sizeof (glpoly_t) + (lnumverts - 4) * VERTEXSIZE * sizeof (float));
	poly->next = surf->polys;
	poly->flags = surf->flags;
	surf->polys = poly;
	poly->numverts = lnumverts;

	for (i = 0; i < lnumverts; i++)
	{
		lindex = mod->surfedges[surf->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = mod->vertexes[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = mod->vertexes[r_pedge->v[1]].position;
		}

		// poly world verts
		poly->verts[i][0] = vec[0];
		poly->verts[i][1] = vec[1];
		poly->verts[i][2] = vec[2];

		// texture s and t
		poly->verts[i][3] = (DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]) / surf->texinfo->texture->width;
		poly->verts[i][4] = (DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]) / surf->texinfo->texture->height;

		// allocate 2 new verts for a 64 * 64 version of the texture, and set their data
		// set up as if the texture was 64 * 64 to keep caustic warps consistent
		// these verts are also used for detail so they are massaged a little to include the
		// necessary scaling for the detail tex so that we're not multiplying or modifying the
		// texture matrix, both of which put extra load on the system
		poly->verts[i][7] = (DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]) / 16.0;
		poly->verts[i][8] = (DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]) / 16.0;

		// optimise caustic warps
		poly->verts[i][9] = poly->verts[i][7] * STMult * 4;
		poly->verts[i][10] = poly->verts[i][8] * STMult * 4;
	}
}


void GL_RSurfInit (void)
{
	int i;
	float j;
	float inc = (2.0 * M_PI) / (float) SINTABLE_SIZE;

	if (gl_textureunits >= 3)
	{
		// pre-set up TMU 2 for caustics but leave it disabled for now
		glActiveTexture (GL_TEXTURE2);
		glMatrixMode (GL_TEXTURE);
		glLoadIdentity ();
		glScalef (CAUSTIC_SCALE, CAUSTIC_SCALE, 1);
		glMatrixMode (GL_MODELVIEW);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		// no need to reset the matrix mode when done...
		glActiveTexture (GL_TEXTURE0);
	}

	for (i = 0, j = 0; i < SINTABLE_SIZE; i++, j += inc)
	{
		sintable[i] = sin (j) * 4.0;
		sintablex2[i] = sin (j) * 8.0;
	}

	STBitMask = SINTABLE_SIZE - 1;

	STMult = (float) SINTABLE_SIZE / (6.0 * M_PI);
	STMultx2 = (float) SINTABLE_SIZE / (2.0 * M_PI);
}

