/*
Copyright (C) 1997 - 2001 Id Software, Inc.

This program is free software; you can redistribute it and / or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.

*/ 

/*
==========================================================================================================================================

MHQUAKE II VERTEX ARRAYS SUBSYSTEM

This was written to enable much the same rendering code to be used for almost everything.  A smallish enough vertex array buffer is
used to enable a balance of fast transfers of smaller chunks of data to the 3D card (with - hopefully - a decent level of amortization)
and a reduced number of API calls.  This system allows up to MAX_VERTS verts to be transferred at once.  In reality though, the transfer
has to take place at least every time the primitive type changes (that's vertex arrays for you).

To use this, just do the following:

1) Call R_InitVArrays with the type (defined in gl_local.h) specified for which arrays you need.
2) Call the relevant drawing routines.  See the example calls in the rest of the engine.
3) Call R_KillVArrays when done.

Hopefully this system should be relatively easy to do as a "drop 'n' go" in your own engine, although you will have to modify parts
of the code, as some of it is MHQuake II specific.

==========================================================================================================================================
*/

#include "gl_local.h"

// this must be equal to MAX_VERTS as it's possible for a single MDL to be composed of only one triangle fan or strip.
// the storage overhead is only in the order of 100K anyway, so it's no big deal.
// add 2 more verts for surface warping
#define MAX_VARRAY_VERTS MAX_VERTS + 2
#define MAX_VARRAY_VERTEX_SIZE 11

float VArrayVerts[MAX_VARRAY_VERTS * MAX_VARRAY_VERTEX_SIZE];

// pointer for dynamic vert allocation
static float *VArray = &VArrayVerts[0];

// number of verts allocated
static int VertexCounter = 0;

// vertex array kill flags
#define KILL_TMU0_POINTER	1
#define KILL_TMU1_POINTER	2
#define KILL_RGBA_POINTER	4

// sizes of our vertexes.  the vertex type can be used as an index into this array
int VertexSizes[] = {5, 5, 7, 7, 9, 11, 5, 3};


int KillFlags;

/*
=================
R_InitVArrays

Sets up the current vertex arrays we're going to use for rendering.
=================
*/
void R_InitVArrays (int varraytype)
{
	// it's assumed that the programmer has already called glDrawArrays for everything before calling this, so
	// here we will just re-init our pointer and counter
	VArray = &VArrayVerts[0];
	VertexCounter = 0;

	// init the kill flags so we'll know what to kill
	KillFlags = 0;

	// all vertex types bring up a vertex pointer
	// uses array indices 0, 1, 2
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (float) * VertexSizes[varraytype], &VArrayVerts[0]);

	// no texturing at all
	if (varraytype == VERT_NO_TEXTURE) return;

	// the simplest possible textured render uses a texcoord pointer for TMU 0
	if (varraytype == VERT_SINGLE_TEXTURED)
	{
		// uses array indices 3, 4
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * VertexSizes[VERT_SINGLE_TEXTURED], &VArrayVerts[3]);

		KillFlags |= KILL_TMU0_POINTER;

		return;
	}

	// 2 TMUs which share texcoords
	if (varraytype == VERT_DUAL_TEXTURED)
	{
		// uses array indices 3, 4
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * VertexSizes[VERT_DUAL_TEXTURED], &VArrayVerts[3]);

		// uses array indices 3, 4
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * VertexSizes[VERT_DUAL_TEXTURED], &VArrayVerts[3]);

		KillFlags |= (KILL_TMU0_POINTER | KILL_TMU1_POINTER);

		return;
	}

	// the bumpmapped setup gets to reuse the verts as it's texcoords for tmu 0
	if (varraytype == VERT_BUMPMAPPED)
	{
		// uses array indices 0, 1, 2
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (3, GL_FLOAT, sizeof (float) * VertexSizes[VERT_BUMPMAPPED], &VArrayVerts[0]);

		// uses array indices 3, 4
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * VertexSizes[VERT_BUMPMAPPED], &VArrayVerts[3]);

		KillFlags |= (KILL_TMU0_POINTER | KILL_TMU1_POINTER);

		return;
	}

	// a standard 2 tmu multitexture needs 2 texcoord pointers.
	if (varraytype == VERT_MULTI_TEXTURED)
	{
		// uses array indices 3, 4
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * VertexSizes[VERT_MULTI_TEXTURED], &VArrayVerts[3]);

		// uses array indices 5, 6
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * VertexSizes[VERT_MULTI_TEXTURED], &VArrayVerts[5]);

		KillFlags |= (KILL_TMU0_POINTER | KILL_TMU1_POINTER);

		return;
	}

	// no texture is used here but we do specify a colour (used in Q2 for laser beams, etc)
	if (varraytype == VERT_COLOURED_UNTEXTURED)
	{
		// uses array indices 3, 4, 5, 6
		glEnableClientState (GL_COLOR_ARRAY);
		glColorPointer (4, GL_FLOAT, sizeof (float) * VertexSizes[VERT_COLOURED_UNTEXTURED], &VArrayVerts[3]);

		KillFlags |= KILL_RGBA_POINTER;

		return;
	}

	// single texture + colour (water, mdls, and so on)
	if (varraytype == VERT_COLOURED_TEXTURED)
	{
		// uses array indices 3, 4
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * VertexSizes[VERT_COLOURED_TEXTURED], &VArrayVerts[3]);

		// uses array indices 5, 6, 7, 8
		glEnableClientState (GL_COLOR_ARRAY);
		glColorPointer (4, GL_FLOAT, sizeof (float) * VertexSizes[VERT_COLOURED_TEXTURED], &VArrayVerts[5]);

		KillFlags |= (KILL_TMU0_POINTER | KILL_RGBA_POINTER);

		return;
	}

	// multi texture + colour
	if (varraytype == VERT_COLOURED_MULTI_TEXTURED)
	{
		// uses array indices 3, 4
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * VertexSizes[VERT_COLOURED_MULTI_TEXTURED], &VArrayVerts[3]);

		// uses array indices 5, 6
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (float) * VertexSizes[VERT_COLOURED_MULTI_TEXTURED], &VArrayVerts[5]);

		// uses array indices 7, 8, 9, 10
		glEnableClientState (GL_COLOR_ARRAY);
		glColorPointer (4, GL_FLOAT, sizeof (float) * VertexSizes[VERT_COLOURED_MULTI_TEXTURED], &VArrayVerts[7]);

		KillFlags |= (KILL_TMU0_POINTER | KILL_TMU1_POINTER | KILL_RGBA_POINTER);

		return;
	}
}


/*
=================
R_KillVArrays

Shuts down the specified vertex arrays.  Again, for maximum flexibility no programmer
hand-holding is done.
=================
*/
void R_KillVArrays (void)
{
	if (KillFlags & KILL_RGBA_POINTER)
		glDisableClientState (GL_COLOR_ARRAY);

	if (KillFlags & KILL_TMU1_POINTER)
	{
		glClientActiveTexture (GL_TEXTURE1);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}

	if (KillFlags & KILL_TMU0_POINTER)
	{
		glClientActiveTexture (GL_TEXTURE0);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}

	// always kill
	glDisableClientState (GL_VERTEX_ARRAY);
}


// originally done for heat haze only, extended to all surfs
#define HH_TEXTURED		1
#define HH_LIGHTMAPPED	2
#define HH_DETAIL		3

extern cvar_t *r_hazewibble;

void R_AddWarpedSurfToVArray (int RenderType, msurface_t *surf, float scroll, float scale)
{
	glpoly_t *p = surf->polys;
	int i;
	int HAZE_VERTEX_SIZE_TYPE;

	// get the scale of the warp
	float warpscale = surf->warpfade * surf->warpscale;

	// scale the warp
	float hazetime = r_newrefdef.time / scale;

	// get the correct vertex size
	switch (RenderType)
	{
	case HH_TEXTURED:
	case HH_DETAIL:
		HAZE_VERTEX_SIZE_TYPE = VERT_SINGLE_TEXTURED;
		break;

	case HH_LIGHTMAPPED:
		HAZE_VERTEX_SIZE_TYPE = VERT_MULTI_TEXTURED;
		break;
	}

	for (; p; p = p->next)
	{
		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		// first polyvert in the trifan is the midpoint.  this is the only one that warps.
		// use the calculated warpscale rather than a constant so that the warp is scaled to the size of the surf.
		VArray[0] = p->midpoint.tv[0] + warpscale * sin (p->midpoint.tv[1] * 0.05 + hazetime) * sin (p->midpoint.tv[2] * 0.05 + hazetime);
		VArray[1] = p->midpoint.tv[1] + warpscale * sin (p->midpoint.tv[0] * 0.05 + hazetime) * sin (p->midpoint.tv[2] * 0.05 + hazetime);
		VArray[2] = p->midpoint.tv[2];

		// the texcoords to use depend on the render type
		switch (RenderType)
		{
		case HH_LIGHTMAPPED:
			// lightmap texture coords
			// fall through to textured for the world st
			VArray[5] = p->midpoint.lm[0];
			VArray[6] = p->midpoint.lm[1];

		case HH_TEXTURED:
			// world texture coords
			VArray[3] = p->midpoint.st[0] + scroll;
			VArray[4] = p->midpoint.st[1];
			break;

		case HH_DETAIL:
			// detail texture coords
			VArray[3] = p->midpoint.st128[0] + scroll;
			VArray[4] = p->midpoint.st128[1];
			break;
		}

		// nothing else is needed
		// increment pointer and counter
		VArray += VertexSizes[HAZE_VERTEX_SIZE_TYPE];
		VertexCounter++;

		// now add the normal verts in order, but do not warp to avoid breakup
		for (i = 0; i < p->numverts; i++)
		{
			// copy in vertex data
			VArray[0] = p->verts[i].tv[0];
			VArray[1] = p->verts[i].tv[1];
			VArray[2] = p->verts[i].tv[2];

			// the texcoords to use depend on the render type
			switch (RenderType)
			{
			case HH_LIGHTMAPPED:
				// lightmap texture coords
				VArray[5] = p->verts[i].lm[0];
				VArray[6] = p->verts[i].lm[1];

			case HH_TEXTURED:
				// world texture coords
				VArray[3] = p->verts[i].st[0] + scroll;
				VArray[4] = p->verts[i].st[1];
				break;

			case HH_DETAIL:
				// detail texture coords
				VArray[3] = p->verts[i].st128[0] + scroll;
				VArray[4] = p->verts[i].st128[1];
				break;
			}

			// nothing else is needed
			// increment pointer and counter
			VArray += VertexSizes[HAZE_VERTEX_SIZE_TYPE];
			VertexCounter++;
		}

		// finally copy the first normal polyvert back to finish the tri fan
		VArray[0] = p->verts[0].tv[0];
		VArray[1] = p->verts[0].tv[1];
		VArray[2] = p->verts[0].tv[2];

		// the texcoords to use depend on the render type
		switch (RenderType)
		{
		case HH_LIGHTMAPPED:
			// lightmap texture coords
			VArray[5] = p->verts[0].lm[0];
			VArray[6] = p->verts[0].lm[1];

		case HH_TEXTURED:
			// world texture coords
			VArray[3] = p->verts[0].st[0] + scroll;
			VArray[4] = p->verts[0].st[1];
			break;

		case HH_DETAIL:
			// detail texture coords
			VArray[3] = p->verts[0].st128[0] + scroll;
			VArray[4] = p->verts[0].st128[1];
			break;
		}

		// nothing else is needed
		// increment pointer and counter
		VArray += VertexSizes[HAZE_VERTEX_SIZE_TYPE];
		VertexCounter++;

		// draw the poly (this will always be a tri fan so the variable type is redundant)
		glDrawArrays (p->QGL_HAZE_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}


/*
=================
R_AddLightMappedSurfToVArray

Adds a surf without texturing to the varray.  The surface can have 1 or more polys, and the VArray is flushed immediately
after each poly is rendered.

It's assumed that the programmer has set up the vertex arrays and other required state properly before calling this, otherwise
weird things may happen!!!
=================
*/
void R_AddVertGeometryToVArray (msurface_t *surf)
{
	glpoly_t *p = surf->polys;
	int i;

	// no warp cases are required as this is only used for stencil buffer testing
	for (; p; p = p->next)
	{
		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		for (i = 0; i < p->numverts; i++)
		{
			// copy in vertex data only
			VArray[0] = p->verts[i].tv[0];
			VArray[1] = p->verts[i].tv[1];
			VArray[2] = p->verts[i].tv[2];

			// nothing else is needed
			// increment pointer and counter
			VArray += VertexSizes[VERT_NO_TEXTURE];
			VertexCounter++;
		}

		// draw the poly
		glDrawArrays (p->QGL_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}


/*
=================
R_AddLightMappedSurfToVArray

Adds a lightmapped surf to the varray.  The surface can have 1 or more polys, and the VArray is flushed immediately
after each poly is rendered.

It's assumed that the programmer has set up the vertex arrays properly before calling this, otherwise weird things may happen!!!
=================
*/
void R_AddLightMappedSurfToVArray (msurface_t *surf, float scroll)
{
	glpoly_t *p = surf->polys;
	int i;

	if (surf->flags & SURF_UNDERWATER)
	{
		R_AddWarpedSurfToVArray (HH_LIGHTMAPPED, surf, scroll, 1);
		return;
	}

	if (surf->hazemask && surf->warpfade > 0.0f && r_hazewibble->value > 0.0f)
	{
		R_AddWarpedSurfToVArray (HH_LIGHTMAPPED, surf, scroll, r_hazewibble->value);
		return;
	}

	for (; p; p = p->next)
	{
		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		for (i = 0; i < p->numverts; i++)
		{
			// copy in vertex data
			VArray[0] = p->verts[i].tv[0];
			VArray[1] = p->verts[i].tv[1];
			VArray[2] = p->verts[i].tv[2];

			// world texture coords
			VArray[3] = p->verts[i].st[0] + scroll;
			VArray[4] = p->verts[i].st[1];

			// lightmap texture coords
			VArray[5] = p->verts[i].lm[0];
			VArray[6] = p->verts[i].lm[1];

			// nothing else is needed
			// increment pointer and counter
			VArray += VertexSizes[VERT_MULTI_TEXTURED];
			VertexCounter++;
		}

		// draw the poly
		glDrawArrays (p->QGL_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}


/*
=================
R_AddTexturedSurfToVArray

Adds a textured surf to the varray.  The surface can have 1 or more polys, and the VArray is flushed immediately
after each poly is rendered.

It's assumed that the programmer has set up the vertex arrays properly before calling this, otherwise weird things may happen!!!
=================
*/
void R_AddTexturedSurfToVArray (msurface_t *surf, float scroll)
{
	glpoly_t *p = surf->polys;
	int i;

	if (surf->flags & SURF_UNDERWATER)
	{
		R_AddWarpedSurfToVArray (HH_TEXTURED, surf, scroll, 1);
		return;
	}

	if (surf->hazemask && surf->warpfade > 0.0f && r_hazewibble->value > 0.0f)
	{
		R_AddWarpedSurfToVArray (HH_TEXTURED, surf, scroll, r_hazewibble->value);
		return;
	}

	for (; p; p = p->next)
	{
		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		for (i = 0; i < p->numverts; i++)
		{
			// copy in vertex data
			VArray[0] = p->verts[i].tv[0];
			VArray[1] = p->verts[i].tv[1];
			VArray[2] = p->verts[i].tv[2];

			// world texture coords
			VArray[3] = p->verts[i].st[0] + scroll;
			VArray[4] = p->verts[i].st[1];

			// nothing else is needed
			// increment pointer and counter
			VArray += VertexSizes[VERT_SINGLE_TEXTURED];
			VertexCounter++;
		}

		// draw the poly
		glDrawArrays (p->QGL_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}


void R_AddTurbulentSurfToVArray (msurface_t *surf, float scroll)
{
	glpoly_t *p = surf->polys;
	int i;
	float st[2];
	float os, ot;

	for (; p; p = p->next)
	{
		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		for (i = 0; i < p->numverts; i++)
		{
			// copy in vertex data
			VArray[0] = p->verts[i].tv[0];
			VArray[1] = p->verts[i].tv[1];
			VArray[2] = p->verts[i].tv[2];

			os = p->verts[i].st[0];
			ot = p->verts[i].st[1];

			st[0] = os + r_turbsin[Q_ftol (((ot * 0.125 + r_newrefdef.time) * TURBSCALE)) & 255];
			st[0] += scroll;
			st[0] *= 0.015625;

			st[1] = ot + r_turbsin[Q_ftol (((os * 0.125 + r_newrefdef.time) * TURBSCALE)) & 255];
			st[1] *= 0.015625;

			// world texture coords
			VArray[3] = st[0];
			VArray[4] = st[1];

			// nothing else is needed
			// increment pointer and counter
			VArray += VertexSizes[VERT_SINGLE_TEXTURED];
			VertexCounter++;
		}

		// draw the poly
		glDrawArrays (p->QGL_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}


/*
=================
R_AddDetailSurfToVArray

Adds a detail textured surf to the varray.  The surface can have 1 or more polys, and the VArray is flushed immediately
after each poly is rendered.

It's assumed that the programmer has set up the vertex arrays properly before calling this, otherwise weird things may happen!!!
=================
*/
void R_AddDetailSurfToVArray (msurface_t *surf)
{
	glpoly_t *p = surf->polys;
	int i;

	if (surf->flags & SURF_UNDERWATER)
	{
		R_AddWarpedSurfToVArray (HH_DETAIL, surf, 0, 1);
		return;
	}

	if (surf->hazemask && surf->warpfade > 0.0f && r_hazewibble->value > 0.0f)
	{
		R_AddWarpedSurfToVArray (HH_DETAIL, surf, 0, r_hazewibble->value);
		return;
	}

	for (; p; p = p->next)
	{
		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		for (i = 0; i < p->numverts; i++)
		{
			// copy in vertex data
			VArray[0] = p->verts[i].tv[0];
			VArray[1] = p->verts[i].tv[1];
			VArray[2] = p->verts[i].tv[2];

			// world texture coords
			VArray[3] = p->verts[i].st128[0];
			VArray[4] = p->verts[i].st128[1];

			// nothing else is needed
			// increment pointer and counter
			VArray += VertexSizes[VERT_SINGLE_TEXTURED];
			VertexCounter++;
		}

		// draw the poly
		glDrawArrays (p->QGL_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}


/*
=================
R_AddColouredMeshToVArray

Adds a coloured but untextured model to the vertex arrays.

It's assumed that the programmer has set up the vertex arrays properly before calling this, otherwise weird things may happen!!!
=================
*/
void R_AddColouredMeshToVArray (vec4_t *verts, int *order, vec3_t shadelight, float alpha)
{
	int count;
	GLenum QGL_PRIMITIVE_TYPE;

	while (1)
	{
		count = *order++;

		// done
		if (!count) return;

		if (count < 0) 
		{
			count = -count;
			QGL_PRIMITIVE_TYPE = GL_TRIANGLE_FAN;
		}
		else QGL_PRIMITIVE_TYPE = GL_TRIANGLE_STRIP;

		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		do
		{
			VArray[0] = verts[order[2]][0];
			VArray[1] = verts[order[2]][1];
			VArray[2] = verts[order[2]][2];

			VArray[3] = shadelight[0];
			VArray[4] = shadelight[1];
			VArray[5] = shadelight[2];
			VArray[6] = alpha;

			order += 3;

			// increment pointer and counter
			VArray += VertexSizes[VERT_COLOURED_UNTEXTURED];
			VertexCounter++;
		} while (--count);

		// draw the poly
		glDrawArrays (QGL_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}


/*
=================
R_AddLitTexturedMeshToVArray

Adds a regular lit and textured model to the vertex arrays.

It's assumed that the programmer has set up the vertex arrays properly before calling this, otherwise weird things may happen!!!
=================
*/
void R_AddLitTexturedMeshToVArray (vec4_t *verts, dtrivertx_t *lightverts, int *order, vec3_t shadelight, float *lightdots, float alpha)
{
	int count;
	GLenum QGL_PRIMITIVE_TYPE;
	float l;

	while (1)
	{
		count = *order++;

		// done
		if (!count) return;

		if (count < 0) 
		{
			count = -count;
			QGL_PRIMITIVE_TYPE = GL_TRIANGLE_FAN;
		}
		else QGL_PRIMITIVE_TYPE = GL_TRIANGLE_STRIP;

		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		do
		{
			l = lightdots[lightverts[order[2]].lightnormalindex];

			VArray[0] = verts[order[2]][0];
			VArray[1] = verts[order[2]][1];
			VArray[2] = verts[order[2]][2];

			VArray[3] = ((float *) order)[0];
			VArray[4] = ((float *) order)[1];

			VArray[5] = l * shadelight[0];
			VArray[6] = l * shadelight[1];
			VArray[7] = l * shadelight[2];
			VArray[8] = alpha;

			order += 3;

			// increment pointer and counter
			VArray += VertexSizes[VERT_COLOURED_TEXTURED];
			VertexCounter++;
		} while (--count);

		// draw the poly
		glDrawArrays (QGL_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}


/*
=================
R_AddLitTexturedMeshToVArray

Adds a textured (only) model to the vertex arrays.

It's assumed that the programmer has set up the vertex arrays properly before calling this, otherwise weird things may happen!!!
=================
*/
void R_AddTexturedMeshToVArray (vec4_t *verts, int *order)
{
	int count;
	GLenum QGL_PRIMITIVE_TYPE;

	while (1)
	{
		count = *order++;

		// done
		if (!count) return;

		if (count < 0) 
		{
			count = -count;
			QGL_PRIMITIVE_TYPE = GL_TRIANGLE_FAN;
		}
		else QGL_PRIMITIVE_TYPE = GL_TRIANGLE_STRIP;

		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		do
		{
			VArray[0] = verts[order[2]][0];
			VArray[1] = verts[order[2]][1];
			VArray[2] = verts[order[2]][2];

			VArray[3] = ((float *) order)[0];
			VArray[4] = ((float *) order)[1];

			order += 3;

			// increment pointer and counter
			VArray += VertexSizes[VERT_SINGLE_TEXTURED];
			VertexCounter++;
		} while (--count);

		// draw the poly
		glDrawArrays (QGL_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}


/*
=================
R_AddHUDMeshToVArray

Adds a multi-textured (but unlit) model to the vertex arrays for the HUD.

It's assumed that the programmer has set up the vertex arrays properly before calling this, otherwise weird things may happen!!!
=================
*/
void R_AddHUDMeshToVArray (daliasframe_t *frame, dtrivertx_t *verts, int *order)
{
	int count;
	GLenum QGL_PRIMITIVE_TYPE;

	while (1)
	{
		count = *order++;

		// done
		if (!count) return;

		if (count < 0) 
		{
			count = -count;
			QGL_PRIMITIVE_TYPE = GL_TRIANGLE_FAN;
		}
		else QGL_PRIMITIVE_TYPE = GL_TRIANGLE_STRIP;

		// reset pointer and counter
		VArray = &VArrayVerts[0];
		VertexCounter = 0;

		do
		{
			VArray[0] = (float) verts[order[2]].v[0] * frame->scale[0] + frame->translate[0];
			VArray[1] = (float) verts[order[2]].v[1] * frame->scale[1] + frame->translate[1];
			VArray[2] = (float) verts[order[2]].v[2] * frame->scale[2] + frame->translate[2];

			VArray[3] = ((float *) order)[0];
			VArray[4] = ((float *) order)[1];

			VArray[5] = ((float *) order)[0];
			VArray[6] = ((float *) order)[1];

			order += 3;

			// increment pointer and counter
			VArray += VertexSizes[VERT_MULTI_TEXTURED];
			VertexCounter++;
		} while (--count);

		// draw the poly
		glDrawArrays (QGL_PRIMITIVE_TYPE, 0, VertexCounter);
	}
}

