/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"
#include "glext.h"


qboolean skythisframe;
qboolean waterthisframe;


extern int lightmap_textures;
extern int stainmap_textures;

int		addSurfs;

// we need a "fake" volume for brush models
volume_t bmvol[2];

qboolean fbChain = false;

#define CAUSTIC_SCALE 0.00390625

int gl_currenttmus;

void R_PrebuildLightmaps (model_t *mod);

void R_PushDlights (mnode_t *node, qboolean bmnode);
void R_PrebuildSurfaceLightmap (msurface_t *surf);

cvar_t r_detailtextures = {"r_detailtextures", "1", true};

/*
===============
Sin approximations.  These work in a similar fashion to the old gl_warpsin.h but are
calculated at engine startup rather than stored in a fixed array.  This lets me keep
a higher precision copy of the table in memory without having to use a great big
include file containing loadsa data, plus I can also tweak the precision.  These are
faster than the library sin function and are significantly faster than JoeQuake's
SINTABLE_APPROX function (which was actually *slower* than the library sin function
when I had it in here...)

SINTABLE_SIZE is currently 1024, which gives a good degree of accuracy without the need
for a massive table lookup.  You can increase it by changing the define in glquake.h,
but be aware that beyond a certain level rounding errors start to come in and the
visual result gets incredibly choppy.  Decreasing it to 256 gives comparable results to
the old gl_warpsin.h method.
===============
*/
float sintable[SINTABLE_SIZE];
float sintablex2[SINTABLE_SIZE];

int STBitMask;

float STMult;
float STMultx2;

float warptime;


extern int detailtexture;
extern int caustictexture;


/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
static texture_t *R_TextureAnimation (texture_t *base)
{
	int		relative;
	int		count;

	// save out the base texture in case we have a bad cycle
	texture_t *bad = base;

	if (currententity->frame)
	{
		if (base->alternate_anims)
			base = base->alternate_anims;
	}

	// no animations
	if (!base->anim_total) return base;

	relative = (int) (realtime * 10) % base->anim_total;

	count = 0;

	while (base->anim_min > relative || base->anim_max <= relative)
	{
		base = base->anim_next;

		// bad animation cycle so just return the base texture
		if (!base) return bad;
		if (++count > 100) return bad;
	}

	return base;
}


/*
===================
R_DrawFullbrightChain

Draws the chain of fullbrights sorted by texture
This should be the very last thing drawn as we don't want fullbrights to be
overdrawn by anything else.  They are drawn as a seperate pass cos of the different BlendFunc
===================
*/
static void R_DrawFullbrightChain (model_t *mod)
{
	int i;
	int j;
	float *v;
	msurface_t *surf;
	texture_t *t;

	// MH - my previous alpha test method didn't work with the quake hi-res luma textures
	// so i changed it.  bleagh.
	glDepthMask (GL_FALSE);
	glEnable (GL_BLEND);
	glBlendFunc (GL_ONE, GL_ONE);

	for (i = 0; i < mod->numtextures; i++)
	{
		if (!mod->textures[i]) continue;
		if (!mod->textures[i]->fbChain) continue;

		t = R_TextureAnimation (mod->textures[i]);

		surf = mod->textures[i]->fbChain;

		glBindTexture (GL_TEXTURE_2D, t->fullbright);

		for (; surf; surf = surf->fbChain)
		{
			glBegin (GL_POLYGON);

			for (j = 0, v = surf->polys->verts[0]; j < surf->polys->numverts; j++, v += VERTEXSIZE)
			{
				glTexCoord2fv (&v[3]);
				glVertex3fv (v);
			}

			glEnd ();
		}

		mod->textures[i]->fbChain = NULL;
	}

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	fbChain = false;
}


static void R_DrawDetailChain (model_t *mod)
{
	int i;
	float *v;
	msurface_t *surf;

	// enable blending for detail textures
	glEnable (GL_BLEND);
	glBlendFunc (GL_DST_COLOR, GL_SRC_COLOR);
	glDepthMask (GL_FALSE);

	// bind the detail texture (same texture is used for all surfs)
	glBindTexture (GL_TEXTURE_2D, detailtexture);

	// scale the texture matrix to make the detail texture smaller
	// (hides pixellation)
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glScalef (1.5, 1.5, 1);
	glMatrixMode (GL_MODELVIEW);

	// draw it
	for (surf = mod->accumSurf; surf; surf = surf->accumSurf)
	{
		glBegin (GL_POLYGON);

		for (i = 0, v = surf->polys->verts[0]; i < surf->polys->numverts; i++, v += VERTEXSIZE)
		{
			// v[7] and v[8] contain texcoords suitable for a 64 x 64 version of the texture
			// this prevents stretching of the detail texture to meet the dimensions of the
			// original surface texture
			glTexCoord2fv (&v[7]);
			glVertex3fv (v);
		}

		glEnd ();
	}

	// restore the texture matrix
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);

	// shut down blending
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	// NULL the model's accumulated surfs
	mod->accumSurf = NULL;
}


static void R_SetSurfaceFog (texture_t *caustic)
{
	glFogi (GL_FOG_MODE, GL_LINEAR);
	glFogfv (GL_FOG_COLOR, caustic->uwater_fog);
	glFogf (GL_FOG_START, 10);
	glFogf (GL_FOG_END, 1000);
	glFogf (GL_FOG_DENSITY, 0.2);
	glEnable (GL_FOG);
}


/*
===============
R_DrawCausticChain

Caustics code ripped from a Mark Kilgard example
===============
*/
static void R_DrawCausticChain (texture_t *caustic, msurface_t *lqChain)
{
	int i;
	float *v;
	msurface_t *surf;
	static float caustictime = 0;
	float sPlane[4];
	float tPlane[4];

	caustictime += frametime * 72;

	if (caustictime >= 72) caustictime -= 72;

	// enable blending for caustic textures
	glEnable (GL_BLEND);
	glBlendFunc (GL_ZERO, GL_SRC_COLOR);
	glDepthMask (GL_FALSE);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// bind the caustic texture (same texture is used for all surfs)
	glBindTexture (GL_TEXTURE_2D, caustictexture + (int) caustictime);

	// scale the texture matrix to make the caustic texture bigger
	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glScalef (CAUSTIC_SCALE, CAUSTIC_SCALE, 1);

	glTexGeni (GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGeni (GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);

	glEnable (GL_TEXTURE_GEN_S);
	glEnable (GL_TEXTURE_GEN_T);

	// draw it
	for (surf = lqChain; surf; surf = surf->lqChain)
	{
		sPlane[1] = sPlane[2] = sPlane[3] = tPlane[0] = tPlane[1] = tPlane[3] = 0.0;
		sPlane[0] = tPlane[2] = 1.0;

		// shift the generated texcoords according to the surfaces plane
		switch (surf->plane->type)
		{
		case PLANE_Y:
			// vertical in line with the world grid
			tPlane[1] = 1.0;
		case PLANE_X:
			// vertical in line with the world grid
			sPlane[1] = 1.0;
			break;

		case PLANE_ANYX:
			// vertical at an angle to the world grid
			// (_ANYY doesn't seem to be used... hmmm, must check...)
			// keep the defaults, nothing else set
			break;

		case PLANE_Z:
			// flat
		default:
			// same as flat
			tPlane[1] = 1.0;
			break;
		}

		glTexGenfv (GL_S, GL_OBJECT_PLANE, sPlane);
		glTexGenfv (GL_T, GL_OBJECT_PLANE, tPlane);

		glBegin (GL_POLYGON);

		for (i = 0, v = surf->polys->verts[0]; i < surf->polys->numverts; i++, v += VERTEXSIZE)
			glVertex3fv (v);

		glEnd ();
	}

	// restore the texture matrix
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);

	// shut down blending
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	glDisable (GL_TEXTURE_GEN_S);
	glDisable (GL_TEXTURE_GEN_T);
}


/*
=====================
R_DrawTextureChains

Sometimes the old ways are still the best...
=====================
*/
extern glpoly_t *stainmap_polys[MAX_LIGHTMAPS];
extern qboolean stainmap_active[MAX_LIGHTMAPS];
void R_BlendStainmaps (void);

static void R_DrawTextureChains (model_t *mod, int volume, texture_t *caustic)
{
	int i;
	int j;
	float *v;
	texture_t *t;
	msurface_t *surf;
	int oldLightmapBind = -1;
	qboolean needStains = false;
	msurface_t *lqChain = NULL;

	// set up GL_TEXTURE1 for lightmapping
	glActiveTexture (GL_TEXTURE1);
	glEnable (GL_TEXTURE_2D);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	memset (stainmap_polys, 0, sizeof (stainmap_polys));

	for (i = 0; i < mod->numtextures; i++)
	{
		// this only happens in e2m3
		if (!mod->textures[i]) continue;

		// not on the draw list
		if (!(surf = mod->textures[i]->texturechain[volume])) continue;

		// get the current texture animation
		t = R_TextureAnimation (mod->textures[i]);

		// bind to TMU 0 (make it active first cos GL_TEXTURE1 will have been previously active, either from
		// initialization or a previous run through this loop)
		glActiveTexture (GL_TEXTURE0);
		glBindTexture (GL_TEXTURE_2D, t->gl_texturenum);

		// make TMU 1 active for lightmaps
		glActiveTexture (GL_TEXTURE1);

		// draw all of the surfs
		for (; surf; surf = surf->drawChain)
		{
			// bind the lightmap texture if it hasn't changed from last time
			if (surf->lightmaptexturenum != oldLightmapBind)
			{
				glBindTexture (GL_TEXTURE_2D, lightmap_textures + surf->lightmaptexturenum);
				oldLightmapBind = surf->lightmaptexturenum;
			}

			// accumulate fullbright textures - only if the current texture animation has fullbrights
			if (t->fullbright)
			{
				surf->fbChain = surf->texinfo->texture->fbChain;
				surf->texinfo->texture->fbChain = surf;

				fbChain = true;
			}

			// accumulate caustic surfaces
			if (caustic)
			{
				surf->lqChain = lqChain;
				lqChain = surf;
			}

			// accumulate the surfaces for the current model
			surf->accumSurf = mod->accumSurf;
			mod->accumSurf = surf;

			// if the stainmap for the surface is active, accumulate polys for it
			// only accumulate polys for the surf with the stain on it, rather than polys for every surf
			// that uses this stainmap!!!
			if (stainmap_active[surf->lightmaptexturenum] && surf->everstained)
			{
				surf->polys->smChain = stainmap_polys[surf->lightmaptexturenum];
				stainmap_polys[surf->lightmaptexturenum] = surf->polys;

				needStains = true;
			}

			// draw the surface polygon
			glBegin (GL_POLYGON);

			for (j = 0, v = surf->polys->verts[0]; j < surf->polys->numverts; j++, v += VERTEXSIZE)
			{
				glMultiTexCoord2fv (GL_TEXTURE0, &v[3]);
				glMultiTexCoord2fv (GL_TEXTURE1, &v[5]);
				glVertex3fv (v);
			}

			glEnd ();
		}

		// explicitly NULL the texture chain
		mod->textures[i]->texturechain[volume] = NULL;
	}

	// take down GL_TEXTURE1 (it will already be active coming out of here)
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glDisable (GL_TEXTURE_2D);
	glActiveTexture (GL_TEXTURE0);

	// draw stainmaps as a seperate pass (we can multitexture all of these later on)
	if (needStains) R_BlendStainmaps ();

	// draw caustics as a seperate pass (we can multitexture all of these later on)
	if (lqChain) R_DrawCausticChain (caustic, lqChain);
}


/*
==================
R_DrawVolumeChains

volume merging and some astutely placed Con_Printfs indicates that
no more than 4 drawing volumes (i.e. volumes after merging) are
used in ID1.  I've upped it to 16 to be utterly safe.

Each "volume" for the purposes of drawing is all the surfs in the map with the same
underwater or abovewater type.  All abovewater surfs are in volume 0 always.  The
underwater type depends on the water texture they are under: if there are 2 different
slime textures in a map, this would give 2 slime volumes.  Volumes are used for setting
the correct fog colour and caustic texture *only* at this stage.
==================
*/
static void R_DrawVolumeChains (model_t *mod)
{
	int i;

	// apply all current stains before doing the lightmaps
	// we don't stain BSP brush models (aside from the world model)
	if (mod == cl.worldmodel)
	{
		R_ApplyStains (mod->nodes);
	}
	else if (mod != cl.worldmodel && mod->name[0] == '*')
	{
		R_ApplyStains (mod->nodes + mod->hulls[0].firstclipnode);
	}

	// prebuild all lightmaps for this model
	R_PrebuildLightmaps (mod);

	// optimise caustic warps
	warptime = STMult * realtime;

	// initialize accumulated surfs
	mod->accumSurf = NULL;

	for (i = 0; i < 16; i++)
	{
		// brush models only have one volume set!!!
		if (i >= mod->numVolumes) break;

		// cos all the volumes are moved to the front of the list, once we hit one with
		// a visframe of -2 we can be sure that the rest will have no surfs
		if (mod->volumes[i].visframe == -2) break;

		// the volume has no drawlist for this frame
		if (mod->volumes[i].visframe != r_framecount) continue;

		R_DrawTextureChains (mod, i, mod->volumes[i].caustic);
	}

	// detail textures are drawn as a seperate pass for the entire model owing to a different BlendFunc
	if (r_detailtextures.value) R_DrawDetailChain (mod);

	// fullbrights are drawn for all volumes at the end of each model.  I *could*
	// multitexture them again but I'd be bringing TMUs up and down like a young
	// lady of the night's undergarments.  Plus I want them to be drawn *after* fog
	if (fbChain) R_DrawFullbrightChain (mod);
}


/*
================
R_DrawWaterSurfaces

Each water texture is given a different scaling factor and blend mode
================
*/
void R_DrawWaterChain (msurface_t *surf);

void R_DrawWaterSurfaces (void)
{
	// Water surfs are chained by texture, although the actual water texture from the BSP is not used
	// by MHQuake any more - instead, I use my caustic texture and derive a colour for it from the
	// "fog" colour of the texture.  This prevents break-up in the water warp effect.  The caustic
	// texture consists of 72 frames animating at 72 FPS.  These textures were generated by
	// Caustics Generator v 1.3 (Copyright (c) 2003-08-06 Kjell Andersson (kand@lysator.liu.se) and
	// Mathias Bergvall) available at http://www.lysator.liu.se/~kand/caustics/.  The resulting bitmaps
	// were decreased to 16bpp and placed in the engine itself as resources.  I would have put the
	// generation algorithm in my own source and made them on the fly when the engine loaded but it's
	// C++ and the code is - to be honest - totally all over the place.  Sorry guys, that's just the way
	// it is... it's what happens when you object-orient stuff that doesn't need to be object-oriented.
	int i;
	msurface_t *surf;
	texture_t *t;
	float scale;
	float alpha;
	static float watertime = 0;

	// update the liquid animation, irrespective of whether or not we draw liquid this frame
	watertime += frametime * 72;

	if (watertime >= 72) watertime -= 72;

	// only go through here if we have to
	if (!waterthisframe) return;

	// go back to the world matrix
    glLoadMatrixf (r_world_matrix);

	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthMask (GL_FALSE);

	glActiveTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glTexGeni (GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glTexGeni (GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);

	glEnable (GL_TEXTURE_GEN_S);
	glEnable (GL_TEXTURE_GEN_T);

	// bind the current warp animation
	glBindTexture (GL_TEXTURE_2D, caustictexture + (int) watertime);
	//glBindTexture (GL_TEXTURE_2D, caustictexture);

	for (i = 0; i < cl.worldmodel->numtextures; i++)
	{
		t = cl.worldmodel->textures[i];
		if (!t) continue;
		if (t->name[0] != '*') continue;
		if (!t->lqChain) continue;

		surf = t->lqChain;

		if (surf->flags & SURF_DRAWSLIME)
		{
			// slime
			scale = 128.0;
			alpha = 0.8;
		}
		else if (surf->flags & SURF_DRAWLAVA)
		{
			// lava
			scale = 96.0;
			alpha = 0.7;
		}
		else
		{
			// the wet stuff
			scale = 192.0;
			alpha = 0.6;
		}

		glColor4f (surf->texinfo->texture->uwater_fog[0],
				   surf->texinfo->texture->uwater_fog[1],
				   surf->texinfo->texture->uwater_fog[2],
				   alpha);

		// do water calculations in hardware - I can't do a global setting of this cos the
		// scale can be different for each liquid type
		glMatrixMode (GL_TEXTURE);
		glLoadIdentity ();
		glScalef (1.0 / scale, 1.0 / scale, 1.0);
		glMatrixMode (GL_MODELVIEW);

		R_DrawWaterChain (surf);

		t->lqChain = NULL;
	}

	glColor4f (1, 1, 1, 1);

	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	glDisable (GL_TEXTURE_GEN_S);
	glDisable (GL_TEXTURE_GEN_T);
}



/*
=================
R_DrawBrushModel
=================
*/
static void R_DrawInlineBModel (entity_t *ent, model_t *clmodel)
{
	int i, k;
	mplane_t *pplane;
	float dotp;
	msurface_t *surf;
	mleaf_t *bLeaf;
	vec3_t neworg;

	// calculate dynamic lighting for bmodel
	if (clmodel->firstmodelsurface != 0 && clmodel->name[0] == '*')
	{
		R_PushDlights (clmodel->nodes + clmodel->hulls[0].firstclipnode, true);
	}

	// here we get or set the leaf the bmodel is in
	if (clmodel->name[0] == '*' && waterthisframe && clmodel->modelleaf)
	{
		// see has the bmodel moved
		if (ent->origin[0] != 0 && ent->origin[1] != 0 && ent->origin[2] != 0)
		{
			// move it up a bit
			neworg[0] = clmodel->origin[0] + ent->origin[0];
			neworg[1] = clmodel->origin[1] + ent->origin[1];
			neworg[2] = clmodel->origin[2] + ent->origin[2] + 8;

			bLeaf = Mod_PointInLeaf (neworg, cl.worldmodel);
		}
		else bLeaf = clmodel->modelleaf;
	}
	else
	{
		// just set to the viewleaf
		bLeaf = r_viewleaf;
	}

	// set up the "fake" volume
	bmvol[0].visframe = r_framecount;
	bmvol[1].visframe = -2;
	bmvol[0].volumenum = 0;

	// ammo and health should never be marked underwater
	if (clmodel->name[0] != '*')
		bmvol[0].caustic = NULL;
	else bmvol[0].caustic = bLeaf->caustic;

	// allocate it to the model
	clmodel->volumes = bmvol;
	clmodel->numVolumes = 2;

	// now get it again cos we need it again
	surf = &clmodel->surfaces[clmodel->firstmodelsurface];

	for (i = 0; i < clmodel->nummodelsurfaces; i++, surf++)
	{
		if (surf->flags & (SURF_DRAWSKY | SURF_DRAWTURB)) continue;

		switch (surf->plane->type)
		{
		case PLANE_X:
			dotp = modelorg[0] - surf->plane->dist;
			break;
		case PLANE_Y:
			dotp = modelorg[1] - surf->plane->dist;
			break;
		case PLANE_Z:
			dotp = modelorg[2] - surf->plane->dist;
			break;
		default:
			dotp = DotProduct (modelorg, surf->plane->normal) - surf->plane->dist;
			break;
		}

		if (((surf->flags & SURF_PLANEBACK) && (dotp < -BACKFACE_EPSILON)) || (!(surf->flags & SURF_PLANEBACK) && (dotp > BACKFACE_EPSILON)))
		{
			// if it's a BSP bmodel don't bother updating light
			if (clmodel->name[0] == '*') R_PrebuildSurfaceLightmap (surf);

			// build the surface drawing chain
			surf->drawChain = surf->texinfo->texture->texturechain[0];
			surf->texinfo->texture->texturechain[0] = surf;

			addSurfs++;
		}
	}

	// and draw, blondie you son of a thousand fathers, all bastards like you!!!
	R_DrawVolumeChains (clmodel);
}


void R_RotateForEntity (entity_t *e, qboolean blendLerp);


void R_DrawBrushModel (entity_t *e)
{
	vec3_t mins, maxs;
	int i, j;
	qboolean rotated;
	model_t	*clmodel;

	clmodel = e->model;

	if (clmodel->nummodelsurfaces == 0) return;

	currententity = e;

	if (e->angles[0] || e->angles[1] || e->angles[2])
	{
		rotated = true;

		for (i = 0; i < 3; i++)
		{
			mins[i] = e->origin[i] - clmodel->radius;
			maxs[i] = e->origin[i] + clmodel->radius;
		}
	}
	else
	{
		rotated = false;

		VectorAdd (e->origin, clmodel->mins, mins);
		VectorAdd (e->origin, clmodel->maxs, maxs);
	}

	if (R_CullBox (mins, maxs)) return;

	glColor3f (1, 1, 1);

	VectorSubtract (r_refdef.vieworg, e->origin, modelorg);

	if (rotated)
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (modelorg, temp);
		AngleVectors (e->angles, forward, right, up);
		modelorg[0] = DotProduct (temp, forward);
		modelorg[1] = -DotProduct (temp, right);
		modelorg[2] = DotProduct (temp, up);
	}

	glPushMatrix ();

	// seems i fixed this somewhere else...  do do dooo do do do doooo (looks away)
	//e->angles[0] = -e->angles[0];	// stupid quake bug
	//e->angles[2] = -e->angles[2];	// stupid quake bug

	R_RotateForEntity (e, true);

	// seems i fixed this somewhere else...  do do dooo do do do doooo (looks away)
	//e->angles[0] = -e->angles[0];	// stupid quake bug
	//e->angles[2] = -e->angles[2];	// stupid quake bug

	R_DrawInlineBModel (e, clmodel);

	glPopMatrix ();
}


/*
=============================================================

	WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
int	addLeafs;

static void R_RecursiveWorldNode (mnode_t *node)
{
	int i;
	int c;
	int side;
	mplane_t *plane;
	msurface_t *surf;
	msurface_t **mark;
	mleaf_t *pleaf;
	double dist;
	int sidebit;

	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;

	// if a leaf node, mark the surfaces as visible
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *) node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				// mark the surfaces in this leaf for drawing
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		// (only used for static entities - torches and suchlike)
		if (pleaf->efrags) R_StoreEfrags (&pleaf->efrags);

		return;
	}

	// node is just a decision point, so go down the apropriate sides
	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dist = modelorg[0] - plane->dist;
		break;
	case PLANE_Y:
		dist = modelorg[1] - plane->dist;
		break;
	case PLANE_Z:
		dist = modelorg[2] - plane->dist;
		break;
	default:
		dist = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dist >= 0)
	{
		// front side
		side = 0;
		sidebit = 0;
	}
	else
	{
		// back side
		side = 1;
		sidebit = SURF_PLANEBACK;
	}

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	if (node->numsurfaces)
	{
		for (c = node->numsurfaces, surf = cl.worldmodel->surfaces + node->firstsurface; c; c--, surf++)
		{
			// not marked as visible
			if (surf->visframe != r_framecount) continue;

			// node backfaces
			if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;

			// check for surf backfacing
			switch (surf->plane->type)
			{
			case PLANE_X:
				dist = modelorg[0] - surf->plane->dist;
				break;
			case PLANE_Y:
				dist = modelorg[1] - surf->plane->dist;
				break;
			case PLANE_Z:
				dist = modelorg[2] - surf->plane->dist;
				break;
			default:
				dist = DotProduct (modelorg, surf->plane->normal) - surf->plane->dist;
				break;
			}

			// surf backfaces
			if (!(((surf->flags & SURF_PLANEBACK) && (dist < -BACKFACE_EPSILON)) || 
				(!(surf->flags & SURF_PLANEBACK) && (dist > BACKFACE_EPSILON)))) continue;

			if (surf->flags & SURF_DRAWSKY)
			{
				skythisframe = true;
			}
			else if (surf->flags & SURF_DRAWTELE)
			{
			}
			else if (surf->flags & SURF_DRAWTURB)
			{
				// store whether water is visible this frame to reduce Mod_PointInLeaf checking
				// for brush models, as well as to determine whether or not we need to draw it!!!
				waterthisframe = true;

				surf->lqChain = surf->texinfo->texture->lqChain;
				surf->texinfo->texture->lqChain = surf;

				addSurfs++;
			}
			else
			{
				// now that we have the lighting info we can update it here.
				R_PrebuildSurfaceLightmap (surf);

				// build a seperate texture chain for each volume
				surf->drawChain = surf->texinfo->texture->texturechain[surf->volume];
				surf->texinfo->texture->texturechain[surf->volume] = surf;

				// mark the volume as visible
				cl.worldmodel->volumes[surf->volume].visframe = r_framecount;

				addSurfs++;
			}
		}
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}


/*
=============
R_DrawWorld
=============
*/
float lighttime = 0;

void R_DrawWorld (void)
{
	entity_t	ent;

	// set up our current tmus for the render - only of value for when i want to
	// test how MHQuake looks on cards with less TMUs than i have
	if (gl_activetmus.value > gl_textureunits) Cvar_SetValue ("gl_activetmus", gl_textureunits);
	if (gl_activetmus.value < 2) Cvar_SetValue ("gl_activetmus", 2);

	gl_currenttmus = (int) gl_activetmus.value;

	memset (&ent, 0, sizeof (ent));
	ent.model = cl.worldmodel;

	// modelorg here is the player's origin in the world
	VectorCopy (r_refdef.vieworg, modelorg);

	currententity = &ent;

	glColor3f (1, 1, 1);

	// only update lighting 10 times per second.  next to no visible difference.
	// this has to be done before anything is either drawn or marked otherwise lighting on
	// the world will go out of sync with lighting on bmodels
	if (lighttime >= LIGHT_UPLOAD) lighttime = 0;
	lighttime += frametime;

	// default values
	skythisframe = false;
	waterthisframe = false;

	// mark dynamic lights for the world
	R_PushDlights (cl.worldmodel->nodes, false);

	// marks surfaces and updates lighting
	R_RecursiveWorldNode (cl.worldmodel->nodes);

	// and draw, Texan!!!
	// (The best jokes work on more than one level.  Here I'm not only punning on the word
	// "draw" by putting it in a completely irrelevant Western context, but also bringing in
	// the fact that ID are based in Texas.  I'll get me coat...)
	R_DrawVolumeChains (cl.worldmodel);

	// we make sure here that we only draw sky when required
	if (skythisframe) R_DrawSkyChain ();
}


/*
===============
AddVisibleLeaf

Adds the specified leaf to the PVS
This is tolerant of adding the same leaf twice...
===============
*/
static void AddVisibleLeaf (mleaf_t *VLeaf)
{
	byte	*vis;
	mnode_t	*node;
	int		i;
	byte	solid[4096];

	// already been added this frame
	if (VLeaf->waterVisFrame == r_visframecount) return;

	if (r_novis.value)
	{
		vis = solid;

		// set it to 255 so the bits always match
		memset (solid, 0xff, (cl.worldmodel->numleafs + 7) >> 3);
	}
	else vis = Mod_LeafPVS (VLeaf, cl.worldmodel);

	addLeafs = 0;

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			addLeafs++;

			// mark the visible nodes
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				// already added
				if (node->visframe == r_visframecount) break;

				node->visframe = r_visframecount;
				node = node->parent;
			}
			while (node);
		}
	}

	// mark it as having been checked
	VLeaf->waterVisFrame = r_visframecount;
}


/*
===============
R_MarkLeaves
===============
*/
extern cvar_t gl_lockpvs;

void R_MarkLeaves (void)
{
	//addSurfs = 0;

	// still in the same leaf
	if (r_oldviewleaf == r_viewleaf) return;

	if (gl_lockpvs.value) return;

	// this only increments when the viewleaf changes
	r_visframecount++;

	// add the viewleaf (we can add PVS for any leaf we want with this function)
	AddVisibleLeaf (r_viewleaf);

	// is this necessary any more???
	if (r_oldviewleaf)
	{
		if (r_oldviewleaf->contents != r_viewleaf->contents)
		{
			// fix water transitions.  If we have translucent water it's probably not
			// much of a bother as the vis for both leafs is likely to be very similar.
			// The only time this could cause bother is if we teleport from above water
			// to underwater or vice versa, although the hit is probably not that bad.
			AddVisibleLeaf (r_oldviewleaf);
		}
	}

	r_oldviewleaf = r_viewleaf;
}



int bound (int lower, int num, int upper)
{
	if (num < lower) return lower;
	if (num > upper) return upper;
	return num;
}


void GL_BuildPolygonFromSurface (model_t *mod, msurface_t *surf)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	float		*vec;
	float		s, t;
	glpoly_t	*poly;

	// reconstruct the polygon
	pedges = mod->edges;
	lnumverts = surf->numedges;

	// draw texture
	poly = Hunk_Alloc (sizeof (glpoly_t) + (lnumverts - 4) * VERTEXSIZE * sizeof (float));
	poly->next = surf->polys;
	poly->flags = surf->flags;
	surf->polys = poly;
	poly->numverts = lnumverts;

	for (i = 0; i < lnumverts; i++)
	{
		lindex = mod->surfedges[surf->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = mod->vertexes[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = mod->vertexes[r_pedge->v[1]].position;
		}

		// poly world verts
		poly->verts[i][0] = vec[0];
		poly->verts[i][1] = vec[1];
		poly->verts[i][2] = vec[2];

		// texture s and t
		poly->verts[i][3] = (DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]) / surf->texinfo->texture->width;
		poly->verts[i][4] = (DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]) / surf->texinfo->texture->height;

		// allocate 2 new verts for a 64 * 64 version of the texture, and set their data
		// set up as if the texture was 64 * 64 to keep caustic warps consistent
		// these verts are also used for detail so they are massaged a little to include the
		// necessary scaling for the detail tex so that we're not multiplying or modifying the
		// texture matrix, both of which put extra load on the system
		poly->verts[i][7] = (DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]) / 16.0;
		poly->verts[i][8] = (DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]) / 16.0;

		// optimise caustic warps
		poly->verts[i][9] = poly->verts[i][7] * STMult * 4;
		poly->verts[i][10] = poly->verts[i][8] * STMult * 4;
	}
}


void R_InitSkyChain (void);

void GL_RSurfInit (void)
{
	int i;
	float j;
	float inc = (2.0 * M_PI) / (float) SINTABLE_SIZE;

	// shut down all texture units except TMU 0 (just make sure...)
	for (i = 1; i < gl_textureunits; i++)
	{
		glActiveTexture (GL_TEXTURE0 + i);
		glDisable (GL_TEXTURE_2D);
	}

	// TMU 0 should always be up
	glActiveTexture (GL_TEXTURE0);
	glEnable (GL_TEXTURE_2D);

	// set up the sin table for warps
	for (i = 0, j = 0; i < SINTABLE_SIZE; i++, j += inc)
	{
		sintable[i] = sin (j) * 4.0;
		sintablex2[i] = sin (j) * 8.0;
	}

	STBitMask = SINTABLE_SIZE - 1;

	STMult = (float) SINTABLE_SIZE / (6.0 * M_PI);
	STMultx2 = (float) SINTABLE_SIZE / (2.0 * M_PI);

	// set up the sky display list
	R_InitSkyChain ();

	// register the detail textures cvar
	Cvar_RegisterVariable (&r_detailtextures);
}

