/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_warp.c -- sky and water polygons

#include "quakedef.h"
#include "glext.h"

extern	model_t	*loadmodel;

int		solidskytexture = 0;
int		alphaskytexture = 0;


msurface_t	*warpface;

extern cvar_t gl_subdivide_size;


// these are incremented in R_RenderView cos we want them always updated even if
// the sky is not being currently drawn
float rotateBack = 0;
float rotateFore = 0;


void BoundPoly (int numverts, float *verts, vec3_t mins, vec3_t maxs)
{
	int		i, j;
	float	*v;

	mins[0] = mins[1] = mins[2] = 9999;
	maxs[0] = maxs[1] = maxs[2] = -9999;

	v = verts;

	for (i=0 ; i<numverts ; i++)
	{
		for (j=0 ; j<3 ; j++, v++)
		{
			if (*v < mins[j])
				mins[j] = *v;
			if (*v > maxs[j])
				maxs[j] = *v;
		}
	}
}


void SubdividePolygon (int numverts, float *verts)
{
	int		i, j, k;
	vec3_t	mins, maxs;
	float	m;
	float	*v;
	vec3_t	front[64], back[64];
	int		f, b;
	float	dist[64];
	float	frac;
	glpoly_t	*poly;
	float	s, t;
	vec3_t	total;
	float	total_s, total_t;

	if (numverts > 60)
		Sys_Error ("numverts = %i", numverts);

	BoundPoly (numverts, verts, mins, maxs);

	for (i=0 ; i<3 ; i++)
	{
		m = (mins[i] + maxs[i]) * 0.5;
		m = gl_subdivide_size.value * floor (m/gl_subdivide_size.value + 0.5);
		if (maxs[i] - m < 8)
			continue;
		if (m - mins[i] < 8)
			continue;

		// cut it
		v = verts + i;
		for (j=0 ; j<numverts ; j++, v+= 3)
			dist[j] = *v - m;

		// wrap cases
		dist[j] = dist[0];
		v-=i;
		VectorCopy (verts, v);

		f = b = 0;
		v = verts;
		for (j=0 ; j<numverts ; j++, v+= 3)
		{
			if (dist[j] >= 0)
			{
				VectorCopy (v, front[f]);
				f++;
			}
			if (dist[j] <= 0)
			{
				VectorCopy (v, back[b]);
				b++;
			}
			if (dist[j] == 0 || dist[j+1] == 0)
				continue;
			if ( (dist[j] > 0) != (dist[j+1] > 0) )
			{
				// clip point
				frac = dist[j] / (dist[j] - dist[j+1]);
				for (k=0 ; k<3 ; k++)
					front[f][k] = back[b][k] = v[k] + frac*(v[3+k] - v[k]);
				f++;
				b++;
			}
		}

		SubdividePolygon (f, front[0]);
		SubdividePolygon (b, back[0]);
		return;
	}

	// add a point in the center to help keep warp valid
	poly = Hunk_Alloc (sizeof(glpoly_t) + ((numverts-4)+2) * VERTEXSIZE*sizeof(float));
	poly->next = warpface->warppolys;
	warpface->warppolys = poly;
	poly->numverts = numverts+2;
	total[0] = total[1] = total[2] = 0;
	total_s = 0;
	total_t = 0;

	for (i=0 ; i<numverts ; i++, verts+= 3)
	{
		VectorCopy (verts, poly->verts[i+1]);
		s = DotProduct (verts, warpface->texinfo->vecs[0]);
		t = DotProduct (verts, warpface->texinfo->vecs[1]);

		total_s += s;
		total_t += t;
		VectorAdd (total, verts, total);

		poly->verts[i + 1][3] = s;
		poly->verts[i + 1][4] = t;

		// MH - store these precalculated to cut down on runtime calcs
		poly->verts[i + 1][7] = s * 0.125 * STMultx2;
		poly->verts[i + 1][8] = t * 0.125 * STMultx2;
	}

	VectorScale (total, (1.0/numverts), poly->verts[0]);

	poly->verts[0][3] = total_s / numverts;
	poly->verts[0][4] = total_t / numverts;

	// copy first vertex to last
	memcpy (poly->verts[i+1], poly->verts[1], sizeof(poly->verts[0]));
}


/*
================
GL_SubdivideSurface

Breaks a polygon up along axial 64 unit
boundaries so that turbulent and sky warps
can be done reasonably.
================
*/
void GL_SubdivideSurface (msurface_t *fa)
{
	vec3_t		verts[64];
	int			numverts;
	int			i;
	int			lindex;
	float		*vec;
	texture_t	*t;
	glpoly_t	*p;

	warpface = fa;

	// convert edges back to a normal polygon
	numverts = 0;

	for (i=0 ; i<fa->numedges ; i++)
	{
		lindex = loadmodel->surfedges[fa->firstedge + i];

		if (lindex > 0)
			vec = loadmodel->vertexes[loadmodel->edges[lindex].v[0]].position;
		else
			vec = loadmodel->vertexes[loadmodel->edges[-lindex].v[1]].position;

		VectorCopy (vec, verts[numverts]);

		numverts++;
	}

	SubdividePolygon (numverts, verts[0]);

	// if it's water, also make polys equal to warppolys
	// in order to catch anything i've missed
	if (warpface->flags & SURF_DRAWTURB)
		warpface->polys = warpface->warppolys;

	// count the number of polys
	for (p = fa->warppolys, fa->numPolys = 0; p; p = p->next)
	{
		fa->numPolys++;
	}
}



/*
======================================================================================

									WATER WARP

======================================================================================
*/


/*
=============
R_DrawWaterChain

Does a water warp on the pre-fragmented glpoly_t chain
=============
*/
extern float warptime;

void WarpVectors (float *Vec1, float *Vec2)
{
	float temp1[4] = {Vec1[0], Vec1[1], Vec1[2], Vec1[3]};
	float temp2[4] = {Vec2[0], Vec2[1], Vec2[2], Vec2[3]};
	int i;

	// warp the coords
	for (i = 0; i < 4; i++)
	{
		if (!Vec1[i] && !Vec2[i]) continue;

		Vec1[i]	= temp1[i] + sin (realtime + temp2[i]);
		Vec2[i]	= temp2[i] + sin (realtime + temp1[i]);
	}
}


void R_DrawWaterChain (msurface_t *surf)
{
	int i;
	float *v;
	glpoly_t *p;
	float sPlane[4];
	float tPlane[4];

	for (; surf; surf = surf->lqChain)
	{
		sPlane[1] = sPlane[2] = sPlane[3] = tPlane[0] = tPlane[1] = tPlane[3] = 0.0;
		sPlane[0] = tPlane[2] = 1.0;

		// shift the generated texcoords according to the surfaces plane
		switch (surf->plane->type)
		{
		case PLANE_Y:
			// vertical in line with the world grid
			tPlane[1] = 1.0;
		case PLANE_X:
			// vertical in line with the world grid
			sPlane[1] = 1.0;
			break;

		case PLANE_ANYX:
			// vertical at an angle to the world grid
			// (_ANYY doesn't seem to be used... hmmm, must check...)
			// keep the defaults, nothing else set
			break;

		case PLANE_Z:
			// flat
		default:
			// same as flat
			tPlane[1] = 1.0;
			break;
		}

		//WarpVectors (sPlane, tPlane);

		glTexGenfv (GL_S, GL_OBJECT_PLANE, sPlane);
		glTexGenfv (GL_T, GL_OBJECT_PLANE, tPlane);

		for (p = surf->warppolys; p; p = p->next)
		{
			glBegin (GL_TRIANGLE_FAN);

			for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
			{
				glVertex3fv (v);
			}

			glEnd ();
		}
	}
}


/*
==============
R_DrawSphere

Draw a sphere!!!  The verts and texcoords are precalculated for extra efficiency.  The sphere is put into
a display list to reduce overhead even further.
==============
*/
float skytexes[440];
float skyverts[660];

int spherelist = -1;

void R_DrawSphere (void)
{
	if (spherelist == -1)
	{
		int i;
		int j;

		int vertspos = 0;
		int texespos = 0;

		// build the sphere display list
		spherelist = glGenLists (1);

		glNewList (spherelist, GL_COMPILE);

		for (i = 0; i < 10; i++)
		{
			glBegin (GL_TRIANGLE_STRIP);

			for (j = 0; j <= 10; j++)
			{
				glTexCoord2fv (&skytexes[texespos]);
				glVertex3fv (&skyverts[vertspos]);

				texespos += 2;
				vertspos += 3;

				glTexCoord2fv (&skytexes[texespos]);
				glVertex3fv (&skyverts[vertspos]);

				texespos += 2;
				vertspos += 3;
			}

			glEnd ();
		}

		glEndList ();
	}

	glCallList (spherelist);
}


/*
==============
R_InitSkyChain

Initialize the sky chain arrays
==============
*/
void R_InitSkyChain (void)
{
	float drho = 0.3141592653589;
	float dtheta = 0.6283185307180;

	float ds = 0.1;
	float dt = 0.1;

	float t = 1.0f;	
	float s = 0.0f;

	int i;
	int j;

	int vertspos = 0;
	int texespos = 0;

	for (i = 0; i < 10; i++)
	{
		float rho = (float) i * drho;
		float srho = (float) (sin (rho));
		float crho = (float) (cos (rho));
		float srhodrho = (float) (sin (rho + drho));
		float crhodrho = (float) (cos (rho + drho));

		s = 0.0f;

		for (j = 0; j <= 10; j++) 
		{
			float theta = (j == 10) ? 0.0f : j * dtheta;
			float stheta = (float) (-sin( theta));
			float ctheta = (float) (cos (theta));

			skytexes[texespos++] = s * 2;
			skytexes[texespos++] = t * 2;

			skyverts[vertspos++] = stheta * srho * 4096.0;
			skyverts[vertspos++] = ctheta * srho * 4096.0;
			skyverts[vertspos++] = crho * 4096.0;

			skytexes[texespos++] = s * 2;
			skytexes[texespos++] = (t - dt) * 2;

			skyverts[vertspos++] = stheta * srhodrho * 4096.0;
			skyverts[vertspos++] = ctheta * srhodrho * 4096.0;
			skyverts[vertspos++] = crhodrho * 4096.0;

			s += ds;
		}

		t -= dt;
	}
}


/*
=================
R_DrawSkyChain

Hot damn Jethro, this sho' is purty!!!
=================
*/
void R_DrawSkyChain (void)
{
	// turn on fogging.  fog looks good on the skies - it gives them a more 
	// "airy" far-away look, and has the knock-on effect of preventing the 
	// old "texture distortion at the poles" problem.
	glFogi (GL_FOG_MODE, GL_LINEAR);
	glFogfv (GL_FOG_COLOR, sky_fog);
	glFogf (GL_FOG_START, 5);
	glFogf (GL_FOG_END, 3000);
	glEnable (GL_FOG);

	// background
	// ==========
	// go to a new matrix
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere
	glScalef (1, 1, 0.5);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateBack, 0, 0, 1);

	// solid sky texture
	glBindTexture (GL_TEXTURE_2D, solidskytexture);

	// draw the sphere
	R_DrawSphere ();

	// restore the previous matrix
	glPopMatrix ();

	// foreground
	// ==========
	// go to a new matrix
	glPushMatrix ();

	// center it on the players position
	glTranslatef (r_origin[0], r_origin[1], r_origin[2]);

	// flatten the sphere and shrink it a little
	glScalef (0.9, 0.9, 0.45);

	// orient it so that the poles are unobtrusive
	glRotatef (-90, 1, 0, 0);

	// make it not always at right angles to the player
	glRotatef (-22, 0 ,1, 0);

	// rotate it around the poles
	glRotatef (-rotateFore, 0, 0, 1);

	// blend mode
	glEnable (GL_BLEND);

	// alpha sky texture
	glBindTexture (GL_TEXTURE_2D, alphaskytexture);

	// draw the sphere
	R_DrawSphere ();

	// back to normal mode
	glDisable (GL_BLEND);

	// restore the previous matrix
	glPopMatrix ();

	// turn off fog
	glDisable (GL_FOG);
}


/*
=============
R_InitSky

A sky texture is 256*128, with the right side being a masked overlay
==============
*/
extern int max_size;

#define SKY_TILE_X 2
#define SKY_TILE_Y 1

void R_InitSky (texture_t *mt)
{
	int			i, j, p;
	byte		*src;
	unsigned	transpix;
	int			r, g, b;
	unsigned	*rgba;
	unsigned	topalpha;
	int			div;

	unsigned	*trans;
	unsigned	*newTrans;

	// initialize our pointers
	trans = malloc (128 * 128 * 4);
	newTrans = malloc ((SKY_TILE_X * 128) * (SKY_TILE_Y * 128) * 4);

	// -----------------
	// solid sky texture
	// -----------------
	src = (byte *)mt + mt->offsets[0];

	// make an average value for the back to avoid
	// a fringe on the top level
	r = g = b = 0;

	for (i=0 ; i<128 ; i++)
	{
		for (j=0 ; j<128 ; j++)
		{
			p = src[i*256 + j + 128];
			rgba = &d_8to24table[p];
			trans[(i*128) + j] = *rgba;

			r += ((byte *)rgba)[0];
			g += ((byte *)rgba)[1];
			b += ((byte *)rgba)[2];
		}
	}

	((byte *)&transpix)[0] = r/(128*128);
	((byte *)&transpix)[1] = g/(128*128);
	((byte *)&transpix)[2] = b/(128*128);
	((byte *)&transpix)[3] = 0;

	if (!solidskytexture) solidskytexture = texture_extension_number++;

	GL_TileTexture (trans, newTrans, 128, 128, SKY_TILE_X, SKY_TILE_Y);

	glBindTexture (GL_TEXTURE_2D, solidskytexture);

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, (SKY_TILE_X * 128), (SKY_TILE_Y * 128), 0, GL_RGBA, GL_UNSIGNED_BYTE, newTrans);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// -----------------
	// alpha sky texture
	// -----------------
	// get another average cos the bottom layer average can be too dark
	div = r = g = b = 0;

	for (i=0 ; i<128 ; i++)
	{
		for (j=0 ; j<128 ; j++)
		{
			p = src[i*256 + j];

			if (p == 0)
				topalpha = transpix;
			else
			{
				rgba = &d_8to24table[p];

				g += ((byte *)rgba)[1];

				r += ((byte *)rgba)[0];
				b += ((byte *)rgba)[2];

				div++;

				topalpha = d_8to24table[p];
			}

			((byte *)&topalpha)[3] = ((byte *)&topalpha)[3] / 2;

			trans[(i*128) + j] = topalpha;
		}
	}

	if (!alphaskytexture) alphaskytexture = texture_extension_number++;

	GL_TileTexture (trans, newTrans, 128, 128, SKY_TILE_X, SKY_TILE_Y);

	glBindTexture (GL_TEXTURE_2D, alphaskytexture);

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, (SKY_TILE_X * 128), (SKY_TILE_Y * 128), 0, GL_RGBA, GL_UNSIGNED_BYTE, newTrans);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// get fog colours for the sky - we halve these
	sky_fog[0] = ((float)r / (float)div) / 512.0;
	sky_fog[1] = ((float)g / (float)div) / 512.0;
	sky_fog[2] = ((float)b / (float)div) / 512.0;
	sky_fog[3] = 0.1;

	// free the used memory
	free (trans);
	free (newTrans);

	// recalc the sphere display list
	spherelist = -1;
}
