/*
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

/*
============================================================================================================================
SOFTWARE TRANSFORMS ENGINE BY MH

This is viable because (a) Quake doesn't particularly tax the CPU anyway, and (b) I want to save the GPU muscle for where
it's really needed.

This is NOT a full software implementation of OpenGL's transforms engine.  It's a cut down version that only implements
the stuff I need.  For a full version, I recommend you look to either Mesa or SGI.  Both are excellent sources for
learning how the innards of OpenGL work.
============================================================================================================================
*/

#include "gl_state.h"

// 32 deep modelview matrix stack to conform with OpenGL spec
static float ModelViewMatrix[32][16];

static int StackDepth = 0;


void swglPushMatrix (void)
{
	// oops.
	if (StackDepth == 31) Sys_Error ("Modelview stack push beyond 32");

	// go to a new matrix
	StackDepth++;

	// copy in the previous
	ModelViewMatrix[StackDepth][0] = ModelViewMatrix[StackDepth - 1][0];
	ModelViewMatrix[StackDepth][1] = ModelViewMatrix[StackDepth - 1][1];
	ModelViewMatrix[StackDepth][2] = ModelViewMatrix[StackDepth - 1][2];
	ModelViewMatrix[StackDepth][3] = ModelViewMatrix[StackDepth - 1][3];

	ModelViewMatrix[StackDepth][4] = ModelViewMatrix[StackDepth - 1][4];
	ModelViewMatrix[StackDepth][5] = ModelViewMatrix[StackDepth - 1][5];
	ModelViewMatrix[StackDepth][6] = ModelViewMatrix[StackDepth - 1][6];
	ModelViewMatrix[StackDepth][7] = ModelViewMatrix[StackDepth - 1][7];

	ModelViewMatrix[StackDepth][8] = ModelViewMatrix[StackDepth - 1][8];
	ModelViewMatrix[StackDepth][9] = ModelViewMatrix[StackDepth - 1][9];
	ModelViewMatrix[StackDepth][10] = ModelViewMatrix[StackDepth - 1][10];
	ModelViewMatrix[StackDepth][11] = ModelViewMatrix[StackDepth - 1][11];

	ModelViewMatrix[StackDepth][12] = ModelViewMatrix[StackDepth - 1][12];
	ModelViewMatrix[StackDepth][13] = ModelViewMatrix[StackDepth - 1][13];
	ModelViewMatrix[StackDepth][14] = ModelViewMatrix[StackDepth - 1][14];
	ModelViewMatrix[StackDepth][15] = ModelViewMatrix[StackDepth - 1][15];
}


void swglPopMatrix (void)
{
	// oops.
	if (StackDepth == 0) Sys_Error ("Modelview stack pop below 0");

	// go back
	StackDepth--;
}


static void UtilLoadIdentity (float *UtilMatrix)
{
	UtilMatrix[0] = 1;
	UtilMatrix[1] = 0;
	UtilMatrix[2] = 0;
	UtilMatrix[3] = 0;

	UtilMatrix[4] = 0;
	UtilMatrix[5] = 1;
	UtilMatrix[6] = 0;
	UtilMatrix[7] = 0;

	UtilMatrix[8] = 0;
	UtilMatrix[9] = 0;
	UtilMatrix[10] = 1;
	UtilMatrix[11] = 0;

	UtilMatrix[12] = 0;
	UtilMatrix[13] = 0;
	UtilMatrix[14] = 0;
	UtilMatrix[15] = 1;
}


void swglMultMatrixf (float *matrix)
{
	int i;
	float MultMatrix[16];
	float *Original = ModelViewMatrix[StackDepth];

	// multiple
	MultMatrix[0] = Original[0] * matrix[0] + Original[1] * matrix[4] + Original[2] * matrix[8] + Original[3] * matrix[12];
	MultMatrix[1] = Original[0] * matrix[1] + Original[1] * matrix[5] + Original[2] * matrix[9] + Original[3] * matrix[13];
	MultMatrix[2] = Original[0] * matrix[2] + Original[1] * matrix[6] + Original[2] * matrix[10] + Original[3] * matrix[14];
	MultMatrix[3] = Original[0] * matrix[3] + Original[1] * matrix[7] + Original[2] * matrix[11] + Original[3] * matrix[15];

	MultMatrix[4] = Original[4] * matrix[0] + Original[5] * matrix[4] + Original[6] * matrix[8] + Original[7] * matrix[12];
	MultMatrix[5] = Original[4] * matrix[1] + Original[5] * matrix[5] + Original[6] * matrix[9] + Original[7] * matrix[13];
	MultMatrix[6] = Original[4] * matrix[2] + Original[5] * matrix[6] + Original[6] * matrix[10] + Original[7] * matrix[14];
	MultMatrix[7] = Original[4] * matrix[3] + Original[5] * matrix[7] + Original[6] * matrix[11] + Original[7] * matrix[15];

	MultMatrix[8] = Original[8] * matrix[0] + Original[9] * matrix[4] + Original[10] * matrix[8] + Original[11] * matrix[12];
	MultMatrix[9] = Original[8] * matrix[1] + Original[9] * matrix[5] + Original[10] * matrix[9] + Original[11] * matrix[13];
	MultMatrix[10] = Original[8] * matrix[2] + Original[9] * matrix[6] + Original[10] * matrix[10] + Original[11] * matrix[14];
	MultMatrix[11] = Original[8] * matrix[3] + Original[9] * matrix[7] + Original[10] * matrix[11] + Original[11] * matrix[15];

	MultMatrix[12] = Original[12] * matrix[0] + Original[13] * matrix[4] + Original[14] * matrix[8] + Original[15] * matrix[12];
	MultMatrix[13] = Original[12] * matrix[1] + Original[13] * matrix[5] + Original[14] * matrix[9] + Original[15] * matrix[13];
	MultMatrix[14] = Original[12] * matrix[2] + Original[13] * matrix[6] + Original[14] * matrix[10] + Original[15] * matrix[14];
	MultMatrix[15] = Original[12] * matrix[3] + Original[13] * matrix[7] + Original[14] * matrix[11] + Original[15] * matrix[15];

	// copy back
	ModelViewMatrix[StackDepth][0] = MultMatrix[0];
	ModelViewMatrix[StackDepth][1] = MultMatrix[1];
	ModelViewMatrix[StackDepth][2] = MultMatrix[2];
	ModelViewMatrix[StackDepth][3] = MultMatrix[3];

	ModelViewMatrix[StackDepth][4] = MultMatrix[4];
	ModelViewMatrix[StackDepth][5] = MultMatrix[5];
	ModelViewMatrix[StackDepth][6] = MultMatrix[6];
	ModelViewMatrix[StackDepth][7] = MultMatrix[7];

	ModelViewMatrix[StackDepth][8] = MultMatrix[8];
	ModelViewMatrix[StackDepth][9] = MultMatrix[9];
	ModelViewMatrix[StackDepth][10] = MultMatrix[10];
	ModelViewMatrix[StackDepth][11] = MultMatrix[11];

	ModelViewMatrix[StackDepth][12] = MultMatrix[12];
	ModelViewMatrix[StackDepth][13] = MultMatrix[13];
	ModelViewMatrix[StackDepth][14] = MultMatrix[14];
	ModelViewMatrix[StackDepth][15] = MultMatrix[15];
}


void swglLoadIdentity (void)
{
	UtilLoadIdentity (ModelViewMatrix[StackDepth]);
}


void swglTranslatef (float x, float y, float z)
{
	float TranslateMatrix[16] = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, x, y, z, 1};

	swglMultMatrixf (TranslateMatrix);
}


void swglRotatef (float ang, float x, float y, float z)
{
	float RotateMatrix[16] = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
	GLfloat xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c;
	float s = sin (ang * DIV180MULTPI);
	float c = cos (ang * DIV180MULTPI);
	float mag;

	// normalize the input vector
	mag = sqrt (x * x + y * y + z * z);
	x /= mag;
	y /= mag;
	z /= mag;

	// get some stuff
	xx = x * x;
	yy = y * y;
	zz = z * z;
	xy = x * y;
	yz = y * z;
	zx = z * x;
	xs = x * s;
	ys = y * s;
	zs = z * s;
	one_c = 1.0F - c;

	// compose the rotation matrix
	RotateMatrix[0] = (one_c * xx) + c;
	RotateMatrix[1] = (one_c * xy) - zs;
	RotateMatrix[2] = (one_c * zx) + ys;

	RotateMatrix[4] = (one_c * xy) + zs;
	RotateMatrix[5] = (one_c * yy) + c;
	RotateMatrix[6] = (one_c * yz) - xs;

	RotateMatrix[8] = (one_c * zx) - ys;
	RotateMatrix[9] = (one_c * yz) + xs;
	RotateMatrix[10] = (one_c * zz) + c;

	// multiply
	swglMultMatrixf (RotateMatrix);
}


void swglLoadMatrixf (float *matrix)
{
	ModelViewMatrix[StackDepth][0] = matrix[0];
	ModelViewMatrix[StackDepth][1] = matrix[1];
	ModelViewMatrix[StackDepth][2] = matrix[2];
	ModelViewMatrix[StackDepth][3] = matrix[3];

	ModelViewMatrix[StackDepth][4] = matrix[4];
	ModelViewMatrix[StackDepth][5] = matrix[5];
	ModelViewMatrix[StackDepth][6] = matrix[6];
	ModelViewMatrix[StackDepth][7] = matrix[7];

	ModelViewMatrix[StackDepth][8] = matrix[8];
	ModelViewMatrix[StackDepth][9] = matrix[9];
	ModelViewMatrix[StackDepth][10] = matrix[10];
	ModelViewMatrix[StackDepth][11] = matrix[11];

	ModelViewMatrix[StackDepth][12] = matrix[12];
	ModelViewMatrix[StackDepth][13] = matrix[13];
	ModelViewMatrix[StackDepth][14] = matrix[14];
	ModelViewMatrix[StackDepth][15] = matrix[15];
}


void swglTransformVerts (float *v, float *vertsout)
{
	float *matrix = ModelViewMatrix[StackDepth];

	// full matrix transformation
	vertsout[0] = matrix[0] * v[0] + matrix[4] * v[1] + matrix[8] * v[2] + matrix[12];
	vertsout[1] = matrix[1] * v[0] + matrix[5] * v[1] + matrix[9] * v[2] + matrix[13];
	vertsout[2] = matrix[2] * v[0] + matrix[6] * v[1] + matrix[10] * v[2] + matrix[14];
}

