
#include <windows.h>
#include <gdiplus.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "gl14.h"

#include "warningdisable.h"


using namespace Gdiplus;


ULONG_PTR gdiplusToken;

BOOL GDIUp = FALSE;

// imports
extern "C" char com_gamedir[];
extern "C" int gl_max_size;

// clsids for the mime types we want to encode
CLSID JPGEncoder;
CLSID PNGEncoder;
CLSID BMPEncoder;

int GetEncoderClsid (const WCHAR *format, CLSID *pClsid)
{
	UINT num = 0;          // number of image encoders
	UINT size = 0;         // size of the image encoder array in bytes

	ImageCodecInfo *pImageCodecInfo = NULL;

	GetImageEncodersSize (&num, &size);

	// failure
	if (!size) return -1;

	pImageCodecInfo = (ImageCodecInfo *) (malloc (size));

	// failure
	if (!pImageCodecInfo) return -1;

	GetImageEncoders (num, size, pImageCodecInfo);

	for (UINT j = 0; j < num; ++j)
	{
		if (!wcscmp (pImageCodecInfo[j].MimeType, format))
		{
			// success
			*pClsid = pImageCodecInfo[j].Clsid;

			free (pImageCodecInfo);
			return j;
		}    
	}

	// failure
	free (pImageCodecInfo);
	return -1;
}


void CharToWChar (char *charin, wchar_t *charout)
{
	// get the length
	int Len = strlen (charin);

	// clear to zero
	memset (charout, 0, (Len + 1) * sizeof (wchar_t));

	// attempt to convert
	MultiByteToWideChar (CP_ACP, 0, charin, Len, charout, Len);
}


void GDI_InitGDI (void)
{
	if (GDIUp) return;

	GdiplusStartupInput gdiplusStartupInput;

	GdiplusStartup (&gdiplusToken, &gdiplusStartupInput, NULL);

	// get our mime type encoders
	GetEncoderClsid (L"image/png", &PNGEncoder);
	GetEncoderClsid (L"image/bmp", &BMPEncoder);
	GetEncoderClsid (L"image/jpeg", &JPGEncoder);

	GDIUp = TRUE;
}


extern "C" void GDI_KillGDI (void)
{
	if (!GDIUp) return;

	GdiplusShutdown (gdiplusToken);
	GDIUp = FALSE;
}


void GDI_UploadBitmapObject (Bitmap *TheBMP)
{
	// get current height and width
	int BitmapWidth = TheBMP->GetWidth ();
	int BitmapHeight = TheBMP->GetHeight ();

	// power of 2 dimensions
	int ScaledWidth;
	int ScaledHeight;

	// get power of 2 dimensions
	for (ScaledWidth = 1; ScaledWidth < BitmapWidth; ScaledWidth *= 2);
	for (ScaledHeight = 1; ScaledHeight < BitmapHeight; ScaledHeight *= 2);

	// clamp to max texture size
	if (ScaledWidth > gl_max_size) ScaledWidth = gl_max_size;
	if (ScaledHeight > gl_max_size) ScaledHeight = gl_max_size;

	BitmapData TheBitmapData;
	Rect TheRect (0, 0, ScaledWidth, ScaledHeight);

	if (ScaledWidth == BitmapWidth && ScaledHeight == BitmapHeight)
	{
		// despite the name, this is actually RGBA with bitshifts and masks
		TheBMP->LockBits (&TheRect, ImageLockModeRead, PixelFormat32bppARGB, &TheBitmapData);

		glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, ScaledWidth, ScaledHeight, 0, GL_BGRA, GL_UNSIGNED_BYTE, TheBitmapData.Scan0);

		TheBMP->UnlockBits (&TheBitmapData);
	}
	else
	{
		// create a new bitmap and some graphics on it
		Bitmap NewBMP (ScaledWidth, ScaledHeight);
		Graphics TheGraphics (&NewBMP);

		// use highest quality interpolation
		TheGraphics.SetInterpolationMode (InterpolationModeHighQualityBilinear);

		// draw the original to it
		TheGraphics.DrawImage (TheBMP, 0, 0, ScaledWidth, ScaledHeight);

		// lock the bits for reading
		NewBMP.LockBits (&TheRect, ImageLockModeRead, PixelFormat32bppARGB, &TheBitmapData);

		// upload as a texture
		glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, ScaledWidth, ScaledHeight, 0, GL_BGRA, GL_UNSIGNED_BYTE, TheBitmapData.Scan0);

		// unlock when done
		NewBMP.UnlockBits (&TheBitmapData);
	}
}


/*
=============
GDI_LoadTGA

TGA Loading stuff

Somebody at Microsoft deserves a good slap for not supporting one of the most widely used and
versatile image formats in their otherwise incredibly nifty graphics library
=============
*/
typedef struct TGAHeader_s
{
	unsigned char id_length;
	unsigned char colormap_type;
	unsigned char image_type;
	unsigned short colormap_index;
	unsigned short colormap_length;
	unsigned char colormap_size;
	unsigned short x_origin;
	unsigned short y_origin;
	unsigned short width;
	unsigned short height;
	unsigned char pixel_size;
	unsigned char attributes;
} TGAHeader_t;


int TGAfgetLittleShort (FILE *f)
{
	unsigned char b1 = fgetc (f);
	unsigned char b2 = fgetc (f);

	return (short) (b1 + b2 * 256);
}


BOOL GDI_LoadTGA (FILE *f)
{
	TGAHeader_t targa_header;
	int columns;
	int rows;
	int numPixels;
	unsigned char *pixbuf;
	int row, column;
	int flipper;
	unsigned char *targa_data;

	// read the header
	targa_header.id_length = fgetc (f);
	targa_header.colormap_type = fgetc (f);
	targa_header.image_type = fgetc (f);

	targa_header.colormap_index = TGAfgetLittleShort (f);
	targa_header.colormap_length = TGAfgetLittleShort (f);
	targa_header.colormap_size = fgetc (f);
	targa_header.x_origin = TGAfgetLittleShort (f);
	targa_header.y_origin = TGAfgetLittleShort (f);
	targa_header.width = TGAfgetLittleShort (f);
	targa_header.height = TGAfgetLittleShort (f);
	targa_header.pixel_size = fgetc (f);
	targa_header.attributes = fgetc (f);

	// they call him flipper, flipper, faster than lightning, no-one you see, is smarter than he.
	flipper = !((targa_header.attributes & 0x20) >> 5);

	// moved up from down below to avoid daft warnings
	Rect TheRect (0, 0, targa_header.width, targa_header.height);

	// validate
	if (targa_header.image_type != 2 && targa_header.image_type != 10) return FALSE;
	if (targa_header.colormap_type != 0 || (targa_header.pixel_size != 32 && targa_header.pixel_size != 24)) return FALSE;

	columns = targa_header.width;
	rows = targa_header.height;
	numPixels = columns * rows;

	// the image will presumably be OK now...
	Bitmap *TheBitmap = new Bitmap (targa_header.width, targa_header.height);

	// create the objects we need to lock the data
	// TheRect moved up...
	BitmapData TheBitmapData;

	// lock the data so we can access it directly
	// the actual pixel format doesn't matter so long as it has 4 components
	TheBitmap->LockBits (&TheRect, ImageLockModeWrite, PixelFormat32bppARGB, &TheBitmapData);

	targa_data = (unsigned char *) TheBitmapData.Scan0;

	// skip TARGA image comment
	if (targa_header.id_length != 0) fseek (f, targa_header.id_length, SEEK_CUR);

	if (targa_header.image_type == 2)
	{
		// Uncompressed, RGB images
		for (row = rows - 1; row >= 0; row--)
		{
			if (flipper)
				pixbuf = targa_data + row * columns * 4;
			else pixbuf = targa_data + ((row - 1) - row) * columns * 4;

			for (column = 0; column < columns; column++)
			{
				unsigned char blue = fgetc (f);
				unsigned char green = fgetc (f);
				unsigned char red = fgetc (f);
				unsigned char alphabyte;

				*pixbuf++ = blue;
				*pixbuf++ = green;
				*pixbuf++ = red;

				switch (targa_header.pixel_size)
				{
				case 24:
					*pixbuf++ = 255;
					break;

				case 32:
					alphabyte = fgetc (f);
					*pixbuf++ = alphabyte;
					break;
				}
			}
		}
	}
	else if (targa_header.image_type == 10)
	{
		// Runlength encoded RGB images
		unsigned char red;
		unsigned char green;
		unsigned char blue;
		unsigned char alphabyte;
		unsigned char packetHeader;
		unsigned char packetSize;
		unsigned char j;

		for (row = rows - 1; row >= 0; row--)
		{
			if (flipper)
				pixbuf = targa_data + row * columns * 4;
			else pixbuf = targa_data + ((row - 1) - row) * columns * 4;

			for (column = 0; column < columns;)
			{
				packetHeader = fgetc (f);
				packetSize = 1 + (packetHeader & 0x7f);

				if (packetHeader & 0x80)
				{
					// run-length packet
					blue = fgetc (f);
					green = fgetc (f);
					red = fgetc (f);

					switch (targa_header.pixel_size)
					{
					case 24:
						alphabyte = 255;
						break;

					case 32:
					default:
						alphabyte = fgetc (f);
						break;
					}
	
					for (j = 0; j < packetSize; j++)
					{
						*pixbuf++ = blue;
						*pixbuf++ = green;
						*pixbuf++ = red;
						*pixbuf++ = alphabyte;

						column++;

						if (column == columns)
						{
							// run spans across rows
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							if (flipper)
								pixbuf = targa_data + row * columns * 4;
							else pixbuf = targa_data + ((row - 1) - row) * columns * 4;
						}
					}
				}
				else
				{
					// non run-length packet
					for (j = 0; j < packetSize; j++)
					{
						blue = fgetc (f);
						green = fgetc (f);
						red = fgetc (f);

						*pixbuf++ = blue;
						*pixbuf++ = green;
						*pixbuf++ = red;

						switch (targa_header.pixel_size)
						{
						case 24:
							*pixbuf++ = 255;
							break;

						case 32:
							alphabyte = fgetc (f);
							*pixbuf++ = alphabyte;
							break;
						}

						column++;

						if (column == columns)
						{
							// pixel packet run spans across rows
							column = 0;

							if (row > 0)
								row--;
							else goto breakOut;

							if (flipper)
								pixbuf = targa_data + row * columns * 4;
							else pixbuf = targa_data + ((row - 1) - row) * columns * 4;
						}
					}
				}
			}

			breakOut:;
		}
	}

	// upload
	fclose (f);

	TheBitmap->UnlockBits (&TheBitmapData);

	// load it as a texture
	GDI_UploadBitmapObject (TheBitmap);

	// release our memory and return
	delete TheBitmap;
	return TRUE;
}


/*
===============
GDI_LoadPCX

PCX Loading

I didn't really bother testing this...
===============
*/
typedef struct pcx_s
{
	char manufacturer;
	char version;
	char encoding;
	char bits_per_pixel;

	unsigned short xmin;
	unsigned short ymin;
	unsigned short xmax;
	unsigned short ymax;

	unsigned short hres;
	unsigned short vres;

	unsigned char palette[48];

	char reserved;
	char color_planes;

	unsigned short bytes_per_line;
	unsigned short palette_type;

	char filler[58];

	unsigned int data;			// unbounded
} pcx_t;


BOOL GDI_LoadPCX (FILE *f)
{
	pcx_t *pcx, pcxbuf;
	unsigned char palette[768];
	unsigned char *pix;
	int x, y;
	int dataByte, runLength;
	int count;
	unsigned char *pcx_data;

	// parse the PCX file
	fread (&pcxbuf, 1, sizeof (pcxbuf), f);

	pcx = &pcxbuf;

	if (pcx->manufacturer != 0x0a || pcx->version != 5 || pcx->encoding != 1 || pcx->bits_per_pixel != 8 || pcx->xmax >= 1024 || pcx->ymax >= 1024)
		return FALSE;

	// seek to palette
	fseek (f, -768, SEEK_END);
	fread (palette, 1, 768, f);

	fseek (f, sizeof (pcxbuf) - 4, SEEK_SET);

	count = (pcx->xmax + 1) * (pcx->ymax + 1);

	// the image will presumably be OK now...
	Bitmap *TheBitmap = new Bitmap (pcx->xmax + 1, pcx->ymax + 1);

	// create the objects we need to lock the data
	BitmapData TheBitmapData;
	Rect TheRect (0, 0, pcx->xmax + 1, pcx->ymax + 1);

	// lock the data so we can access it directly
	// the actual pixel format doesn't matter so long as it has 4 components
	TheBitmap->LockBits (&TheRect, ImageLockModeWrite, PixelFormat32bppARGB, &TheBitmapData);

	pcx_data = (unsigned char *) TheBitmapData.Scan0;

	for (y = 0; y <= pcx->ymax; y++)
	{
		pix = pcx_data + 4 * y * (pcx->xmax + 1);

		for (x = 0; x <= pcx->xmax;)
		{
			dataByte = fgetc (f);

			if ((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				dataByte = fgetc (f);
			}
			else runLength = 1;

			while (runLength-- > 0)
			{
				pix[2] = palette[dataByte * 3];
				pix[1] = palette[dataByte * 3 + 1];
				pix[0] = palette[dataByte * 3 + 2];

				// PCX files never have alpha
				pix[3] = 255;
				pix += 4;
				x++;
			}
		}
	}

	// upload
	fclose (f);

	TheBitmap->UnlockBits (&TheBitmapData);

	// load it as a texture
	GDI_UploadBitmapObject (TheBitmap);

	// release our memory and return
	delete TheBitmap;
	return TRUE;
}


/*
==================
GDI_HackCharset

Evil hack alert.

The menu slider characters need the rightmost item and the slider indicator swapped around because otherwise the
indicator will alpha-blend into the slider body at the left and right sides.
==================
*/
BOOL CharSetHack = FALSE;

void GDI_HackCharset (char *identifier, Bitmap *TheBMP)
{
	if (CharSetHack) return;

	if (identifier[0] != 'w') return;
	if (identifier[1] != 'a') return;
	if (identifier[2] != 'd') return;
	if (identifier[3] != '/') return;
	if (identifier[4] != 'c') return;
	if (identifier[5] != 'h') return;
	if (identifier[6] != 'a') return;
	if (identifier[7] != 'r') return;
	if (identifier[8] != 's') return;
	if (identifier[9] != 'e') return;
	if (identifier[10] != 't') return;

	// get width and height
	int BMPWidth = TheBMP->GetWidth ();
	int BMPHeight = TheBMP->GetHeight ();

	// get width and height of an individual character
	int CharWidth = BMPWidth / 16;
	int CharHeight = BMPHeight / 16;

	Color Swap1;
	Color Swap2;

	for (int x = 0; x < CharWidth; x++)
	{
		for (int y = 0; y < CharHeight; y++)
		{
			// read the pixels in
			TheBMP->GetPixel (CharWidth * 2 + x, CharHeight * 8 + y, &Swap1);
			TheBMP->GetPixel (CharWidth * 3 + x, CharHeight * 8 + y, &Swap2);

			// set them back
			TheBMP->SetPixel (CharWidth * 3 + x, CharHeight * 8 + y, Swap1);
			TheBMP->SetPixel (CharWidth * 2 + x, CharHeight * 8 + y, Swap2);
		}
	}

	// it's hacked now
	CharSetHack = TRUE;
}


/*
=================
LoadGDIBitmap

Loads a generic GDI+ bitmap object that supports the BMP, GIF, JPEG, PNG, TIFF, Exif, WMF and EMF data types.
=================
*/
extern "C" unsigned char LoadGDIBitmap (char *searchpath, char *texturename, char *extension)
{
	GDI_InitGDI ();

	// build the full name
	char fullname[128];
	FILE *TheFile;

	sprintf (fullname, "%s\\%s.%s", searchpath, texturename, extension);

	// attempt to open it directly.  this saves on creating all of the GDI+ objects before we open it
	TheFile = fopen (fullname, "rb");

	// file doesn't exist
	if (!TheFile) return 0;

	// TGA and PCX files need to be handled here as GDI+ doesn't support these formats (BAD BAD Bill!).
	// we might as well do them here as anywhere else since we already have the file open.
	// just send them straight through the functions and let header checking sort them out!
	if (GDI_LoadTGA (TheFile)) return 1;
	if (GDI_LoadPCX (TheFile)) return 1;

	// file exists, close it and carry on
	fclose (TheFile);

	// step 1 - GDI+ functions use the Unicode data types for strings, so we need to convert
	wchar_t UnicodeName[128];
	CharToWChar (fullname, UnicodeName);

	// create a bitmap object
	Bitmap *TheBitmap = new Bitmap (UnicodeName);

	// did it work?
	// this needs tested for as well as direct opening as the file format may be bad.
	if (TheBitmap->GetLastStatus () != Ok)
	{
		delete TheBitmap;
		return 0;
	}

	// check for a bitmap charset hack
	GDI_HackCharset (texturename, TheBitmap);

	// upload the object
	GDI_UploadBitmapObject (TheBitmap);

	// release our memory
	delete TheBitmap;

	// now it's loaded OK
	return 1;
}


extern "C" void LoadGDIData (char *identifier, unsigned char *Data, unsigned int *Palette, int Width, int Height)
{
	int i;

	GDI_InitGDI ();

	Bitmap TheBMP (Width, Height);
	BitmapData TheBitmapData;
	Rect TheRect (0, 0, Width, Height);

	TheBMP.LockBits (&TheRect, ImageLockModeWrite, PixelFormat32bppARGB, &TheBitmapData);

	unsigned int *Scaled = (unsigned int *) TheBitmapData.Scan0;

	if (Palette)
	{
		for (i = 0; i < (Width * Height); i++)
		{
			Scaled[i] = Palette[Data[i]];
		}
	}
	else
	{
		// convert pointer to 32 bpp
		unsigned int *iData = (unsigned int *) Data;

		// copy straight in
		for (i = 0; i < (Width * Height); i++)
		{
			Scaled[i] = iData[i];
		}
	}

	TheBMP.UnlockBits (&TheBitmapData);

	// check for a bitmap charset hack
	GDI_HackCharset (identifier, &TheBMP);

	GDI_UploadBitmapObject (&TheBMP);
}


extern "C" void GDI_MakeFont (int FontTexnum)
{
	GDI_InitGDI ();

#define BMPSIZE		512
#define CELLSIZE	(BMPSIZE / 16)
#define FONTSIZE	((CELLSIZE) / 4 * 3)
#define USCOREOFS	((FONTSIZE) / 4)
	// create a bitmap and some graphics to hold our font
	Bitmap TheBMP (BMPSIZE, BMPSIZE);
	Graphics TheGraphics (&TheBMP);

	// strings for writing - this doesn't need to be 16 - 2 is sufficient, but hey! :D
	wchar_t WConChars[16];

	// clear to black
	TheGraphics.Clear (Color (0, 0, 0));

	// starting positions
	int StartX = 0;
	int StartY = 0;

	// set up string writing parameters
	FontFamily  TheFontFamily (L"Courier New");
	Font TheFont (&TheFontFamily, FONTSIZE, FontStyleBold, UnitPixel);

	SolidBrush WhiteBrush (Color (194, 194, 194));
	SolidBrush BrownBrush (Color (219, 125, 60));
	SolidBrush YallaBrush (Color (179, 140, 45));
	SolidBrush *CurrentBrush;

	TheGraphics.SetTextRenderingHint (TextRenderingHintAntiAliasGridFit);

	// colour components we're going to use
	Color TheColor;
	int myAlpha;
	int myR;
	int myG;
	int myB;

	// always
	WConChars[1] = 0;

	// build the charset
	for (int i = 0; i < 256; i++)
	{
		// build our string to write
		int Charnum = (i < 128) ? i : i - 128;

		CurrentBrush = &WhiteBrush;

		if (i >= 128) CurrentBrush = &BrownBrush;
		if (Charnum < 32) CurrentBrush = &YallaBrush;

		// hack the characters because this uses unicode rather than ascii...
		if (Charnum == 0 || Charnum == 5 || Charnum == 14 || Charnum == 15 || Charnum == 28)
		{
			WConChars[0] = 0x00b7;

			if (i < 128) CurrentBrush = &WhiteBrush;
		}
		else if (Charnum == 11)
		{
			WConChars[0] = 0x2593;

			if (i < 128)
				CurrentBrush = &WhiteBrush;
			else CurrentBrush = &BrownBrush;
		}
		else if (Charnum == 13)
		{
			WConChars[0] = 0x25ba;
			CurrentBrush = &BrownBrush;
		}
		else if (Charnum < 16)
			WConChars[0] = L' ';
		else if (Charnum == 16)
			WConChars[0] = L'[';
		else if (Charnum == 17)
			WConChars[0] = L']';
		else if (Charnum < 28)
			WConChars[0] = L'0' + Charnum - 18;
		else if (Charnum < 32)
			WConChars[0] = L' ';
		else if (Charnum == 127)
			WConChars[0] = 0x2190;
		else WConChars[0] = Charnum;

		// hack for the underscore in the courier charset
		TheGraphics.DrawString (WConChars, -1, &TheFont, PointF (StartX, StartY - ((WConChars[0] == L'_') ? USCOREOFS : 0)), CurrentBrush);

		// accumulate the scales
		StartX += CELLSIZE;

		if (StartX >= BMPSIZE)
		{
			// jump to next line
			StartX = 0;
			StartY += CELLSIZE;
		}
	}

	// now do the alpha transformation
	for (int x = 0; x < BMPSIZE; x++)
	{
		for (int y = 0; y < BMPSIZE; y++)
		{
			TheBMP.GetPixel (x, y, &TheColor);

			// colour components
			myAlpha = 0;
			myR = TheColor.GetR ();
			myG = TheColor.GetG ();
			myB = TheColor.GetB ();

			if (!myR && !myG && !myB)
			{
				// standard full alpha
				TheBMP.SetPixel (x, y, Color (0, 0, 0, 0));
			}
			else
			{
				// take the highest of the 3 as the alpha value
				if (myR > myAlpha) myAlpha = myR;
				if (myG > myAlpha) myAlpha = myG;
				if (myB > myAlpha) myAlpha = myB;

				// double and clamp it
				myAlpha *= 2;
				if (myAlpha > 255) myAlpha = 255;

				// set back
				TheBMP.SetPixel (x, y, Color (myAlpha, myR, myG, myB));
			}
		}
	}

	// upload it as a texture
	GDI_UploadBitmapObject (&TheBMP);

	// test - save it out
	// TheBMP.Save (L"newconchars.png", &PNGEncoder, NULL);
}


extern "C" void GDI_Screenshot (char *ssname, unsigned char *data, int w, int h)
{
	GDI_InitGDI ();

	// create a bitmap object with the correct resolution
	Bitmap TheBMP (w, h);

	// data and rectangle
	BitmapData TheBitmapData;
	Rect TheRect (0, 0, w, h);

	// lock for writing
	TheBMP.LockBits (&TheRect, ImageLockModeWrite, PixelFormat24bppRGB, &TheBitmapData);

	// get a byte pointer to the data and write it in
	unsigned char *BMPData = (unsigned char *) TheBitmapData.Scan0;

	for (int i = 0; i < w * h * 3; i++)
		BMPData[i] = data[i];

	// unlock
	TheBMP.UnlockBits (&TheBitmapData);

	// because opengl works from bottom-left, we need to invert the bitmap
	TheBMP.RotateFlip (RotateNoneFlipY);

	// set jpeg quality
	EncoderParameters encoderParameters;
	ULONG quality;

	encoderParameters.Count = 1;
	encoderParameters.Parameter[0].Guid = EncoderQuality;
	encoderParameters.Parameter[0].Type = EncoderParameterValueTypeLong;
	encoderParameters.Parameter[0].NumberOfValues = 1;

	// Save the image as a JPEG with quality level 90 for an acceptable size/quality compromise
	quality = 90;
	encoderParameters.Parameter[0].Value = &quality;

	// convert name to unicode
	wchar_t WSSName[128];
	CharToWChar (ssname, WSSName);

	// save as a jpeg
	TheBMP.Save (WSSName, &JPGEncoder, &encoderParameters);
}


