/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_mesh.c: triangle model functions

#include "quakedef.h"

/*
=================================================================

ALIAS MODEL DISPLAY LIST GENERATION

=================================================================
*/

model_t		*aliasmodel;
aliashdr_t	*paliashdr;

qboolean	used[8192];

// the command list holds counts and s/t values that are valid for
// every frame
int		commands[8192];
int		numcommands;

// all frames will have their vertexes rearranged and expanded
// so they are in the order expected by the command list
int		vertexorder[8192];
int		numorder;

int		allverts, alltris;

int		stripverts[128];
int		striptris[128];
int		stripcount;

/*
================
StripLength
================
*/
int	StripLength (int starttri, int startv)
{
	int			m1, m2;
	int			j;
	mtriangle_t	*last, *check;
	int			k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv)%3];
	stripverts[1] = last->vertindex[(startv+1)%3];
	stripverts[2] = last->vertindex[(startv+2)%3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv+2)%3];
	m2 = last->vertindex[(startv+1)%3];

	// look for a matching triangle
nexttri:
	for (j=starttri+1, check=&triangles[starttri+1] ; j<pheader->numtris ; j++, check++)
	{
		if (check->facesfront != last->facesfront)
			continue;
		for (k=0 ; k<3 ; k++)
		{
			if (check->vertindex[k] != m1)
				continue;
			if (check->vertindex[ (k+1)%3 ] != m2)
				continue;

			// this is the next part of the fan

			// if we can't use this triangle, this tristrip is done
			if (used[j])
				goto done;

			// the new edge
			if (stripcount & 1)
				m2 = check->vertindex[ (k+2)%3 ];
			else
				m1 = check->vertindex[ (k+2)%3 ];

			stripverts[stripcount+2] = check->vertindex[ (k+2)%3 ];
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:

	// clear the temp used flags
	for (j=starttri+1 ; j<pheader->numtris ; j++)
		if (used[j] == 2)
			used[j] = 0;

	return stripcount;
}

/*
===========
FanLength
===========
*/
int	FanLength (int starttri, int startv)
{
	int		m1, m2;
	int		j;
	mtriangle_t	*last, *check;
	int		k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv)%3];
	stripverts[1] = last->vertindex[(startv+1)%3];
	stripverts[2] = last->vertindex[(startv+2)%3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv+0)%3];
	m2 = last->vertindex[(startv+2)%3];


	// look for a matching triangle
nexttri:
	for (j=starttri+1, check=&triangles[starttri+1] ; j<pheader->numtris ; j++, check++)
	{
		if (check->facesfront != last->facesfront)
			continue;
		for (k=0 ; k<3 ; k++)
		{
			if (check->vertindex[k] != m1)
				continue;
			if (check->vertindex[ (k+1)%3 ] != m2)
				continue;

			// this is the next part of the fan

			// if we can't use this triangle, this tristrip is done
			if (used[j])
				goto done;

			// the new edge
			m2 = check->vertindex[ (k+2)%3 ];

			stripverts[stripcount+2] = m2;
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:

	// clear the temp used flags
	for (j=starttri+1 ; j<pheader->numtris ; j++)
		if (used[j] == 2)
			used[j] = 0;

	return stripcount;
}


/*
================
BuildTris

Generate a list of trifans or strips
for the model, which holds for all frames
================
*/
void BuildTris (void)
{
	int		i, j, k;
	int		startv;
	float	s, t;
	int		len, bestlen, besttype;
	int		bestverts[1024];
	int		besttris[1024];
	int		type;

	//
	// build tristrips
	//
	numorder = 0;
	numcommands = 0;
	memset (used, 0, sizeof(used));
	for (i=0 ; i<pheader->numtris ; i++)
	{
		// pick an unused triangle and start the trifan
		if (used[i])
			continue;

		bestlen = 0;
		besttype = 0;

		for (type = 0 ; type < 2 ; type++)
//	type = 1;
		{
			for (startv =0 ; startv < 3 ; startv++)
			{
				if (type == 1)
					len = StripLength (i, startv);
				else
					len = FanLength (i, startv);
				if (len > bestlen)
				{
					besttype = type;
					bestlen = len;
					for (j=0 ; j<bestlen+2 ; j++)
						bestverts[j] = stripverts[j];
					for (j=0 ; j<bestlen ; j++)
						besttris[j] = striptris[j];
				}
			}
		}

		// mark the tris on the best strip as used
		for (j=0 ; j<bestlen ; j++)
			used[besttris[j]] = 1;

		if (besttype == 1)
			commands[numcommands++] = (bestlen+2);
		else
			commands[numcommands++] = -(bestlen+2);

		for (j=0 ; j<bestlen+2 ; j++)
		{
			// emit a vertex into the reorder buffer
			k = bestverts[j];
			vertexorder[numorder++] = k;

			// emit s/t coords into the commands stream
			s = stverts[k].s;
			t = stverts[k].t;
			if (!triangles[besttris[0]].facesfront && stverts[k].onseam)
				s += pheader->skinwidth / 2;	// on back side
			s = (s + 0.5) / pheader->skinwidth;
			t = (t + 0.5) / pheader->skinheight;

			*(float *)&commands[numcommands++] = s;
			*(float *)&commands[numcommands++] = t;
		}
	}

	commands[numcommands++] = 0;		// end of list marker

	Con_DPrintf ("%3i tri %3i vert %3i cmd\n", pheader->numtris, numorder, numcommands);

	allverts += numorder;
	alltris += pheader->numtris;
}


/*
================
GL_MakeAliasModelDisplayLists
================
*/
void GL_MakeAliasModelDisplayLists (model_t *m, aliashdr_t *hdr)
{
	int		i, j;
	int			*cmds;
	drawvertx_t	*verts;
	int count;
	int		*order;

	aliasmodel = m;
	paliashdr = hdr;

	// build every time
	// requires more load time but avoids undying ugliness in the longer term
	BuildTris ();

	// save the data out
	paliashdr->poseverts = numorder;

	cmds = Hunk_Alloc (numcommands * 4);
	paliashdr->commands = (byte *)cmds - (byte *)paliashdr;
	memcpy (cmds, commands, numcommands * 4);

	verts = Hunk_Alloc (paliashdr->numposes * paliashdr->poseverts * sizeof (drawvertx_t));
	paliashdr->posedata = (byte *) verts - (byte *) paliashdr;

	for (i = 0; i < paliashdr->numposes; i++)
	{
		for (j = 0; j < numorder; j++)
		{
			// prescale the verts at load time.  this requires moving over to floats for storing verts, which
			// gives a * 4 memory overhead but cuts down on a LOT of run-time calculations
			// i've tried converting back to bytes but STRANGE things happen...
			verts->v[0] = ((float) poseverts[i][vertexorder[j]].v[0]) * paliashdr->scale[0] + paliashdr->scale_origin[0];
			verts->v[1] = ((float) poseverts[i][vertexorder[j]].v[1]) * paliashdr->scale[1] + paliashdr->scale_origin[1];
			verts->v[2] = ((float) poseverts[i][vertexorder[j]].v[2]) * paliashdr->scale[2] + paliashdr->scale_origin[2];

			verts->lightnormalindex = poseverts[i][vertexorder[j]].lightnormalindex;
			verts->nodraw = false;

			verts++;
		}
	}

	// here's a nasty hack for you!
	// code for elimination of muzzleflashes on viewmodels.
	// is it a view model???
	if ((m->name[6] == 'v' || m->name[6] == 'V') && m->name[7] == '_' && hdr->numframes > 2)
	{
		// get pointers to the verts
		drawvertx_t *vertsf0 = (drawvertx_t *) ((byte *) hdr + hdr->posedata);
		drawvertx_t *vertsf1 = (drawvertx_t *) ((byte *) hdr + hdr->posedata);

		// setthese pointers to the 0th and 1st frames
		vertsf0 += hdr->frames[0].firstpose * hdr->poseverts;
		vertsf1 += hdr->frames[1].firstpose * hdr->poseverts;

		// now go through them and compare.  we expect that (a) the animation is sensible and there's no major
		// difference between the 2 frames to be expected, and (b) any verts that do exhibit a major difference
		// can be assumed to belong to the muzzleflash
		for (j = 0; j < numorder; j++)
		{
			// get difference in front to back movement
			float vdiff = vertsf1->v[0] - vertsf0->v[0];

			// if it's above a certain treshold, assume a muzzleflash and mark for nodraw
			// 10 is the approx lowest range of visible front to back in a view model, so that seems reasonable to work with
			if (vdiff > 10)
				vertsf0->nodraw = true;
			else vertsf0->nodraw = false;

			// next set of verts
			vertsf0++;
			vertsf1++;
		}
	}

	// now we mark every other vert in the model as per the instructions in the first frame
	for (i = 1; i < paliashdr->numframes; i++)
	{
		// get pointers to the verts
		drawvertx_t *vertsf0 = (drawvertx_t *) ((byte *) hdr + hdr->posedata);
		drawvertx_t *vertsfi = (drawvertx_t *) ((byte *) hdr + hdr->posedata);

		// set these pointers to the 0th and i'th frames
		vertsf0 += hdr->frames[0].firstpose * hdr->poseverts;
		vertsfi += hdr->frames[i].firstpose * hdr->poseverts;

		for (j = 0; j < numorder; j++)
		{
			// just copy it across
			vertsfi->nodraw = vertsf0->nodraw;

			if (vertsf0->nodraw)
			{
				vertsfi->v[0] = vertsf0->v[0];
				vertsfi->v[1] = vertsf0->v[1];
				vertsfi->v[2] = vertsf0->v[2];
			}

			// next set of verts
			vertsf0++;
			vertsfi++;
		}
	}

	Con_DPrintf ("Meshed %s (%i cmds", m->name, numcommands);

	j = 0;

	for (order = commands; ;)
	{
		// get the vertex count and primitive type
		count = *order++;

		// done
		if (!count) break;

		// set primitive type
		if (count < 0) count = -count;

		j += count;

		do
		{
			order += 2;
		} while (--count);
	}

	Con_DPrintf ("  %i verts)\n", j);

	// allocate more verts
	if (j > maxverts)
	{
		maxverts = j;
		if (maxverts < 4) maxverts = 4;

		if (surfverts) free (surfverts);

		surfverts = (glvertex_t *) malloc (maxverts * sizeof (glvertex_t));
	}
}

