/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_surf.c: surface-related refresh code

#include "quakedef.h"

// lighting imports
void R_ModifySurfaceLightmap (msurface_t *surf);
void R_SurfaceLightmapTexCoords (msurface_t *surf, float *verts, float *vec);
void R_BlendLightmap (int i);
qboolean R_CullBox (vec3_t mins, vec3_t maxs);
void R_StoreEfrags (efrag_t **ppefrag);
void R_DrawSkyChain (qboolean skydraw);
void R_LightPoint (vec3_t p, vec3_t colour);
void R_PushDlights (mnode_t *headnode);
void R_PushDlightsNOVIS (mnode_t *headnode);

// misc textures
extern gltexture_t *blacktexture;

// surface vertex array
glvertex_t *surfverts = NULL;

// cache of vertexes for models
glvertex_t *vertcache = NULL;

// water caustics textures
extern gltexture_t *watercaustics[];

// global passflags
int globalpassflags = 0;

// framecount for lighting
int r_lightframecount = 0;

// sintable for warps
#define SINTABLE_SIZE 1024
float sintable[SINTABLE_SIZE];
int STBitMask;
float STMult;

extern gltexture_t Atten1D;
extern gltexture_t Atten2D;


/*
===============
R_TextureAnimation

Returns the proper texture for a given time and base texture
===============
*/
texture_t *R_TextureAnimation (entity_t *ent, texture_t *base)
{
	int		reletive;
	int		count;
	texture_t *original = base;

	if (!ent) return base;

	if (ent->frame)
		if (base->alternate_anims)
			base = base->alternate_anims;
	
	if (!base->anim_total)
		return base;

	reletive = (int) (cl.time * 10) % base->anim_total;

	count = 0;	

	while (base->anim_min > reletive || base->anim_max <= reletive)
	{
		base = base->anim_next;

		if (!base) return original;
		if (++count > 100) return original;
	}

	return base;
}


/*
=============================================================

	BRUSH MODELS

=============================================================
*/


extern	int		solidskytexture;
extern	int		alphaskytexture;
extern	float	speedscale;		// for top sky and bottom sky


// global vertex array flags
int VAFlags;


/*
========================
R_ConfigureVertexArrays

configures vertex arrays for the specified render.  note that it's possible to do a conflicting specification here.
if this occurs, the order of precedence is as per the if statements below.
========================
*/
void R_ConfigureVertexArrays (int flags)
{
	// vertex
	if ((flags & VA_VERTEX) && !(VAFlags & VA_VERTEX))
	{
		// bring up vertex pointer
		glEnableClientState (GL_VERTEX_ARRAY);
		glVertexPointer (3, GL_FLOAT, sizeof (glvertex_t), surfverts->v);
	}
	else if (!(flags & VA_VERTEX) && (VAFlags & VA_VERTEX))
	{
		// take down vertex pointer
		glDisableClientState (GL_VERTEX_ARRAY);
	}

	// diffuse
	if ((flags & VA_TEXTURE0_ST) && !(VAFlags & VA_TEXTURE0_ST))
	{
		// bring up diffuse texture array
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->st);
	}
	else if (!(flags & VA_TEXTURE0_ST) && (VAFlags & VA_TEXTURE0_ST))
	{
		// take down diffuse array
		glClientActiveTexture (GL_TEXTURE0);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}

	// lightmap
	if ((flags & VA_TEXTURE1_LM) && !(VAFlags & VA_TEXTURE1_LM))
	{
		// bring up lightmap texture array
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->lm);
	}
	else if (!(flags & VA_TEXTURE1_LM) && (VAFlags & VA_TEXTURE1_LM))
	{
		// take down diffuse array
		glClientActiveTexture (GL_TEXTURE1);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}

	// water caustics
	if ((flags & VA_TEXTURE0_WC) && !(VAFlags & VA_TEXTURE0_WC))
	{
		// bring up water caustics texture array
		glClientActiveTexture (GL_TEXTURE0);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->wc);
	}
	else if (!(flags & VA_TEXTURE0_WC) && (VAFlags & VA_TEXTURE0_WC))
	{
		// take down water caustics array
		glClientActiveTexture (GL_TEXTURE0);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}

	// luma in tmu1
	if ((flags & VA_TEXTURE1_ST) && !(VAFlags & VA_TEXTURE1_ST))
	{
		// bring up diffuse texture array
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->st);
	}
	else if (!(flags & VA_TEXTURE1_ST) && (VAFlags & VA_TEXTURE1_ST))
	{
		// take down diffuse array
		glClientActiveTexture (GL_TEXTURE1);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	}

	// colour array
	if ((flags & VA_COLOUR) && !(VAFlags & VA_COLOUR))
	{
		// bring up colour array
		glEnableClientState (GL_COLOR_ARRAY);
		glColorPointer (4, GL_FLOAT, sizeof (glvertex_t), surfverts->rgba);
	}
	else if (!(flags & VA_COLOUR) && (VAFlags & VA_COLOUR))
	{
		// take down colour array
		glDisableClientState (GL_COLOR_ARRAY);
	}

	// set the new flags
	VAFlags = flags;
}


/*
=============
EmitWaterPolys

Does a water warp on the pre-fragmented glpoly_t chain
these are pretransformed by the world matrix so we don't need to transform them by hand
=============
*/
void EmitWaterPolys (msurface_t *surf)
{
	glpoly_t	*p;
	int			i;
	float		sadd, tadd;
	glvertex_t *verts;

	for (p = surf->polys; p; p = p->next)
	{
		// initialize each time because we can have multiple polys and each must start at 0
		verts = surfverts;

		for (i = 0; i < p->numverts; i++, verts++)
		{
			// calculate the amount to add for the warp
			sadd = sintable[(int) ((p->verts[i].v[surf->warpindex[1]] * 0.025 + realtime) * STMult) & STBitMask];
			tadd = sintable[(int) ((p->verts[i].v[surf->warpindex[0]] * 0.025 + realtime) * STMult) & STBitMask];

			// set the warp verts
			// fix water warping - derive the texcoords from the verts!!!  mad, but it works!!!
			// warpindex is precalculated in gl_model.c
			verts->st[0] = p->verts[i].v[surf->warpindex[0]] + sadd;
			verts->st[1] = p->verts[i].v[surf->warpindex[1]] + tadd;

			// copy untransformed verts to the vertex array
			verts->v[0] = p->verts[i].v[0];
			verts->v[1] = p->verts[i].v[1];
			verts->v[2] = p->verts[i].v[2];
		}

		// draw the array
		glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
	}
}


/*
================
R_DrawWaterSurfaces
================
*/
void R_DrawWaterSurfaces (void)
{
	msurface_t	*s;
	gltexture_t	*t;

	// go back to the world matrix
	glLoadMatrixf (r_worldentity.matrix);

	// no liquid
	if (!(globalpassflags & PASS_LIQUID)) return;

	// bring up vertex arrays for the render
	R_ConfigureVertexArrays (VA_VERTEX | VA_TEXTURE0_ST);

	glDepthMask (GL_FALSE);

	glEnable (GL_BLEND);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glMatrixMode (GL_TEXTURE);
	glLoadIdentity ();
	glScalef (0.015625, 0.015625, 0.015625);

	for (t = gltextures; t; t = t->next)
	{
		// texture not included in liquid pass
		if (!(t->passflags & PASS_LIQUID)) continue;

		s = t->texturechain;

		if (!s) continue;

		if (!(s->flags & SURF_DRAWTURB)) continue;

		if (s->flags & SURF_DRAWLAVA)
			glColor4f (1, 1, 1, 1);
		else if (s->flags & SURF_DRAWTELE)
			glColor4f (1, 1, 1, 1);
		else if (s->flags & SURF_DRAWSLIME)
			glColor4f (1, 1, 1, 0.9);
		else glColor4f (1, 1, 1, 0.65);

		GL_BindTexture (t->texnum);

		for (; s; s = s->texturechain)
		{
			// surface not included in liquid pass
			if (!(s->passflags & PASS_LIQUID)) continue;

			EmitWaterPolys (s);
		}
	}

	glLoadIdentity ();
	glMatrixMode (GL_MODELVIEW);

	R_ConfigureVertexArrays (VA_KILL);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glColor4f (1, 1, 1, 1);
	glDisable (GL_BLEND);

	glDepthMask (GL_TRUE);
}


void GL_CombineRGBScale2 (void)
{
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
	glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 2.0f);
}


void GL_CombineRGBScale4 (void)
{
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
	glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
	glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
	glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
	glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 4.0f);
}


void GL_KillTMU (GLenum TMU_TO_KILL)
{
	// make sure everything is dead
	GL_ActivateTexture (TMU_TO_KILL);

	if (TMU_TO_KILL != GL_TEXTURE0)
	{
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glDisable (GL_TEXTURE_2D);
	}

	glClientActiveTexture (TMU_TO_KILL);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
}


void GL_EnableTMU (GLenum TMU_TO_ENABLE, float *TexCoordPtr)
{
	if (TMU_TO_ENABLE != GL_TEXTURE0)
	{
		GL_ActivateTexture (TMU_TO_ENABLE);
		glEnable (GL_TEXTURE_2D);
	}

	// and vertex arrays
	glClientActiveTexture (TMU_TO_ENABLE);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), TexCoordPtr);
}


void ToggleWaterFog (void)
{
	// testing
	float FogColor[4] = {0.5, 0.3, 0.1, 1.0};

	// switch on underwater fog
	// note that we don't set colour at this stage cos it's different for each surf
	glFogi (GL_FOG_MODE, GL_LINEAR);

	// this looks like a Thing That Should Not Be, but it's actually what we want to do, to
	// ensure a minimum amount of fogging.  this will not look good when we're in lava or 
	// slime, but by that point we've probably got all sorts of other things on our mind...
	glFogf (GL_FOG_START, -1500);
	glFogf (GL_FOG_END, 1900);

	glFogf (GL_FOG_DENSITY, 0.2);
	glEnable (GL_FOG);

	// testing
	glFogfv (GL_FOG_COLOR, FogColor);
}


qboolean R_CheckViewEnt (void)
{
	if (!r_drawviewmodel.value) return false;
	if (chase_active.value) return false;
	if (cl.items & IT_INVISIBILITY) return false;
	if (cl.stats[STAT_HEALTH] <= 0) return false;
	if (!cl.viewent.model) return false;

	return true;
}

void R_MarkSurfaceLight (lightent_t *light, mnode_t *node)
{
	int			c, side;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	double		dot;
	int			j;
	glpoly_t *p;
	glvertex_t *verts, *v;

	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;
	if (node->lightvisframe != r_lightframecount) return;

	// if a leaf node, mark surfaces
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *) node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->lightvisframe = r_lightframecount;
				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		//if (pleaf->efrags) R_StoreEfrags (&pleaf->efrags);

		return;
	}

	// node is just a decision point, so go down the apropriate sides
	// find which side of the node we are on
	/*
	switch (node->plane->type)
	{
	case PLANE_X:
		dot = light->origin[0] - node->plane->dist;
		break;
	case PLANE_Y:
		dot = light->origin[1] - node->plane->dist;
		break;
	case PLANE_Z:
		dot = light->origin[2] - node->plane->dist;
		break;
	default:
		dot = DotProduct (light->origin, node->plane->normal) - node->plane->dist;
		break;
	}
	*/

		dot = DotProduct (light->origin, node->plane->normal) - node->plane->dist;

	// light is in range so determine which side we're on
	if (dot >= 0)
		side = 0;
	else side = 1;

	// recurse down the children, front side first
	R_MarkSurfaceLight (light, node->children[side]);

	// draw stuff
	c = node->numsurfaces;

	if (c) // && dot > -light->intensity && dot < light->intensity)
	{
		surf = cl.worldmodel->surfaces + node->firstsurface;

		// damned if i can see the sense behind this particular bit of code
		//if (dot < 0 - BACKFACE_EPSILON)
		//	side = SURF_PLANEBACK;
		//else if (dot > BACKFACE_EPSILON)
		//	side = 0;

		for (; c; c--, surf++)
		{
			// not visible
			if (surf->visframe != r_framecount) continue;
			if (surf->lightvisframe != r_lightframecount) continue;
			if (surf->flags & SURF_DRAWTURB) continue;
			if (surf->flags & SURF_DRAWSKY) continue;
			if (!surf->polys->matrix) continue;

			dot = DotProduct (light->origin, surf->plane->normal) - surf->plane->dist;

			if (dot < -light->intensity) continue;
			if (dot > light->intensity) continue;

			// build a transformed vertex array for the poly
			p = surf->polys;

			for (j = 0, verts = surfverts, v = p->verts; j < p->numverts; j++, verts++, v++)
			{
				// build transformed verts directly into the vertex array
				verts->v[0] = p->matrix[0x0] * v->v[0] + p->matrix[0x4] * v->v[1] + p->matrix[0x8] * v->v[2] + p->matrix[0xc];
				verts->v[1] = p->matrix[0x1] * v->v[0] + p->matrix[0x5] * v->v[1] + p->matrix[0x9] * v->v[2] + p->matrix[0xd];
				verts->v[2] = p->matrix[0x2] * v->v[0] + p->matrix[0x6] * v->v[1] + p->matrix[0xa] * v->v[2] + p->matrix[0xe];

				verts->rgba[0] = ((p->verts[j].v[0] - light->origin[0]) / light->intensity) / 2.0f + 0.5f;
				verts->rgba[1] = ((p->verts[j].v[1] - light->origin[1]) / light->intensity) / 2.0f + 0.5f;
				verts->rgba[2] = ((p->verts[j].v[2] - light->origin[2]) / light->intensity) / 2.0f + 0.5f;
			}

			glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
		}
	}

	// recurse down the back side
	R_MarkSurfaceLight (light, node->children[!side]);
}


void R_RenderSurfaceLight (msurface_t *surf, lightent_t *light)
{
	float		dot;
	int			j;
	glpoly_t *p;
	glvertex_t *verts, *v;

	if (surf->visframe != r_framecount) return;
	if (surf->flags & SURF_DRAWTURB) return;
	if (surf->flags & SURF_DRAWSKY) return;
	if (!surf->polys->matrix) return;

	dot = DotProduct (light->origin, surf->plane->normal) - surf->plane->dist;

	if (((surf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) || 
		(!(surf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)) ||
		(surf->flags & SURF_BMODEL))
	{
		// build a transformed vertex array for the poly
		p = surf->polys;

		for (j = 0, verts = surfverts, v = p->verts; j < p->numverts; j++, verts++, v++)
		{
			// build transformed verts directly into the vertex array
			verts->v[0] = p->matrix[0x0] * v->v[0] + p->matrix[0x4] * v->v[1] + p->matrix[0x8] * v->v[2] + p->matrix[0xc];
			verts->v[1] = p->matrix[0x1] * v->v[0] + p->matrix[0x5] * v->v[1] + p->matrix[0x9] * v->v[2] + p->matrix[0xd];
			verts->v[2] = p->matrix[0x2] * v->v[0] + p->matrix[0x6] * v->v[1] + p->matrix[0xa] * v->v[2] + p->matrix[0xe];

			verts->rgba[0] = ((p->verts[j].v[0] - light->origin[0]) / light->intensity) / 2.0f + 0.5f;
			verts->rgba[1] = ((p->verts[j].v[1] - light->origin[1]) / light->intensity) / 2.0f + 0.5f;
			verts->rgba[2] = ((p->verts[j].v[2] - light->origin[2]) / light->intensity) / 2.0f + 0.5f;
		}

		glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
	}
}


void R_MarkSurfaceLight666 (lightent_t *light, mnode_t *node)
{
	float		dist;
	msurface_t	*surf;
	int			i;

	if (node->contents < 0) return;
	if (node->visframe != r_visframecount) return;
	if (node->lightvisframe != r_lightframecount) return;

	dist = DotProduct (light->origin, node->plane->normal) - node->plane->dist;

	// out of range so just go down the appropriate side
	if (dist > light->intensity)
	{
		R_MarkSurfaceLight666 (light, node->children[0]);
		return;
	}

	// out of range so just go down the appropriate side
	if (dist < -light->intensity)
	{
		R_MarkSurfaceLight666 (light, node->children[1]);
		return;
	}

	// mark the polygons
	surf = cl.worldmodel->surfaces + node->firstsurface;

	for (i = 0; i < node->numsurfaces; i++, surf++)
		R_RenderSurfaceLight (surf, light);

	// go down both sides
	R_MarkSurfaceLight666 (light, node->children[0]);
	R_MarkSurfaceLight666 (light, node->children[1]);
}


void R_MarkSurfaceLights (lightent_t *light)
{
	byte *vis;
	int i;
	mnode_t *node;

	vis = Mod_LeafPVS (light->lightleaf, cl.worldmodel);

	// advance the framecount
	r_lightframecount++;

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		//if (cl.worldmodel->leafs[i].visframe != r_visframecount) continue;

		if (vis[i >> 3] & (1 << (i & 7)))
		{
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				if (node->lightvisframe == r_lightframecount) break;

				node->lightvisframe = r_lightframecount;
				node = node->parent;
			} while (node);
		}
	}

	R_MarkSurfaceLight666 (light, cl.worldmodel->nodes);
}


void R_RenderNewSurfaces (void)
{
	gltexture_t *tex;
	msurface_t *surf;
	glpoly_t *p;
	glvertex_t *verts, *v;
	int i;
	lightent_t *le;
	lightent_t lent;

	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (glvertex_t), surfverts->v);

	// render everything in black
	glDisable (GL_TEXTURE_2D);
	glColor3f (0, 0, 0);

	// go through all the textures in the world
	for (tex = gltextures; tex; tex = tex->next)
	{
		// no surfs chained for this texture
		if (!tex->texturechain) continue;

		// no lighting info
		if (tex->texturechain->flags & SURF_DRAWTURB) continue;
		if (tex->texturechain->flags & SURF_DRAWSKY) continue;

		// go through the surfs that use this texture
		for (surf = tex->texturechain; surf; surf = surf->texturechain)
		{
			// build a transformed vertex array for the poly
			p = surf->polys;

			for (i = 0, verts = surfverts, v = p->verts; i < p->numverts; i++, verts++, v++)
			{
				// build transformed verts directly into the vertex array
				verts->v[0] = p->matrix[0x0] * v->v[0] + p->matrix[0x4] * v->v[1] + p->matrix[0x8] * v->v[2] + p->matrix[0xc];
				verts->v[1] = p->matrix[0x1] * v->v[0] + p->matrix[0x5] * v->v[1] + p->matrix[0x9] * v->v[2] + p->matrix[0xd];
				verts->v[2] = p->matrix[0x2] * v->v[0] + p->matrix[0x6] * v->v[1] + p->matrix[0xa] * v->v[2] + p->matrix[0xe];
			}

			glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
		}
	}

	glColor3f (1, 1, 1);
	glEnable (GL_TEXTURE_2D);

	glEnable (GL_BLEND);
	glDepthMask (GL_FALSE);

	glBlendFunc (GL_ONE, GL_ONE);

	// texcoords arrays
	glClientActiveTexture (GL_TEXTURE0);
	glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->rgba);

	// hack - use RGBA for the texcoord pointer here
	glClientActiveTexture (GL_TEXTURE1);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (1, GL_FLOAT, sizeof (glvertex_t), &surfverts->rgba[2]);

	GL_ActivateTexture (GL_TEXTURE1);
	glDisable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_1D);
	glBindTexture (GL_TEXTURE_1D, Atten1D.texnum);

	// modulate mode on TMU 1 - this tmu will store the actual texture colour.
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// switch to tmu 0 for the 2D texture
	GL_ActivateTexture (GL_TEXTURE0);
	GL_BindTexture (Atten2D.texnum);

	// modulate mode on TMU0
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	// make sure TMU 1 is active coming in here
	GL_ActivateTexture (GL_TEXTURE1);

	for (le = lightents; le; le = le->next)
	{
		// light is not visible this frame
		// if (le->lightleaf->visframe != r_visframecount) continue;

		glColor3f (le->colour[0] * ((float) d_lightstylevalue[le->style] / 1024.0f),
			le->colour[1] * ((float) d_lightstylevalue[le->style] / 1024.0f), 
			le->colour[2] * ((float) d_lightstylevalue[le->style] / 1024.0f));

		R_MarkSurfaceLights (le);
	}

	// add in dynamic lights
	for (i = 0; i < MAX_DLIGHTS; i++)
	{
		// dead light
		if (cl_dlights[i].die < cl.time) continue;

		lent.origin[0] = cl_dlights[i].origin[0];
		lent.origin[1] = cl_dlights[i].origin[1];
		lent.origin[2] = cl_dlights[i].origin[2];

		lent.intensity = cl_dlights[i].radius;

		lent.lightleaf = Mod_PointInLeaf (lent.origin, cl.worldmodel);

		glColor3f (cl_dlights[i].rgb[0] / 4.0f, cl_dlights[i].rgb[1] / 4.0f, cl_dlights[i].rgb[2] / 4.0f);

		R_MarkSurfaceLights (&lent);
	}

	// hack to keep tracking up to date
	glDisable (GL_TEXTURE_1D);
	glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glEnable (GL_TEXTURE_2D);
	GL_BindTexture (Atten2D.texnum);
	glDisable (GL_TEXTURE_2D);

	glClientActiveTexture (GL_TEXTURE1);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glColor3f (1, 1, 1);

	GL_ActivateTexture (GL_TEXTURE0);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable (GL_BLEND);
	glDepthMask (GL_TRUE);

	glClientActiveTexture (GL_TEXTURE0);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	glDisableClientState (GL_VERTEX_ARRAY);
}


void R_RenderSingleEntityLighting (entity_t *e, glvertex_t *cache)
{
	int i;
	int j;
	glpoly_t *p;
	glvertex_t *verts;
	msurface_t *surf;

	// go through all of the surfs in the entity
	for (surf = e->model->surfaces, i = 0; i < e->model->numsurfaces; i++, surf++)
	{
		if (surf->flags & SURF_DRAWSKY) continue;
		if (surf->flags & SURF_DRAWTURB) continue;

		// bsp models will have a single poly, alias models will have them chained
		for (p = surf->polys; p; p = p->next)
		{
			// build a transformed vertex array for the poly
			for (j = 0, verts = surfverts; j < p->numverts; j++, verts++, cache++)
			{
				// build transformed verts directly into the vertex array
				verts->v[0] = e->matrix[0x0] * cache->v[0] + e->matrix[0x4] * cache->v[1] + e->matrix[0x8] * cache->v[2] + e->matrix[0xc];
				verts->v[1] = e->matrix[0x1] * cache->v[0] + e->matrix[0x5] * cache->v[1] + e->matrix[0x9] * cache->v[2] + e->matrix[0xd];
				verts->v[2] = e->matrix[0x2] * cache->v[0] + e->matrix[0x6] * cache->v[1] + e->matrix[0xa] * cache->v[2] + e->matrix[0xe];

				// colour
				verts->rgba[0] = cache->rgba[0] * e->lightpoint[0];
				verts->rgba[1] = cache->rgba[1] * e->lightpoint[1];
				verts->rgba[2] = cache->rgba[2] * e->lightpoint[2];
				verts->rgba[3] = 1;
			}

			glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
		}
	}
}


void R_RenderEntityLighting (void)
{
	int i;
	entity_t *e;

	// add entity lighting first - we don't assume that there are always entities visible
	if (globalpassflags & PASS_ENTITY)
	{
		glDisable (GL_TEXTURE_2D);
		glEnableClientState (GL_COLOR_ARRAY);
		glColorPointer (4, GL_FLOAT, sizeof (glvertex_t), surfverts->rgba);

		for (i = 0; i < cl_numvisedicts; i++)
		{
			e = cl_visedicts[i];

			// entity has no model
			if (!e->model) continue;

			// inline bmodels are handled with the world
			if (e->model->name[0] == '*') continue;

			// entity has no passflags as it has been culled
			if (!e->PassFlags) continue;

			R_RenderSingleEntityLighting (e, vertcache + (maxverts * i));
		}

		if (R_CheckViewEnt ())
		{
			glDepthRange (0, 0.3);
			R_RenderSingleEntityLighting (&cl.viewent, vertcache + (maxverts * MAX_VISEDICTS));
			glDepthRange (0, 1);
		}

		glDisableClientState (GL_COLOR_ARRAY);
		glEnable (GL_TEXTURE_2D);
	}
}


void R_RenderLightmaps (void)
{
	int i;
	glpoly_t *p;
	glvertex_t *v;
	glvertex_t *verts;
	gllightmap_t *lm;
	extern gllightmap_t *gllightmaps;

	// render all lightmaps
	for (lm = gllightmaps; lm; lm = lm->next)
	{
		// no polys in the chain
		if (!lm->Polys) continue;

		// bind lightmap texture
		GL_BindTexture (lm->texnum);

		// check for modification and re-upload
		R_BlendLightmap (lm);

		// go through the poly chain for the lightmap
		for (p = lm->Polys; p; p = p->chain)
		{
			// build a transformed vertex array for the poly
			for (i = 0, verts = surfverts, v = p->verts; i < p->numverts; i++, verts++, v++)
			{
				// build transformed verts directly into the vertex array
				verts->v[0] = p->matrix[0x0] * v->v[0] + p->matrix[0x4] * v->v[1] + p->matrix[0x8] * v->v[2] + p->matrix[0xc];
				verts->v[1] = p->matrix[0x1] * v->v[0] + p->matrix[0x5] * v->v[1] + p->matrix[0x9] * v->v[2] + p->matrix[0xd];
				verts->v[2] = p->matrix[0x2] * v->v[0] + p->matrix[0x6] * v->v[1] + p->matrix[0xa] * v->v[2] + p->matrix[0xe];

				// lightmap texcoords
				verts->lm[0] = v->lm[0];
				verts->lm[1] = v->lm[1];
			}

			glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
		}

		// clear the polys
		lm->Polys = NULL;
	}
}


void R_RenderSingleEntityDLight (entity_t *e, dlight_t *dl, glvertex_t *cache)
{
	int i;
	int j;
	glpoly_t *p;
	glvertex_t *verts;

	// texture and surface pointers
	msurface_t *surf;

	// go through all of the surfs in the entity
	for (surf = e->model->surfaces, i = 0; i < e->model->numsurfaces; i++, surf++)
	{
		if (surf->flags & SURF_DRAWSKY) continue;
		if (surf->flags & SURF_DRAWTURB) continue;

		// bsp models will have a single poly, alias models will have them chained
		for (p = surf->polys; p; p = p->next)
		{
			// build a transformed vertex array for the poly
			for (j = 0, verts = surfverts; j < p->numverts; j++, verts++, cache++)
			{
				// build transformed verts directly into the vertex array
				verts->v[0] = e->matrix[0x0] * cache->v[0] + e->matrix[0x4] * cache->v[1] + e->matrix[0x8] * cache->v[2] + e->matrix[0xc];
				verts->v[1] = e->matrix[0x1] * cache->v[0] + e->matrix[0x5] * cache->v[1] + e->matrix[0x9] * cache->v[2] + e->matrix[0xd];
				verts->v[2] = e->matrix[0x2] * cache->v[0] + e->matrix[0x6] * cache->v[1] + e->matrix[0xa] * cache->v[2] + e->matrix[0xe];

				// dlight texcoords
				verts->rgba[0] = ((cache->v[0] + e->origin[0] - dl->origin[0]) / dl->radius) / 2.0f + 0.5f;
				verts->rgba[1] = ((cache->v[1] + e->origin[1] - dl->origin[1]) / dl->radius) / 2.0f + 0.5f;
				verts->rgba[2] = ((cache->v[2] + e->origin[2] - dl->origin[2]) / dl->radius) / 2.0f + 0.5f;
			}

			glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
		}
	}
}


void R_RenderDynamicLight (dlight_t *dl, int lnum)
{
	int i;
	glpoly_t *p;
	glvertex_t *v;
	glvertex_t *verts;
	entity_t *e;
	vec3_t end;

	// texture and surface pointers
	gltexture_t *tex;
	msurface_t *surf;

	// update the dlight's colour
	// this is faster than re-uploading the 1d texture with the new colours, and gives us the ability to
	// use a larger attenuation map for better detail
	glColor3f (dl->rgb[0] / 4.0f, dl->rgb[1] / 4.0f, dl->rgb[2] / 4.0f);

	// go through all the textures in the world
	for (tex = gltextures; tex; tex = tex->next)
	{
		// no surfs chained for this texture
		if (!tex->texturechain) continue;

		// no lighting info
		if (tex->texturechain->flags & SURF_DRAWTURB) continue;
		if (tex->texturechain->flags & SURF_DRAWSKY) continue;

		// go through the surfs that use this texture
		for (surf = tex->texturechain; surf; surf = surf->texturechain)
		{
			// no dlight this frame
			if (surf->dlightframe != r_framecount) continue;

			// not marked for this light
			if (!(surf->dlightbits[(lnum >> 3)] & (1 << (lnum & 7)))) continue;

			// build a transformed vertex array for the poly
			p = surf->polys;

			for (i = 0, verts = surfverts, v = p->verts; i < p->numverts; i++, verts++, v++)
			{
				// build transformed verts directly into the vertex array
				verts->v[0] = p->matrix[0x0] * v->v[0] + p->matrix[0x4] * v->v[1] + p->matrix[0x8] * v->v[2] + p->matrix[0xc];
				verts->v[1] = p->matrix[0x1] * v->v[0] + p->matrix[0x5] * v->v[1] + p->matrix[0x9] * v->v[2] + p->matrix[0xd];
				verts->v[2] = p->matrix[0x2] * v->v[0] + p->matrix[0x6] * v->v[1] + p->matrix[0xa] * v->v[2] + p->matrix[0xe];

				verts->rgba[0] = ((p->verts[i].v[0] - dl->origin[0]) / dl->radius) / 2.0f + 0.5f;
				verts->rgba[1] = ((p->verts[i].v[1] - dl->origin[1]) / dl->radius) / 2.0f + 0.5f;
				verts->rgba[2] = ((p->verts[i].v[2] - dl->origin[2]) / dl->radius) / 2.0f + 0.5f;
			}

			glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
		}
	}

	if (R_CheckViewEnt ())
	{
		// check dlight relative to the view entity
		VectorSubtract (cl.viewent.origin, dl->origin, end);

		// dlight is too far
		if ((dl->radius - Length (end)) > 0)
		{
			glDepthRange (0, 0.3);
			R_RenderSingleEntityDLight (&cl.viewent, dl, vertcache + (maxverts * MAX_VISEDICTS));
			glDepthRange (0, 1);
		}
	}

	// no entities
	if (!globalpassflags & PASS_ENTITY) return;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		e = cl_visedicts[i];

		// entity has no model
		if (!e->model) continue;

		// inline bmodels are handled with the world
		if (e->model->name[0] == '*') continue;

		// entity has no passflags as it has been culled
		if (!e->PassFlags) continue;

		// check dlight relative to the entity
		VectorSubtract (e->origin, dl->origin, end);

		// dlight is too far
		if ((dl->radius - Length (end)) <= 0) continue;

		R_RenderSingleEntityDLight (e, dl, vertcache + (maxverts * i));
	}
}


void R_RenderSingleTexturedEntity (entity_t *e, glvertex_t *cache, int PassFlags)
{
	int i;
	int j;
	glpoly_t *p;
	glvertex_t *verts;
	gltexture_t *mdltex = NULL;

	// texture and surface pointers
	msurface_t *surf;

	// go through all of the surfs in the entity
	for (surf = e->model->surfaces, i = 0; i < e->model->numsurfaces; i++, surf++)
	{
		if (surf->flags & SURF_DRAWSKY) continue;
		if (surf->flags & SURF_DRAWTURB) continue;

		// surf passflags don't match
		if (!(surf->passflags & PassFlags))
		{
			// hack - need to skip over the cache here so that it's in the proper place for the next surf
			for (p = surf->polys; p; p = p->next)
				for (j = 0; j < p->numverts; j++)
					cache++;

			// don't draw
			continue;
		}

		// the texture to select depends on what type of model it is
		if (e->model->type == mod_alias)
		{
			int anim;
			extern	gltexture_t	playertextures[];

			// get the correct texture number
			anim = (int) (cl.time * 10) & 3;
			mdltex = ((aliashdr_t *) e->model->cache.data)->gl_texture[e->skinnum][anim];

			// we can't dynamically colormap textures, so they are cached
			// seperately for the players.  Heads are just uncolored.
			if (e->colormap != vid.colormap && !gl_nocolors.value)
			{
				j = e - cl_entities;

				if (j >= 1 && j <= cl.maxclients) mdltex = &playertextures[j - 1];
			}
		}
		else
		{
			// brush model so get the correct animation
			texture_t *tex = R_TextureAnimation (e, surf->texinfo->texture);
			mdltex = tex->gl_texture;
		}

		// check which pass we're on
		if (PassFlags & PASS_LUMA)
		{
			// bind the luma texture
			GL_BindTexture (mdltex->luma->texnum);
		}
		else
		{
			// bind the regular texture
			GL_BindTexture (mdltex->texnum);

			// check for a luma and flags as appropriate
			if (mdltex->luma)
			{
				// set for the surf, the entity and globally, so that we'll
				// know to call this in the main render too.
				surf->passflags |= PASS_LUMA;
				e->PassFlags |= PASS_LUMA;
				globalpassflags |= PASS_LUMA;
			}
		}

		// bsp models will have a single poly, alias models will have them chained
		for (p = surf->polys; p; p = p->next)
		{
			// build a transformed vertex array for the poly
			for (j = 0, verts = surfverts; j < p->numverts; j++, verts++, cache++)
			{
				// build transformed verts directly into the vertex array
				verts->v[0] = e->matrix[0x0] * cache->v[0] + e->matrix[0x4] * cache->v[1] + e->matrix[0x8] * cache->v[2] + e->matrix[0xc];
				verts->v[1] = e->matrix[0x1] * cache->v[0] + e->matrix[0x5] * cache->v[1] + e->matrix[0x9] * cache->v[2] + e->matrix[0xd];
				verts->v[2] = e->matrix[0x2] * cache->v[0] + e->matrix[0x6] * cache->v[1] + e->matrix[0xa] * cache->v[2] + e->matrix[0xe];

				// texcoords
				verts->st[0] = cache->st[0];
				verts->st[1] = cache->st[1];

				verts->wc[0] = cache->wc[0];
				verts->wc[1] = cache->wc[1];
			}

			glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
		}
	}
}


/*
====================
R_RenderTexturedSurfaces

Renders all textured surfaces.  Takes a PassFlags arg to make it reusable for both textures and lumas
====================
*/
void R_RenderTexturedSurfaces (int PassFlags)
{
	int i;
	glpoly_t *p;
	glvertex_t *v;
	glvertex_t *verts;

	// texture and surface pointers
	gltexture_t *tex;
	msurface_t *surf;

	entity_t *e;

	// go through all the textures in the world
	for (tex = gltextures; tex; tex = tex->next)
	{
		// no surfs chained for this texture
		if (!tex->texturechain) continue;

		// no lighting info
		if (tex->texturechain->flags & SURF_DRAWTURB) continue;
		if (tex->texturechain->flags & SURF_DRAWSKY) continue;

		// not in this pass
		if (!(tex->passflags & PassFlags)) continue;

		if (PassFlags & PASS_LUMA)
			GL_BindTexture (tex->luma->texnum);
		else GL_BindTexture (tex->texnum);

		// go through the surfs that use this texture
		for (surf = tex->texturechain; surf; surf = surf->texturechain)
		{
			// not in this pass
			if (!(surf->passflags & PassFlags)) continue;

			// build a transformed vertex array for the poly
			p = surf->polys;

			for (i = 0, verts = surfverts, v = p->verts; i < p->numverts; i++, verts++, v++)
			{
				// build transformed verts directly into the vertex array
				verts->v[0] = p->matrix[0x0] * v->v[0] + p->matrix[0x4] * v->v[1] + p->matrix[0x8] * v->v[2] + p->matrix[0xc];
				verts->v[1] = p->matrix[0x1] * v->v[0] + p->matrix[0x5] * v->v[1] + p->matrix[0x9] * v->v[2] + p->matrix[0xd];
				verts->v[2] = p->matrix[0x2] * v->v[0] + p->matrix[0x6] * v->v[1] + p->matrix[0xa] * v->v[2] + p->matrix[0xe];

				verts->st[0] = v->st[0];
				verts->st[1] = v->st[1];

				verts->wc[0] = v->wc[0];
				verts->wc[1] = v->wc[1];
			}

			glDrawArrays (p->PRIMITIVE_TYPE, 0, p->numverts);
		}
	}

	if (R_CheckViewEnt () && (cl.viewent.PassFlags & PassFlags))
	{
		glDepthRange (0, 0.3);
		R_RenderSingleTexturedEntity (&cl.viewent, vertcache + (maxverts * MAX_VISEDICTS), PassFlags);
		glDepthRange (0, 1);
	}

	// no entities
	if (!globalpassflags & PASS_ENTITY) return;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		e = cl_visedicts[i];

		// entity has no model
		if (!e->model) continue;

		// inline bmodels are handled with the world
		if (e->model->name[0] == '*') continue;

		// entity has no passflags as it has been culled
		if (!e->PassFlags) continue;

		if (!(e->PassFlags & PassFlags)) continue;

		R_RenderSingleTexturedEntity (e, vertcache + (maxverts * i), PassFlags);
	}
}


/*
===================
R_RenderSurfaces

Handles all state changes (except some texture changes and some TMU switching) and calling the individual render
functions.  render functions do not handle state, with the exception of some texture changes where appropriate.
===================
*/
void R_RenderSurfaces (void)
{
	int i;
	qboolean ActiveDLights = false;

	R_RenderNewSurfaces ();

	// bring up vertex arrays for the render - common to all
	glEnableClientState (GL_VERTEX_ARRAY);
	glVertexPointer (3, GL_FLOAT, sizeof (glvertex_t), surfverts->v);

	// render entity lighting
	//R_RenderEntityLighting ();

	// render lightmaps in TMU 0
	glClientActiveTexture (GL_TEXTURE0);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->lm);

	// add in static lightmaps to the world
	//R_RenderLightmaps ();

	glEnable (GL_BLEND);

	// add in dynamic lights
	for (i = 0; i < MAX_DLIGHTS; i++)
	{
		/*
		// dead light
		if (cl_dlights[i].die < cl.time) continue;

		// dlight hasn't marked any surfs this frame
		if (cl_dlights[i].frame != r_framecount) continue;

		if (!ActiveDLights)
		{
			// bring up state
			glBlendFunc (GL_ONE, GL_ONE);

			// texcoords arrays
			glClientActiveTexture (GL_TEXTURE0);
			glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->rgba);

			// hack - use RGBA for the texcoord pointer here
			glClientActiveTexture (GL_TEXTURE1);
			glEnableClientState (GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer (1, GL_FLOAT, sizeof (glvertex_t), &surfverts->rgba[2]);

			GL_ActivateTexture (GL_TEXTURE1);
			glDisable (GL_TEXTURE_2D);
			glEnable (GL_TEXTURE_1D);
			glBindTexture (GL_TEXTURE_1D, Atten1D.texnum);

			// modulate mode on TMU 1 - this tmu will store the actual texture colour.
			// it's a 1d texture for faster uploads, and will be double-scaled because we halve the colours at 
			// upload time (this is to allow us to specify double scales at allocation time, so that 0.5|0.5|0.5 
			// is white, and we can add intensity to each channel as well as subtract.
			// the above is no longer valid, since good old glColor3f makes uploads totally unnecessary.
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
			glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
			glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
			glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
			glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
			glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 2.0f);

			// switch to tmu 0 for the 2D texture
			GL_ActivateTexture (GL_TEXTURE0);

			GL_BindTexture (Atten2D.texnum);

			// add mode on TMU0
			glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

			// make sure TMU 1 is active coming in here
			GL_ActivateTexture (GL_TEXTURE1);

			ActiveDLights = true;
		}

		// render the light
		R_RenderDynamicLight (&cl_dlights[i], i);
		*/
	}

	// take down dlight state
	if (ActiveDLights)
	{
		// hack to keep tracking up to date
		glDisable (GL_TEXTURE_1D);
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glEnable (GL_TEXTURE_2D);
		GL_BindTexture (Atten2D.texnum);
		glDisable (GL_TEXTURE_2D);

		glClientActiveTexture (GL_TEXTURE1);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);

		glColor3f (1, 1, 1);
	}

	// now set up for the textured pass
	glClientActiveTexture (GL_TEXTURE0);
	glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->st);
	GL_ActivateTexture (GL_TEXTURE0);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// render the textured pass.
	glBlendFunc (GL_DST_COLOR, GL_SRC_COLOR);

		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
		glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
		glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
		glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
		glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
		glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 2.0f);

	R_RenderTexturedSurfaces (PASS_TEXTURED);

		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	if (globalpassflags & PASS_CAUSTIC)
	{
		// render the caustic pass.  retain the same blendfunc but do a 2 x modulate in tmu 1 for the
		// correct caustic texture for this frame
		int CausticFrame = (int) (realtime * 15) % 32;

		// set up tmu
		glClientActiveTexture (GL_TEXTURE1);
		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, sizeof (glvertex_t), surfverts->wc);

		GL_ActivateTexture (GL_TEXTURE1);
		glEnable (GL_TEXTURE_2D);
		GL_BindTexture (watercaustics[CausticFrame]->texnum);

		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
		glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB, GL_MODULATE);
		glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE0_RGB, GL_PREVIOUS);
		glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB, GL_SRC_COLOR);
		glTexEnvi (GL_TEXTURE_ENV, GL_SOURCE1_RGB, GL_TEXTURE);
		glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB, GL_SRC_COLOR);
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 2.0f);

		// back to tmu 0 to get correct binds for the render
		GL_ActivateTexture (GL_TEXTURE0);

		// render pass
		R_RenderTexturedSurfaces (PASS_CAUSTIC);

		// restore state
		glClientActiveTexture (GL_TEXTURE1);
		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
		glClientActiveTexture (GL_TEXTURE0);

		GL_ActivateTexture (GL_TEXTURE1);
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE, 1.0f);
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glDisable (GL_TEXTURE_2D);

		GL_ActivateTexture (GL_TEXTURE0);
	}

	// check for other passes
	if (globalpassflags & PASS_LUMA)
	{
		// render the luma pass.
		glBlendFunc (GL_ONE, GL_ONE);
		R_RenderTexturedSurfaces (PASS_LUMA);
	}

	// kill tmu 0 texcoord array
	glClientActiveTexture (GL_TEXTURE0);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);

	// kill vertex array
	glDisableClientState (GL_VERTEX_ARRAY);
	glDisable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}


void R_SetPassFlags (msurface_t *surf, gltexture_t *tex)
{
	if (surf->passframe != r_framecount)
	{
		// initialize the surface
		surf->passflags = 0;
		surf->passframe = r_framecount;
	}

	if ((surf->flags & SURF_UNDERWATER) && watercaustics[0])
	{
		// underwater
		tex->passflags |= PASS_CAUSTIC;
		surf->passflags |= PASS_CAUSTIC;
	}
	else
	{
		// plain ordinary textured only
		tex->passflags |= PASS_TEXTURED;
		surf->passflags |= PASS_TEXTURED;
	}

	if (tex->luma)
	{
		// lumamaps
		tex->passflags |= PASS_LUMA;
		surf->passflags |= PASS_LUMA;
	}

	if (surf->dlightframe == r_framecount)
	{
		// dynamic light
		tex->passflags |= PASS_DYNAMIC;
		surf->passflags |= PASS_DYNAMIC;
	}

	// always lightmapped
	tex->passflags |= PASS_LIGHTMAPPED;
	surf->passflags |= PASS_LIGHTMAPPED;

	// accumulate into the global passflags
	globalpassflags |= tex->passflags;
}


// mark all nodes in a bmodel visible
void R_MakeVisible (mnode_t *node)
{
}


void R_AddBrushModelToList (entity_t *e)
{
	vec3_t mins, maxs;
	int i;
	msurface_t *surf;
	float dot;
	mplane_t *pplane;
	model_t *clmodel;
	qboolean rotated;
	texture_t *tex;
	vec3_t originhack;
	mleaf_t *entleaf;
	mleaf_t *surfleaf;

	clmodel = e->model;

	if (e->angles[0] || e->angles[1] || e->angles[2])
	{
		rotated = true;

		for (i = 0; i < 3; i++)
		{
			mins[i] = e->origin[i] - clmodel->radius;
			maxs[i] = e->origin[i] + clmodel->radius;
		}
	}
	else
	{
		rotated = false;
		VectorAdd (e->origin, clmodel->mins, mins);
		VectorAdd (e->origin, clmodel->maxs, maxs);
	}

	if (R_CullBox (mins, maxs))
		return;

	VectorSubtract (r_refdef.vieworg, e->origin, modelorg);

	if (rotated)
	{
		vec3_t	temp;
		vec3_t	forward, right, up;

		VectorCopy (modelorg, temp);
		AngleVectors (e->angles, forward, right, up);
		modelorg[0] = DotProduct (temp, forward);
		modelorg[1] = -DotProduct (temp, right);
		modelorg[2] = DotProduct (temp, up);
	}

	// hack the origin to prevent bmodel z fighting
	VectorCopy (e->origin, originhack);
	e->origin[0] -= DIST_EPSILON;
	e->origin[1] -= DIST_EPSILON;
	e->origin[2] -= DIST_EPSILON;

	// transform and store the matrix.  these transform relative to the world matrix
	swLoadMatrixf (r_worldentity.matrix);

	swTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);

    swRotatef (e->angles[1],  0, 0, 1);
    swRotatef (-e->angles[0],  0, 1, 0);
    swRotatef (e->angles[2],  1, 0, 0);

	swGetMatrixf (e->matrix);

	// un-hack the origin
	VectorCopy (originhack, e->origin);

	// get the midpoint of the bbox
	originhack[0] = (mins[0] + maxs[0]) / 2.0f;
	originhack[1] = (mins[1] + maxs[1]) / 2.0f;
	originhack[2] = (mins[2] + maxs[2]) / 2.0f;

	// figure which leaf it's in
	entleaf = Mod_PointInLeaf (originhack, cl.worldmodel);

	// force to visible
	R_MakeVisible (clmodel->nodes + clmodel->hulls[0].firstclipnode);

	// draw texture
	for (i = 0, surf = &clmodel->surfaces[clmodel->firstmodelsurface]; i < clmodel->nummodelsurfaces; i++, surf++)
	{
		// don't bother drawing these
		// this slightly breaks hipnotic compatability, but the thing it breaks was a pretty fucking filthy hack anyway
		if (surf->flags & SURF_DRAWTURB) continue;
		if (surf->flags & SURF_DRAWSKY) continue;

		// find which side of the node we are on
		pplane = surf->plane;

		dot = DotProduct (modelorg, pplane->normal) - pplane->dist;

		// draw the polygon
		if (((surf->flags & SURF_PLANEBACK) && (dot < -BACKFACE_EPSILON)) || (!(surf->flags & SURF_PLANEBACK) && (dot > BACKFACE_EPSILON)))
		{
			// get the proper animation
			tex = R_TextureAnimation (e, surf->texinfo->texture);

			// if sorting by texture, just store it out
			surf->texturechain = tex->gl_texture->texturechain;
			tex->gl_texture->texturechain = surf;

			// get the lightmap modifications
			R_ModifySurfaceLightmap (surf);

			// use an originhack on the surface midpoint too
			originhack[0] = surf->polys->midpoint[0] + e->origin[0];
			originhack[1] = surf->polys->midpoint[1] + e->origin[1];
			originhack[2] = surf->polys->midpoint[2] + e->origin[2];

			// figure which leaf the surf is in
			surfleaf = Mod_PointInLeaf (originhack, cl.worldmodel);

			// these surfs can sometimes be embedded in solid objects as a result of plat movement, so use the
			// entity position as a fallback if we find one that is.  fixme - cache this in the entity and the
			// surfaces at load time, and only recalculate it if we determine that the entity has moved.
			if (surfleaf->contents == CONTENTS_SOLID)
			{
				// determine underwater by entity position
				if (entleaf->contents == CONTENTS_SOLID || entleaf->contents == CONTENTS_EMPTY)
				{
					// ugly but it works
					surf->flags |= SURF_UNDERWATER;
					surf->flags ^= SURF_UNDERWATER;
				}
				else surf->flags |= SURF_UNDERWATER;
			}
			else if (surfleaf->contents == CONTENTS_EMPTY)
			{
				// ugly but it works
				surf->flags |= SURF_UNDERWATER;
				surf->flags ^= SURF_UNDERWATER;
			}
			else surf->flags |= SURF_UNDERWATER;

			// set passflags
			R_SetPassFlags (surf, tex->gl_texture);

			// chain in the lightmap
			surf->polys->chain = surf->lightmap->Polys;
			surf->lightmap->Polys = surf->polys;

			// set the entity for this surf
			surf->entity = e;
			surf->polys->matrix = e->matrix;

			// mark as visible
			surf->visframe = r_framecount;

			// mark as a bmodel (fixme - only ever do this once)
			surf->flags |= SURF_BMODEL;
		}
	}

	// calculate dynamic lighting for bmodel if it's not an instanced model
	// FIXME - isn't this always true?
	// the push is left till here because we need to mark surfs as visible as we pass them
	if (clmodel->firstmodelsurface != 0) R_PushDlightsNOVIS (clmodel->nodes + clmodel->hulls[0].firstclipnode);
}


void R_AddBrushModels (void)
{
	int i;

	for (i = 0; i < cl_numvisedicts; i++)
	{
		// skip conditions
		if (!cl_visedicts[i]->model) continue;
		if (cl_visedicts[i]->model->type != mod_brush) continue;
		if (cl_visedicts[i]->model->name[0] != '*') continue;

		// add to the draw list
		R_AddBrushModelToList (cl_visedicts[i]);
	}
}


void R_SetupAliasVerts (entity_t *e, glvertex_t *cache);

void R_PrepareSingleEntityModel (entity_t *e, glvertex_t *verts)
{
	int i, j;
	vec3_t originhack;
	vec3_t mins, maxs;
	glvertex_t *v;
	msurface_t *surf;
	glpoly_t *p;
	mleaf_t *entleaf;

	// bounding box cull
	VectorAdd (e->origin, e->model->mins, mins);
	VectorAdd (e->origin, e->model->maxs, maxs);

	if (e != &cl.viewent)
	{
		if (R_CullBox (mins, maxs))
		{
			e->PassFlags = PASS_NODRAW;
			return;
		}
	}

	// set up origin for lighting
	// move it up a bit cos these have their origin at a bottom corner
	originhack[0] = (mins[0] + maxs[0]) / 2.0f;
	originhack[1] = (mins[1] + maxs[1]) / 2.0f;
	originhack[2] = (mins[2] + maxs[2]) / 2.0f;

	// get lighting
	R_LightPoint (originhack, e->lightpoint);

	// keep the lighting scale consistent with the world
	if (e == &cl.viewent)
	{
		// always give the gun some light
		e->lightpoint[0] = (e->lightpoint[0] + 16.0f) / 271.0f;
		e->lightpoint[1] = (e->lightpoint[1] + 16.0f) / 271.0f;
		e->lightpoint[2] = (e->lightpoint[2] + 16.0f) / 271.0f;
	}
	else if (e->model->flags & EF_ROTATE)
	{
		// make pickups easier to see
		e->lightpoint[0] = (e->lightpoint[0] + 128.0f) / 383.0f;
		e->lightpoint[1] = (e->lightpoint[1] + 128.0f) / 383.0f;
		e->lightpoint[2] = (e->lightpoint[2] + 128.0f) / 383.0f;
	}
	else if ((e - cl_entities) >= 1 && (e - cl_entities) <= cl.maxclients)
	{
		// never allow players to go totally black
		e->lightpoint[0] = (e->lightpoint[0] + 8.0f) / 263.0f;
		e->lightpoint[1] = (e->lightpoint[1] + 8.0f) / 263.0f;
		e->lightpoint[2] = (e->lightpoint[2] + 8.0f) / 263.0f;
	}
	else
	{
		// standard 0 to 1 scale
		e->lightpoint[0] /= 255.0f;
		e->lightpoint[1] /= 255.0f;
		e->lightpoint[2] /= 255.0f;
	}

	// for alias models use the mins or maxs of the bbox to determine if the entity is underwater
	if (e->model->type == mod_alias)
	{
		if (r_viewleaf->contents == CONTENTS_EMPTY || r_viewleaf->contents == CONTENTS_SOLID)
			originhack[2] = maxs[2];
		else originhack[2] = mins[2];
	}

	entleaf = Mod_PointInLeaf (originhack, cl.worldmodel);

	// flag for entities present
	globalpassflags |= PASS_ENTITY;

	if (entleaf->contents == CONTENTS_EMPTY || entleaf->contents == CONTENTS_SOLID || !watercaustics[0])
		e->PassFlags |= (PASS_LIGHTMAPPED | PASS_TEXTURED);
	else
	{
		e->PassFlags |= (PASS_LIGHTMAPPED | PASS_CAUSTIC);
		globalpassflags |= PASS_CAUSTIC;
	}

	// load into the vertex cache
	if (e->model->type == mod_brush)
	{
		for (i = 0, surf = e->model->surfaces; i < e->model->numsurfaces; i++, surf++)
		{
			// set passflags directly
			if (entleaf->contents == CONTENTS_EMPTY || entleaf->contents == CONTENTS_SOLID || !watercaustics[0])
				surf->passflags = PASS_TEXTURED;
			else surf->passflags = PASS_CAUSTIC;

			for (j = 0, p = surf->polys, v = surf->polys->verts; j < surf->polys->numverts; j++)
			{
				// copy to the cache
				verts->v[0] = v->v[0];
				verts->v[1] = v->v[1];
				verts->v[2] = v->v[2];

				verts->st[0] = v->st[0];
				verts->st[1] = v->st[1];

				verts->wc[0] = v->wc[0];
				verts->wc[1] = v->wc[1];

				// load colour into the cache.
				verts->rgba[0] = 1;
				verts->rgba[1] = 1;
				verts->rgba[2] = 1;
				verts->rgba[3] = 1;

				verts++;
				v++;
			}
		}
	}
	else if (e->model->type == mod_alias)
	{
		// set passflags directly
		if (entleaf->contents == CONTENTS_EMPTY || entleaf->contents == CONTENTS_SOLID || !watercaustics[0])
			e->model->surfaces->passflags = PASS_TEXTURED;
		else e->model->surfaces->passflags = PASS_CAUSTIC;

		// the alias vert setup routine is a bit to involved to include in-line here, so it's been split off
		R_SetupAliasVerts (e, verts);
	}

	// now get it relative to the correct world matrix
	swLoadMatrixf (r_worldentity.matrix);

    swTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);

    swRotatef (e->angles[1],  0, 0, 1);
    swRotatef (-e->angles[0],  0, 1, 0);
    swRotatef (e->angles[2],  1, 0, 0);

	swGetMatrixf (e->matrix);
}


void R_PrepareEntityModels (void)
{
	int ent;
	entity_t *e;

	for (ent = 0; ent < cl_numvisedicts; ent++)
	{
		e = cl_visedicts[ent];

		if (!e->model) continue;
		if (e->model->name[0] == '*') continue;

		R_PrepareSingleEntityModel (e, vertcache + (maxverts * ent));
	}

	// check for viewent visibility
	if (!R_CheckViewEnt ()) return;

	// add in the view entity
	R_PrepareSingleEntityModel (&cl.viewent, vertcache + (maxverts * MAX_VISEDICTS));
}


/*
=============================================================

	WORLD MODEL

=============================================================
*/

/*
================
R_RecursiveWorldNode
================
*/
void R_RecursiveWorldNode (mnode_t *node)
{
	int			c, side;
	mplane_t	*plane;
	msurface_t	*surf, **mark;
	mleaf_t		*pleaf;
	double		dot;

	texture_t *tex;

	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;
	if (R_CullBox (node->minmaxs, node->minmaxs + 3)) return;

	// if a leaf node, draw stuff
	if (node->contents < 0)
	{
		pleaf = (mleaf_t *)node;

		mark = pleaf->firstmarksurface;
		c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		// deal with model fragments in this leaf
		if (pleaf->efrags) R_StoreEfrags (&pleaf->efrags);

		return;
	}

	// node is just a decision point, so go down the apropriate sides
	// find which side of the node we are on
	plane = node->plane;

	switch (plane->type)
	{
	case PLANE_X:
		dot = modelorg[0] - plane->dist;
		break;
	case PLANE_Y:
		dot = modelorg[1] - plane->dist;
		break;
	case PLANE_Z:
		dot = modelorg[2] - plane->dist;
		break;
	default:
		dot = DotProduct (modelorg, plane->normal) - plane->dist;
		break;
	}

	if (dot >= 0)
		side = 0;
	else side = 1;

	// recurse down the children, front side first
	R_RecursiveWorldNode (node->children[side]);

	// draw stuff
	c = node->numsurfaces;

	if (c)
	{
		surf = cl.worldmodel->surfaces + node->firstsurface;

		// damned if i can see the sense behind this particular bit of code
		//if (dot < 0 - BACKFACE_EPSILON)
		//	side = SURF_PLANEBACK;
		//else if (dot > BACKFACE_EPSILON)
		//	side = 0;

		for (; c; c--, surf++)
		{
			// not visible
			if (surf->visframe != r_framecount) continue;

			// test for water
			if (surf->flags & SURF_DRAWTURB)
			{
				// include liquid globally
				globalpassflags |= PASS_LIQUID;

				// chain for liquid drawing
				surf->texturechain = surf->texinfo->texture->gl_texture->texturechain;
				surf->texinfo->texture->gl_texture->texturechain = surf;

				// set these explicitly to just liquid
				surf->passflags = PASS_LIQUID;
				surf->texinfo->texture->gl_texture->passflags = PASS_LIQUID;

				// do no more testing
				continue;
			}

			// test for sky
			if (surf->flags & SURF_DRAWSKY)
			{
				globalpassflags |= PASS_SKY;
				continue;
			}

			// don't backface underwater surfaces, because they warp
			if (!(surf->flags & SURF_UNDERWATER) && ((dot < 0) ^ !!(surf->flags & SURF_PLANEBACK))) continue;

			// get the proper animation
			tex = R_TextureAnimation (&r_worldentity, surf->texinfo->texture);

			// if sorting by texture, just store it out
			surf->texturechain = tex->gl_texture->texturechain;
			tex->gl_texture->texturechain = surf;

			// get the lightmap modifications
			R_ModifySurfaceLightmap (surf);

			// set passflags
			R_SetPassFlags (surf, tex->gl_texture);

			// chain in the lightmap
			surf->polys->chain = surf->lightmap->Polys;
			surf->lightmap->Polys = surf->polys;

			// set the world entity to use this surf
			surf->entity = &r_worldentity;
			surf->polys->matrix = r_worldentity.matrix;
		}
	}

	// recurse down the back side
	R_RecursiveWorldNode (node->children[!side]);
}


void R_BeginSurfaceRender (void)
{
	gltexture_t *tex;

	// clear down texturechains
	for (tex = gltextures; tex; tex = tex->next)
	{
		// clear both chain and passflags
		tex->texturechain = NULL;
		tex->passflags = PASS_NODRAW;
	}

	// nothing drawn initially
	globalpassflags = PASS_NODRAW;

	// make sure no vertex arrays are up initially
	R_ConfigureVertexArrays (VA_KILL);
}


/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (void)
{
	VectorCopy (r_refdef.vieworg, modelorg);

	glColor3f (1, 1, 1);
	glShadeModel (GL_SMOOTH);
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	// clear down the texture chains and set other initial states before drawing anything
	R_BeginSurfaceRender ();

	// add the world
	R_RecursiveWorldNode (cl.worldmodel->nodes);

	// push dynamic lighting for the world
	R_PushDlights (cl.worldmodel->nodes);

	// add inline brush model entities to the main render
	R_AddBrushModels ();

	// prepare box model entites for rendering
	R_PrepareEntityModels ();

	// go back to identity
	glLoadIdentity ();

	// render surfaces
	R_RenderSurfaces ();
	//R_RenderNewSurfaces ();

	// go back to the world matrix
	glLoadMatrixf (r_worldentity.matrix);

	// always draw this because we need to update the rotation each frame whether or not it has sky surfs
	R_DrawSkyChain ((globalpassflags & PASS_SKY));
}


/*
===============
R_MarkLeaves
===============
*/
void R_AddLeafToPVS (mleaf_t *leaf)
{
	byte	*vis;
	mnode_t	*node;
	int		i;

	vis = Mod_LeafPVS (leaf, cl.worldmodel);

	for (i = 0; i < cl.worldmodel->numleafs; i++)
	{
		if (vis[i >> 3] & (1 << (i & 7)))
		{
			node = (mnode_t *) &cl.worldmodel->leafs[i + 1];

			do
			{
				if (node->visframe == r_visframecount) break;
				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}
}


void R_MarkLeaves (void)
{
	// no transition
	if (r_oldviewleaf == r_viewleaf) return;

	// gone to a new PVS
	r_visframecount++;

	// add the viewleaf
	R_AddLeafToPVS (r_viewleaf);

	// r_oldviewleaf is NULL on the first frame
	if (!r_oldviewleaf) r_oldviewleaf = r_viewleaf;

	// check for a contents transition and add the old leaf if we've had one
	if (r_viewleaf->contents != r_oldviewleaf->contents) R_AddLeafToPVS (r_oldviewleaf);

	// set the current viewleaf as the new oldviewleaf
	r_oldviewleaf = r_viewleaf;
}



mvertex_t	*r_pcurrentvertbase;
model_t		*currentmodel;

int	nColinElim;

/*
================
BuildSurfaceDisplayList
================
*/
int maxverts = 0;
int vertcount = 0;

vec3_t farmin;
vec3_t farmax;

void BuildSurfaceDisplayList (msurface_t *surf)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	float		*vec;
	glpoly_t	*poly;

	// reconstruct the polygon
	pedges = currentmodel->edges;
	lnumverts = surf->numedges;

	// draw texture
	poly = Hunk_Alloc (sizeof (glpoly_t));
	poly->verts = Hunk_Alloc (lnumverts * sizeof (glvertex_t));
	poly->next = surf->polys;
	poly->flags = surf->flags;
	surf->polys = poly;
	poly->numverts = lnumverts;

	// init the midpoint
	poly->midpoint[0] = 0;
	poly->midpoint[1] = 0;
	poly->midpoint[2] = 0;

	for (i = 0; i < lnumverts; i++)
	{
		lindex = currentmodel->surfedges[surf->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = r_pcurrentvertbase[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = r_pcurrentvertbase[r_pedge->v[1]].position;
		}

		// set verts
		poly->verts[i].v[0] = vec[0];
		poly->verts[i].v[1] = vec[1];
		poly->verts[i].v[2] = vec[2];

		// check against far clipping plane dimensions
		if (poly->verts[i].v[0] < farmin[0]) farmin[0] = poly->verts[i].v[0];
		if (poly->verts[i].v[1] < farmin[1]) farmin[1] = poly->verts[i].v[1];
		if (poly->verts[i].v[2] < farmin[2]) farmin[2] = poly->verts[i].v[2];

		if (poly->verts[i].v[0] > farmax[0]) farmax[0] = poly->verts[i].v[0];
		if (poly->verts[i].v[1] > farmax[1]) farmax[1] = poly->verts[i].v[1];
		if (poly->verts[i].v[2] > farmax[2]) farmax[2] = poly->verts[i].v[2];

		// accumulate verts into midpoint
		poly->midpoint[0] += vec[0];
		poly->midpoint[1] += vec[1];
		poly->midpoint[2] += vec[2];

		// texcoords
		poly->verts[i].st[0] = (DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]) / surf->texinfo->texture->width;
		poly->verts[i].st[1] = (DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]) / surf->texinfo->texture->height;

		// caustic texcoords
		poly->verts[i].wc[0] = (DotProduct (vec, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3]) / 192.0f;
		poly->verts[i].wc[1] = (DotProduct (vec, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3]) / 192.0f;

		// lightmap
		R_SurfaceLightmapTexCoords (surf, poly->verts[i].lm, vec);
	}

	// set number of verts
	poly->numverts = lnumverts;

	// take average for midpoint
	poly->midpoint[0] /= (float) poly->numverts;
	poly->midpoint[1] /= (float) poly->numverts;
	poly->midpoint[2] /= (float) poly->numverts;

	// if the max number of verts increases, reallocate storage for the surface verts
	if (poly->numverts > maxverts)
	{
		maxverts = poly->numverts;
		if (maxverts < 4) maxverts = 4;

		if (surfverts) free (surfverts);

		surfverts = (glvertex_t *) malloc (maxverts * sizeof (glvertex_t));
	}

	// set the primitive type
	if (poly->numverts < 3)
		Sys_Error ("Poly with < 3 verts");
	else if (poly->numverts == 3)
		poly->PRIMITIVE_TYPE = GL_TRIANGLES;
	else if (poly->numverts == 4)
		poly->PRIMITIVE_TYPE = GL_QUADS;
	else poly->PRIMITIVE_TYPE = GL_TRIANGLE_FAN;

	// clear entity for the surf
	surf->entity = NULL;

	surf->passframe = -1;

	vertcount += poly->numverts;
}


void R_InitSinTable (void)
{
	int i;
	float j;
	float inc = (2.0 * M_PI) / (float) SINTABLE_SIZE;

	// set up the sin table for warps
	for (i = 0, j = 0; i < SINTABLE_SIZE; i++, j += inc)
		sintable[i] = sin (j) * 8.0;

	STBitMask = SINTABLE_SIZE - 1;

	STMult = (float) SINTABLE_SIZE / (2.0 * M_PI);
}
