/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "quakedef.h"

int		gl_max_size;
int		gl_num_tmus;

// MD5 driver interface
void MD5_Checksum (unsigned char *data, int dataLen, unsigned char *checksum);
char MD5_Compare (unsigned char *digest1, unsigned char *digest2);
void MD5_Copy (unsigned char *srcdigest, unsigned char *dstdigest);

// GDI+ interface
unsigned char LoadGDIBitmap (char *searchpath, char *texturename, char *extension);
void LoadGDIData (char *identifier, unsigned char *Data, unsigned int *Palette, int Width, int Height);

gltexture_t	*gltextures = NULL;

char searchpaths[4][64];

// other textures
gltexture_t *blacktexture;
gltexture_t *particletexture;	// little dot for particles

// these aren't stored in the cache or sent through the regular loader
gltexture_t Atten1D;
gltexture_t Atten2D;

qboolean NewParticles;

byte	dottexture[8][8] =
{
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,1,1,0,0,0},
	{0,0,1,1,1,1,0,0},
	{0,0,1,1,1,1,0,0},
	{0,0,0,1,1,0,0,0},
	{0,0,0,0,0,0,0,0},
	{0,0,0,0,0,0,0,0},
};


byte NoData[16] =
{
	55, 23, 55, 23,
	23, 55, 23, 55,
	55, 23, 55, 23,
	23, 55, 23, 55
};

// basic state management for textures
int CurrentTMU = GL_TEXTURE0;
int CurrentTexture[32] = {-2};

void GL_ActivateTexture (int TMU)
{
	if (CurrentTMU == TMU) return;

	CurrentTMU = TMU;
	glActiveTexture (CurrentTMU);
}

void GL_BindTexture (int texnum)
{
	if (CurrentTexture[0] == -2)
	{
		int i;

		for (i = 0; i < 32; i++)
		{
			CurrentTexture[i] = -1;
		}
	}

	if (CurrentTexture[CurrentTMU - GL_TEXTURE0] == texnum) return;

	CurrentTexture[CurrentTMU - GL_TEXTURE0] = texnum;
	glBindTexture (GL_TEXTURE_2D, texnum);
}


void myRemoveExtension (char *FilePath)
{
	int Len = strlen (FilePath) - 1;

	for (; Len; Len--)
	{
		if (FilePath[Len] == '.')
		{
			FilePath[Len] = 0;
			return;
		}
	}
}


int MaxTexture = -1;

void GL_AllocTexture (int *texnum)
{
	glGenTextures (1, (GLuint *) texnum);

	// store the maximum texture number in use
	if (*texnum > MaxTexture) MaxTexture = *texnum;
}


/*
================
GL_FindTexture
================
*/
int GL_FindTexture (char *identifier)
{
	gltexture_t	*glt;

	for (glt = gltextures; glt; glt = glt->next)
	{
		if (!strcmp (identifier, glt->identifier))
			return glt->texnum;
	}

	return -1;
}


void WriteTGA (char *name, int w, int h, byte *data, int bpp)
{
	FILE *f;
	int i;
	byte buffer[18];

	if (!(bpp == 8 || bpp == 32)) return;
	f = fopen (name, "wb");
	if (!f) return;

	memset (buffer, 0, 18);
	buffer[2] = 2;
	buffer[12] = w & 255;
	buffer[13] = w >> 8;
	buffer[14] = h & 255;
	buffer[15] = h >> 8;
	buffer[16] = 32;
	buffer[17] = 32;

	fwrite (buffer, 18, 1, f);

	if (bpp == 32)
		fwrite (data, w * h * 4, 1, f);
	else
	{
		for (i = 0; i < w * h; i++)
		{
			// bgra palette for tga
			fwrite (&d_8to24tableBGRA[data[i]], 4, 1, f);
		}
	}

	fclose (f);
}


qboolean GL_LoadExternalTextures (char *identifier)
{
	int i;
	int j;
	char *extensions[] = {"tga", "bmp", "png", "jpg", "pcx", NULL};

	for (i = 0; i < 4; i++)
		for (j = 0; extensions[j]; j++)
			if (LoadGDIBitmap (searchpaths[i], identifier, extensions[j])) return true;

	return false;
}


/*
===============
GL_UploadTexture

Split out as we need a separate entry point for direct access
===============
*/
qboolean GL_UploadTexture (gltexture_t *glt)
{
	// bind for uploading
	GL_BindTexture (glt->texnum);

	// set hardware mipmapping (OpenGL 1.4 or higher an absolute must) 
	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, (glt->flags & TEX_MIPMAP) ? GL_TRUE : GL_FALSE);

	// apply proper filtering
	glHint (GL_GENERATE_MIPMAP_HINT, GL_NICEST);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (glt->flags & TEX_MIPMAP) ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	if (gl_max_size > 256)
	{
		// try for an external texture, but only if our card can handle it
		if (GL_LoadExternalTextures (glt->identifier)) return true;

		if (glt->flags & TEX_BSP)
		{
			// first pass included the wad name in the identifier, so now we strip that out and make another pass
			int i;

			for (i = 0; ; i++)
			{
				if (glt->identifier[i] == '/') break;
				if (glt->identifier[i] == '\\') break;
				if (!glt->identifier[i]) break;
			}

			// this should never evaluate to false as we ensured that the wadname always had a length at load time
			if (glt->identifier[i])
			{
				if (GL_LoadExternalTextures (&glt->identifier[i + 1])) return true;
			}
		}
	}

	if (!glt->data) return false;

	if (glt->flags & TEX_32BPP)
		LoadGDIData (glt->identifier, glt->data, NULL, glt->width, glt->height);
	else LoadGDIData (glt->identifier, glt->data, d_8to24tableBGRA, glt->width, glt->height);

	return true;
}


/*
================
GL_LoadTexture
================
*/
gltexture_t *GL_LoadTexture (char *identifier, int width, int height, byte *data, int flags)
{
	int i;
	gltexture_t	*glt;
	byte md5digest[16];
	byte ShotgunShells[] = {202, 6, 69, 163, 17, 112, 190, 234, 102, 56, 225, 242, 212, 175, 27, 187};

	// make an md5, preferably on the data but if there is none we need to use the identifier
	if (data)
		MD5_Checksum (data, (width * height), md5digest);
	else MD5_Checksum ((unsigned char *) identifier, strlen (identifier), md5digest);

	// fix white line at base of shotgun shells box
	// done before the cache check to prevent reloads of this texture...
	if (MD5_Compare (md5digest, ShotgunShells))
		Q_memcpy (data, data + 32 * 31, 32);

	// see if the texture is allready present
	for (glt = gltextures; glt; glt = glt->next)
	{
		if (MD5_Compare (md5digest, glt->md5digest) && !strcmp (identifier, glt->identifier))
		{
			Con_DPrintf ("Matched texture %s\n", identifier);

			glt->resident = true;

			return glt;
		}
	}

	// find an unused slot
	for (glt = gltextures; glt; glt = glt->next)
		if (!glt->texnum) break;

	if (!glt)
	{
		// didn't find a slot so allocate a new texture object and add it to the list
		glt = (gltexture_t *) malloc (sizeof (gltexture_t));

		// link up the texture
		glt->next = gltextures;
		gltextures = glt;
	}

	// store the md5
	MD5_Copy (md5digest, glt->md5digest);

	// set the name
	strcpy (glt->identifier, identifier);

	// remove the extension from the identifier
	myRemoveExtension (glt->identifier);

	// replace * with #
	// we can't just do [0] because the identifier will now be prefixed with the wad name
	for (i = 0; ; i++)
	{
		if (glt->identifier[i] == '*') glt->identifier[i] = '#';
		if (!glt->identifier[i]) break;
	}

	// copy in the data
	GL_AllocTexture (&glt->texnum);
	glt->width = width;
	glt->height = height;
	glt->bpp = 8;
	glt->data = data;

	// set flags
	glt->flags = flags;

	// upload it
	if (!GL_UploadTexture (glt))
	{
		// mark as not resident
		glt->resident = false;

		// used to identify a free texture slot
		glt->texnum = 0;

		// required so we'll never get a cache match by accident
		glt->identifier[0] = 0;

		// let god (the calling function) sort em out
		return NULL;
	}

	glt->resident = true;

	return glt;
}


void GL_DeleteTextures (void)
{
	gltexture_t	*glt;
	int numres = 0;

	for (glt = gltextures; glt; glt = glt->next)
	{
		// delete non-resident textures (only bsp textures here)
		// we'll keep the glt slot so that we can reuse it
		if (!glt->resident && (glt->flags & TEX_BSP))
		{
			// delete the texture object
			glDeleteTextures (1, (GLuint *) &glt->texnum);

			// used to identify a free texture slot
			glt->texnum = 0;

			// required so we'll never get a cache match by accident
			glt->identifier[0] = 0;

			// kill the luma pointer
			glt->luma = NULL;
		}
		else numres++;

		// always mark as non-resident in advance of the next load
		glt->resident = false;
	}

	Con_DPrintf ("%i resident textures\n", numres);
}

// size of the attenuation map
#define ATTENRADIUS 256
#define ATTENSIZE (ATTENRADIUS * 2)

void MakeAttenuationMaps (void)
{
	unsigned char *AttenTex;

	// loop control
	int x, y, z;

	GL_AllocTexture (&Atten1D.texnum);
	GL_AllocTexture (&Atten2D.texnum);

	// allocate temp memory for 2D map
	AttenTex = (unsigned char *) malloc (ATTENSIZE);

	// 1D map
	for (z = 0; z < ATTENSIZE; z++)
	{
		int TheRZ = z - ATTENRADIUS;
		float TheZResult;
		int TheFinalResult;
		unsigned char OutResult;

		// make sure that the edges are black
		if (z == 0 || z == (ATTENSIZE - 1))
		{
			AttenTex[z] = 0;
			continue;
		}

		// divide Z by ATTENRADIUS and square it
		TheZResult = ((float) TheRZ / (float) ATTENRADIUS);
		TheZResult *= TheZResult;

		// for the final result, multiply by 255 for an intensity value.
		// we need to make this an int (a short would be better, but - hey! - memory
		// ain't an issue here) cos if it's > 255 or < 0 and unsigned char would wrap
		TheFinalResult = (int) (TheZResult * 255);

		// clamp in 0 to 255 range (is this necessary?  probly not, but better to be safe)
		if (TheFinalResult < 0) TheFinalResult = 0;
		if (TheFinalResult > 255) TheFinalResult = 255;

			TheFinalResult = 255 - TheFinalResult;
			TheFinalResult *= TheFinalResult;
			TheFinalResult /= 255;
			TheFinalResult *= TheFinalResult;
			TheFinalResult /= 255;

		// store in an unsigned char now
		OutResult = (unsigned char) TheFinalResult;

		// store the base intensity in the alpha channel (we'll write it to the rgba channels at update time...)
		AttenTex[z] = OutResult;
	}

	// switch to 1D
	glDisable (GL_TEXTURE_2D);
	glEnable (GL_TEXTURE_1D);

	// upload to OpenGL
	// we can't send this through GL_BindTexture as that explicitly binds as a 2D object, so we have to do it by hand
	// for this pass, then send the 2D one through GL_BindTexture to ensure that tracking is kept up to date
	glBindTexture (GL_TEXTURE_1D, Atten1D.texnum);

	glTexParameteri (GL_TEXTURE_1D, GL_GENERATE_MIPMAP, GL_TRUE);

	glTexImage1D (GL_TEXTURE_1D, 
				  0, 
				  GL_LUMINANCE, 
				  ATTENSIZE, 
				  0, 
				  GL_LUMINANCE, 
				  GL_UNSIGNED_BYTE, 
				  &AttenTex[0]);

	glTexParameteri (GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri (GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);

	// switch back to 2D
	glDisable (GL_TEXTURE_1D);
	glEnable (GL_TEXTURE_2D);

	// allocate temp memory for 2D map
	free (AttenTex);
	AttenTex = (unsigned char *) malloc (ATTENSIZE * ATTENSIZE);

	// 2D map
	for (y = 0, z = 0; y < ATTENSIZE; y++)
	{
		for (x = 0; x < ATTENSIZE; x++, z++)
		{
			// find x and y as a -ATTENRADIUS to ATTENRADIUS scale
			int TheRX = x - ATTENRADIUS;
			int TheRY = y - ATTENRADIUS;
			float TheXResult;
			float TheYResult;
			int TheFinalResult;
			unsigned char OutResult;

			// make sure the edges are black
			if (x == 0 || x == (ATTENSIZE - 1) || y == 0 || y == (ATTENSIZE - 1))
			{
				AttenTex[z] = 0;
				continue;
			}

			// divide X by ATTENRADIUS and square it
			TheXResult = ((float) TheRX / (float) ATTENRADIUS);
			TheXResult *= TheXResult;

			// divide Y by ATTENRADIUS and square it
			TheYResult = ((float) TheRY / (float) ATTENRADIUS);
			TheYResult *= TheYResult;

			// add them for the final result, and multiply by 255 for an intensity value
			// we need to make this an int (a short would be better, but - hey! - memory
			// ain't an issue here) cos if it's > 255 or < 0 and unsigned char would wrap
			TheFinalResult = (int) ((TheXResult + TheYResult) * 255);

			// clamp in 0 to 255 range (is this necessary?  probly not, but better to be safe)
			if (TheFinalResult < 0) TheFinalResult = 0;
			if (TheFinalResult > 255) TheFinalResult = 255;

			TheFinalResult = 255 - TheFinalResult;
			TheFinalResult *= TheFinalResult;
			TheFinalResult /= 255;
			TheFinalResult *= TheFinalResult;
			TheFinalResult /= 255;

			// store in an unsigned char now
			OutResult = (unsigned char) TheFinalResult;

			// write to texture
			AttenTex[z] = OutResult;
		}
	}

	// upload to OpenGL
	GL_BindTexture (0);
	GL_BindTexture (Atten2D.texnum);

	glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);

	glTexImage2D (GL_TEXTURE_2D, 
				  0, 
				  GL_LUMINANCE, 
				  ATTENSIZE, 
				  ATTENSIZE, 
				  0, 
				  GL_LUMINANCE, 
				  GL_UNSIGNED_BYTE, 
				  &AttenTex[0]);

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	// don't need you anymore
	free (AttenTex);
}


/*
==================
R_InitTextures
==================
*/
void R_InitTextures (void)
{
	int		x,y, m;
	byte	*dest;
	byte data[4];

	// create a simple checkerboard texture for the default
	r_notexture_mip = Hunk_AllocName (sizeof(texture_t) + 16*16+8*8+4*4+2*2, "notexture");

	r_notexture_mip->width = r_notexture_mip->height = 16;
	r_notexture_mip->offsets[0] = sizeof(texture_t);
	r_notexture_mip->offsets[1] = r_notexture_mip->offsets[0] + 16*16;
	r_notexture_mip->offsets[2] = r_notexture_mip->offsets[1] + 8*8;
	r_notexture_mip->offsets[3] = r_notexture_mip->offsets[2] + 4*4;

	for (m=0 ; m<4 ; m++)
	{
		dest = (byte *)r_notexture_mip + r_notexture_mip->offsets[m];
		for (y=0 ; y< (16>>m) ; y++)
			for (x=0 ; x< (16>>m) ; x++)
			{
				if (  (y< (8>>m) ) ^ (x< (8>>m) ) )
					*dest++ = 0;
				else
					*dest++ = 0xff;
			}
	}

	r_notexture_mip->gl_texture = GL_LoadTexture ("notexture", 16, 16, (byte *) (r_notexture_mip + 1), TEX_MIPMAP);

	// black texture
	data[0] = data[1] = data[2] = data[3] = 0;

	blacktexture = GL_LoadTexture ("gfx/blacktexture", 2, 2, data, TEX_MIPMAP);

	// set up search paths
	sprintf (searchpaths[0], "%s\\textures", com_gamedir);
	sprintf (searchpaths[1], "id1\\textures");
	sprintf (searchpaths[2], "%s", com_gamedir);
	sprintf (searchpaths[3], "id1");

	//MakeAttenuationMaps ();
}


qboolean R_InitNewParticles (void);

void R_InitParticleTextures (void)
{
	int		x,y;
	byte	data[8][8][4];

	for (x = 0; x < 8; x++)
	{
		for (y = 0; y < 8; y++)
		{
			data[y][x][0] = 255;
			data[y][x][1] = 255;
			data[y][x][2] = 255;
			data[y][x][3] = dottexture[x][y] * 255;
		}
	}

	particletexture = GL_LoadTexture ("particles/particle", 8, 8, data, (TEX_32BPP | TEX_MIPMAP));

	NewParticles = R_InitNewParticles ();

	MakeAttenuationMaps ();
}

