/*
====================================================================================================================================

		SOFTWARE MATRIX CALCULATIONS

	in order to reduce the burden on GPUs, and in particular as this engine is specifically targetted at machines with low end
	GPUs but relatively high-end CPUs, and let them get on with their primary role of rendering triangles, texturing and blending,
	we do all of our matrix calculations in software explicitly.  shame on OpenGL for not allowing software vertex processing.

	the functions in here pretty much mimic the similarly named OpenGL functions.  it's not exhaustive - only the modelview matrix
	is emulated here, and not everything has an equivalent.

	in all cases, it's viewed as the responsibility of the calling code to ensure that the stack is not pushed or popped too far,
	and to ensure that any external buffers used have sufficient space to hold the data requested from them or pushed out to them.
	in some cases a silent fail is written, in others you can be prepared for the dreaded "behaviour is undefined".  the code is 
	intended to address a specific requirement in a controlled environment rather than be generally robust or managed in any way.

====================================================================================================================================
*/

#include <math.h>


// define our basic 4x4 matrix
typedef float SW_Matrix[16];

// create a modelview matrix stack 32 deep
#define MAX_MODELVIEW_MATRIX_STACK		32

SW_Matrix ModelViewMatrixStack[MAX_MODELVIEW_MATRIX_STACK];

// current level in the stack is 0
int CurrentModelViewMatrix = 0;

// go to the next level in the stack and copy the previous up to it
void swPushMatrix (void)
{
	// it's the presponsibility of the caller, not the matrix library, to manage the stack depth, so fail silently
	// if we try to push it too far
	if (CurrentModelViewMatrix >= (MAX_MODELVIEW_MATRIX_STACK - 1)) return;

	// go to the next in the stack
	CurrentModelViewMatrix++;

	// copy up
	ModelViewMatrixStack[CurrentModelViewMatrix][0x0] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x0];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x1] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x1];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x2] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x2];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x3] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x3];

	ModelViewMatrixStack[CurrentModelViewMatrix][0x4] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x4];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x5] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x5];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x6] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x6];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x7] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x7];

	ModelViewMatrixStack[CurrentModelViewMatrix][0x8] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x8];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x9] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0x9];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xa] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0xa];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xb] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0xb];

	ModelViewMatrixStack[CurrentModelViewMatrix][0xc] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0xc];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xd] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0xd];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xe] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0xe];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xf] = ModelViewMatrixStack[CurrentModelViewMatrix - 1][0xf];
}


// jump back to the previous level on the stack
void swPopMatrix (void)
{
	// it's the presponsibility of the caller, not the matrix library, to manage the stack depth, so fail silently
	// if we try to push it too far
	if (!CurrentModelViewMatrix) return;

	// go to the previous in the stack
	CurrentModelViewMatrix--;
}


// load the specified matrix on to the stack
void swLoadMatrixf (float *Matrix)
{
	// copy in
	ModelViewMatrixStack[CurrentModelViewMatrix][0x0] = Matrix[0x0];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x1] = Matrix[0x1];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x2] = Matrix[0x2];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x3] = Matrix[0x3];

	ModelViewMatrixStack[CurrentModelViewMatrix][0x4] = Matrix[0x4];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x5] = Matrix[0x5];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x6] = Matrix[0x6];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x7] = Matrix[0x7];

	ModelViewMatrixStack[CurrentModelViewMatrix][0x8] = Matrix[0x8];
	ModelViewMatrixStack[CurrentModelViewMatrix][0x9] = Matrix[0x9];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xa] = Matrix[0xa];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xb] = Matrix[0xb];

	ModelViewMatrixStack[CurrentModelViewMatrix][0xc] = Matrix[0xc];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xd] = Matrix[0xd];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xe] = Matrix[0xe];
	ModelViewMatrixStack[CurrentModelViewMatrix][0xf] = Matrix[0xf];
}


// this one's new - simply copy the current stack level back to the specified array
void swGetMatrixf (float *Matrix)
{
	// copy back
	Matrix[0x0] = ModelViewMatrixStack[CurrentModelViewMatrix][0x0];
	Matrix[0x1] = ModelViewMatrixStack[CurrentModelViewMatrix][0x1];
	Matrix[0x2] = ModelViewMatrixStack[CurrentModelViewMatrix][0x2];
	Matrix[0x3] = ModelViewMatrixStack[CurrentModelViewMatrix][0x3];

	Matrix[0x4] = ModelViewMatrixStack[CurrentModelViewMatrix][0x4];
	Matrix[0x5] = ModelViewMatrixStack[CurrentModelViewMatrix][0x5];
	Matrix[0x6] = ModelViewMatrixStack[CurrentModelViewMatrix][0x6];
	Matrix[0x7] = ModelViewMatrixStack[CurrentModelViewMatrix][0x7];

	Matrix[0x8] = ModelViewMatrixStack[CurrentModelViewMatrix][0x8];
	Matrix[0x9] = ModelViewMatrixStack[CurrentModelViewMatrix][0x9];
	Matrix[0xa] = ModelViewMatrixStack[CurrentModelViewMatrix][0xa];
	Matrix[0xb] = ModelViewMatrixStack[CurrentModelViewMatrix][0xb];

	Matrix[0xc] = ModelViewMatrixStack[CurrentModelViewMatrix][0xc];
	Matrix[0xd] = ModelViewMatrixStack[CurrentModelViewMatrix][0xd];
	Matrix[0xe] = ModelViewMatrixStack[CurrentModelViewMatrix][0xe];
	Matrix[0xf] = ModelViewMatrixStack[CurrentModelViewMatrix][0xf];
}


// load the identity matrix into the current stack position
void swLoadIdentity (void)
{
	ModelViewMatrixStack[CurrentModelViewMatrix][0x0] = 1;
	ModelViewMatrixStack[CurrentModelViewMatrix][0x1] = 0;
	ModelViewMatrixStack[CurrentModelViewMatrix][0x2] = 0;
	ModelViewMatrixStack[CurrentModelViewMatrix][0x3] = 0;

	ModelViewMatrixStack[CurrentModelViewMatrix][0x4] = 0;
	ModelViewMatrixStack[CurrentModelViewMatrix][0x5] = 1;
	ModelViewMatrixStack[CurrentModelViewMatrix][0x6] = 0;
	ModelViewMatrixStack[CurrentModelViewMatrix][0x7] = 0;

	ModelViewMatrixStack[CurrentModelViewMatrix][0x8] = 0;
	ModelViewMatrixStack[CurrentModelViewMatrix][0x9] = 0;
	ModelViewMatrixStack[CurrentModelViewMatrix][0xa] = 1;
	ModelViewMatrixStack[CurrentModelViewMatrix][0xb] = 0;

	ModelViewMatrixStack[CurrentModelViewMatrix][0xc] = 0;
	ModelViewMatrixStack[CurrentModelViewMatrix][0xd] = 0;
	ModelViewMatrixStack[CurrentModelViewMatrix][0xe] = 0;
	ModelViewMatrixStack[CurrentModelViewMatrix][0xf] = 1;
}


// multiplies the current modelview matrix by the specified one
// cribbed from mesa
#define A(row,col)  a[(col << 2) + row]
#define B(row,col)  b[(col << 2) + row]
#define P(row,col)  product[(col << 2) + row]

static void matmul4 (float *product, const float *a, const float *b)
{
	int i;

	for (i = 0; i < 4; i++)
	{
		const float ai0 = A (i, 0), ai1= A (i, 1), ai2 = A (i, 2), ai3 = A (i, 3);

		P (i, 0) = ai0 * B (0, 0) + ai1 * B (1, 0) + ai2 * B (2, 0) + ai3 * B (3, 0);
		P (i, 1) = ai0 * B (0, 1) + ai1 * B (1, 1) + ai2 * B (2, 1) + ai3 * B (3, 1);
		P (i, 2) = ai0 * B (0, 2) + ai1 * B (1, 2) + ai2 * B (2, 2) + ai3 * B (3, 2);
		P (i, 3) = ai0 * B (0, 3) + ai1 * B (1, 3) + ai2 * B (2, 3) + ai3 * B (3, 3);
	}
}

void swMultMatrixf (float *Matrix)
{
	matmul4 (ModelViewMatrixStack[CurrentModelViewMatrix], ModelViewMatrixStack[CurrentModelViewMatrix], Matrix);
}


void swTranslatef (float x, float y, float z)
{
	SW_Matrix TranslateMatrix = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};

	// compose the translation matrix
	TranslateMatrix[0xc] = x;
	TranslateMatrix[0xd] = y;
	TranslateMatrix[0xe] = z;

	// multiply it by the current matrix
	swMultMatrixf (TranslateMatrix);
}


void swRotatef (float ang, float x, float y, float z)
{
	float RotateMatrix[16] = {1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
	float xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c;
	float s = (float) sin (-ang * 0.017453292519943295769236907684886f);
	float c = (float) cos (-ang * 0.017453292519943295769236907684886f);
	float mag;

	// normalize the input vector
	mag = (float) sqrt (x * x + y * y + z * z);
	x /= mag;
	y /= mag;
	z /= mag;

	// get some stuff
	xx = x * x;
	yy = y * y;
	zz = z * z;
	xy = x * y;
	yz = y * z;
	zx = z * x;
	xs = x * s;
	ys = y * s;
	zs = z * s;
	one_c = 1.0F - c;

	// compose the rotation matrix
	RotateMatrix[0] = (one_c * xx) + c;
	RotateMatrix[1] = (one_c * xy) - zs;
	RotateMatrix[2] = (one_c * zx) + ys;

	RotateMatrix[4] = (one_c * xy) + zs;
	RotateMatrix[5] = (one_c * yy) + c;
	RotateMatrix[6] = (one_c * yz) - xs;

	RotateMatrix[8] = (one_c * zx) - ys;
	RotateMatrix[9] = (one_c * yz) + xs;
	RotateMatrix[10] = (one_c * zz) + c;

	// multiply
	swMultMatrixf (RotateMatrix);
}
