/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// alias model drawing

#include "quakedef.h"
#include "gl_mtex.h"

extern int gl_fogenabled;
extern cvar_t gl_aliasglow;
extern cvar_t gl_aliastrans;
extern cvar_t gl_aliaslightlerp;

extern int gl_useclight;

int R_LightPoint (vec3_t p);
int *R_LightPointColour (vec3_t p);

int fullbright_skin;
extern cvar_t gl_fullbright_alias;
extern float lerp_mody;

#define LIGHT_DIV 124.0

// used for testing glRotatef values for the viewmodel
#define VIEW_TEST

#ifdef VIEW_TEST
cvar_t view_hand_x = {"view_hand_x", "0.0"};
cvar_t view_hand_y = {"view_hand_y", "0.0"};
cvar_t view_hand_z = {"view_hand_z", "0.0"};
#endif


void R_InitAliasMod (void)
{
#ifdef VIEW_TEST
	Cvar_RegisterVariable (&view_hand_x);
	Cvar_RegisterVariable (&view_hand_y);
	Cvar_RegisterVariable (&view_hand_z);
#endif
}

/*
=================
R_CullBox

Returns true if the box is completely outside the frustom
=================
*/
qboolean R_CullBox (vec3_t mins, vec3_t maxs)
{
	int		i;

	for (i=0 ; i<4 ; i++)
		if (BoxOnPlaneSide (mins, maxs, &frustum[i]) == 2)
			return true;

	return false;
}


void R_RotateForEntity (entity_t *e)
{
    glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);

    glRotatef (e->angles[1],  0, 0, 1);
    glRotatef (-e->angles[0],  0, 1, 0);
    glRotatef (e->angles[2],  1, 0, 0);
}


/*
=============
R_BlendedRotateForEntity

fenix@io.com: model transform interpolation
=============
*/
void R_BlendedRotateForEntity (entity_t *e, int shadow)
{
	float timepassed;
	float blend;
	vec3_t d;
	int i;

	// positional interpolation
	timepassed = (realtime - e->translate_start_time) * lerp_mody;

	if (e->translate_start_time == 0 || timepassed > 1)
	{
		e->translate_start_time = realtime;
		VectorCopy (e->origin, e->origin1);
		VectorCopy (e->origin, e->origin2);
	}

	if (!VectorCompare (e->origin, e->origin2))
	{
		e->translate_start_time = realtime;
		VectorCopy (e->origin2, e->origin1);
		VectorCopy (e->origin,  e->origin2);
		blend = 0;
	}
	else
	{
		blend =  timepassed / 0.1;

		if (cl.paused || blend > 1) blend = 1;
	}

	VectorSubtract (e->origin2, e->origin1, d);

	glTranslatef (e->origin1[0] + (blend * d[0]),
		e->origin1[1] + (blend * d[1]),
		e->origin1[2] + (blend * d[2]));

	// orientation interpolation (Euler angles, yuck!)
	timepassed = (realtime - e->rotate_start_time) * lerp_mody;

	if (e->rotate_start_time == 0 || timepassed > 1)
	{
		e->rotate_start_time = realtime;
		VectorCopy (e->angles, e->angles1);
		VectorCopy (e->angles, e->angles2);
	}

	if (!VectorCompare (e->angles, e->angles2))
	{
		e->rotate_start_time = realtime;
		VectorCopy (e->angles2, e->angles1);
		VectorCopy (e->angles,  e->angles2);
		blend = 0;
	}
	else
	{
		blend = timepassed / 0.1;
 
		if (cl.paused || blend > 1) blend = 1;
	}

	VectorSubtract (e->angles2, e->angles1, d);

	// always interpolate along the shortest path
	for (i = 0; i < 3; i++) 
	{
		if (d[i] > 180) 
			d[i] -= 360;
		else if (d[i] < -180) d[i] += 360;
	}

	glRotatef (e->angles1[1] + ( blend * d[1]),  0, 0, 1);

	if (!shadow)
	{
		glRotatef (-e->angles1[0] + (-blend * d[0]),  0, 1, 0);
		glRotatef ( e->angles1[2] + ( blend * d[2]),  1, 0, 0);
	}
}


#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] = 
{
	#include "anorms.h"
};

vec3_t	shadevector;
float	shadelight, ambientlight;
float	shadelightcolour[4];


// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] = 
#include "anorm_dots.h"
;

float	*shadedots = r_avertexnormal_dots[0];

// light lerping by pox
float *shadedots2 = r_avertexnormal_dots[0];
float lightlerpoffset;

int	lastposenum;
int lastposenum0;

extern int notexture_num;


float Calc_LightLerp (float l1, float l2)
{
	float l, diff;

	if (l1 != l2)
	{
		if (l1 > l2)
			diff = (l1 - l2) * -1;
		else diff = l2 - l1;

		diff *= lightlerpoffset;
		l = l1 + diff;
	}
	else l = l1;

	// add contrast - this is a nasty hack from Pox's tut which simulates specular lighting
	// mainly introduced cos at the time he didn't know the norms.  however, my engine does
	// calculate the norms at model load time (although it doesn't really use them) so i
	// should be able to do proper specular lighting... if only i knew how :-(
	l *= l;

	return l;
}


/*
=============
GL_DrawAliasFrame
=============
*/
void GL_DrawAliasFrame (aliashdr_t *paliashdr, int posenum, entity_t *e)
{
	float 	l;
	trivertx_t	*verts;
	int		*order;
	int		count;
	float	trans_lev;
	vec3_t	drawverts;
	float l1, l2;

	lastposenum = posenum;

	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	if (paliashdr->trans_level < 1.0 && gl_aliastrans.value)
	{
		glEnable(GL_BLEND);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		trans_lev = paliashdr->trans_level;
	}
	else trans_lev = 1.0;

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;

		if (!count)
			break;		// done

		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
			order += 2;

			drawverts[0] = verts->v[0];
			drawverts[1] = verts->v[1];
			drawverts[2] = verts->v[2];

			if (gl_useclight && !gl_aliaslightlerp.value)
			{
				// coloured light, no lerping
				l = shadedots[verts->lightnormalindex] * shadelightcolour[3];

				// weird, huh?  let's me have contrast on the models without lerping
				l = Calc_LightLerp (l, l);

				glColor4f (l * shadelightcolour[0], l * shadelightcolour[1], 
					l * shadelightcolour[2], trans_lev);
			}
			else if (!gl_useclight && !gl_aliaslightlerp.value)
			{
				// white light, no lerping
				l = shadedots[verts->lightnormalindex] * shadelight;

				// weird, huh?  let's me have contrast on the models without lerping
				l = Calc_LightLerp (l, l);

				glColor4f (l, l, l, trans_lev);
			}
			else if (gl_useclight && gl_aliaslightlerp.value)
			{
				// coloured light, lerping
				l1 = shadedots[verts->lightnormalindex] * shadelightcolour[3];
				l2 = shadedots2[verts->lightnormalindex] * shadelightcolour[3];

				l = Calc_LightLerp (l1, l2);

				glColor4f (l * shadelightcolour[0], l * shadelightcolour[1], 
					l * shadelightcolour[2], trans_lev);
			}
			else
			{
				// white light, lerping
				l1 = shadedots[verts->lightnormalindex] * shadelight;
				l2 = shadedots2[verts->lightnormalindex] * shadelight;

				l = Calc_LightLerp (l1, l2);

				glColor4f (l, l, l, trans_lev);
			}

			glVertex3fv (&drawverts[0]);
			verts++;
		} while (--count);

		glEnd ();
	}

	if (paliashdr->trans_level < 1.0 && gl_aliastrans.value)
	{
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glDisable (GL_BLEND);
	}

	glColor4f (1,1,1,1);

	if (fullbright_skin == -1 || !gl_fullbright_alias.value)
		return;

	if (gl_fogenabled) glDisable (GL_FOG);

	// draw it again with a fullbright skin - fixme - translucency on this???
	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	glBindTexture (GL_TEXTURE_2D, fullbright_skin);
	oldtexture = fullbright_skin;

	glEnable(GL_BLEND);

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;

		if (!count)
			break;		// done

		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
			order += 2;

			glVertex3f (verts->v[0], verts->v[1], verts->v[2]);
			verts++;
		} while (--count);

		glEnd ();
	}

	glDisable (GL_BLEND);

	if (gl_fogenabled) glEnable (GL_FOG);
}


/*
=============
GL_DrawAliasBlendedFrame

fenix@io.com: model animation interpolation
=============
*/
void GL_DrawAliasBlendedFrame (aliashdr_t *paliashdr, int pose1, int pose2, float blend, entity_t *e)
{
	float       l;
	trivertx_t* verts1;
	trivertx_t* verts2;
	int*        order;
	int         count;
	vec3_t      d;
	float		trans_lev;
	float		light_blend;

	vec3_t v;
	float d2, l1, l2;

	lastposenum0 = pose1;
	lastposenum  = pose2;
        
	verts1  = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts2  = verts1;

	verts1 += pose1 * paliashdr->poseverts;
	verts2 += pose2 * paliashdr->poseverts;

	order = (int *)((byte *)paliashdr + paliashdr->commands);
            
	if (paliashdr->trans_level < 1.0 && gl_aliastrans.value)
	{
		glEnable(GL_BLEND);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		trans_lev = paliashdr->trans_level;
	}
	else trans_lev = 1.0;

	for (;;)
	{
		// get the vertex count and primitive type
		count = *order++;

		if (!count) break;
        
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
			order += 2;

			// normals and vertexes come from the frame list
			// blend the light intensity from the two frames together
			if (!gl_aliaslightlerp.value)
			{
				light_blend = shadedots[verts2->lightnormalindex] - 
					shadedots[verts1->lightnormalindex];
			}

			VectorSubtract(verts2->v, verts1->v, d);

			// blend the vertex positions from each frame together
			v[0] = verts1->v[0] + (blend * d[0]);
			v[1] = verts1->v[1] + (blend * d[1]);
			v[2] = verts1->v[2] + (blend * d[2]);

			// fx flags are in there but not used any more
			if (gl_useclight && !gl_aliaslightlerp.value)
			{
				// coloured light, no lerping
				l = shadelightcolour[3] * (shadedots[verts1->lightnormalindex] + 
					(blend * light_blend));

				// weird, huh?  let's me have contrast on the models without lerping
				l = Calc_LightLerp (l, l);

				glColor4f (l * shadelightcolour[0], l * shadelightcolour[1], 
					l * shadelightcolour[2], trans_lev);
			}
			else if (!gl_useclight && !gl_aliaslightlerp.value)
			{
				// white light, no lerping
				l = shadelight * (shadedots[verts1->lightnormalindex] + (blend * light_blend));

				// weird, huh?  let's me have contrast on the models without lerping
				l = Calc_LightLerp (l, l);

				glColor4f (l, l, l, trans_lev);
			}
			else if (gl_useclight && gl_aliaslightlerp.value)
			{
				// coloured light, lerping
				d[0] = shadedots[verts2->lightnormalindex] - shadedots[verts1->lightnormalindex];
				d2 = shadedots2[verts2->lightnormalindex] - shadedots2[verts1->lightnormalindex];

				l1 = shadelightcolour[3] * (shadedots[verts1->lightnormalindex] + (blend * d[0]));
				l2 = shadelightcolour[3] * (shadedots2[verts1->lightnormalindex] + (blend * d2));

				l = Calc_LightLerp (l1, l2);

				glColor4f (l * shadelightcolour[0], l * shadelightcolour[1], 
					l * shadelightcolour[2], trans_lev);
			}
			else
			{
				// white light, lerping
				d[0] = shadedots[verts2->lightnormalindex] - shadedots[verts1->lightnormalindex];
				d2 = shadedots2[verts2->lightnormalindex] - shadedots2[verts1->lightnormalindex];

				l1 = shadelight * (shadedots[verts1->lightnormalindex] + (blend * d[0]));
				l2 = shadelight * (shadedots2[verts1->lightnormalindex] + (blend * d2));

				l = Calc_LightLerp (l1, l2);

				glColor4f (l, l, l, trans_lev);
			}

			glVertex3fv (v);

			verts1++;
			verts2++;
		} while (--count);

		glEnd ();
	}

	if (paliashdr->trans_level < 1.0 && gl_aliastrans.value)
	{
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glDisable (GL_BLEND);
	}

	glColor4f (1,1,1,1);

	if (fullbright_skin == -1 || !gl_fullbright_alias.value)
		return;

	if (gl_fogenabled) glDisable (GL_FOG);

	// draw it again with a fullbright skin - fixme: translucency on this?
	verts1  = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts2  = verts1;

	verts1 += pose1 * paliashdr->poseverts;
	verts2 += pose2 * paliashdr->poseverts;

	order = (int *)((byte *)paliashdr + paliashdr->commands);
            
	glBindTexture (GL_TEXTURE_2D, fullbright_skin);
	oldtexture = fullbright_skin;

	glEnable(GL_BLEND);

	for (;;)
	{
		// get the vertex count and primitive type
		count = *order++;

		if (!count) break;
        
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			glTexCoord2f (((float *)order)[0], ((float *)order)[1]);
			order += 2;

			VectorSubtract(verts2->v, verts1->v, d);

			// blend the vertex positions from each frame together
			v[0] = verts1->v[0] + (blend * d[0]);
			v[1] = verts1->v[1] + (blend * d[1]);
			v[2] = verts1->v[2] + (blend * d[2]);

			glVertex3fv (v);

			verts1++;
			verts2++;
		} while (--count);

		glEnd ();
	}

	glDisable (GL_BLEND);

	if (gl_fogenabled) glEnable (GL_FOG);
}


/*
=============
GL_DrawAliasShadow
=============
*/
extern	vec3_t			lightspot;
qboolean SV_RecursiveHullCheck (hull_t *hull, int num, float p1f, float p2f, vec3_t p1, vec3_t p2, trace_t *trace);

void GL_DrawAliasShadow (aliashdr_t *paliashdr, int posenum)
{
	trivertx_t	*verts;
	int		*order;
	vec3_t	point;
	float	height, lheight;
	int		count;

	// muff@yakko.globalnet.co.uk - 17 Mar 2000
	trace_t		downtrace;
	vec3_t		downmove;
	float		s1,c1; //this will store the sin and cos values as they are used so heavily
	//end of muff

	lheight = currententity->origin[2] - lightspot[2];

	height = 0;
	verts = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts += posenum * paliashdr->poseverts;
	order = (int *)((byte *)paliashdr + paliashdr->commands);

	height = -lheight ;//+ 1.0; //was 1.0 - muff

	// muff@yakko.globalnet.co.uk - 17 Mar 2000
	// better shadowing, now takes angle of ground into account	
	// cast a traceline into the floor directly below the player
	// and gets normals from this
	VectorCopy (currententity->origin, downmove);
	downmove[2] = downmove[2] - 4096;
	memset (&downtrace, 0, sizeof(downtrace));

	SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, currententity->origin, downmove, &downtrace);

	// calculate the all important angles to keep speed up
	s1 = sin( currententity->angles[1] / 180*M_PI);
	c1 = cos( currententity->angles[1] / 180*M_PI);
	// end of muff

	while (1)
	{
		// get the vertex count and primitive type
		count = *order++;
		if (!count)
			break;		// done
		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			// (skipped for shadows) glTexCoord2fv ((float *)order);
			order += 2;

			// normals and vertexes come from the frame list
			point[0] = verts->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
			point[1] = verts->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
			point[2] = verts->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];

			point[0] -= shadevector[0]*(point[0]); 
			point[1] -= shadevector[1]*(point[1]);
			point[2] -= shadevector[2]*(point[2]);

			//drop it down to floor
			point[2] = point[2] - (currententity->origin[2] - downtrace.endpos[2]) ;

			// now adjust the point with respect to all the normals of the tracepoint
			// I'm probably missing an easy piece of maths to do this
			// it reads messy, but works ;)
			point[2] +=	((point[1] * (s1 * downtrace.plane.normal[0])) -
					     (point[0] * (c1 * downtrace.plane.normal[0])) -
						 (point[0] * (s1 * downtrace.plane.normal[1])) - 
						 (point[1] * (c1 * downtrace.plane.normal[1]))) 
						 +  ((1.0 - downtrace.plane.normal[2])*20) + 0.2 ;
	        //end of muff

			glVertex3fv (point);

			verts++;
		} while (--count);

		glEnd ();
	}	
}


void GL_DrawAliasBlendedShadow (aliashdr_t *paliashdr, int pose1, int pose2, entity_t* e)
{
	trivertx_t* verts1;
	trivertx_t* verts2;
	int*        order;
	vec3_t      point1;
	vec3_t      point2;
	vec3_t      d;
	float       height;
	float       lheight;
	int         count;
	float       blend;

	// muff@yakko.globalnet.co.uk - 17 Mar 2000
	trace_t		downtrace;
	vec3_t		downmove;
	float		s1,c1; //this will store the sin and cos values as they are used so heavily
	//end of muff

	blend = (realtime - e->frame_start_time) / e->frame_interval;

	blend = blend * lerp_mody;

	if (blend > 1) blend = 1;

	lheight = e->origin[2] - lightspot[2];
	height  = -lheight;// + 1.0; //was 1.0 - muff

	VectorCopy (currententity->origin, downmove);
	downmove[2] = downmove[2] - 4096;
	memset (&downtrace, 0, sizeof(downtrace));

	// only checks the worldmodel, so shadows are not being cast on brush models (plats, etc)
	SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, currententity->origin, downmove, &downtrace);

	// calculate the all important angles
	s1 = sin( currententity->angles[1]/180*M_PI);
	c1 = cos( currententity->angles[1]/180*M_PI);
	// end of muff

	verts1 = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts2 = verts1;

	verts1 += pose1 * paliashdr->poseverts;
	verts2 += pose2 * paliashdr->poseverts;

	order = (int *)((byte *)paliashdr + paliashdr->commands);

	for (;;)
	{
		// get the vertex count and primitive type
		count = *order++;

		if (!count) break;

		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			order += 2;

			point1[0] = verts1->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
			point1[1] = verts1->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
			point1[2] = verts1->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];
        
			point1[0] -= shadevector[0]*(point1[2]+lheight);
			point1[1] -= shadevector[1]*(point1[2]+lheight);

			point2[0] = verts2->v[0] * paliashdr->scale[0] + paliashdr->scale_origin[0];
			point2[1] = verts2->v[1] * paliashdr->scale[1] + paliashdr->scale_origin[1];
			point2[2] = verts2->v[2] * paliashdr->scale[2] + paliashdr->scale_origin[2];

			point2[0] -= shadevector[0]*(point2[2]+lheight); 
			point2[1] -= shadevector[1]*(point2[2]+lheight); 

			VectorSubtract(point2, point1, d);


			// muff@yakko.globalnet.co.uk - 17 Mar 2000
			point1[0] = point1[0] + (blend * d[0]);  //+lheight); //muff
			point1[1] = point1[1] + (blend * d[1]);  //+lheight); //muff
			point1[2] = point1[2] + (blend * d[2]);  //+lheight); //muff

			//drop it down to floor
			point1[2] =  - (currententity->origin[2] - downtrace.endpos[2]) ;

			//now move the z-coordinate as appropriate
			point1[2] += ((point1[1] * (s1 * downtrace.plane.normal[0])) -
				          (point1[0] * (c1 * downtrace.plane.normal[0])) -
					      (point1[0] * (s1 * downtrace.plane.normal[1])) - 
					      (point1[1] * (c1 * downtrace.plane.normal[1]))) 
						  + ((1.0 - downtrace.plane.normal[2])*20) + 0.2 ;

			// now load the point we have just calculated
			glVertex3fv (point1);

			verts1++;
			verts2++;
		} while (--count);

		glEnd ();
	}       
}


/*
=================
R_SetupAliasFrame

=================
*/
void R_SetupAliasFrame (int frame, aliashdr_t *paliashdr, entity_t *e)
{
	int				pose, numposes;
	float			interval;

	if ((frame >= paliashdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
		frame = 0;
	}

	pose = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		interval = paliashdr->frames[frame].interval;
		pose += (int)(cl.time / interval) % numposes;
	}

	// save this out for later
	e->pose1 = pose;

	GL_DrawAliasFrame (paliashdr, pose, e);
}


/*
=================
R_SetupAliasBlendedFrame

fenix@io.com: model animation interpolation
=================
*/
void R_SetupAliasBlendedFrame (int frame, aliashdr_t *paliashdr, entity_t* e)
{
	int   pose;
	int   numposes;
	float blend;

	if ((frame >= paliashdr->numframes) || (frame < 0))
	{
		Con_DPrintf ("R_AliasSetupFrame: no such frame %d\n", frame);
		frame = 0;
	}

	pose = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		e->frame_interval = paliashdr->frames[frame].interval;
		pose += (int)(cl.time / e->frame_interval) % numposes;
	}
	else 
	{
		/* One tenth of a second is a good for most Quake animations.
		If the nextthink is longer then the animation is usually meant to pause
		(e.g. check out the shambler magic animation in shambler.qc).  If its
		shorter then things will still be smoothed partly, and the jumps will be
		less noticable because of the shorter time.  So, this is probably a good
		assumption. */
		e->frame_interval = 0.1;
	}

	if (e->pose2 != pose)
	{
		e->frame_start_time = realtime;
		e->pose1 = e->pose2;
		e->pose2 = pose;
		blend = 0;
	}
	else blend = (realtime - e->frame_start_time) / e->frame_interval;

	blend = blend * lerp_mody;

	// wierd things start happening if blend passes 1
	if (cl.paused || blend > 1) blend = 1;

	// mh - super nailgun bug fix... this is a hack
	// there's sometimes a weird effect with the sng when switching from active to inactive
	// in mid-spin - this is a QC bug - it should go through the full spin cycle before becoming
	// inactive rather than doing so straight away.  we'll fix it by not interpolating if the
	// current framenumber is 0.  the sng spin cycle uses frames 1 to 8 so this is safe to do.
	if (!paliashdr->interpolate && frame == 0)
		GL_DrawAliasFrame (paliashdr, pose, e);
	else GL_DrawAliasBlendedFrame (paliashdr, e->pose1, e->pose2, blend, e);

	e->lastframe = frame;
}


/*
=================
R_DrawAliasModel

=================
*/
void R_SetAliasLight (entity_t *e, aliashdr_t *paliashdr, vec3_t org)
{
	int			i;
	int			*jcolours;
	int			lnum;
	vec3_t		dist;
	float		add;

	// just make the damn thing fullbright as appropriate, rather than fucking around finding the
	// light level and then finally after all that, checking if it should be fullbright.  stupid
	// stupid stupid stupid.  this should drastically speed up model rendering.

	if (paliashdr->is_fullbright && !gl_useclight)
		ambientlight = shadelight = paliashdr->is_fullbright;
	else if (paliashdr->is_fullbright && gl_useclight)
	{
		ambientlight = 
		shadelightcolour[0] = 
		shadelightcolour[1] = 
		shadelightcolour[2] = 
		shadelightcolour[3] = paliashdr->is_fullbright;
	}
	else if (gl_useclight)
	{
		// get lighting information
		jcolours = R_LightPointColour (org);
	    shadelightcolour[0] = (float) jcolours[0];
		shadelightcolour[1] = (float) jcolours[1];
	    shadelightcolour[2] = (float) jcolours[2];
		shadelightcolour[3] = (float) jcolours[3];
	    ambientlight = shadelightcolour[3];

		// allways give the gun some light
		if (e == &cl.viewent && ambientlight < 24)
			ambientlight = shadelightcolour[3] = 24;

		for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
		{
			if (!r_dynamic.value)
				break;

			if (cl_dlights[lnum].die >= cl.time)
			{
				VectorSubtract (org, cl_dlights[lnum].origin, dist);
				add = cl_dlights[lnum].radius - VLength(dist);

				if (add > 0) 
				{
					ambientlight += add;
					shadelightcolour[0] += (cl_dlights[lnum].colour[0] * add);
					shadelightcolour[1] += (cl_dlights[lnum].colour[1] * add);
					shadelightcolour[2] += (cl_dlights[lnum].colour[2] * add);
					shadelightcolour[3] += add;
				}
			}
		}

		// clamp lighting so it doesn't overbright as much
		if (ambientlight > 164) ambientlight = 164;
		if (shadelightcolour[0] > 164) shadelightcolour[0] = 164;
		if (shadelightcolour[1] > 164) shadelightcolour[1] = 164;
		if (shadelightcolour[2] > 164) shadelightcolour[2] = 164;
		if (shadelightcolour[3] > 164) shadelightcolour[3] = 164;

		// ZOID: never allow players to go totally black
		i = currententity - cl_entities;
		if (i >= 1 && i<=cl.maxclients)
			if (ambientlight < 8)
				ambientlight = shadelightcolour[3] = 8;
	}
	else
	{
		// get lighting information
		ambientlight = shadelight = R_LightPoint (org);

		// allways give the gun some light
		if (e == &cl.viewent && ambientlight < 24) ambientlight = shadelight = 24;

		for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
		{
			if (cl_dlights[lnum].die >= cl.time)
			{
				VectorSubtract (org, cl_dlights[lnum].origin, dist);
				add = cl_dlights[lnum].radius - VLength(dist);

				if (add > 0) 
				{
					ambientlight += add;
					shadelight += add;
				}
			}
		}

		// clamp lighting so it doesn't overbright as much
		if (ambientlight > 164) ambientlight = 164;
		if (shadelight > 164) shadelight = 164;

		// ZOID: never allow players to go totally black
		i = currententity - cl_entities;
		if (i >= 1 && i<=cl.maxclients)
			if (ambientlight < 8)
				ambientlight = shadelight = 8;
	}

	if (gl_useclight)
	{
	    shadelightcolour[0] = shadelightcolour[0] / 164.0;
	    shadelightcolour[1] = shadelightcolour[1] / 164.0;
	    shadelightcolour[2] = shadelightcolour[2] / 164.0;
	    shadelightcolour[3] = shadelightcolour[3] / LIGHT_DIV;
	}
	else shadelight = shadelight / LIGHT_DIV;

	// light smoothing by pox
	// mh - slight bugfix - last_shadelight is always 0 for the first frame, so don't
	// include it in the calcs if it is 0.
	if (currententity->last_shadelight)
		shadelight = (shadelight + currententity->last_shadelight) / 2;

	// if shadelight is genuinely calculated to 0 (rare, but you never know) the above
	// will cock up, so set last_shadelight to 1 rather than to shadelight - this will
	// look virtually identical.
	if (shadelight)
		currententity->last_shadelight = shadelight;
	else currententity->last_shadelight = 1;
}


void R_FlameSmokes (vec3_t org);
extern int num_amodel_glows;
extern int num_amodel_shadows;
extern cvar_t view_hand;

void R_DrawAliasModel (entity_t *e, aliashdr_t *paliashdr)
{
	// i'm probably doubling up on some variables here
	int			i;
	model_t		*clmodel;
	vec3_t		mins, maxs;
	float		an;
	int			anim;
	float ang_ceil, ang_floor;
	float vm_pos;

	clmodel = currententity->model;

	VectorAdd (currententity->origin, clmodel->mins, mins);
	VectorAdd (currententity->origin, clmodel->maxs, maxs);

	// the paliashdr->nocull variable is obsolete - remove it sometime
	if (R_CullBox (mins, maxs) && !paliashdr->nocull) return;

	VectorCopy (currententity->origin, r_entorigin);
	VectorSubtract (r_origin, r_entorigin, modelorg);

	R_SetAliasLight (e, paliashdr, currententity->origin);

	// light lerping by pox
	if (!gl_aliaslightlerp.value)
	{
		shadedots = r_avertexnormal_dots[((int)(e->angles[1] * 
			(SHADEDOT_QUANT / 360.0))) & (SHADEDOT_QUANT - 1)];
	}
	else
	{
		// add pitch angle so lighting changes when looking up/down (mainly for viewmodel)
		lightlerpoffset = (e->angles[1]+e->angles[0]) * (SHADEDOT_QUANT / 360.0);

		ang_ceil = ceil(lightlerpoffset);
		ang_floor = floor(lightlerpoffset);

		lightlerpoffset = ang_ceil - lightlerpoffset;

		shadedots = r_avertexnormal_dots[(int)ang_ceil & (SHADEDOT_QUANT - 1)];
		shadedots2 = r_avertexnormal_dots[(int)ang_floor & (SHADEDOT_QUANT - 1)];
	}
 	
	an = e->angles[1] / 180*M_PI;
	shadevector[0] = cos(-an);
	shadevector[1] = sin(-an);
	shadevector[2] = 1;
	VectorNormalize (shadevector);

	// draw all the triangles
    glPushMatrix ();

	// transform should interpolate for all models (except the viewmodel, of course...)
	// BUG! doesn't interpolate the player model
	if (r_interpolate_alias.value)
	{
		// FIXED!
		if (paliashdr->is_playermodel)
			R_BlendedRotateForEntity (e, 0);
		else if (e == &cl.viewent)
			R_RotateForEntity (e);
		else R_BlendedRotateForEntity (e, 0);
	}
	else R_RotateForEntity (e);

	// FIXME! - add in glTranslatef and glRotatef here to get the proper left/right/center
	// orientation for the viewmodel
	if (view_hand.value && e == &cl.viewent)
	{
		if (view_hand.value > 1.0)
		{
			glRotatef (25.0, 1, 0, 0);
			// translate here
#ifdef VIEW_TEST
			glTranslatef (view_hand_x.value, view_hand_y.value, view_hand_z.value);
#endif
		}
		else
		{
			glRotatef (-25.0, 1, 0, 0);
			// translate here
#ifdef VIEW_TEST
			glTranslatef (view_hand_x.value, view_hand_y.value, view_hand_z.value);
#endif
		}

		// fov adjustments - don't adjust the viewmodel for fov of 90 or less, for higher fov
		// rotate it in the y axis to retain correct positioning at the bottom center.  this
		// formula is good for most values of fov.  and yes, i use rotate rather than the more
		// obvious seeming translate - hell, it works.
		if (r_refdef.fov_x > 90.0)
		{
			vm_pos = (r_refdef.fov_x - 90.0) * 0.66666666;
			glRotatef (vm_pos, 0, 1, 0);
		}
	}

	// i'm no longer doubling the size of the eyes cos i wanna push invisible player models
	// to the attached client for any other clients that have the ring in deathmatch.
	glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
	glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);

	anim = (int)(cl.time * 10) & 3;

	// we don't mind checking for oldtexture here, although we do have to set it
    glBindTexture (GL_TEXTURE_2D, paliashdr->gl_texturenum[currententity->skinnum][anim]);
	oldtexture = paliashdr->gl_texturenum[currententity->skinnum][anim];

	// certain model skins are always fullbright anyway
	if (paliashdr->is_fullbright)
		fullbright_skin = -1;
	else fullbright_skin = paliashdr->fullbright[currententity->skinnum][anim];

	// we can't dynamically colormap textures, so they are cached
	// seperately for the players.  Heads are just uncolored.
	if (currententity->colormap != vid.colormap && !gl_nocolors.value)
	{
		i = currententity - cl_entities;

		if (i >= 1 && i <= cl.maxclients)
		{
			// we don't mind checking for oldtexture here, although we do have to set it
		    glBindTexture (GL_TEXTURE_2D, playertextures - 1 + i);
			oldtexture = playertextures - 1 + i;

			// oh dear...
			fullbright_skin = -1;
		}
	}

	if (e == &cl.viewent)
	{
		// this fucks up scripted translucency for a viewmodel - unfortunately, there's nothing
		// we can really do about it without making a godawful mess.
		if (cl.items & IT_INVISIBILITY)
			paliashdr->trans_level = 0.5;
		else paliashdr->trans_level = 1;

		// don't draw a fullbright skin for the inactive animation of a viewmodel
		if (currententity->frame == 0) fullbright_skin = -1;
	}

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	if (r_interpolate_alias.value)
		R_SetupAliasBlendedFrame (currententity->frame, paliashdr, currententity);
	else R_SetupAliasFrame (currententity->frame, paliashdr, currententity);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glPopMatrix ();

	// these models should give a smoking effect...
	if (!strncmp (clmodel->name, "progs/flame", 11) || !strcmp(clmodel->name, "progs/candle.mdl"))
		R_FlameSmokes (currententity->origin);

	// check for glows and shadows so we don't have to go through the loop again if we have none
	if (paliashdr->is_glowmap && gl_aliasglow.value) num_amodel_glows++;
	if (r_shadows.value && paliashdr->cast_shadow) num_amodel_shadows++;
}


extern float PowerupGlowRadius;
extern float TorchGlowFlicker;
extern float BoltGlowFlicker;

void R_DrawAliasGlowFX (entity_t *e, aliashdr_t *paliashdr)
{
	model_t	*clmodel;
	vec3_t lightorigin;
	vec3_t v;
	float radius;
	int k;
	int i;

	if (paliashdr->is_glowmap && gl_aliasglow.value)
	{
		clmodel = currententity->model;

        VectorCopy(currententity->origin, lightorigin);

        // set radius based on what model we are doing here - we wanna increase the radius for a
		// large flame, but i haven't found a way to do this yet
		if (paliashdr->is_torch || paliashdr->is_missile)
			radius = TorchGlowFlicker;	// flickering torch
		else if (paliashdr->is_bolt)
			radius = BoltGlowFlicker;	// flickering lightning bolt
		else if (paliashdr->is_powerup)
			radius = PowerupGlowRadius;	// throb
		else radius = 30.0f;	// steady

        VectorSubtract(lightorigin, r_origin, v);

        glPushMatrix();

		if (paliashdr->is_torch)
			glTranslatef(0.0f, 0.0f, 8.0f);
		else if (paliashdr->is_powerup)
			glTranslatef(0.0f, 0.0f, 20.0f);
		else if (paliashdr->is_missile)
        {
			// move the glow behind the rocket - fixme - do a fast sin/cos lookup table for this
			// would warpsin do? - actually, what's the point since there's really only ever gonna
			// be a handful of these per frame...?
			glTranslatef (cos( e->angles[1] / 180*M_PI)*(-20.0f),
				sin( e->angles[1] / 180*M_PI)*(-20.0f),
				sin( e->angles[0] / 180*M_PI)*(-20.0f));
        }
		else if (!strcmp (e->model->name, "progs/s_light.mdl"))
		{
			// shambler magic quick and dirty bugfix
			glTranslatef(0.0f, 0.0f, 75.0f);
			radius = 60.0f;
		}

        glBegin(GL_TRIANGLE_FAN);

        // set the colour of the glow - muff
		// mh - do it quicker!
		glColor3fv (paliashdr->glow_rgb);

        for (i=0 ; i<3 ; i++)
            v[i] = lightorigin[i] - vpn[i]*radius;

        glVertex3fv (v);
        glColor3f(0.0f, 0.0f, 0.0f);

        for (i=16; i>=0; i--) 
		{
			for (k=0; k<3; k++) v[k] = lightorigin[k] + vright[k]*glowcos[i]*radius + vup[k]*glowsin[i]*radius;
			glVertex3fv (v);
        }

        glEnd();

        // Restore previous matrix! KH
	    glPopMatrix();
	}
}


void R_DrawAliasShadowFX (entity_t *e, aliashdr_t *paliashdr)
{
	trace_t		downtrace;
	vec3_t		downmove;
	vec3_t		mins;

	if (r_shadows.value && paliashdr->cast_shadow)
	{
		VectorAdd (currententity->origin, currententity->model->mins, mins);

		// can't do a single global push and pop for all shadows cos each shadow will have
		// different translatef and rotatef modifications.  shit.
		glPushMatrix ();

		// this determines the shadow vector relative to origin of entity
		// changed always be directly below entity for the moment.
		// will alter later (hopefully) when I got proper light source shadows working
		shadevector[0] = 0;
		shadevector[1] = 0;
		shadevector[2] = 1;
		VectorNormalize (shadevector);

		// this translates the location of the shadow, and rotates in horizontal plane only
		// this means that shadows now work properly for chase_cam, grenade etc
		// hence removal of code above
		// mh - don't do a blended rotate for entity on the viewmodel
		// (it's not casting a shadow right now anyway, but just in case we ever change the code...)
		if (paliashdr->interpolate && e != &cl.viewent && r_interpolate_alias.value)
			R_BlendedRotateForEntity (e, 1); // we've re-written this to work properly		 
		else
		{
			glTranslatef (e->origin[0],  e->origin[1],  e->origin[2]);
			glRotatef (e->angles[1],  0, 0, 1);
		}

		// this is so we can calculate height of the ground accurately
		VectorCopy (e->origin, downmove);

		downmove[2] = downmove[2] - 4096;
		memset (&downtrace, 0, sizeof(downtrace));
		SV_RecursiveHullCheck (cl.worldmodel->hulls, 0, 0, 1, e->origin, downmove, &downtrace);

		glDisable (GL_TEXTURE_2D);
		glEnable (GL_BLEND);

		// set the shade of the shadow based on height off ground
		// mh - this was divided by 60 so we divide by 30 instead to get a lighter shadow
		// (looks nicer...)
		glColor4f (0,0,0,1.0 - ((mins[2]-downtrace.endpos[2])/30));  //60)); 

		// BUG! - lastposenum0 and lastposenum are different for each model - since we now draw
		// shadows seperately to the model, we need to store them for the model otherwise
		// the model frame will get out of sync with the shadow frame... for interpolated models
		// they are stored in pose1 and pose2, for non-interpolated models i just use pose1.
		if (paliashdr->interpolate && r_interpolate_alias.value)
			GL_DrawAliasBlendedShadow (paliashdr, currententity->pose1, currententity->pose2, currententity); //lastposenum0, lastposenum, currententity);
		else GL_DrawAliasShadow (paliashdr, currententity->pose1); //lastposenum);

		glEnable (GL_TEXTURE_2D);
		glDisable (GL_BLEND);
		glColor4f (1,1,1,1);
		glPopMatrix ();
	}
}


/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (void)
{
	float		ambient[4], diffuse[4];
	int			j;
	int			lnum;
	vec3_t		dist;
	float		add;
	dlight_t	*dl;
	int			ambientlight, shadelight;
	int			*jcolour;
	int			shadelightcolour[4];

	if (!r_drawviewmodel.value) return;
	if (chase_active.value) return;
	if (!r_drawentities.value) return;
	if (cl.stats[STAT_HEALTH] <= 0) return;
	currententity = &cl.viewent;
	if (!currententity->model) return;

	if (gl_useclight)
	{
		jcolour = R_LightPointColour (currententity->origin);

		// allways give some light on gun
		if (jcolour[3] < 24) jcolour[3] = 24;

		ambientlight = jcolour[3];
		shadelightcolour[0] = jcolour[0];
		shadelightcolour[1] = jcolour[1];
		shadelightcolour[2] = jcolour[2];
		shadelightcolour[3] = jcolour[3];
	}
	else 
	{
		j = R_LightPoint (currententity->origin);

		// allways give some light on gun
		if (j < 24) j = 24;

		ambientlight = j;
		shadelight = j;
	}

	// add dynamic lights		
	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		if (!r_dynamic.value) break;
		dl = &cl_dlights[lnum];
		if (!dl->radius) continue;
		if (!dl->radius) continue;
		if (dl->die < cl.time) continue;

		VectorSubtract (currententity->origin, dl->origin, dist);
		add = dl->radius - VLength(dist);

		if (add > 0)
		{
			if (gl_useclight)
			{
				shadelightcolour[0] += (dl->colour[0] * add);
				shadelightcolour[1] += (dl->colour[1] * add);
				shadelightcolour[2] += (dl->colour[2] * add);
				shadelightcolour[3] += add;
			}
			else shadelight += add;	// id left this out...

			ambientlight += add;
		}
	}

	ambient[0] = ambient[1] = ambient[2] = ambient[3] = (float)ambientlight / 128;

	if (gl_useclight)
		diffuse[0] = diffuse[1] = diffuse[2] = diffuse[3] = (float)shadelightcolour[3] / 128;
	else diffuse[0] = diffuse[1] = diffuse[2] = diffuse[3] = (float)shadelight / 128;

	// hack the depth range to prevent view model from poking into walls
	glDepthRange (gldepthmin, gldepthmin + 0.3*(gldepthmax-gldepthmin));
	R_DrawAliasModel (currententity, (aliashdr_t *)Mod_Extradata (currententity->model));
	glDepthRange (gldepthmin, gldepthmax);
}


