/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_mesh.c: triangle model functions
#include "quakedef.h"

model_t		*aliasmodel;
aliashdr_t	*paliashdr;

qboolean	used[8192];

// the command list holds counts and s/t values that are valid for
// every frame
int		commands[8192];
int		numcommands;

// all frames will have their vertexes rearranged and expanded
// so they are in the order expected by the command list
int		vertexorder[8192];
int		numorder;

int		allverts, alltris;

int		stripverts[128];
int		striptris[128];
int		stripcount;

/*
================
StripLength
================
*/
int	StripLength (int starttri, int startv)
{
	int			m1, m2;
	int			j;
	mtriangle_t	*last, *check;
	int			k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv)%3];
	stripverts[1] = last->vertindex[(startv+1)%3];
	stripverts[2] = last->vertindex[(startv+2)%3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv+2)%3];
	m2 = last->vertindex[(startv+1)%3];

	// look for a matching triangle
nexttri:
	for (j=starttri+1, check=&triangles[starttri+1] ; j<pheader->numtris ; j++, check++)
	{
		if (check->facesfront != last->facesfront) continue;

		for (k=0 ; k<3 ; k++)
		{
			if (check->vertindex[k] != m1) continue;
			if (check->vertindex[ (k+1)%3 ] != m2) continue;

			// this is the next part of the fan

			// if we can't use this triangle, this tristrip is done
			if (used[j]) goto done;

			// the new edge
			if (stripcount & 1)
				m2 = check->vertindex[ (k+2)%3 ];
			else m1 = check->vertindex[ (k+2)%3 ];

			stripverts[stripcount+2] = check->vertindex[ (k+2)%3 ];
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:

	// clear the temp used flags
	for (j=starttri+1 ; j<pheader->numtris ; j++)
		if (used[j] == 2) used[j] = 0;

	return stripcount;
}

/*
===========
FanLength
===========
*/
int	FanLength (int starttri, int startv)
{
	int		m1, m2;
	int		j;
	mtriangle_t	*last, *check;
	int		k;

	used[starttri] = 2;

	last = &triangles[starttri];

	stripverts[0] = last->vertindex[(startv)%3];
	stripverts[1] = last->vertindex[(startv+1)%3];
	stripverts[2] = last->vertindex[(startv+2)%3];

	striptris[0] = starttri;
	stripcount = 1;

	m1 = last->vertindex[(startv+0)%3];
	m2 = last->vertindex[(startv+2)%3];


	// look for a matching triangle
nexttri:
	for (j=starttri+1, check=&triangles[starttri+1] ; j<pheader->numtris ; j++, check++)
	{
		if (check->facesfront != last->facesfront) continue;

		for (k=0 ; k<3 ; k++)
		{
			if (check->vertindex[k] != m1) continue;
			if (check->vertindex[ (k+1)%3 ] != m2) continue;

			// this is the next part of the fan

			// if we can't use this triangle, this tristrip is done
			if (used[j]) goto done;

			// the new edge
			m2 = check->vertindex[ (k+2)%3 ];

			stripverts[stripcount+2] = m2;
			striptris[stripcount] = j;
			stripcount++;

			used[j] = 2;
			goto nexttri;
		}
	}
done:

	// clear the temp used flags
	for (j=starttri+1 ; j<pheader->numtris ; j++)
		if (used[j] == 2) used[j] = 0;

	return stripcount;
}

/*
================
BuildTris

Generate a list of trifans or strips
for the model, which holds for all frames
================
*/
void BuildTris (void)
{
	int		i, j, k;
	int		startv;
	float	s, t;
	int		len, bestlen, besttype;
	int		bestverts[1024];
	int		besttris[1024];
	int		type;

	//
	// build tristrips
	//
	numorder = 0;
	numcommands = 0;
	memset (used, 0, sizeof(used));

	for (i=0 ; i<pheader->numtris ; i++)
	{
		// pick an unused triangle and start the trifan
		if (used[i]) continue;

		bestlen = 0;

		for (type = 0 ; type < 2 ; type++)
		{
			for (startv =0 ; startv < 3 ; startv++)
			{
				if (type == 1)
					len = StripLength (i, startv);
				else len = FanLength (i, startv);

				if (len > bestlen)
				{
					besttype = type;
					bestlen = len;

					for (j=0 ; j<bestlen+2 ; j++) bestverts[j] = stripverts[j];
					for (j=0 ; j<bestlen ; j++) besttris[j] = striptris[j];
				}
			}
		}

		// mark the tris on the best strip as used
		for (j=0 ; j<bestlen ; j++) used[besttris[j]] = 1;

		if (besttype == 1)
			commands[numcommands++] = (bestlen+2);
		else commands[numcommands++] = -(bestlen+2);

		for (j=0 ; j<bestlen+2 ; j++)
		{
			// emit a vertex into the reorder buffer
			k = bestverts[j];
			vertexorder[numorder++] = k;

			// emit s/t coords into the commands stream
			s = stverts[k].s;
			t = stverts[k].t;

			if (!triangles[besttris[0]].facesfront && stverts[k].onseam)
				s += pheader->skinwidth / 2;	// on back side

			s = (s + 0.5) / pheader->skinwidth;
			t = (t + 0.5) / pheader->skinheight;

			*(float *)&commands[numcommands++] = s;
			*(float *)&commands[numcommands++] = t;
		}
	}

	commands[numcommands++] = 0;		// end of list marker

	allverts += numorder;
	alltris += pheader->numtris;
	
}

/*
=============
GL_GetTriNeighbors

Get the neighbors for a given triangle (meaning, said neighbor
shares an edge with this triangle).
-rww
=============
*/
void GL_GetTriNeighbors(aliashdr_t *hdr, mtriangle_t *tris, mtriangle_t *tri, int *neighbors)
{
	static mtriangle_t *n;
	int i;
	int numN;
	static int j, k;
	static int vertA1, vertA2, vertB1, vertB2;

	i = 0;
	numN = 0;
	
	neighbors[0] = neighbors[1] = neighbors[2] = -1;

	while (i < hdr->numtris)
	{
		n = &tris[i];

		if (n != tri)
		{
			j = 0;
			while (j < 3)
			{
				vertA1 = tri->vertindex[j];
				vertA2 = tri->vertindex[(j+1)%3];

				k = 0;
				while (k < 3)
				{
					vertB1 = n->vertindex[k];
					vertB2 = n->vertindex[(k+1)%3];

					if ((vertA1 == vertB1 && vertA2 == vertB2) || (vertA1 == vertB2 && vertA2 == vertB1))
					{
						if (neighbors[j] == -1)
						{
							neighbors[j] = i;
						}
						else
						{ //this edge is shared by more than 2 tris apparently.
						  //this seems to actually happen at times, which is why
						  //we are bothering to check for it.
							neighbors[j] = -2;
						}
					}

					k++;
				}

				j++;
			}
		}

		i++;
	}

	j = 0;
	while (j < 3)
	{ //mark any -2 as -1 now since we're done
		if (neighbors[j] == -2)
		{
			neighbors[j] = -1;
		}
		j++;
	}
}

/*
=============
GL_TriPrecalc

Do whatever precalc'ing we can (in this case just get neighbors).
We could also pre-determine planeEq's for each tri*each frame in
here if we desired.
-rww
=============
*/
void GL_TriPrecalc(aliashdr_t *hdr, mtriangle_t *tris, trivertx_t *verts)
{
	int			i;
	mtriangle_t *tri;

	i  = 0;

	while (i < hdr->numtris)
	{
		tri = &tris[i];
		GL_GetTriNeighbors(hdr, tris, tri, tri->neighbors);
		i++;
	}
}

/*
================
GL_MakeAliasModelDisplayLists
================
*/
void GL_MakeAliasModelDisplayLists (model_t *m, aliashdr_t *hdr)
{
	int			i, j;
	int			*cmds;
	mtriangle_t *tris;
	trivertx_t	*verts;
	trivert2	*norms;

	//First save the original vertex data for reference by our tris
	verts = Hunk_Alloc (hdr->numposes * hdr->numverts * sizeof(trivertx_t));
	hdr->baseposedata = (byte *)verts - (byte *)hdr;
	
	for (i=0 ; i<hdr->numposes ; i++)
	{
		for (j=0 ; j<hdr->numverts ; j++)
		{
			*verts++ = poseverts[i][vertexorder[j]];	
		}
	}

	//Now save the triangles as we will need them for shadow volumes
	tris = Hunk_Alloc (hdr->numtris * sizeof(mtriangle_t));
	hdr->triangles = (byte *)tris - (byte *)hdr;
	memcpy(tris, triangles, hdr->numtris * sizeof(mtriangle_t));

	//Calculate some stuff for the tris ahead of time
	GL_TriPrecalc(hdr, tris, verts);

	Con_DPrintf ("loading %s...\n",m->name);

	BuildTris ();

	// save the data out
	hdr->poseverts = numorder;

	cmds = Hunk_Alloc (numcommands * 4);
	hdr->commands = (byte *)cmds - (byte *)hdr;
	memcpy (cmds, commands, numcommands * 4);

	verts = Hunk_Alloc (hdr->numposes * hdr->poseverts * sizeof(trivertx_t));
	norms = Hunk_Alloc (hdr->numposes * hdr->poseverts * sizeof(trivert2));

	hdr->posedata = (byte *)verts - (byte *)hdr;
	hdr->normdata = (byte *)norms - (byte *)hdr;

	for (i=0 ; i<hdr->numposes ; i++)
	{
		for (j=0 ; j<numorder ; j++)
		{
			*verts++ = poseverts[i][vertexorder[j]];
			*norms++ = posenorms[i][vertexorder[j]];
		}
	}
}



