/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// models.c -- model loading and caching

// models are the only shared resource between a client and server running
// on the same machine.

#include "quakedef.h"
#include "mod_common.h"

cvar_t gl_colouredlight = {"gl_colouredlight", "0", true};
cvar_t gl_subdivide_size = {"gl_subdivide_size", "128", true};

// fullbright colours - common to brush and alias models
int FindFullbrightTexture (byte *pixels, int num_pix)
{
	int i;

	// 224 to 254 are the fullbright colours in the quake palette.  255 is the transparent colour
	// fixme: if all the pixels are fullbright, then return -1 and only draw the fullbright
	// texture with no light modification
	for (i = 0; i < num_pix; i++)
	{
	    if (pixels[i] > 223 && pixels[i] != 255)
		{
			return 1;
		}
	}return 0;
}


void ConvertPixels (byte *pixels, int num_pixels)
{
	int i;

	for (i = 0; i < num_pixels; i++)
	{
		if (pixels[i] < 224)
		{
			pixels[i] = 255;
		}
	}
}


/*
===============
Mod_Init
===============
*/
void Mod_Init (void)
{
	Cvar_RegisterVariable (&gl_subdivide_size);
	Cvar_RegisterVariable (&gl_colouredlight);
	memset (mod_novis, 0xff, sizeof(mod_novis));
}


/*
===============
Mod_Extradata

Caches the data if needed
===============
*/
void *Mod_Extradata (model_t *mod)
{
	void	*r;

	r = Cache_Check (&mod->cache);

	if (r)
	{
		return r;
	}
	Mod_LoadModel (mod, true);

	if (!mod->cache.data)
	{
		Sys_Error ("Mod_Extradata: caching failed");
	}
	return mod->cache.data;
}


/*
===============
Mod_PointInLeaf
===============
*/
mleaf_t *Mod_PointInLeaf (vec3_t p, model_t *model)
{
	mnode_t *node;
	float d;
	mplane_t *plane;

	if (!model || !model->nodes)
	{
		Sys_Error ("Mod_PointInLeaf: bad model");
	}
	
	node = model->nodes;

	while (1)
	{
		if (node->contents < 0)
		{
			return (mleaf_t *)node;
		}
		
		plane = node->plane;
		d = DotProduct (p, plane->normal) - plane->dist;

		if (d > 0)
		{
			node = node->children[0];
		}
		else
		{
			node = node->children[1];
		}
	}
	return NULL;	// never reached
}


/*
===================
Mod_DecompressVis
===================
*/
byte *Mod_DecompressVis (byte *in, model_t *model)
{
	static byte	decompressed[MAX_MAP_LEAFS/8];
	int			c;
	byte		*out;
	int			row;

	row = (model->numleafs+7)>>3;	
	out = decompressed;

	if (!in)
	{	// no vis info, so make all visible
		while (row)
		{
			*out++ = 0xff;
			row--;
		}
		return decompressed;		
	}

	do
	{
		if (*in)
		{
			*out++ = *in++;
			continue;
		}
	
		c = in[1];
		in += 2;

		while (c)
		{
			*out++ = 0;
			c--;
		}
	} while (out - decompressed < row);
	
	return decompressed;
}


byte *Mod_LeafPVS (mleaf_t *leaf, model_t *model)
{
	if (leaf == model->leafs)
	{
		return mod_novis;
	}
	return Mod_DecompressVis (leaf->compressed_vis, model);
}


/*
===================
Mod_ClearAll
===================
*/
void Mod_ClearAll (void)
{
	int		i;
	model_t	*mod;
	
	for (i = 0, mod = mod_known; i < mod_numknown; i++, mod++)
	{
		if (mod->type != mod_alias)
		{
			mod->needload = true;
		}
	}
}


/*
==================
Mod_FindName

==================
*/
model_t *Mod_FindName (char *name)
{
	int		i;
	model_t	*mod;
	
	// should never happen
	if (!name[0])
	{
		Sys_Error ("Mod_ForName: NULL name");
	}
	// search the currently loaded models
	for (i=0 , mod=mod_known ; i<mod_numknown ; i++, mod++)
	{
		if (!strcmp (mod->name, name))
		{
			break;
		}
	}
	if (i == mod_numknown)
	{
		if (mod_numknown == MAX_MOD_KNOWN)
		{
			Sys_Error ("mod_numknown == MAX_MOD_KNOWN");
		}
		strcpy (mod->name, name);
		mod->needload = true;
		mod_numknown++;
	}
	return mod;
}


/*
==================
Mod_TouchModel

==================
*/
void Mod_TouchModel (char *name)
{
	model_t	*mod;
	
	mod = Mod_FindName (name);
	
	if (!mod->needload)
	{
		if (mod->type == mod_alias)
		{
			Cache_Check (&mod->cache);
		}
	}
}


/*
==================
Mod_LoadModel

Loads a model into the cache
==================
*/
model_t *Mod_LoadModel (model_t *mod, qboolean crash)
{
	void		*d;
	unsigned	*buf;
	byte		stackbuf[1024];		// avoid dirtying the cache heap

	if (!mod->needload)
	{
		if (mod->type != mod_sprite && mod->type != mod_brush && mod->type != mod_md3)
		{
			d = Cache_Check (&mod->cache);
			if (d) return mod;
		}
		else 
		{
			return mod;		// not cached at all
		}
	}

	// because the world is so huge, load it one piece at a time
	if (!crash)
	{
	}

	// load the file
	buf = (unsigned *)COM_LoadStackFile (mod->name, stackbuf, sizeof(stackbuf));

	if (!buf)
	{
		if (crash)
		{
			Sys_Error ("Mod_NumForName: %s not found", mod->name);
		}
		return NULL;
	}

	// allocate a new model
	COM_FileBase (mod->name, loadname);
	
	// fill it in
	loadmodel = mod;

	// call the apropriate loader
	mod->needload = false;

	switch (LittleLong(*(unsigned *)buf))
	{
	case IDPOLYHEADER:
		Mod_LoadAliasModel (mod, buf);
		break;

	case MD3IDHEADER:
		Mod_LoadQ3Model (mod, buf);
		break;

	case IDSPRITEHEADER:
		Mod_LoadSpriteModel (mod, buf);
		break;

	default:
		Mod_LoadBrushModel (mod, buf);
		break;
	}
	return mod;
}


/*
==================
Mod_ForName

Loads in a model for the given name
==================
*/
model_t *Mod_ForName (char *name, qboolean crash)
{
	model_t	*mod;

 	mod = Mod_FindName (name);
	
	return Mod_LoadModel (mod, crash);
}


/*
================
Mod_Print
================
*/
void Mod_Print (void)
{
	int		i;
	model_t	*mod;

	Con_Printf ("Cached models:\n");

	for (i=0, mod=mod_known ; i < mod_numknown ; i++, mod++)
	{
		Con_Printf ("%8p : %s\n",mod->cache.data, mod->name);
	}
}


