/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// common script routines - these probably duplicate a lot of stuff already in the engine - i'll
// clean it up someday.  these routines may not be very robust, so try to keep your script layout
// kinda neat.
// script files are always opened as ascii.

#include "quakedef.h"

#ifdef USE_SCRIPTS
static qboolean dirty_hack;

int mh_getc (FILE *scriptfile)
{
	char c;

	if ((c = getc (scriptfile)) == EOF) return -1;
	return c;
}


// open the script file as ascii
FILE *Script_Open (char *path)
{
	char basedir[MAX_OSPATH];
	FILE *scriptfile;

	// get the full path to the id1 directory
	sprintf (basedir, "%s/%s", host_parms.basedir, GAMENAME);

	// check current game directory - if not present, check id1
	scriptfile = fopen (va ("%s/scripts/%s.mhs", com_gamedir, path), "r");
	if (!scriptfile) scriptfile = fopen (va ("%s/scripts/%s.mhs", basedir, path), "r");
	return scriptfile;
}


// returns one of a key or value read from the scriptfile as a string
qboolean Script_Read (FILE *scriptfile, char *Script_Buffer)
{
	int i;

	// fscanf may be evil, but it's easier to code this using it.
	if (fscanf (scriptfile, " %s ", Script_Buffer) == EOF) return false;

	// inelegant, but it works
	if (!dirty_hack)
		for (i = 0; i < strlen (Script_Buffer); i++)
			Script_Buffer[i] = Script_Buffer[i + 1];

	dirty_hack = false;

	Script_Buffer[strlen (Script_Buffer) - 1] = '\0';
	return true;
}


// reads until we get an opening brace (ie. skips everything else), then reads to the first
// non-whitespace character.
// will return an error if we get to the eof without an opening brace
qboolean Script_ParseOpeningBrace (FILE *scriptfile)
{
	char c;

	while (1)
	{
		if ((c = mh_getc (scriptfile)) == -1) goto bad_script;
		if (c == '{') break;
	}

	return true;

bad_script:
	return false;
}


// checks if the next non-whitespace character is a closing brace or double quote
int Script_ParseClosingBrace (FILE *scriptfile)
{
	char c;

	while (1)
	{
		if ((c = mh_getc (scriptfile)) == -1) goto bad_script;
		if (c == 34)
			break;
		else if (c == '}') return 1;
	}

	dirty_hack = true;
	return 0;

bad_script:
	return -1;
}


// to prevent people from acting the bollocks, all scripts must begin with the characters
// {SCRIPT  - this validates the script and returns true if it's valid
qboolean Validate_Script (FILE *scriptfile)
{
	char Script_Header[9];
	int i;
	char c;

	// yeah, i know - so shoot me!
	sprintf (Script_Header, "%s", "{SCRIPT ");

	for (i = 0; i < 8; i++)
	{
		if ((c = mh_getc (scriptfile)) == -1) goto bad_script;
		if (c != Script_Header[i]) goto bad_script;
	}

	return true;

bad_script:
	return false;
}


// skip over the script header so that we don't confuse the parsing - it's already been
// validated so we don't need to do so again.
qboolean Skip_Validation (FILE *scriptfile)
{
	char c;

	while (1)
	{
		if ((c = mh_getc (scriptfile)) == -1) return false;
		if (c == '}') return true;
	}
}


void Alias_Script_Execute (char *scriptname, FILE *scriptfile);
void Map_Script_Execute (char *scriptname, FILE *scriptfile);

void Script_f (void)
{
	FILE *scriptfile;
	char scripttype[4];
	qboolean complain;

	// 2 args - 1st is script type (currently only alias), second is name of script
	// (not including the path and extension)
	if (Cmd_Argc () < 2) goto no_script;

	if (Cmd_Argc () == 3)
	{
		if (!strcmp (Cmd_Argv (2), "false"))
			complain = true;
		else complain = false;
	}
	else complain = false;

	// open the script to validate and determine the type
	scriptfile = Script_Open (Cmd_Argv (1));
	if (!scriptfile) goto no_script;
	if (!Validate_Script (scriptfile)) goto bad_script;

	// the file pointer is now at the first character of the script type identifier, so we
	// can just read it in - it's always gonna be 3 uppercase chars, so getc will do.
	// however, using the mh_getc function will do a lot of eof testing for us also, so that's
	// preferable.  here we also check that the script type is followed by a closing brace.
	if ((scripttype[0] = mh_getc (scriptfile)) == -1) goto bad_script;
	if ((scripttype[1] = mh_getc (scriptfile)) == -1) goto bad_script;
	if ((scripttype[2] = mh_getc (scriptfile)) == -1) goto bad_script;
	if (mh_getc (scriptfile) != '}') goto bad_script;

	scripttype[3] = '\0';

	// now that we have the script type, we can call the appropriate parsing routine for the
	// currently supported script types, or exit for an unsupported one.
	if (!strcmp ("MDL", scripttype))
		Alias_Script_Execute (Cmd_Argv (1), scriptfile);
	else if (!strcmp ("MAP", scripttype))
		Map_Script_Execute (Cmd_Argv (1), scriptfile);
	else goto bad_type;

	// we don't want no steenkin' error messages here
	goto finish_up;

	// error handling
bad_script:
	if (complain) Con_Printf ("scripts/%s.mhs is not a valid MHScript\n", Cmd_Argv (1));
	goto finish_up;

bad_type:
	if (complain) Con_Printf ("%s scripts are currently unsupported\n", scripttype);

finish_up:
	fclose (scriptfile);
	return;

no_script:
	if (complain) Con_Printf ("script <name>: run an mhquake script\n");
	return;
}


void Mod_f (void);


void Script_Init (void)
{
	Cmd_AddCommand ("script", Script_f);
	Cmd_AddCommand ("mod", Mod_f);
}
#endif
