// gl_image.c - texture loading and handling

#include "quakedef.h"
#include "gl_frontend_shared.h"
#include "glext.h"


int		texture_extension_number = 1;

int		gl_filter_min = GL_LINEAR_MIPMAP_LINEAR;
int		gl_filter_max = GL_LINEAR;


typedef struct gltexture_s
{
	int				texnum;
	char			identifier[64];
	int				width, height;
	qboolean		mipmap;
	unsigned short	crc;
	byte			*data;
	int				usage;
} gltexture_t;


#define	MAX_GLTEXTURES	1024


gltexture_t	gltextures[MAX_GLTEXTURES];
int			numgltextures;


typedef struct glmode_s
{
					char *name;
					int	minimize, maximize;
} glmode_t;


glmode_t modes[] =
{
	{"GL_NEAREST", GL_NEAREST, GL_NEAREST},
	{"GL_LINEAR", GL_LINEAR, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_NEAREST", GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_NEAREST", GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_LINEAR", GL_NEAREST_MIPMAP_LINEAR, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_LINEAR", GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR}
};


void GL_TexDump_f (void)
{
	int i;
	int freetex;
	FILE *f;

	f = fopen ("texdump.txt", "w");

	for (i = 0, freetex = 0; i < MAX_GLTEXTURES; i++)
	{
		if (!gltextures[i].crc)
		{
			freetex++;
			continue;
		}

		fprintf (f, "%s\n", gltextures[i].identifier);
		fprintf (f, "width [%i]  height [%i]\n\n", gltextures[i].width, gltextures[i].height);
	}

	fprintf (f, "%i free textures\n", freetex);
	fclose (f);
}



/*
===============
Draw_TextureMode_f
===============
*/
void Draw_TextureMode_f (void)
{
	int i;
	gltexture_t	*glt;

	if (Cmd_Argc() == 1)
	{
		for (i = 0; i < 6; i++)
		{
			if (gl_filter_min == modes[i].minimize)
			{
				Con_Printf ("%s\n", modes[i].name);
				return;
			}
		}

		Con_Printf ("current filter is unknown???\n");

		return;
	}

	for (i = 0; i < 6; i++)
	{
		if (!Q_strcasecmp (modes[i].name, Cmd_Argv(1)))
		{
			break;
		}
	}

	if (i == 6)
	{
		Con_Printf ("bad filter name\n");
		return;
	}

	gl_filter_min = modes[i].minimize;
	gl_filter_max = modes[i].maximize;

	// change all the existing mipmap texture objects
	for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
	{
		if (glt->mipmap)
		{
			glBindTexture (GL_TEXTURE_2D, glt->texnum);
			glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
			glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
		}
	}
}


/*
===============
BuildIntensityTable

Got this from Quake II and if you don't like it it's just too bad
===============
*/
// we have a gamma table in Quake 1 already
static byte intensitytable[256];
qboolean intensitytabledone = false;


void BuildIntensityTable (void)
{
	int i, j;

	// we have to do it this way cos we get a WEIRD internal compiler error if
	// we reuse the exact same code from my Quake engine (which works just fine)
	for (i = 0; i < 128; i++)
	{
		j = i * 2;

		intensitytable[i] = j;
	}

	for (i = 128; i < 256; i++)
		intensitytable[i] = 255;

	intensitytabledone = true;
}


/*
===============
GL_LightScaleTexture

Got this from Quake II and if you don't like it it's just too bad
===============
*/
void GL_LightScaleTexture (unsigned *texin, int inwidth, int inheight)
{
	int	i, c;
	byte *p;

	if (!intensitytabledone) BuildIntensityTable ();

	p = (byte *) texin;

	c = inwidth * inheight;

	for (i = 0; i < c; i++, p += 4)
	{
		p[0] = gammatable[intensitytable[p[0]]];
		p[1] = gammatable[intensitytable[p[1]]];
		p[2] = gammatable[intensitytable[p[2]]];
	}
}



/*
================
GL_ResampleTexture
================
*/
void GL_ResampleTexture (unsigned *texIn, int inwidth, int inheight, unsigned *texOut, int outwidth, int outheight)
{
	int i, j;
	unsigned *inrow, *inrow2;
	unsigned theFrac, fracstep;
	unsigned p1[1024], p2[1024];
	byte *pix1, *pix2, *pix3, *pix4;

	if (inwidth == outwidth && inheight == outheight)
	{
		// do a straight copy
		for (i = 0; i < inwidth * inheight; i++)
			texOut[i] = texIn[i];

		return;
	}

	fracstep = inwidth * 0x10000 / outwidth;
	theFrac = fracstep >> 2;

	for (i = 0; i < outwidth; i++)
	{
		p1[i] = 4 * (theFrac >> 16);
		theFrac += fracstep;
	}

	theFrac = 3 * (fracstep >> 2);

	for (i = 0; i < outwidth; i++)
	{
		p2[i] = 4 * (theFrac >> 16);
		theFrac += fracstep;
	}

	for (i = 0; i < outheight; i++, texOut += outwidth)
	{
		inrow = texIn + inwidth * (int) ((i + 0.25) * inheight / outheight);
		inrow2 = texIn + inwidth * (int) ((i + 0.75) * inheight / outheight);

		// this isn't actually *used* any more
		theFrac = fracstep >> 1;

		for (j = 0; j < outwidth; j++)
		{
			pix1 = (byte *) inrow + p1[j];
			pix2 = (byte *) inrow + p2[j];
			pix3 = (byte *) inrow2 + p1[j];
			pix4 = (byte *) inrow2 + p2[j];

			((byte *) (texOut + j))[0] = (pix1[0] + pix2[0] + pix3[0] + pix4[0]) >> 2;
			((byte *) (texOut + j))[1] = (pix1[1] + pix2[1] + pix3[1] + pix4[1]) >> 2;
			((byte *) (texOut + j))[2] = (pix1[2] + pix2[2] + pix3[2] + pix4[2]) >> 2;
			((byte *) (texOut + j))[3] = (pix1[3] + pix2[3] + pix3[3] + pix4[3]) >> 2;
		}
	}
}


/*
===============
GL_Upload32
===============
*/
void GL_Upload32 (unsigned *data, int width, int height, qboolean mipmap, qboolean alpha)
{
	int w;
	int h;
	static unsigned scaled[1024 * 1024];

	// get the power of 2 equal to or above the texture size
	for (w = 1; w < width; w *= 2);
	for (h = 1; h < height; h *= 2);

	// cap at 1024
	if (w > 1024) w = 1024;
	if (h > 1024) h = 1024;

	// cap at max texture size
	if (w > maxGLSize) w = maxGLSize;
	if (h > maxGLSize) h = maxGLSize;

	// either scale it to the new size or do a straight copy
	GL_ResampleTexture (data, width, height, scaled, w, h);

	if (mipmap)
	{
		GL_LightScaleTexture (scaled, w, h);

		gluBuild2DMipmaps (GL_TEXTURE_2D, 
						   (alpha ? GL_RGBA : GL_RGB), 
						   w, 
						   h, 
						   GL_RGBA, 
						   GL_UNSIGNED_BYTE, 
						   scaled);

		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	else
	{
		glTexImage2D (GL_TEXTURE_2D, 
					  0, 
					  (alpha ? GL_RGBA : GL_RGB), 
					  w, 
					  h, 
					  0, 
					  GL_RGBA, 
					  GL_UNSIGNED_BYTE, 
					  scaled);

		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
}


/*
===============
GL_Upload8
===============
*/
qboolean watertex;

void GL_Upload8 (byte *data, int width, int height, qboolean mipmap, qboolean alpha)
{
	static unsigned trans[1024 * 1024];
	int i;
	int s;
	int p;
	byte *rgba;

	s = width * height;
	alpha = false;

	for (i = 0; i < s; i++)
	{
		p = data[i];

		if (p == 255)
		{
			alpha = true;
			trans[i] = 0;
		}
		else trans[i] = d_8to24table[p];

		if (watertex)
		{
			// set the alpha component for liquid texes
			rgba = (byte *) &trans[i];
			rgba[3] = 128;
			alpha = true;

			// cos GL_Upload8 is not always called from GL_LoadTexture
			watertex = false;
		}
	}

	GL_Upload32 (trans, width, height, mipmap, alpha);
}


/*
================
GL_LoadTexture
================
*/
int Argh = 0;

int GL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha)
{
	qboolean	noalpha;
	int			i, p, s;
	gltexture_t	*glt;
	unsigned short crc;

	watertex = false;

	crc = CRC_Block (data, (width * height));

	// see if the texture is allready present
	if (identifier[0])
	{
		// hack to see if we're in water
		if (identifier[0] == '*') watertex = true;

		for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
		{
			// the names may match, but the data may not
			if (!strcmp (identifier, glt->identifier))
			{
				// the data doesn't match so reload it
				if (crc != glt->crc)
					goto CacheMismatchSetup;

				// update the data pointer and return the tex num
				gltextures[i].data = data;
				if (glt->usage != -1) gltextures[i].usage++;
				return gltextures[i].texnum;
			}
		}
	}
	else
	{
		// some Quake textures are loaded without names, so we find one that matches.
		// first lets give it a dummy name so we have something to call it by
		sprintf (identifier, "Argh_%4i", Argh++);

		for (i = 0, glt = gltextures; i < numgltextures; i++, glt++)
		{
			if (crc == glt->crc)
			{
				// update the data pointer and return the tex num
				gltextures[i].data = data;
				if (glt->usage != -1) gltextures[i].usage++;
				return gltextures[i].texnum;
			}
		}
	}

	// having a non-bounds checked static array is great!!!  it means that i get to
	// write code like this...
	if (numgltextures == MAX_GLTEXTURES)
		return 1;

	glt = &gltextures[numgltextures];
	numgltextures++;

CacheMismatchSetup:;
	strcpy (glt->identifier, identifier);
	glt->texnum = texture_extension_number;
	glt->width = width;
	glt->height = height;
	glt->mipmap = mipmap;

	glt->data = data;
	glt->crc = CRC_Block (glt->data, (width * height));

	// this is the first time this particular texture has been used, so set
	// it's usage count to 0.  when we start to run out of textures we'll
	// reuse one of the least used ones.
	if (glt->identifier[0] == 'p' && glt->identifier[1] == 'r' && glt->identifier[2] == 'o' &&
		glt->identifier[3] == 'g' && glt->identifier[4] == 's' && glt->identifier[5] == '/')
	{
		// we don't want to reuse certain classes of texture however.  model
		// skins and 2D screen graphics will have to be always kept.
		glt->usage = -1;
	}
	else if (mipmap)
	{
		glt->usage = 0;
	}
	else
	{
		// 2D screen graphics so keep
		glt->usage = -1;
	}

	glBindTexture (GL_TEXTURE_2D, texture_extension_number);

	GL_Upload8 (data, width, height, mipmap, alpha);

	texture_extension_number++;

	return texture_extension_number - 1;
}


/*
================
GL_LoadPicTexture
================
*/
int GL_LoadPicTexture (char *name, qpic_t *pic)
{
	return GL_LoadTexture (name, pic->width, pic->height, pic->data, false, true);
}


